--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tBlock = {

["AddBlock"] = function (self, ssid, btype, vnick, vip)
	if btype and vip and vip ~= "" and vnick and vnick ~= "" then
		local btypes = {
			["mute"] = {"tMute", "User is already muted", "User: %s with ip: %s is now muted."},
			["mutestealth"] = {"tMuteStealth", "User is already silently muted", "User: %s with ip: %s is now silently muted."},
			["kenny"] = {"tKenny",  "User is already kennylized", "User: %s with ip: %s is now kennylized."},
			["muppet"] = {"tMuppet", "User is already muppetized", "User: %s with ip: %s is now muppetized."},
			["connects"] = {"tConnects", "User's connections are already blocked", "User: %s with ip: %s is now blocked from making new connections."},
			["pm"] = {"tPm", "User's private messages are already blocked", "User: %s with ip: %s is now blocked from sending personal messages."},
			["mc"] = {"tMc", "User already can't receive mainchat messages", "User: %s with ip: %s can't receive mainchat messages anymore."},
		}
		
		if btypes[btype] then
			local tname = btypes[btype][1]
			if not tBlocks[tname][vip] then
				tBlocks[tname][vip] = vnick
				tFuncs:SaveSettings("blocks")
				return true, Lang(ssid, 0, btypes[btype][3], vnick, vip)
			else
				return true, Lang(ssid, 0, btypes[btype][2])
			end
		else
			return false, Lang(ssid, 0, "Error: Unknown blocktype: %s", btype)
		end
	else
		return false, Lang(ssid, 0, "AddBlock Error: invalid btype, vip or vnick")
	end
end,

["DelBlock"] = function (self, ssid, btype, value)
	if btype and value and value ~= "" then
		local btypes = {
			["mute"] = {"tMute", "User is already muted", "User: %s with ip: %s is now unmuted."},
			["mutestealth"] = {"tMuteStealth", "User is already silently muted", "User: %s with ip: %s is now silently unmuted."},
			["kenny"] = {"tKenny",  "User is already kennylized", "User: %s with ip: %s is now unkennylized."},
			["muppet"] = {"tMuppet", "User is already muppetized", "User: %s with ip: %s is now unmuppetized."},
			["connects"] = {"tConnects", "User's connections are already blocked", "User: %s with ip: %s can make new connections again."},
			["pm"] = {"tPm", "User's private messages are already blocked", "User: %s with ip: %s can send personal messages again."},
			["mc"] = {"tMc", "User already can't receive mainchat messages", "User: %s with ip: %s will now receive mainchat messages again."},
		}

		local vip = false
		local vnick = false

		if btypes[btype] then
			local tname = btypes[btype][1]
			local found = false
			if tBlocks[tname][value] then	--// ip
				vnick = tBlocks[tname][value]
				vip = value
				tBlocks[tname][value] = nil
				found = true
			else
				vnick = value
				for tip,vnick in pairs(tBlocks[tname]) do
					if value:lower() == vnick:lower() then
						vip = tip
						tBlocks[tname][tip] = nil
						found = true
					end
				end
			end
			if found then
				tFuncs:SaveSettings("blocks")
				return true, Lang(ssid, 0, btypes[btype][3], vnick, vip)
			else
				return false, Lang(ssid, 0, "User or IP wasn't found.")
			end
		else
			return false, Lang(ssid, 0, "Error: Unknown blocktype: %s", btype)
		end
	else
		return false, Lang(ssid, 0, "AddBlock Error: invalid btype or value")
	end
end,
}

tBan = {

["AddBan"] = function(self, slist, snick, sip, stype, sreason, stime, bysid, real, cat)
	local bysid = bysid or "AAAA"
	local bynick = tUser:GetNick(bysid)
	local stime = stime or tSecurity.General.DefaultTimeban
	local snick = snick or ""
	local var = ""
	local success, msg = true, ""
--~ 	local pnicks ={}	-- protected nicks

	--// Check for protected users:
	local protected, why = tUser:IsProtected(nil, snick, sip)
	if protected then return false, why end

	--// Get multiple nicks from IP
	local nick, tnicks, nicks = tIp:GetNick(sip)
	if not nick then nick = snick  tnicks = {snick}  	nicks = snick end
	
	if slist == "ipban" then	--// use snick for rangeban handle 'startip-endip'
		var = sip
		if bysid and tClient[bysid] and tClient[bysid].sIP == sip then
			return false, "You can't ban your own IP."
		end
	elseif type(snick) == "string" then
		var = snick:lower()
	end
	
	if not tBans[slist] then
		return false, "[AddBan] Error: slist '"..slist.."' doesn't exist."
	end

	local bantable = tBans[slist][var] or {}

	--// slist = nickban / ipban / rangeban / warn
	if var and var ~= "" then
		bantable.sNick = nicks
		bantable.sIP = sip or ""
		bantable.sType = stype
		bantable.sReason = sreason
		bantable.iDate = os.time()
		bantable.sBy = bynick
		bantable.sReal = real		--// Real reason, sreason is used to display to user as kick/banmessage, but real reason will show in baninfo for operators

		if stime > 0 then
			bantable.iExpires = os.time()+stime
		elseif stime == 0 then
			bantable.iExpires = os.time()+tSecurity.General.MaxBan
		else
			bantable.iExpires = stime
		end

		if stype == "kick" or stype == "warn" then
			bantable.iCount = (bantable.iCount or 0) + 1
		end

		if slist == "rangeban" then
			local _, _, ip1, ip2 = sip:find("(%d+)-(%d+)")
			if ip1 and ip2 then
				bantable.sStart = tonumber(ip1)
				bantable.sEnd= tonumber(ip2)
			else
				return false, Lang(bysid, 0, "[AddBan] Invalid banrange specified.")
			end
		end

		tBans[slist][var] = bantable

		local reason = real or sreason
		if not cat then	-- cat is when the warn comes from Deflood or pattern
			tFuncs:Debug("[AddBan] "..bynick.." added "..stype.." '"..var.."' "..(bantable.iExpires>0 and "until: "..ToDate(bantable.iExpires) or "permanently")..(reason and reason~="" and " because: "..reason or "")..".", "FI", 0)
		else
			tFuncs:Debug("[AddBan] "..bynick.." added "..stype.." '"..var.."' "..(bantable.iExpires>0 and "until: "..ToDate(bantable.iExpires) or "permanently")..(reason and reason~="" and " because: "..reason or "")..".", "FI", 2)
		end

		return true, "Added "..stype.." '"..var.."' "..(bantable.iExpires>0 and "until: "..ToDate(bantable.iExpires) or "permanently")..(reason and reason~="" and " because: "..reason or "")
	else
		return false, "[AddBan] Error: no var to create the bantable"
	end
end,

["DropUser"] = function(self, sid, msg, bysid, cat)
	local bysid = bysid or "AAAA"
	local nick = tUser:GetNick(sid) or "<nick unknown>"
	local ip = tUser:GetIp(sid)

	--// Protected check
	if cat then	--// Manually dropping a protected user (with hubcommand) is still allowed!
		local protected, why = tUser:IsProtected(nil, nick, ip)
		if protected then return false, why end
	end
	
	--// Permissions check
	if not tUser:IsHigherUser(bysid,sid) then
		return false, "This user is too powerful for you!"
	end
	
	--// Textual
	local msg = msg or ""
	local msgadd = (msg~="" and "You are disconnected because: " or "")
	local replyadd = (msg~="" and " because: " or "")

	--// Execute
	local res, err = tUser:Disconnect(sid, msgadd..msg)
	if not res then return false, err end

	--// Return result
	if cat then	--// Was started by DeFlood
		return true, "User disconnected."
	else			--// Was started by hubcommand
		tFuncs:Debug("[DropUser] "..tUser:GetShortData(sid).." was dropped by <"..tUser:GetNick(bysid) ..">"..replyadd ..msg)
		return true, "You dropped user: "..tUser:GetShortData(sid)
	end
end,

["RediUser"] = function(self, sid, msg, bysid, address, cat)
	local bysid = bysid or "AAAA"
	local nick = tUser:GetNick(sid) or "<nick unknown>"
	local ip = tUser:GetIp(sid)
	local msg = msg or ""

	--// Protected check
	local protected, why = tUser:IsProtected(nil, nick, ip)
	if protected then return false, why end
	
	--// Permissions check
	if not tUser:IsHigherUser(bysid,sid) then
		return false, "This user is too powerful for you!"
	end
	
	--// Disconnect the user
	local res, err = tUser:Redirect(sid, msg, address)	
	if not res then return false, err end

	--// Return result
	if cat then	--// Was started by DeFlood
		return true, "User redirected to "..tHubSettings.tRedirects.sFlood
	else			--// Was started by hubcommand
		tFuncs:Debug("[Redirect] "..tUser:GetShortData(sid).." was redirected to: "..address.." by <"..tUser:GetNick(bysid) ..">"..(msg~= "" and " because: "..msg or ""))
		return true, "You redirected user: "..tUser:GetShortData(sid).." to address: "..address..(msg~= "" and " because: "..msg or "")
	end
end,

["NotifyUser"] = function(self, sid, msg, bysid, cat)
	if not msg or msg == "" then return false, "[NotifyUser] There's no message." end
	local bysid = bysid or "AAAA"
	local nick = tUser:GetNick(sid) or "<nick unknown>"
	local ip = tUser:GetIp(sid)

	--// Permissions check
	if not tUser:IsHigherUser(bysid,sid) then
		return false, "This user is too powerful for you!"
	end

	--// Execute
	tSend:ChatToUser(msg, "AAAA", sid)

	--// Return result
	if cat then		--// Was started by DeFlood
		return true, "User notified."
	else				--// Was started by hubcommand	
		tFuncs:Debug("[NotifyUser] "..tUser:GetShortData(sid).." was notified because: "..msg, "FI" , 0)
		return true, "You notified user: "..tUser:GetShortData(sid).." because: "..msg
	end
end,

["NotifyOpchat"] = function(self, sid, msg, bysid, cat)
	if not msg or msg == "" then return false, "[NotifyOpchat] There's no message." end

	--// Execute
	tSend:ChatRoom(msg, "AAAA", "BBBB")
	
	--// Return result
	if cat then		--// Was started by DeFlood
		return true, "Opchat notified."
	else				--// Was started by hubcommand	
		tFuncs:Debug("[NotifyOpchat] Opchat was notified about user: "..tUser:GetShortData(sid).."because: "..msg, "FI" , 0)
		return true, "You notified opchat about user: "..tUser:GetShortData(sid).."because: "..msg
	end
	
	
	return true, "Opchat notified."
end,

["UnWarnUser"] = function(self, nick, bysid)
	local bysid = bysid or "AAAA"
	local lnick = nick:lower()

	--// Execution
	if tBans.warn[lnick] then
		local warns = (tBans.warn[lnick].iCount or 0)
		tBans.warn[lnick] = nil
		if warns > 0 then
			tFuncs:SaveSettings("bans")
			tFuncs:Debug("[UnWarn] "..tClient[bysid].sNick.." has reset the warnings for nickname <"..nick..">.")
			return true, "You've reset the warnings of <"..nick .."> from "..warns .." to 0."
		else
			return true, "User <"..nick .."> hasn't been warned."
		end
	else
		return true, "User <"..nick .."> hasn't been warned."
	end
end,

["WarnUser"] = function(self, sid, msg, bysid, cat)
	local bysid = bysid or "AAAA"
	local nick = tUser:GetNick(sid) or "<nick unknown>"
	local lnick = nick:lower()
	local ip = tUser:GetIp(sid)

	--// Protected check
	local protected, why = tUser:IsProtected(sid)
	if protected then return false, why end

	--// Permissions check
	if not tUser:IsHigherUser(bysid,sid) then
		return false, "This user is too powerful for you!"
	end

	--// Textual
	local msg = msg or ""
	local msgadd = (msg~="" and "You are being warned because: " or "")
	local replyadd = (msg~="" and " because: " or "")

	--// Execution
	local result, retmsg = tBan:AddBan("warn", nick, ip, "warn", msg, 0, bysid, nil, cat)
	
	--// Return result
	if result then
		if tBans.warn[lnick].iCount < tSecurity.General.WarnMax then
			--// Normal warning
			tSend:ChatToUser(msgadd..msg, "AAAA", sid)
			if tSecurity.General.bShowWarnings then tSend:ChatToAll("User <"..nick.."> is warned"..replyadd..msg) end
			tFuncs:SaveSettings("bans")
			if cat then
				return true, "User was warned, warnings: "..tBans.warn[lnick].iCount
			else
				tFuncs:Debug("[Warn] <"..nick .."> was warned by <"..tUser:GetNick(bysid) ..">"..replyadd ..msg)
				return true, "You've warned <"..nick .."> "..tBans.warn[lnick].iCount .." time(s)"..replyadd..msg
			end
		else
			--// Too many warnings, it's a kick
			local res, retmsg = tBan:KickUser(sid, "You've had too many warnings: "..msg, bysid, stype, cat)

			if res then
				if tSecurity.General.bShowWarnings then tSend:ChatToAll("User <"..nick.."> has been warned enough and is now kicked"..replyadd..msg) end
				tBans.warn[lnick] = nil    --// reset the amount of warnings
				tFuncs:SaveSettings("bans")
				if cat then
					return true, "Maximum warnings reached, kicked: "..retmsg
				else
					tFuncs:Debug("[Warn] Maximum warnings for <"..nick.."> reached. , user was kicked"..replyadd..msg..". "..retmsg, "FI", 0)
					return true, "Maximum warnings for <"..nick.."> reached. , user was kicked"..replyadd ..msg..". "..retmsg
				end
			else
				tFuncs:SaveSettings("bans")
				if cat then
					return false, "Maximum warnings reached, but kick failed: "..retmsg
				else
					tFuncs:Debug("[Warn] Maximum warnings for <"..nick.."> reached, but kick failed: "..retmsg, "FI", 0)
					return false, "Maximum warnings for <"..nick.."> reached, but kick failed: "..retmsg
				end
			end
		end
	else
		return false, retmsg
	end
end,

["KickUser"] = function(self, sid, msg, bysid, cat)
	local bysid = bysid or "AAAA"
	local nick = tUser:GetNick(sid) or "<nick unknown>"
	local ip = tUser:GetIp(sid)
	local msg = msg or ""

	--// Protected check
	local protected, why = tUser:IsProtected(sid)
	if protected then return false, why end

	--// Check for regged users using this IP:
	local rnick, rdata = tIp:GetHighestReg(ip)
	if rnick and tClient[bysid].iProfile <= rdata.iLevel and tClient[bysid].iProfile ~= 10 then
		return false, "This IP is in use by <"..rnick .."> and has an equal or higher profile level. You can't kick this user.", true
	end

	--~ 						tBan:AddBan(slist, snick, sip, stype, sreason, stime, bysid, real, cat)
	local result, retmsg = tBan:AddBan("ipban", nick, ip, "kick", msg, tSecurity.General.KickTime, bysid, nil, cat)

	if result then
		tFuncs:SaveSettings("bans")

		local dnicks = {}
		local banmsg, banmsg2, banned = "", "", false
		if tBans.ipban[ip] and (tBans.ipban[ip].iCount or 0) >= tSecurity.General.KickMax then
			local res, retmsg, dnicks, regged = tBan:BanIp(ip, msg, nil, bysid, "Kick")
			banmsg = "Maximum kicks for: "..nick.." is reached: "
			if res then
				banned = true
				tBans.ipban[ip].iCount = 0	-- reset the amount of kicks
				tFuncs:SaveSettings("bans")
				banmsg = banmsg.." IP banned."
				banmsg2 = "Maximum kicks for this user is reached. "..retmsg
				if not cat then
					tFuncs:Debug("Maximum kicks for this user is reached. "..retmsg)
				end
			else
				if not cat then
					tFuncs:Debug("Maximum kicks for this user is reached, but IP ban failed: "..retmsg)
				end
				banmsg = banmsg.." IP ban failed: "..retmsg
				banmsg2 = "Maximum kicks for this user is reached, but kick failed: "..retmsg
			end
		end
		local dnicks = dnicks or {}

		local res, err = tUser:Disconnect(sid, "You are kicked by <"..tUser:GetNick(bysid) .."> because: "..msg)
		if res then table.insert(dnicks, nick) end

		if cat then
			return true, "Kicked. "..banmsg, (dnicks or {}), (regged or {})
		else
--~ 			tFuncs:Debug("[Kick] <"..tUser:GetNick(bysid) .."> kicked user <"..nick .."> with IP: "..ip..(msg~=""  and " because: "..msg or ""))
			return true, "You kicked user: "..nick.." with IP: "..ip..(msg~=""  and " because: "..msg or "")..(banmsg2~="" and "\n"..banmsg2 or "")
		end
	else
		if cat then
			return false, retmsg
		else
			return false, "You can't kick user: "..nick.." with IP: "..ip.." because: "..retmsg
		end
	end
end,

["KillUser"] = function(self, sid, msg, stime, bysid, cat)
	local stime = stime or tSecurity.General.DefaultTimeban
	if stime > tSecurity.General.MaxBan then stime = tSecurity.General.MaxBan end
	local ttime = tFuncs:Duration(stime)
	local bysid = bysid or "AAAA"
	local nick = tUser:GetNick(sid) or "<nick unknown>"
	local result = ""
	local msg = msg or ""
	local ip = (tClient[sid] and tClient[sid].sIP)

	if not ip then return false, "User not found." end	--// this means kill only works for online users

	local result1, retmsg1, dnicks1, regged1 = tBan:BanIp(ip, msg, stime, bysid, "Kill")
	if not result1 then return false, retmsg1 end
	result = result..retmsg1.." / "

	local result2, retmsg2, dnicks2, regged2 = tBan:BanNick(nick, msg, stime, bysid, "Kill")
	result = result..retmsg2
	
	local dnicks = {}
	for _,data in ipairs(dnicks1 or {}) do table.insert(dnicks, data) end
	for _,data in ipairs(dnicks2 or {}) do table.insert(dnicks, data) end
	local regged = {}
	for _,data in ipairs(regged1 or {}) do table.insert(regged, data) end
	for _,data in ipairs(regged2 or {}) do table.insert(regged, data) end
	
	if not result2 then --partial failure
		if cat then
			return false, result, dnicks
		else
			tFuncs:Debug("[Kill] <"..tUser:GetNick(bysid) .."> tried killing user: "..nick .." with IP: "..ip ..": "..result, "FI" , 0) 
			return false, "Killing failed, nick: "..nick.." was banned. IP: "..ip.." wasn't banned: "..result2, dnicks
		end
	else
		if cat then
			return true,  "Killed IP: "..ip.." and nick: "..nick..(stime>0 and " for "..ttime or " permanently")..(#dnicks > 0 and  ". Disconnected: <"..table.concat(dnicks, ">, <")..">" or ""), dnicks
		else
--~ 			tFuncs:Debug("[Kill] <"..tUser:GetNick(bysid) .."> killed nick: "..nick .." and IP: "..ip..(stime>0 and " for "..ttime or " permanently")..(msg~=""  and " because: "..msg or ""), "FI", 0)
			return true, "You killed the nick: "..nick.." and IP: "..ip..(stime>0 and ttime and " for "..ttime or " permanently")..(msg~=""  and " because: "..msg or "")..(#dnicks > 0 and  "\nDisconnected: <"..table.concat(dnicks, ">, <")..">" or "")..(#regged>0 and "\nThe following killed user(s) are still registered, you may want to delreg: <"..table.concat(regged, "> , <")..">" or "")
		end
	end
end,


["BanIp"] = function(self, ip, msg, stime, bysid, cat)
-- desired behavior: stime -1 = permban / stime 0 = maxban / stime >0 = timeban until os.time()+stime
-- maxban is taken care of in AddBan, not here
	local stime = stime or tSecurity.General.DefaultTimeban
	if stime > tSecurity.General.MaxBan then stime = tSecurity.General.MaxBan end
	local ttime = tFuncs:Duration(stime)
	local bysid = bysid or "AAAA"
	local by = tUser:GetNick(bysid)
	local nick = "<nick unknown>"
	local msg = msg or ""

	--// Check for regged users using this IP:
	local rnick, rdata = tIp:GetHighestReg(ip)
	if rnick and tClient[bysid].iProfile <= rdata.iLevel and tClient[bysid].iProfile ~= 10 then
		return false, "This IP is in use by <"..rnick .."> and has an equal or higher profile level. IP can't banned."
	end
--~ 	if rnick then nick = rnick end

	--// tBan:AddBan(slist, snick, sip, stype, sreason, stime, bysid, real, cat)
	local result, retmsg = tBan:AddBan("ipban", nick, ip, "ipban", msg, stime, bysid, nil, cat)

	if result then
		tFuncs:SaveSettings("bans")

		--// Disconnect all users with that IP:
		local sids = tIp:GetSids(ip)
		local dnicks = {}
		if sids then
			for _, vsid in ipairs(sids) do
				local res, err = tUser:Disconnect(vsid, "You are banned by <"..by ..">"..(msg~= "" and " because: "..msg or ""))
				if res then 
					table.insert(dnicks, (tClient[vsid].sNick or "<nick unknown>"))
				end
			end
			
		end
		
		--// Report regged users with banned IP:
		local regs, regged = tIp:GetRegs(ip)
		if regged then tFuncs:Debug("The user(s) with banned IP: "..ip.." are still registered, you may want to delreg: <"..table.concat(regged, "> , <")..">") end

		if cat then
			return true, "Banned IP: "..ip..(stime>0 and " for "..ttime or " permanently")..(#dnicks > 0 and  ". Disconnected: <"..table.concat(dnicks, ">, <")..">" or ""), dnicks, regged
		else
			if #dnicks>0 then tFuncs:Debug("Disconnected user(s) with banned IP: "..ip ..": <"..table.concat(dnicks, ">, <")..">" or "") end
			return true, "You banned the IP: "..ip..(stime>0 and " for "..ttime or " permanently")..(msg~=""  and " because: "..msg or "")..(#dnicks > 0 and  "\nDisconnected: <"..table.concat(dnicks, ">, <")..">" or "")..(info3 and "\n"..info3 or "")
		end
	else
		return false, retmsg, dnicks
	end
end,

["BanNick"] = function(self, nick, msg, stime, bysid, cat)
-- desired behavior: stime -1 = permban / stime 0 = maxban / stime >0 = timeban until os.time()+stime
-- maxban is taken care of in AddBan, not here
	local stime = stime or tSecurity.General.DefaultTimeban
	if stime > tSecurity.General.MaxBan then stime = tSecurity.General.MaxBan end
	local ttime = tFuncs:Duration(stime)
	local bysid = bysid or "AAAA"
	local by = tUser:GetNick(bysid)
	local profile = false
	local vsid = tNick:GetSid(nick)

	--// Check for protected users:
	local protected, why = tUser:IsProtected(vsid, nick)
	if protected then return false, "Ban nick failed: "..why end

	--// Check if nick is regged with higher profile
	local reg, tnick = tNick:GetReg(nick)
	if reg and (reg.iLevel >= tClient[bysid].iProfile) and tClient[bysid].iProfile ~= 10 then
		return false, "Ban nick failed: This user is too powerful for you!"
	elseif reg and tnick then
		nick = tnick
		profile = reg.sProfile
	end

	--// tBan:AddBan(slist, snick, sip, stype, sreason, stime, bysid)
	local vip = tUser:GetIp(vsid) or false
	local result, retmsg = tBan:AddBan("nickban", nick, vip, "nickban", msg, stime, bysid, nil, cat)

	if result then
		tFuncs:SaveSettings("bans")

		--// Disconnect the user(s):        --// TODO: make this work for banned nicks containing wildcards
		local dnicks = {}
		if vsid then
			local res, err = tUser:Disconnect(vsid, "You are banned by <"..by .."> because: "..msg)
			if res then 
				table.insert(dnicks, (tClient[vsid].sNick or "<nick unknown>"))
			end
		end

		--// Is there an account to remove?
		local regged = {}
		if profile then
			regged = {nick}
			tFuncs:Debug("The user with banned nick: "..nick.." and profile: "..profile.." is still registered, you may want to delreg this user.")
		end
		
		if cat then
			return true, "Banned nick: "..nick..(stime>0 and " for "..ttime or " permanently")..(#dnicks > 0 and  ". Disconnected: <"..table.concat(dnicks, ">, <")..">" or ""), dnicks, regged
		else
			if #dnicks>0 then tFuncs:Debug("Disconnected user(s) with banned nick: "..nick ..": <"..table.concat(dnicks, ">, <")..">" or "") end
			return true, "You banned the nick: "..nick..(stime>0 and " for "..ttime or " permanently")..(msg~=""  and " because: "..msg or "")..(#dnicks > 0 and  "\nDisconnected: <"..table.concat(dnicks, ">, <")..">" or "")..(info3 and "\n"..info3 or "")
		end		
	else
		return false, "Ban nick failed: "..retmsg
	end
end,

["BanRange"] = function(self, sip, eip, msg, stime, bysid, reply)	--//stime in hours	(not avaible from floodactions, no cat needed)
	local stime = stime or tSecurity.General.DefaultTimeban
	if stime > tSecurity.General.MaxBan then stime = tSecurity.General.MaxBan end
	local ttime = tFuncs:Duration(stime)
	local bysid = bysid or "AAAA"
	local by = tUser:GetNick(bysid)
	local ip1 = tFuncs:IP2DEC(sip)
	local ip2 = tFuncs:IP2DEC(eip)
	local own = tFuncs:IP2DEC(tUser:GetIp(bysid))

	--// Check for valid range
	if ip1 == -1 or ip2 == -1 or ip1 > ip2 then
		return false, "Invalid IP range, use a range between 0.0.0.0-255.255.255.255"
	end

	--// Check if you're not rangebanning your own IP
	if own >= ip1 and own <= ip2 then
		return false, "You can't ban a range that includes your own IP."
	end

	--// Check for regged users using this IP:
	for vnick,data in pairs(tAccounts) do
		if data.sLastIP and tProfiles[data.sProfile] and tProfiles[data.sProfile].iLevel >= tClient[bysid].iProfile and tClient[bysid].iProfile ~= 10 then
			local ip = tFuncs:IP2DEC(data.sLastIP) 
			if ip >= ip1 and ip <= ip2 then
				return false, "You can't ban this range, it includes a user with equal or higher profile: "..data.sNick.." with IP: "..data.sLastIP
			end
		end
	end

	--// Check to see if no protected users are within the rangeban
	for vsid,data in pairs(tClient) do
		local ip = tFuncs:IP2DEC(data.sIP) 
		if ip >= ip1 and ip <= ip2 then
			if tUser:IsProtected(vsid) then
				return false, "You can't ban this range, it includes a protected user: "..data.sNick.." with IP: "..data.sIP
			elseif not tUser:IsHigher(bysid, vsid) then
				return false, Lang(bysid, 0, "This user is too powerful for you!")
			end
		end
	end

	local rip = ip1.."-"..ip2
	local result, retmsg = tBan:AddBan("rangeban", sip .."-"..eip, rip, "rangeban", msg, stime, bysid)

	--// Disconnect all users within that range:
	if result then
		local bannedusers = ""
		for vsid,data in pairs(tClient) do
			local vip = tFuncs:IP2DEC(data.sIP)
			if ip1 <= vip and ip2 >= vip then
				bannedusers = bannedusers.."\n\t"..(tUser:GetNick(vsid) or "nick unknown").." with IP: "..data.sIP
				tUser:Disconnect(vsid, "You are banned by <"..by.."> because: "..msg)
			end
		end
		if bannedusers ~= "" then bannedusers = "\nUser(s) within this range disconnected: "..bannedusers end

		local retmsg = "<"..by .."> rangebanned the IP's: "..sip.."-"..eip ..(stime>0 and " for "..ttime or " permanently").." because: "..(msg~=""  and " because: "..msg or "")..bannedusers
		tFuncs:Debug("[BanRange] "..retmsg, "FI")
		tFuncs:SaveSettings("bans")
		return true, "You banned the IP range: "..sip .."-"..eip..(stime>0 and " for "..ttime or " permanently")..(msg~=""  and " because: "..msg or "")..bannedusers
	else
		return false, "Rangeban "..sip .."-"..eip .." wasn't added because: "..retmsg
	end
end,

["Unban"] = function(self, original, reason, bysid, nosave, cat)
	local var = original:lower()
	local reason = reason or ""
	local by = tUser:GetNick((bysid or "AAAA")) or "Hub"
	local retmsg = ""
	local found = false
	local removed = false
	local cat = cat or false

	for _,slist in ipairs({"ipban", "nickban", "rangeban", "autoban"}) do
		found = false

		if tBans[slist][var] then
			--// todo: add a nice overview of this ban, to display the content before removing.
			found = var
		elseif slist == "nickban" then	--// if this line is changed to a single else, it will unban everything that matches and not just nickbans
			for val, data in pairs(tBans[slist]) do
				if string.find(string.lower(data.sNick), var, 1, true) then
					local tNick = tFuncs:Split(string.lower(data.sNick), ", ")
					for i, v in pairs (tNick) do
						if v == var then
							found = val
							retmsg = retmsg.."Indirect match: '"..original.."' found in '".. data.sNick.."'"
						end
					end
				end
			end
		end
		if found then
			tBans[slist][found] = nil
			if retmsg ~= "" then retmsg = retmsg.."\n\t" end
			retmsg = retmsg.."The ban: "..found.." was removed from "..slist.."s."
			if not cat then
				tFuncs:Debug("[Unban] <"..by.."> removed ban: "..found..(reason~= "" and " because: "..reason or ""), "FI", 0)
			end
			removed = true
		end
	end

	if removed then
		if not nosave then
			tFuncs:SaveSettings("bans")
		end
		return true, retmsg
	else
		return false, "The value: "..original.." wasn't found in any banlist."
	end
end,

}
