--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tDeflood = {

["tText"] = {
	--// NOTE: only security categories mentioned here will 

	["XX"] = { ["Name"] = "unspecified", ["FloodWarn"] = "You're flooding, ", ["FloodBan"] = "Don't flood."},
	["MA"] = { ["Name"] = "BINF", ["FloodWarn"] = "Sending too much clientinfo (BINF), ", ["FloodBan"] = "Don't flood with clientinfo (BINF)"},
	["MN"] = { ["Name"] = "MyINFO", ["FloodWarn"] = "Sending too much clientinfo (MyINFO), ", ["FloodBan"] = "Don't flood with clientinfo (MyINFO)"},
	["RE"] = { ["Name"] = "report", ["FloodWarn"] = "Don't abuse the report command, ", ["FloodBan"] = "Don't abuse +report"},
	["HA"] = { ["Name"] = "reconnect", ["FloodWarn"] = "Don't reconnect so fast, ", ["FloodBan"] = "Hammering."},

	["SA"] = { ["Name"] = "search", ["FloodWarn"] = "Don't search too fast, ", ["FloodBan"] = "Flooding with searches."},
	["SP"] = { ["Name"] = "search (passive)", ["FloodWarn"] = "Don't search too fast, ", ["FloodBan"] = "Flooding with passive searches."},
	["SR"] = { ["Name"] = "searchreply", ["FloodWarn"] = "Too many searchreplies, ", ["FloodBan"] = "Flooding with searchreplies."},
	["PR"] = { ["Name"] = "searchreply", ["FloodWarn"] = "Too many searchreplies received, ", ["FloodBan"] = ""},
	["SG"] = { ["Name"] = "nickname", ["FloodWarn"] = "Ignoring more searchresults for this user, ", ["FloodBan"] = ""},

	["MC"] = { ["Name"] = "mainchat", ["FloodWarn"] = "Don't flood mainchat, ", ["FloodBan"] = "Flooding mainchat."},
	["HC"] = { ["Name"] = "hubcommand", ["FloodWarn"] = "Don't flood hubcommands, ", ["FloodBan"] = "Flooding hubcommands."},
	["PM"] = { ["Name"] = "PM", ["FloodWarn"] = "Don't flood PM, ", ["FloodBan"] = "Flooding PM."},
	["CT"] = { ["Name"] = "ConnectToMe", ["FloodWarn"] = "Sending too many connection requests, ", ["FloodBan"] = "Flooding connection requests."},
	["RC"] = { ["Name"] = "RevConnectToMe", ["FloodWarn"] = "Sending too many Passive connection requests", ["FloodBan"] = "Flooding reverse connection requests."},
	["MY"] = { ["Name"] = "MyINFO", ["FloodWarn"] = "Sending too many MyInfo's", ["FloodBan"] = "Flooding MyINFOs."},
	["NL"] = { ["Name"] = "Nicklists", ["FloodWarn"] = "Requesting nicklists too often, ", ["FloodBan"] = "Flooding nicklist requests."},
	["NI"] = { ["Name"] = "nickname", ["FloodWarn"] = "Sending Nick Validation too often, ", ["FloodBan"] = "Flooding Nick Validations."},
},

["BandWidth"] = function(self, fsid, stype, data, tosid)
	local ctime = os.time()
	local rule = tSecurity.BW

	if not tUser:CanOverride(fsid, "securitybw") then
		--// Create new stype table if it doesn't exist
		if not tFlood[fsid] then tFlood[fsid] = {} end
		if tFlood[fsid]["BW"] then
			user = tFlood[fsid].BW
		else
			user = {
				["iBytes1"] = 0,
				["iBytes2"] = 0,
				["iUnTime1"] = ctime + (rule["iTime1"] or 0),
				["iUnTime2"] = ctime + (rule["iTime2"] or 0),
			}
		end
	end
end,

["Main"] = function(self, fsid, stype, data, tosid)
	local ctime = os.time()
	local res, options, result, actionok, actionresult = false, false, false, false, false
	local stype = tDeflood.tText[stype] and stype or "XX"
	local rule = tSecurity[stype]
	local text = tDeflood.tText[stype]
	local user = {}

	if not tSecurity[stype] then return true, data end

	if not tUser:CanOverride(fsid, "security"..stype:lower()) then
		--// Create new stype table if it doesn't exist
		if not tFlood[fsid] then tFlood[fsid] = {} end
		if tFlood[fsid][stype] then
			user = tFlood[fsid][stype]
		else
			user = {
				["iUnTime1"] = ctime + (rule["iTime1"] or 0),
				["iUnTime2"] = ctime + (rule["iTime2"] or 0),
				["iCounter1"] = 0,
				["iCounter2"] = 0,
				["bHad1"] = false,
				["bHad2"] = false,
			}
		end

		--Frequency check:
		for a = 2,1,-1 do										--// First check the second (most severe) floodfrequency, if that's ok, see if the first exceeds a maximum.
			if rule["iTime"..a] and (ctime < user["iUnTime"..a]) then		--// new command within timeframe detected
				user["iCounter"..a] = user["iCounter"..a] + 1
				if user["iCounter"..a] > rule["iMaxCount"..a] and ctime <= user["iUnTime"..a] then
					--// Show debug
					if not rule["sAction"..a] == "Ignore" then
						tFuncs:Debug("[Security] ["..stype.."] <"..(tUser:GetNick(fsid) or tClient[fsid].sIP).."> floods: "..(user["iCounter"..a]-rule["iMaxCount"..a]).." time(s). Action: "..rule["sAction"..a])
					end

					user["bHad"..a] = ctime
					local msg
					local msgshort = false
					if rule["sAction"..a] == "Notify" or rule["sAction"..a] == "Warn" then
						msg = text.FloodWarn.."maximum of "..rule["iMaxCount"..a].." each "..rule["iTime"..a].." second(s)."
						if stype == "PM" then
							tSend:PmToUser(msg, "AAAA", fsid, tosid)
						elseif stype == "SE" or stype == "SP" then
							tSend:SearchResult(msg, "AAAA", fsid)
						end
					elseif rule["sAction"..a] == "NotifyOpchat" or rule["sAction"..a] == "LockMain" then
						msg = "[Security] ["..stype.."] ["..a.."] ["..rule["sAction"..a].."] "..tUser:GetShortData(fsid).."/ Reason: "..text.FloodWarn.."maximum of "..rule["iMaxCount"..a].." each "..rule["iTime"..a].." second(s)."
						msgshort = text.FloodWarn.."maximum of "..rule["iMaxCount"..a].." each "..rule["iTime"..a].." second(s)."
					else
						msg = text.FloodBan	-- // FloodBan is empty for some types- copy from warn or make up new ones??
					end

					--// PERFORM ACTION   res = allow msg or not
					res, result, actionok, actionresult = tDeflood.Action[rule["sAction"..a]](fsid, msg, data, stype)
					--// Output to feed
					if actionresult then
						local debuglevels = { ["Disabled"] = 2, ["Ignore"] = 2, ["Notify"] = 1, ["NotifyOpchat"] = 1, ["Replace"] = 1, ["ReplaceLine"] = 1}
						local debuglevel = debuglevels[rule["sAction"..a]] or 0

						if not actionok then
							tFuncs:Debug("[Security] ["..stype.."] ["..a.."] ["..rule["sAction"..a].."] "..tUser:GetShortData(fsid).."/ Reason: "..(msgshort or msg).." / FAILED: "..tostring(actionresult), "FI", debuglevel)
						else
							tFuncs:Debug("[Security] ["..stype.."] ["..a.."] ["..rule["sAction"..a].."] "..tUser:GetShortData(fsid).."/ Reason: "..(msgshort or msg).." / OK: "..tostring(actionresult), "FI", debuglevel)
						end
					end

					--// Output to proper part
					if not res then
						tFlood[fsid][stype] = user       --// always set the modified user table back to global before any return !!!!!
						return false, result
					end
				end
			elseif rule["iTime"..a] then     --// waited long enough, counter is set to 1 and untime reset
				user["iCounter"..a] = 1
				user["iUnTime"..a] = ctime + rule["iTime"..a]
			end
		end

		tFlood[fsid][stype] = user		--// always set the modified user table back to global before any return !!!!!

		if tosid and not tClient[tosid] then return false, "User not found." end
	end

	-- Specific stype checks/blockchecks:
--~	if stype and tDeflood.Type[stype] and not tUser:CanOverride(fsid, "check"..stype) then
	if stype and tDeflood.Type[stype] then
		res, result, moddata = tDeflood.Type[stype](fsid, data, tosid)
		if not res then
			return false, result
		elseif res and moddata then
			data = moddata
		end
	end

	-- Length checks:
	if not tUser:CanOverride(fsid, "security"..stype:lower()) then
		if rule.MaxLength and #data > rule.MaxLength then
			local debuglevels = { ["Disabled"] = 2, ["Ignore"] = 2, ["Notify"] = 1, ["NotifyOpchat"] = 1, ["Replace"] = 1, ["ReplaceLine"] = 1}
			local debuglevel = debuglevels[rule.LengthAction] or 0

			res, result = tDeflood.Action[rule.LengthAction](fsid, "Your "..text.Name.." is too long, maximum length = "..rule.MaxLength.." characters.", data, "length check")
			tFuncs:Debug("[Security] ["..stype.."] [MaxLength] ["..rule.LengthAction.."] "..tUser:GetShortData(fsid).."/ Reason: The "..text.Name.." is too long, maximum length = "..rule.MaxLength.." characters. / "..tostring(result), "FI", (debuglevel or 0))
			if not res then return false, result end
		end
		if rule.MinLength and #data < rule.MinLength then
			local debuglevels = { ["Disabled"] = 2, ["Ignore"] = 2, ["Notify"] = 1, ["NotifyOpchat"] = 1, ["Replace"] = 1, ["ReplaceLine"] = 1}
			local debuglevel = debuglevels[rule.LengthAction] or 0

			res, result = tDeflood.Action[rule.LengthAction](fsid, "Your "..text.Name.." is too short, minimum length = "..rule.MinLength.." characters.", data, "length check")
			tFuncs:Debug("[Security] ["..stype.."] [MinLength] ["..rule.LengthAction.."] "..tUser:GetShortData(fsid).."/ Reason: The "..text.Name.." is too short, minimum length = "..rule.MinLength.." characters. / "..tostring(result), "FI", (debuglevel or 0))
			if not res then return false, result end
		end
	end

	return true, data
end,

["Check"] = {
	["Buffer"] = function(fsid, data)
		if not tUser:CanOverride(fsid, "maxbuffer") then
			local max = tSecurity.General.MaxParseBuffer or 2048
			if #data > max and (not tClient[fsid].bOperator) then
				return false, "Sending too much data at once."
			end
		end
		return true
	end,
},

["Type"] = {

	["CT"] = function(fsid, data)
		if not tUser:CanDownload(fsid, true) then return false, data end		--// Note: this will block if an allowed passive user is trying to download from a blocked active user..
		return true, data
	end,
	["RC"] = function(fsid, data)
		if not tUser:CanDownload(fsid, true) then return false, data end
		return true, data
	end,
	["SA"] = function(fsid, data)
		if not tUser:CanDownload(fsid, true) then return false, data end
		return true, data
	end,
	["SP"] = function(fsid, data)
		if not tUser:CanDownload(fsid, true) then return false, data end
		return true, data
	end,
	["PR"] = function(fsid, data)
		if not tUser:CanDownload(fsid) then return false, data end
		return true, data
	end,
	["MC"] = function(fsid, data)
		local moddata = false
		local ret = true
--~ 		tFuncs:Debug("Starting MC type check: "..socket.gettime(), "X")
		if not tUser:CanOverride(fsid, "securitymc", false) then
			local rules = tSecurity.MC
			local ip = tClient[fsid].sIP
			local nick = tClient[fsid].sNick

			--// Mainchat locked check
			if tSecurity.General.MainLocked and not tUser:IsOp(fsid) and not tUser:CanOverride(fsid, "lockmc", false)then
				tSend:ChatToUser("Mainchat is locked, you're not able to speak now.", "AAAA", fsid)
				return false, data, true     --// change the last var to 'true' if you don't want debug in feed of it
			end

			--// Multiline truncation
			local _, counter = string.gsub(data, "\n", "")
			--local allowed = false
			local ctime = socket.gettime()
			if counter > (rules.MaxLines or 4) then
				local allowed = data:match(("^("..string.rep(".-\n", (rules.MaxLines or 4))..")")) or ""
				tFuncs:Debug(tUser:GetDataLine(fsid).."[Security] [checkmc] Message contained "..#data.." characters and "..counter.." lines. Message truncated",1)
				moddata = allowed.."\n...message truncated, contained too many lines."
--~ 				tFuncs:Debug("Multiline trunc duration: "..(socket.gettime() - ctime).. " seconds.", "X")
			end

--[[
			local check = string.rep(".-\n", (rules.MaxLines or 4))
			local ctime = socket.gettime()

			local numlines = string.gmatch(data, "\n")
			if numlines and #numlines > (rules.Maxlines or 4) then 
				tFuncs:Debug("Found too many lines!..."..socket.gettime(), "X")
			end

			tFuncs:Debug("Performing find..."..ctime, "X")
--~ 			local posa, posb, allowed, toomuch = data:find(("^("..check..")(.+)$"))
			tFuncs:Debug("Find took: "..(socket.gettime() - ctime).." seconds.", "X")
			if toomuch and toomuch ~= "" then
				local counter = 0
				tFuncs:Debug("Starting gmatch loop..."..socket.gettime(), "X")

				for w in string.gmatch(data, "\n") do
					counter = counter+1
				end
				tFuncs:Debug("gmatch loop finished..."..socket.gettime(), "X")
				tFuncs:Debug(tUser:GetDataLine(fsid).."[Security] [checkmc] Message contained "..#data.." characters and "..counter.." lines. Message truncated",1)
				moddata = allowed.."\n...message truncated, contained too many lines."
			end

			tFuncs:Debug("Multiline trunc duration: "..(socket.gettime() - ctime).. " seconds.", "X")
]]--
			--// Mute check
			if tBlocks.tMute[ip] then
				tSend:ChatToUser("You are still muted.", "AAAA", fsid)
				return false, "User still muted"
			end

			--// Stealth mute check --// This will bypass patternchecks
			if tBlocks.tMuteStealth[ip] then
				tSend:ChatToUser((moddata or data), fsid, fsid)
				return false, "User still muted"
			end
		end

--~ 		tFuncs:Debug("Finished MC type check: "..socket.gettime(), "X")

		return ret, data, moddata
	end,
	["HC"] = function(fsid, data)	--// Hubcommand
		local client = tClient[fsid]
		local ret = true
		local ip = client.sIP
		local _,_,cmd = data:find("^.(%S+)")
		if cmd then
			if not tUser:CanOverride(fsid, "securityhc", false) then

				cmd = cmd:lower()
				local forbidden = {["topic"] = true, ["report"] = true, ["say"] = true, ["mc"] = true}
				
				--// Muted check
				if tBlocks.tMute[ip] and forbidden[cmd] then
					tSend:ChatToUser("You are still muted.", "AAAA", fsid)
					return false, Lang(sid, 0, "You are still muted.")
				elseif tBlocks.tMuteStealth[ip] and forbidden[cmd] then
					return false, Lang(sid, 0, "You are not allowed to use this command.")
				end

				--// Mainchat locked check
				if tSecurity.General.MainLocked and not tUser:IsOp(fsid) and forbidden[cmd] then
					tSend:ChatToUser("Mainchat is locked, you're not able to use these commands now.", "AAAA", fsid)
					return false, data, true     --// change the last var to 'true' if you don't want debug in feed of it
				end
			end
		end
		return ret, data
	end,
	["PM"] = function(fsid, data, tosid)
		if not tUser:CanOverride(fsid, "securitypm", false) then
			local client = tClient[fsid]
			local ip = client.sIP

			--// TODO: add #line check

			--// BLock check
			if (tSecurity.General.PmLocked and not tUser:IsOp(fsid) and not tUser:CanOverride(fsid, "lockpm", false)) or tBlocks.tPm[ip] then
				--\\["PmToUser"] = function(self, msg, fromsid, tosid, roomsid, me)  
				tSend:PmToUser("Private messages are blocked.", "AAAA", fsid, tosid)
				return false, "Private messages are blocked."
			end
		end
		return true, data
	end,
},

["Action"] = {

	["Replace"] = function(fsid, msg, pat, repl)
		msg = msg:gsub(pat, (repl and tFuncs:Escape(repl) or ""))
		return true, msg, true, "Match replaced."
	end,

	["ReplaceLine"] = function(fsid, msg, pat, repl)
		return true, repl, true, "Message replaced."
	end,

	["Disabled"] = function(fsid, msg, data, cat, realreason)
		return true, msg
	end,

	["Ignore"] = function(fsid, msg, data, cat, realreason)
		return false, msg, true, false
	end,

	["Notify"] = function(fsid, msg, data, cat, realreason)
		local res, retmsg = tBan:NotifyUser(fsid, msg, nil, cat)
		return false, retmsg, res, retmsg
	end,

	["NotifyOpchat"] = function(fsid, msg, data, cat, realreason)
		local res, retmsg = tBan:NotifyOpchat(fsid, (realreason or msg), nil, cat)
		return true, msg, res, retmsg
	end,

	["Disconnect"] = function(fsid, msg, data, cat, realreason)
		local res, retmsg = tBan:DropUser(fsid, msg, nil, cat)
		return false, msg, res, retmsg
	end,

	["LockMain"] = function(fsid, msg, data, cat, realreason)
		SetManager:Set("lock_main", true)
		tSend:ChatToAll("Mainchat is now locked.")
		return false, msg, true, "User caused a mainchat lock."
	end,

	["Ban"] = function(fsid, msg, data, cat, realreason)
		--~ ["BanIp"] = function(self, ip, msg, stime, bysid, cat)
		local res, retmsg = tBan:BanIp(tClient[fsid].sIP, msg, 0, nil, cat)
		return false, msg, res, retmsg
	end,

	["Redirect"] = function(fsid, msg, data, cat, realreason)
		--~ ["BanIp"] = function(self, ip, msg, stime, bysid, cat)
		local res, retmsg = tBan:RediUser(fsid, msg, nil, "flood", cat)
		return false, msg, res, retmsg
	end,

	["Kick"] = function(fsid, msg, data, cat, realreason)
		--~ ["KickUser"] = function(self, sid, msg, bysid, cat)
		local res, retmsg = tBan:KickUser(fsid, msg, nil, cat)
		return false, msg, res, retmsg
	end,

	["Mute"] = function(fsid, msg, data, cat, realreason)
		local vnick = (tClient[fsid].sNick or "<unknown>")
		local vip = tClient[fsid].sIP
		local res, retmsg = tBlock:AddBlock("AAAA", "mute", vnick, vip)
		return false, msg, res, retmsg
	end,

	["Warn"] = function(fsid, msg, data, cat, realreason)
		local res, retmsg = tBan:WarnUser(fsid, msg, nil, cat)
		return false, msg, res, retmsg
	end,

},

}
