--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

if not _NOGUI then
	tCtrls = {
		xmlResource = wx.wxXmlResource(),
		tDeadThreads = {},                -- The threads, which can be destroyed
	}

	tCtrls["xmlResource"]:InitAllHandlers()
	meta = {}
	meta.__index = function(t,k)
		if k and not rawget(t,k) and k:sub(1,4) == "wxID" or k:sub(1,3) == "ID_" then
			local NewId = tCtrls.xmlResource.GetXRCID(k)
			if not NewId then NewId = wx.wxNewId() end -- Not used now
			rawset(t,k,NewId)
			return NewId
		end
	end
	setmetatable(tCtrls,meta)

	--~ tCtrls["xmlResource"]:SetFlags(wx.wxXRC_USE_LOCALE)
end


function main()
	if not _NOGUI then

		local xrcFilename = "./res/FlexHub-Console.xrc"
		-- try to load the resource and ask for path to it if not found
		while not tCtrls["xmlResource"]:Load(_APPPATH..xrcFilename) do
			-- must unload the file before we try again
			tCtrls["xmlResource"]:Unload(xrcFilename)
			Notify("Error loading xrc resources, please choose the path to 'FlexHub-Console.xrc'.","FlexHub Multi Protocol Hub",wx.wxOK + wx.wxICON_EXCLAMATION)
			local fileDialog = wx.wxFileDialog(wx.NULL,"Open 'FlexHub-Console.xrc' resource file","","FlexHub-Console.xrc","XRC files (*.xrc)|*.xrc|All files (*)|*",wx.wxOPEN + wx.wxFILE_MUST_EXIST)
			if fileDialog:ShowModal() == wx.wxID_OK then
				xrcFilename = fileDialog:GetPath()
			else
				return -- quit
			end
		end
		tCtrls["app"] = wx.wxFrame()
		if not tCtrls["xmlResource"]:LoadFrame(tCtrls["app"],wx.NULL,"MainFrame") then
			Notify("Error loading xrc resources! Please re-download the application","FlexHub Multi Protocol Hub",wx.wxOK + wx.wxICON_EXCLAMATION)
			return -- quit program
		end

		tCtrls["accounts"] = wx.wxDialog()
		if not tCtrls["xmlResource"]:LoadDialog(tCtrls["accounts"],wx.NULL,"ID_di_accounts") then
			Notify("Error loading xrc resources! Please re-download the application","FlexHub Multi Protocol Hub",wx.wxOK + wx.wxICON_EXCLAMATION)
			return -- quit program
		end

		tCtrls["import"] = wx.wxDialog()
		if not tCtrls["xmlResource"]:LoadDialog(tCtrls["import"],wx.NULL,"ID_di_import") then
			Notify("Error loading xrc resources! Please re-download the application","FlexHub Multi Protocol Hub",wx.wxOK + wx.wxICON_EXCLAMATION)
			return -- quit program
		end

		tCtrls["xmlResource"]:InitAllHandlers()

		--// Configure statusbar
		tCtrls["statusbar"] = tFuncs:GetCtrl("ID_ss_statusbar")
		tCtrls["statusbar"]:SetFieldsCount(4)
		tCtrls["statusbar"]:SetStatusWidths({-2,-1,-1,-2})

		--// Add taskbar icon
		tCtrls["taskbar"] = wx.wxTaskBarIcon()
		bitmap = tFuncs:GetImage("flexhub")
		local icon = wx.wxIcon()
		icon:CopyFromBitmap(bitmap)
		tCtrls["app"]:SetIcon(icon)
		tCtrls["app"]:SetTitle(tHubSettings.Main.sName.." on ".._APPNAME.." ".._APPVER.." svn:".._REV.." - by FlipFlop and Daywalker")
		tCtrls["taskbar"]:SetIcon(icon, tHubSettings.Main.sName.." (FlexHub)")

		icon:delete()
		bitmap:delete()

		local toolid = tCtrls["xmlResource"].GetXRCID("ID_to_stop")
		tFuncs:GetCtrl("ID_tb_main"):ToggleTool(toolid, true)  --// Temp disabled for linux

		tFuncs:GetCtrl("ID_cb_autostart"):SetValue(tHubSettings.Console.AutoStartHub)
		tFuncs:GetCtrl("ID_cb_confirm"):SetValue(tHubSettings.Console.WarnOnClose)
		tFuncs:GetCtrl("ID_tc_hubname"):SetValue(tHubSettings.Main.sName)
		tFuncs:GetCtrl("ID_sc_maxusers"):SetValue(tHubSettings.tRules.iMaxUsersHard or 20000)
		tFuncs:GetCtrl("ID_sc_debuglevel"):SetValue(tHubSettings.Console.DebugLevel or -1)

		if tHubSettings.Console.bMainchat then
			tFuncs:GetCtrl("ID_st_mainchat"):Show(true)
			tFuncs:GetCtrl("ID_tc_mainchat"):Show(true)
			tFuncs:GetCtrl("ID_tc_mainchatsend"):Show(true)
		end
	end

	local rport, sport, gport = "","",""
	for port,data in pairs(tHubSettings.Main.tPortList.Regular) do
		rport = rport..port..", "
	end
	rport = rport:sub(1,-3)
	for port,data in pairs(tHubSettings.Main.tPortList.Secure) do
		sport = sport..port..", "
	end
	sport = sport:sub(1,-3)
	for port,data in pairs(tHubSettings.Main.tPortList.Gui) do
		gport = gport..port..", "
	end
	gport = gport:sub(1,-3)

	if not _NOGUI then
		if _OS == "Linux" or tHubSettings.Main.bForceTLS then
			tFuncs:GetCtrl("ID_tc_portssecure"):Enable(true)
			tFuncs:GetCtrl("ID_bu_createcert"):Enable(true)
		end

		tFuncs:GetCtrl("ID_tc_portsregular"):SetValue(rport)
		tFuncs:GetCtrl("ID_tc_portssecure"):SetValue(sport)
		tFuncs:GetCtrl("ID_tc_portsgui"):SetValue(gport)
		
			

		wx.wxHandleFatalExceptions()
	end

--~ 	local result, err = pcall(tScript["Init"])      -- initialize ScriptManager
--~ 	if not result then tFuncs:Debug("ERROR on execution of tScript:Init, err: "..err) end
end

main()

if tSettings.Debug.EnablePrint then
	RedefinePrint()
end

function NotifyGUI(msg, title, icon)
	WriteException("["..(title or "Notify").."] "..msg)
 	if not _NOGUI then
 		wx.wxMessageBox(tostring(msg),(title or "Information"),(icon or wx.wxICON_INFORMATION))
 	else
		print(os.date("[%c] ")..tostring(msg.."\n"))
 	end
end

if not _NOGUI then 
	dofile(_APPPATH..'./core/Hub-GuiFunc.lua')

	--// This executes every second from the main/gui thread
	tCtrls["app"].timer = wx.wxTimer(tCtrls["app"], 100)
	tCtrls["app"]:Connect(100, wx.wxEVT_TIMER, 
		function(event)
			tFuncs:DoMainTimer()
			tFuncs:UpdateStatusBar()
			tFuncs:CheckShutdown()
		end
	)
	tCtrls["app"].timer:Start(1000)
	if not tCtrls["app"].timer then tFuncs:SafeDebug("Failed to start the GUI timer.") else tFuncs:SafeDebug("Started GUI timer.") end

	tCtrls["app"]:Centre()
	tCtrls["app"]:Show(true)
end

--~ tFuncs:SetStatusText("GUI loaded...")
tStats.Console.Status = "Running"

--// Add master account
if tArgs.nick and tArgs.nick ~= "" then
	local reg, vnick, lnick = tNick:GetReg(tArgs.nick)
	local prof, err = ProfileManager:GetName("AAAA", (tArgs.profile or "Hubowner"))
	if not prof then tFuncs:Debug(err, "FI", -1) return end
	if not tArgs.pass or tArgs.pass == "" then
		tFuncs:Debug("No pass specified for master account '"..tArgs.nick.."', exiting.", "FI", -1)
		return
	else
		local res, err = tCheck:PasswordChars(tArgs.pass)
		if not res then
			tFuncs:Debug("No valid pass specified for master account '"..tArgs.nick.."', error : "..err, "FI", -1)
			return
		end
	end

	local ip = (tArgs.nickip and tArgs.nickip ~= "" and tArgs.nickip or false)

	if not reg then -- user isn't registered yet
		tAccounts[lnick] = {
				["sNick"] = tArgs.nick,
				["sProfile"] = prof,
				["sPassword"] = tArgs.pass,
				["iLastLogin"] = 0,
				["iLastSpoke"] = 0,
				["iTotalOnlineTime"] = 0,	-- total onlinetime in seconds
				["sReggedBy"] = "Command line",
				["iRegDate"] = os.time(),	--// store os.time() here, unformatted for easy sorting
				["sComment"] = "",
				["sGreeting"]= "",
				["sBindIP"] = ip,
		}
		tFuncs:Debug("Master account created: Nick: "..tArgs.nick.." / Pass: "..tArgs.pass.." / Profile: "..prof..(ip and " / Bound to IP: "..ip or "").." / By: Command line", "FI", -1)
		tFuncs:SaveSettings("accounts")	
	else	--// User is already registered, modify details
		tAccounts[lnick].sProfile = prof
		tAccounts[lnick].sBindIP = ip
		tAccounts[lnick].sPassword = tArgs.pass
		tFuncs:SaveSettings("accounts")
		tFuncs:Debug("Master account updated: Nick: "..tArgs.nick.." / Pass: "..tArgs.pass.." / Profile: "..prof..(ip and " / Bound to IP: "..ip or "").." / By: Command line", "FI", -1)
	end
else
	local acc = 0
	for name,data in pairs(tAccounts) do
		acc = acc+1
	end
	if acc < 1 then
		if _NOGUI then 
			tFuncs:Debug("No accounts available, use commandline options or edit FlexHub.lua before running the hub to create a master account", "FI", -1)
		else
			tFuncs:Debug("No accounts available, use 'Edit master accounts..' to add accounts", "FI", -1)
		end
	end
end

--// Start GUI and hub
if not _NOGUI then
	if tHubSettings.Console.AutoStartHub then
		local tPortList, err = tGui:ReadConsole()
		if not tPortList then
			Notify(err,"WARNING!",wx.wxICON_EXCLAMATION)
			return
		else
			tFuncs:StartHub(tPortList)
		end
	end
	collectgarbage("collect")

	--// wx.wxGetApp():OnFatalException needs to be defined here to allow wx to handle exceptions

	wx.wxGetApp():MainLoop()

else	--// no GUI, use commandline ports:
	local portcounter = 0
	local portlist = { ["Regular"] = {}, ["Secure"] = {}, ["Gui"] = {}, }
	--// Add regular ports
	if tArgs.port and tArgs.port ~= "" then
		for port in string.gmatch(tArgs.port, "%d+") do
			portlist["Regular"][tostring(port)] = true
			portcounter = portcounter+1
		end
	end
	--// Add secure ports
	if tArgs.sport and tArgs.sport ~= "" then
		for port in string.gmatch(tArgs.sport, "%d+") do
			if not portlist["Regular"][tostring(port)] then
				portlist["Secure"][tostring(port)] = true
				portcounter = portcounter+1
			else
				print("Error: You can't use the same port as secure and regular port, port: "..port)
			end
		end
	end
	--// Add Remote GUI port
	if tArgs.gport and tArgs.gport ~= "" then
		for port in string.gmatch(tArgs.sport, "%d+") do
			if not portlist["Regular"][tostring(port)] and not portlist["Secure"][tostring(port)] then
				portlist["Gui"][tostring(port)] = true
				portcounter = portcounter+1
			elseif portlist["Regular"][tostring(port)] then
				print("Error: You can't use the same port as regular and dedicated gui port, port: "..port)
			else
				print("Error: You can't use the same port as secure and dedicated gui port, port: "..port)
			end
		end
	end

	if portcounter > 0 then
		tPortList = portlist
		tFuncs:StartHub(portlist)
		collectgarbage("collect")
		while true do
			ThreadWait(1000)
			tFuncs:DoMainTimer()
			if not tFuncs:CheckShutdown() then break end
		end
	else
		print("Error: You need to specify at least one regular, secure or gui port.")
	end
end

--// FlipFlop --//-- Daywalker
