--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tGui = {

["ExecuteConsoleEvent"] = function()
	if not _NOGUI then
		local events = {}

		ThreadLockMutex(mutexconsole)
		for num, data in ipairs(tConsoleBuffer.Messages) do
			table.insert(events, data)
		end
		tConsoleBuffer.Messages = {}
		ThreadUnlockMutex(mutexconsole)

		--// Actual handling of GUI events outside of the console mutex
		for num, data in ipairs(events) do
			tGui:ConsoleEvent(data.Name, data.Flags, data.Message)
		end
	end
end,

["ConsoleEvent"] = function(self, ename, flags, msg)
	if ename and msg and string.sub(ename, 4,5) then
		local etype = string.sub(ename, 4,5)
		local control = tFuncs:GetCtrl(ename) -- control through ID
		if control and etype then
			if etype == "st" then
				local res, err = pcall(control.SetLabel,control,msg)
				if not res then OnError(err) end
			elseif etype == "wt" then --// Window title update
				--// Update tooltip and windowtitle
				local bitmap = tFuncs:GetImage("flexhub")
				local icon = wx.wxIcon()
				icon:CopyFromBitmap(bitmap)
				tCtrls["app"]:SetIcon(icon)
				tCtrls["taskbar"]:SetIcon(icon,tHubSettings.Main.sName.." (FlexHub)")
				tCtrls["app"]:SetTitle(tHubSettings.Main.sName.." on ".._APPNAME.." ".._APPVER.." svn:".._REV.." - by FlipFlop and Daywalker")
			elseif etype == "sc" then
				local num = tonumber(msg)
				local res, err = pcall(control.SetValue,control,num)
			elseif etype == "ga" then
				local num = tonumber(msg)
				local res, err = pcall(control.SetValue,control,num)
				if not res then OnError(err) end
			elseif etype == "tc" then
				msg = string.gsub(msg, "ERROR:", "Error:")
				msg = string.gsub(msg, "WARNING:", "Warning:")
			
				-- check if maxlines doens't exceed tSettings.MaxLinesDebug
				local res, numlines = pcall(control.GetNumberOfLines,control)
				if not res then OnError(numlines) end

				local removelines = numlines - tSettings.Gui.GeneralMaxLines
				if removelines > 0 then
					local removepos = 1 -- always include the first newline
					for linenum = 0, removelines do
						local res, linelen = pcall(control.GetLineLength,control, linenum)
						if not res then OnError(linelen) end
						removepos = removepos+linelen+1 --// including newline char
					end
					local res, err = pcall(control.Remove,control,0,removepos+1)
					if not res then OnError(err) end
--					control:Remove(0,removepos+1)
				end

				if flags and flags == "1" then		-- 0 or nil = append, 1 = new value
					local res, err = pcall(control.SetValue,control,msg)
					if not res then OnError(err) end
--						control:SetValue(msg)
				else
					local res, err = pcall(control.AppendText,control,msg)
					if not res then OnError(err) end
--						control:AppendText(msg)
				end

--~				local res, multi = pcall(control.IsMultiLine, control)
--~				if not res then OnError(multi) end
--~					if multi and tSettings.Gui.AutoScroll == 1 then
--~						local res, lastpos = pcall(control.GetLastPosition,control)
--~						if not res then OnError(err) end
--~						local res, err = pcall(control.ShowPosition,control,lastpos)
--~						if not res then OnError(err) end
--~					end

			elseif etype == "lc" then
				if flags ~= "del" then
					local item
					local counter = 0
					for value in string.gmatch(msg, "(.-)||") do
						if counter == 0 then
							local res, itemcount = pcall(control.GetItemCount,control)
							if not res then OnError(itemcount) end
							res, item = pcall(control.InsertItem, control, itemcount, value)
							if not res then OnError(item) end
--							item = control:InsertItem(control:GetItemCount(),value)
							counter = counter +1
						else
							local res, err = pcall(control.SetItem, control, item, counter, value)
							if not res then OnError(err) end
--							control:SetItem(item, counter, value)
							counter = counter +1
						end
					end
					local res, itemcount = pcall(control.GetItemCount, control)
					if not res then OnError(itemcount) end
					itemcount = itemcount -1
					local res, err = pcall(control.SetItemData, control, item, itemcount)
					if not res then OnError(err) end
--					control:SetItemData(item, control:GetItemCount()-1)
				else		--// flag del was given, msg = value in first column
					local res, item = pcall(control.FindItem, control, -1, msg)
					if not res then OnError(item) end
--					local item = control:FindItem(-1, msg)
					if item then
						local res, err = pcall(control.DeleteItem, control, item)
						if not res then OnError(err) end
--						control:DeleteItem(item)
					else
						tFuncs:GetCtrl("ID_tc_debug"):AppendText("Error: GUIFunc.lua: Value to delete from userlist not found: "..msg)
					end
				end

			elseif etype == "ss" then
				if flags ~= "" then flags = tonumber(flags) else flags = 0 end
				local res, err = pcall(control.SetStatusText, control, msg, flags)
				if not res then OnError(err) end
--				control:SetStatusText(msg, flags)
			end
		else
			tFuncs:GetCtrl("ID_tc_debug"):AppendText("*** wxID_GUI_UPDATE Gui element "..tostring(ename).." not found for data: "..tostring(msg))
		end
	else
		tFuncs:GetCtrl("ID_tc_debug"):AppendText("*** wxID_GUI_UPDATE No proper ename and/or msg were sent to the GUI: "..tostring(ename))
	end
	collectgarbage("collect")
end,

["Codes"] = {
	["H"] = "tHubSettings",

	["HD"] = "Security",

	["HDGE"] = "General",
	["HDLO"] = "Login",
	["HDPA"] = "Passive",	--// myinfo
	["HDNI"] = "NI",		--// nick
	["HDNL"] = "NL",		--// nick
	["HDSE"] = "SE",		--// search
	["HDSR"] = "SE",		--// searchresults
	["HDMI"] = "MI",		--// myinfo
	["HDCT"] = "CT",		--// connecttome
	["HDRC"] = "RC",		--// revconnecttome
	["HDMC"] = "MC",		--// mainchat
	["HDPM"] = "PM",		--// pm
},

--~ ["UpdateTaskbarTooltip"] = function(self, msg)
--~ 	if not _NOGUI then
--~ 		bitmap = tFuncs:GetImage("flexhub")
--~ 		local icon = wx.wxIcon()
--~ 		icon:CopyFromBitmap(bitmap)
--~ 		tCtrls["app"]:SetIcon(icon)
--~ 		tCtrls["app"]:SetTitle(tHubSettings.Main.sName.." on ".._APPNAME.." ".._APPVER.." - by FlipFlop and Daywalker")
--~ 		tCtrls["taskbar"]:SetIcon(icon,tHubSettings.Main.sName)
--~ 		icon:delete()
--~ 		bitmap:delete()
--~ 		collectgarbage("collect")
--~ 	end
--~ end,

["Gc"] = function(self)
	local c1 = collectgarbage("count")
	local c2 = c1
--	collectgarbage("collect")	-- garbage collection every second?? experimental... now disabled for checking mem-leaks
--	local c2 = collectgarbage("count")
	return c1, c2
end,

["ReadConsole"] = function(self)
	local tPortList, err = tGui:GetPorts()
	if tPortList then
		tHubSettings.Main.tPortList = tPortList

		local level = tonumber(tFuncs:GetCtrl("ID_sc_maxusers"):GetValue())
		--// Userlimit changed
		if tHubSettings.tRules.iMaxUsersHard ~= level then
			tHubSettings.tRules.iMaxUsersHard = level
			tFuncs:Debug("Host userlimit (console) changed to: "..level)
			tFuncs:SaveSettings("settings")
			local res, err, msg = SetManager:Set("maxusers", level)
			if not res then tFuncs:Debug(err) end
		end

		local newhubname = tFuncs:GetCtrl("ID_tc_hubname"):GetValue()
		--// Hubname changed
		if newhubname and newhubname ~= "" and tHubSettings.Main.sName ~= newhubname then

			--// Update tooltip and windowtitle
			local bitmap = tFuncs:GetImage("flexhub")
			local icon = wx.wxIcon()
			icon:CopyFromBitmap(bitmap)
			tCtrls["app"]:SetIcon(icon)
			tCtrls["taskbar"]:SetIcon(icon,tHubSettings.Main.sName.." (FlexHub)")
			tCtrls["app"]:SetTitle(tHubSettings.Main.sName.." on ".._APPNAME.." ".._APPVER.." svn:".._REV.." - by FlipFlop and Daywalker")

			if tFuncs:GetCtrl("ID_bu_startstop"):GetLabel() == "Stop Hub" then
				local _, res = SetCmd.sethubname.cmd("AAAA", "!sethubname "..newhubname)
				tFuncs:GetCtrl("ID_tc_debug"):AppendText((res or "Error when changing Hub name").."\n")
			else
				tHubSettings.Main.sName = newhubname
				tFuncs:SaveSettings("settings")
			end
		end

		tHubSettings.Console.AutoStartHub = tFuncs:GetCtrl("ID_cb_autostart"):IsChecked()
		tHubSettings.Console.WarnOnClose = tFuncs:GetCtrl("ID_cb_confirm"):IsChecked()
		tFuncs:SaveSettings("settings")
		tFuncs:SaveSettings("accounts")
	end
	return tPortList, err
end,

["SetAccountFocus"] = function(self, name, profile)
		local focus = 1

		local profile = profile or tFuncs:GetCtrl("ID_co_guiprofile", tCtrls["accounts"]):GetValue() or "Gui"
		local guiaccounts = AccountManager:GetProfileAccounts("AAAA", profile)

		if guiaccounts then
			local sortnicks, guinicks, guidata = {}, {}, {}
			for nick in pairs(guiaccounts) do 
				table.insert(sortnicks, nick)
			end
			table.sort(sortnicks)
			for num, nick in ipairs(sortnicks) do
				table.insert(guinicks, guiaccounts[nick].sNick)
				table.insert(guidata, guiaccounts[nick])
				if name and name:lower() == nick then focus = #guinicks end
			end
			if #guinicks > 0 then
				tFuncs:GetCtrl("ID_co_guiname", tCtrls["accounts"]):Clear()
				tFuncs:GetCtrl("ID_co_guiname", tCtrls["accounts"]):Append(guinicks)
				tFuncs:GetCtrl("ID_co_guiname", tCtrls["accounts"]):SetSelection(focus-1)
				tFuncs:GetCtrl("ID_tc_guipass", tCtrls["accounts"]):SetValue(guidata[focus].sPassword)
				tFuncs:GetCtrl("ID_co_guiprofile", tCtrls["accounts"]):SetValue(guidata[focus].sProfile)
				tFuncs:GetCtrl("ID_cb_guipid", tCtrls["accounts"]):SetValue(guidata[focus].bUsePID)
				tFuncs:GetCtrl("ID_tc_guipid", tCtrls["accounts"]):SetValue(guidata[focus].sPID or "")
				tFuncs:GetCtrl("ID_cb_guiremote", tCtrls["accounts"]):SetValue(guidata[focus].bAllowRemote)
				tFuncs:GetCtrl("ID_cb_guiclient", tCtrls["accounts"]):SetValue(guidata[focus].bAllowClient)
			else
				tFuncs:GetCtrl("ID_co_guiname", tCtrls["accounts"]):Clear()
				tFuncs:GetCtrl("ID_tc_guipass", tCtrls["accounts"]):SetValue("")
				tFuncs:GetCtrl("ID_co_guiprofile", tCtrls["accounts"]):SetValue(profile)
				tFuncs:GetCtrl("ID_cb_guipid", tCtrls["accounts"]):SetValue(false)
				tFuncs:GetCtrl("ID_tc_guipid", tCtrls["accounts"]):SetValue("")
				tFuncs:GetCtrl("ID_cb_guiremote", tCtrls["accounts"]):SetValue(true)
				tFuncs:GetCtrl("ID_cb_guiclient", tCtrls["accounts"]):SetValue(false)
			end
		else
			tFuncs:GetCtrl("ID_co_guiname", tCtrls["accounts"]):Clear()
			tFuncs:GetCtrl("ID_tc_guipass", tCtrls["accounts"]):SetValue("")
			tFuncs:GetCtrl("ID_co_guiprofile", tCtrls["accounts"]):SetValue(profile)
			tFuncs:GetCtrl("ID_cb_guipid", tCtrls["accounts"]):SetValue(false)
			tFuncs:GetCtrl("ID_tc_guipid", tCtrls["accounts"]):SetValue("")
			tFuncs:GetCtrl("ID_cb_guiremote", tCtrls["accounts"]):SetValue(true)
			tFuncs:GetCtrl("ID_cb_guiclient", tCtrls["accounts"]):SetValue(false)

			--NotifyGUI("No GUI account has been made yet","Gui account(s) not found",wx.wxICON_EXCLAMATION)
			tFuncs:GetCtrl("ID_st_accountstatus", tCtrls["accounts"]):SetLabel("No GUI account has been made yet")
		end
		if profile ~= "Gui" then
			tFuncs:GetCtrl("ID_cb_guiremote", tCtrls["accounts"]):Enable(false)
			tFuncs:GetCtrl("ID_cb_guiclient", tCtrls["accounts"]):Enable(false)
		else
			tFuncs:GetCtrl("ID_cb_guiremote", tCtrls["accounts"]):Enable(true)
			tFuncs:GetCtrl("ID_cb_guiclient", tCtrls["accounts"]):Enable(true)
		end
end,

-- // commented out on 2011-10-22 by 1dat, we'll se if its in use
-- // if its put back in work, add tCheck:Passwordschar to passwords here
--["ReadAccounts"] = function(self)
--	tHubSettings.Gui.AllowRemote = tFuncs:GetCtrl("ID_cb_guiremote"):IsChecked()
--	tHubSettings.Gui.AllowMultiple = tFuncs:GetCtrl("ID_cb_guimultiple"):IsChecked()
--	tHubSettings.Gui.UseTLS = tFuncs:GetCtrl("ID_cb_guitls"):IsChecked()
--	if tAccounts.flexgui then
--		tAccounts.flexgui.sPassword = tFuncs:GetCtrl("ID_tc_guipass"):GetValue()
--	end
--end,

["GetPorts"] = function(self)
	local portlist = { ["Regular"] = {}, ["Secure"] = {}, ["Gui"] = {}, }
	local rports = tFuncs:GetCtrl("ID_tc_portsregular"):GetValue()
	local sports = tFuncs:GetCtrl("ID_tc_portssecure"):GetValue()
	local gports = tFuncs:GetCtrl("ID_tc_portsgui"):GetValue()
	local gsecure = tFuncs:GetCtrl("ID_cb_securegui"):IsChecked()
	local portcounter = 0
	for port in string.gmatch(rports, "%d+") do
		portlist["Regular"][tostring(port)] = true
		portcounter = portcounter+1
	end
	for port in string.gmatch(sports, "%d+") do
		if not portlist["Regular"][tostring(port)] then
			portlist["Secure"][tostring(port)] = true
			portcounter = portcounter+1
		else
			return false, "You can't use the same port as regular and secure port."
		end
	end
	for port in string.gmatch(gports, "%d+") do
		if portlist["Regular"][tostring(port)] and gsecure then
			return false, "You can't use the same port as regular and secure GUI port."
		elseif portlist["Secure"][tostring(port)] and not gsecure then
			return false, "You can't use the same port as secure and regular GUI port."
		elseif portlist["Regular"][tostring(port)] or portlist["Secure"][tostring(port)] then
			NotifyGUI("When using the same port for the GUI, you won't be able to connect to the hub when it's offline","Warning!",wx.wxICON_EXCLAMATION)
			portlist["Gui"][tostring(port)] = true
			portcounter = portcounter+1
		else
			portlist["Gui"][tostring(port)] = true
			portcounter = portcounter+1
		end
	end

	if portcounter == 0 then
		return false, "You need to specify at least one regular or secure port."
	else
		return portlist
	end
end,
}

-- Real event handlers
tCtrls["app"]:Connect(wx.wxEVT_TIMER,
function(event)
	local c1, c2 = tGui:Gc()		-- garbage collection
	local control = tFuncs:GetCtrl("ID_ss_statusbar")
	local res, err = pcall(control.SetStatusText, control, "U:"..(tStats.Hub.Users.Total or "0"), 1)
	if not res then OnError(err) end
	local res, err = pcall(control.SetStatusText, control, "H:"..(tLogs.Counters.Ddos.Total or "0"), 2)
	if not res then OnError(err) end
	local res, err = pcall(control.SetStatusText, control, "M:"..tFuncs:ConvertShare(math.ceil(c2*1024)), 3)
	if not res then OnError(err) end
	c1,c2,control = nil
end)

tCtrls["app"]:Connect(wx.wxID_ANY,wx.wxEVT_COMMAND_SPINCTRL_UPDATED,
function(event)
	local tSpinEvtIds = {
	[tCtrls["ID_sc_debuglevel"]] = function()
		local level = tFuncs:GetCtrl("ID_sc_debuglevel"):GetValue()
		tHubSettings.Console.DebugLevel = tonumber(level)
		tFuncs:SaveSettings("settings")
	end,

	[tCtrls["ID_sc_maxusers"]] = function()
--~ 		local level = tonumber(tFuncs:GetCtrl("ID_sc_maxusers"):GetValue())
--~ 		tHubSettings.tRules.iMaxUsersHard = level
--~ 		tFuncs:Debug("Host userlimit (console) changed to: "..level)
--~ 		tFuncs:SaveSettings("settings")

--~ 		local res, err, msg = SetManager:Set("maxusers", level)
--~ 		if not res then tFuncs:Debug(err) end
--~ 		if msg then tFuncs:Debug(msg, "FI", 0) end
	end,
	}
	if tSpinEvtIds[event:GetId()] then
		tSpinEvtIds[event:GetId()]()
	else
		print("wx.wxEVT_SCROLL: Id not found for GetId: "..event:GetId())
	end
end)

tCtrls["app"]:Connect(wx.wxID_ANY,wx.wxEVT_COMMAND_TEXT_ENTER,
function(event)
	local tButtonEvtIds = {
		[tCtrls["ID_tc_mainchatsend"]] = function()
			local newline = tFuncs:GetCtrl("ID_tc_mainchatsend"):GetValue()
			tFuncs:SafeExecute("tSend:Mainchat(tClient[\"AAAA\"], \"AAAA\", "..string.format("%q",newline)..")")
			tFuncs:GetCtrl("ID_tc_mainchatsend"):Clear()
		end,
	}
	if tButtonEvtIds[event:GetId()] then
		tButtonEvtIds[event:GetId()]()
	else
		print("wx.wxEVT_COMMAND_TEXT_ENTER: Id not found for GetId: "..event:GetId())
	end
end)

tCtrls["app"]:Connect(wx.wxID_ANY,wx.wxEVT_COMMAND_BUTTON_CLICKED,
function(event)
	local tButtonEvtIds = {
	[tCtrls["ID_bu_startstop"]] = function()
		if tFuncs:GetCtrl("ID_bu_startstop"):GetLabel() == "Stop Hub" then
			tFuncs:StopHub()
		else
			local tPortList, err = tGui:ReadConsole()
			if not tPortList then
				NotifyGUI(err,"Warning!",wx.wxICON_EXCLAMATION)
				return
			else
				tFuncs:StartHub(tPortList)
			end
		end
	end,
	[tCtrls["ID_bu_updatesettings"]] = function()
		local tPortList, err = tGui:ReadConsole()
		if not tPortList then
			NotifyGUI(err,"Warning!",wx.wxICON_EXCLAMATION)
			return
		else
			if tFuncs:GetCtrl("ID_bu_startstop"):GetLabel() == "Stop Hub" then
				tFuncs:UpdatePorts(tPortList)
			end
		end
	end,
	[tCtrls["ID_bu_createcert"]] = function()
		--// Check for existing certificate
		local cert = io.open("certificates/cert.pem", "r")
		local key = io.open("certificates/key.pem", "r")
		if cert or key then
			if cert then cert:close() end
			if key then key:close() end
			
			local res = wx.wxMessageBox("A certificate already exists, are you sure you want to create a new one?", "TLS Certificate exists", wx.wxYES_NO)
			if res == wx.wxNO then return end
		end
		
		if _OS == "Linux" then
			local res, retmsg = tFuncs:CreateCertificate(path, true)
			tFuncs:GetCtrl("ID_tc_debug"):AppendText(retmsg.."\n")
		else
			--// Get path to openssl.exe
			local path = wx.wxFileSelector("Select the file openssl.exe from your OpenSSL installation", "", "openssl.exe")
			if path then
				path = path:gsub("openssl.exe$", "")
				local res, retmsg = tFuncs:CreateCertificate(path, true)
				tFuncs:GetCtrl("ID_tc_debug"):AppendText(retmsg.."\n")
			end
		end

	end,
	
	[tCtrls["ID_bu_startgui"]] = function()
		if tStats.Gui.StartTime and tStats.Gui.StartTime < (os.time() - 15) then
			tFuncs:GetCtrl("ID_tc_debug"):AppendText("Gui initializing, one moment please.\n")
			tStats.Gui.StartTime = os.time()
			wx.wxExecute("../FlexGui/FlexHub_Gui")
		else
			local wait = tStats.Gui.StartTime + 15 - os.time()
			tFuncs:GetCtrl("ID_tc_debug"):AppendText("Gui already started, please wait "..wait.." seconds.\n")
		end
	end,
	[tCtrls["ID_bu_accounts"]] = function()
--~ 		tFuncs:GetCtrl("ID_cb_guiremote", tCtrls["accounts"]):SetValue(tHubSettings.Gui.AllowRemote)
		tFuncs:GetCtrl("ID_cb_guimultiple", tCtrls["accounts"]):SetValue(tHubSettings.Gui.AllowMultiple)
		tFuncs:GetCtrl("ID_cb_guitls", tCtrls["accounts"]):SetValue(tHubSettings.Gui.UseTLS)

		local profs = ProfileManager:GetByLevel("AAAA", 10)
		if profs then 
			tFuncs:GetCtrl("ID_co_guiprofile", tCtrls["accounts"]):Clear()
			tFuncs:GetCtrl("ID_co_guiprofile", tCtrls["accounts"]):Append(profs)
			tFuncs:GetCtrl("ID_co_guiprofile", tCtrls["accounts"]):SetSelection(0)

		else
			NotifyGUI("No profiles with level 10 available, accounts can't be created here.","Missing profiles",wx.wxICON_EXCLAMATION)
			return
		end

		tGui:SetAccountFocus()

		tCtrls["accounts"]:Centre()
		tCtrls["accounts"]:Show(true)
	end,

	}
	if tButtonEvtIds[event:GetId()] then
		tButtonEvtIds[event:GetId()]()
	else
		print("wx.wxEVT_COMMAND_TOOL_CLICKED: Id not found for GetId: "..event:GetId())
	end
end)

tCtrls["app"]:Connect(wx.wxID_ANY, wx.wxEVT_COMMAND_CHECKBOX_CLICKED,
function(event)
	local tToolEvtIds = {
	[tCtrls["ID_cb_confirm"]] = function()
		tHubSettings.Console.WarnOnClose = tFuncs:GetCtrl("ID_cb_confirm"):IsChecked()
		tFuncs:SaveSettings("settings")
	end,
	[tCtrls["ID_cb_autostart"]] = function()
		tHubSettings.Console.AutoStartHub = tFuncs:GetCtrl("ID_cb_autostart"):IsChecked()
		tFuncs:SaveSettings("settings")
	end,

	}
	if tToolEvtIds[event:GetId()] then
		tToolEvtIds[event:GetId()]()
	else
		print("wx.wxEVT_COMMAND_TOOL_CLICKED: Id not found for GetId: "..event:GetId())
	end
end)

tCtrls["app"]:Connect(wx.wxID_ANY,wx.wxEVT_COMMAND_TOOL_CLICKED,
function(event)
	local tToolEvtIds = {

	[tCtrls["ID_to_start"]] = function()
		if tFuncs:GetCtrl("ID_bu_startstop"):GetLabel() == "Start Hub" then
			local tPortList, err = tGui:ReadConsole()
			if not tPortList then
				NotifyGUI(err,"Warning!",wx.wxICON_EXCLAMATION)
				return
			else
				tFuncs:StartHub(tPortList)
			end
		end
	end,

	[tCtrls["ID_to_stop"]] = function()
		if tFuncs:GetCtrl("ID_bu_startstop"):GetLabel() == "Stop Hub" then
			tFuncs:StopHub()
			tFuncs:GetCtrl("ID_bu_startstop"):SetLabel("Start Hub")
		end
	end,

	[tCtrls["ID_to_import"]] = function()
		--// Disable the view log button
--~		tFuncs:GetCtrl("ID_bu_viewlog", tCtrls["import"]):Disable()

		tCtrls["import"]:Centre()
		tCtrls["import"]:Show(true)
	end,

	}
	if tToolEvtIds[event:GetId()] then
		tToolEvtIds[event:GetId()]()
	else
		print("wx.wxEVT_COMMAND_TOOL_CLICKED: Id not found for GetId: "..event:GetId())
	end
end)


--// Account dialog events

tCtrls["accounts"]:Connect(wx.wxEVT_CLOSE_WINDOW,
function(event)
	tCtrls["accounts"]:Show(false) 
end)

tCtrls["accounts"]:Connect(wx.wxID_ANY,wx.wxEVT_COMMAND_COMBOBOX_SELECTED,
function(event)
	local tComboEvtIds = {
	[tCtrls["ID_co_guiname"]] = function()
		local nick = tFuncs:GetCtrl("ID_co_guiname", tCtrls["accounts"]):GetValue()
		local profile = tFuncs:GetCtrl("ID_co_guiprofile", tCtrls["accounts"]):GetValue()
		tGui:SetAccountFocus(nick, profile)
	end,
	[tCtrls["ID_co_guiprofile"]] = function()
		local profile = tFuncs:GetCtrl("ID_co_guiprofile", tCtrls["accounts"]):GetValue()
		tGui:SetAccountFocus(nil, profile)
	end,

	}
	if tComboEvtIds[event:GetId()] then
		tComboEvtIds[event:GetId()]()
	else
		print("wx.wxEVT_COMMAND_BUTTON_CLICKED: Id not found for GetId: "..event:GetId())
	end

end)

tCtrls["accounts"]:Connect(wx.wxID_ANY,wx.wxEVT_COMMAND_BUTTON_CLICKED,
function(event)
	local tAccountButtonEvtIds = {
	[tCtrls["ID_bu_guisave"]] = function()
		local nick = tFuncs:GetCtrl("ID_co_guiname", tCtrls["accounts"]):GetValue()
		local pass = tFuncs:GetCtrl("ID_tc_guipass", tCtrls["accounts"]):GetValue()
		local profile = tFuncs:GetCtrl("ID_co_guiprofile", tCtrls["accounts"]):GetValue()
		local usepid = tFuncs:GetCtrl("ID_cb_guipid", tCtrls["accounts"]):GetValue()
		local pid = tFuncs:GetCtrl("ID_tc_guipid", tCtrls["accounts"]):GetValue()
		local allowremote = tFuncs:GetCtrl("ID_cb_guiremote", tCtrls["accounts"]):GetValue()
		local allowclient = tFuncs:GetCtrl("ID_cb_guiclient", tCtrls["accounts"]):GetValue()

		local res,_,_,err = tCheck:Nick(nil, nil, nick, 10)
		if not res then 
			--NotifyGUI(err,"Invalid nickname",wx.wxICON_EXCLAMATION)
			tFuncs:GetCtrl("ID_st_accountstatus", tCtrls["accounts"]):SetLabel("Invalid nickname : "..err)
			return
		end

		local res, err = tCheck:PasswordChars(pass)
		if not res then
		--	NotifyGUI(err, "Invalid password",wx.wxICON_EXCLAMATION)
			tFuncs:GetCtrl("ID_st_accountstatus", tCtrls["accounts"]):SetLabel("Invalid password : "..err)
			return
		end

		local reg, vnick, lnick = tNick:GetReg(nick)
		if reg and reg.sProfile ~= profile then
			--NotifyGUI("This nickname is already in use with profile: "..tostring(reg.sProfile), "Invalid nickname",wx.wxICON_EXCLAMATION)
			tFuncs:GetCtrl("ID_st_accountstatus", tCtrls["accounts"]):SetLabel("This nickname is already in use with profile: "..tostring(reg.sProfile))
			return
		elseif reg then		--// Existing account
			tAccounts[lnick].sProfile = profile
			tAccounts[lnick].sPassword = pass
			tAccounts[lnick].bUsePID = usepid
			tAccounts[lnick].sPID = pid
			if profile == "Gui" then
				tAccounts[lnick].bAllowRemote = allowremote
				tAccounts[lnick].bAllowClient = allowclient
			end
			tFuncs:GetCtrl("ID_st_accountstatus", tCtrls["accounts"]):SetLabel("Account '"..nick.."' updated.")
		else				--// New account
			local res, err = AccountManager:Add("AAAA", nick, pass, profile)
			if not res or not tAccounts[lnick] then
				--NotifyGUI("Couldn't create account for '"..nick.."': "..err,"Error creating account",wx.wxICON_EXCLAMATION)
				tFuncs:GetCtrl("ID_st_accountstatus", tCtrls["accounts"]):SetLabel("Couldn't create account for '"..nick.."': "..err)
				return
			end
			tAccounts[lnick].bUsePID = usepid
			tAccounts[lnick].sPID = pid
			if profile == "Gui" then
				tAccounts[lnick].bAllowRemote = allowremote
				tAccounts[lnick].bAllowClient = allowclient
			end
			tFuncs:GetCtrl("ID_st_accountstatus", tCtrls["accounts"]):SetLabel("Account '"..nick.."' created.")
		end

		tFuncs:SaveSettings("accounts")
		tGui:SetAccountFocus(nick, profile)

	end,
	[tCtrls["ID_bu_guidelete"]] = function()
		local nick = tFuncs:GetCtrl("ID_co_guiname", tCtrls["accounts"]):GetValue()
		local reg, vnick, lnick = tNick:GetReg(nick)
		if reg and reg.iLevel == 10 then
			local res, err = AccountManager:Delete("AAAA", vnick)
			tGui:SetAccountFocus()	

			if not res then
				--NotifyGUI("Couldn't delete account: "..err,"Error deleting account",wx.wxICON_EXCLAMATION)
				tFuncs:GetCtrl("ID_st_accountstatus", tCtrls["accounts"]):SetLabel("Couldn't delete account: "..err)
				return
			end
			tFuncs:GetCtrl("ID_st_accountstatus", tCtrls["accounts"]):SetLabel("Account '"..vnick.."' deleted.")

		else
		--	NotifyGUI("An account with nickname '"..nick.."' doesn't exist.","Account doesn't exist",wx.wxICON_EXCLAMATION)
			tFuncs:GetCtrl("ID_st_accountstatus", tCtrls["accounts"]):SetLabel("An account with nickname '"..nick.."' doesn't exist.")
			return
		end

	end,
	}
	if tAccountButtonEvtIds[event:GetId()] then
		tAccountButtonEvtIds[event:GetId()]()
	else
		print("wx.wxEVT_COMMAND_BUTTON_CLICKED: Id not found for GetId: "..event:GetId())
	end
end)

--// Import dialog events

tCtrls["import"]:Connect(wx.wxEVT_CLOSE_WINDOW,
function(event)
	tCtrls["import"]:Show(false)
end)

tCtrls["import"]:Connect(wx.wxID_ANY,wx.wxEVT_COMMAND_BUTTON_CLICKED,
function(event)
	local tButtonEvtIds = {
		[tCtrls["ID_bu_import"]] = function()
			local filename = tFuncs:GetCtrl("ID_fp_importfile", tCtrls["import"]):GetPath()
			if not filename then
				NotifyGUI("You need to select a valid file for importing!", "Warning!",wx.wxICON_EXCLAMATION)
			else
				local selected = tFuncs:GetCtrl("ID_rb_hubsoft", tCtrls["import"]):GetSelection()
				if selected == 0 then
					local ok, res, status, importlog = pcall(tImport.Ynhub, true, filename)
					if not ok then NotifyGUI("Error during Ynhub accounts import: "..res) return end
					if status then
						tFuncs:GetCtrl("ID_st_status", tCtrls["import"]):SetLabel(status)
						NotifyGUI(status.."\nCheck the file: ".._LOGDIR.."Import.log for details.", "Ynhub accounts import results",wx.wxICON_EXCLAMATION)
					end
				elseif selected == 1 then
					local ok, res, status, importlog = pcall(tImport.Ptokax, true, filename)
					if not ok then NotifyGUI("Error during Ptokax accounts import: "..res) return end
					if status then
						tFuncs:GetCtrl("ID_st_status", tCtrls["import"]):SetLabel(status)
						NotifyGUI(status.."\nCheck the file: ".._LOGDIR.."Import.log for details.", "Ptokax accounts import results",wx.wxICON_EXCLAMATION)
					end
				elseif selected == 2 then
					local ok, res, status, importlog = pcall(tImport.ADCH, true, filename)
					if not ok then NotifyGUI("Error during ADCH++ accounts import: "..res) return end
					if status then
						tFuncs:GetCtrl("ID_st_status", tCtrls["import"]):SetLabel(status)
						NotifyGUI(status.."\nCheck the file: ".._LOGDIR.."Import.log for details.", "ADCH++ accounts import results",wx.wxICON_EXCLAMATION)
					end
				end
			end

		end,
		[tCtrls["ID_bu_viewlog"]] = function()

		end,
	}
	if tButtonEvtIds[event:GetId()] then
		tButtonEvtIds[event:GetId()]()
	else
		print("wx.wxEVT_COMMAND_BUTTON_CLICKED: Id not found for GetId: "..event:GetId())
	end
end)

--[[
--// Custom event handlers
tCtrls["app"]:Connect(tCtrls["wxID_EVENT_GUI"],wx.wxEVT_NULL,
function(event)
--~ 	ThreadLockMutex(mutexconsole)
--~ 	ThreadWait(10)
	local sString = event:DynamicCast("wxCommandEvent"):GetString()
--~ 	ThreadUnlockMutex(mutexconsole)
	local _,_,ename,flags,msg = string.find(sString, "^#(.-)#(.-)#(.*)")	-- ename can be the controlname either in XRC or direct, or the ID as stored in the tCtrls.GuiElements table

	if ename and msg and string.sub(ename, 4,5) then
		local etype = string.sub(ename, 4,5)
		local control = tFuncs:GetCtrl(ename) -- control through ID
		if control and etype then
			if etype == "st" then
				local res, err = pcall(control.SetLabel,control,msg)
				if not res then OnError(err) end
			elseif etype == "wt" then --// Window title update
				--// Update tooltip and windowtitle
				local bitmap = tFuncs:GetImage("flexhub")
				local icon = wx.wxIcon()
				icon:CopyFromBitmap(bitmap)
				tCtrls["app"]:SetIcon(icon)
				tCtrls["taskbar"]:SetIcon(icon,tHubSettings.Main.sName.." (FlexHub)")
				tCtrls["app"]:SetTitle(tHubSettings.Main.sName.." on ".._APPNAME.." ".._APPVER.." svn:".._REV.." - by FlipFlop and Daywalker")
			elseif etype == "sc" then
				local num = tonumber(msg)
				local res, err = pcall(control.SetValue,control,num)
			elseif etype == "ga" then
				local num = tonumber(msg)
				local res, err = pcall(control.SetValue,control,num)
				if not res then OnError(err) end
			elseif etype == "tc" then
				-- check if maxlines doens't exceed tSettings.MaxLinesDebug
				local res, numlines = pcall(control.GetNumberOfLines,control)
				if not res then OnError(numlines) end

				local removelines = numlines - tSettings.Gui.GeneralMaxLines
				if removelines > 0 then
					local removepos = 1 -- always include the first newline
					for linenum = 0, removelines do
						local res, linelen = pcall(control.GetLineLength,control, linenum)
						if not res then OnError(linelen) end
						removepos = removepos+linelen+1 --// including newline char
					end
					local res, err = pcall(control.Remove,control,0,removepos+1)
					if not res then OnError(err) end
--					control:Remove(0,removepos+1)
				end

				if flags and flags == "1" then		-- 0 or nil = append, 1 = new value
					local res, err = pcall(control.SetValue,control,msg)
					if not res then OnError(err) end
--						control:SetValue(msg)
				else
					local res, err = pcall(control.AppendText,control,msg)
					if not res then OnError(err) end
--						control:AppendText(msg)
				end

--~				local res, multi = pcall(control.IsMultiLine, control)
--~				if not res then OnError(multi) end
--~					if multi and tSettings.Gui.AutoScroll == 1 then
--~						local res, lastpos = pcall(control.GetLastPosition,control)
--~						if not res then OnError(err) end
--~						local res, err = pcall(control.ShowPosition,control,lastpos)
--~						if not res then OnError(err) end
--~					end

			elseif etype == "lc" then
				if flags ~= "del" then
					local item
					local counter = 0
					for value in string.gmatch(msg, "(.-)||") do
						if counter == 0 then
							local res, itemcount = pcall(control.GetItemCount,control)
							if not res then OnError(itemcount) end
							res, item = pcall(control.InsertItem, control, itemcount, value)
							if not res then OnError(item) end
--							item = control:InsertItem(control:GetItemCount(),value)
							counter = counter +1
						else
							local res, err = pcall(control.SetItem, control, item, counter, value)
							if not res then OnError(err) end
--							control:SetItem(item, counter, value)
							counter = counter +1
						end
					end
					local res, itemcount = pcall(control.GetItemCount, control)
					if not res then OnError(itemcount) end
					itemcount = itemcount -1
					local res, err = pcall(control.SetItemData, control, item, itemcount)
					if not res then OnError(err) end
--					control:SetItemData(item, control:GetItemCount()-1)
				else		--// flag del was given, msg = value in first column
					local res, item = pcall(control.FindItem, control, -1, msg)
					if not res then OnError(item) end
--					local item = control:FindItem(-1, msg)
					if item then
						local res, err = pcall(control.DeleteItem, control, item)
						if not res then OnError(err) end
--						control:DeleteItem(item)
					else
						tFuncs:GetCtrl("ID_tc_debug"):AppendText("Error: GUIFunc.lua: Value to delete from userlist not found: "..msg)
					end
				end

			elseif etype == "ss" then
				if flags ~= "" then flags = tonumber(flags) else flags = 0 end
				local res, err = pcall(control.SetStatusText, control, msg, flags)
				if not res then OnError(err) end
--				control:SetStatusText(msg, flags)
			end
		else
			tFuncs:GetCtrl("ID_tc_debug"):AppendText("*** wxID_GUI_UPDATE Gui element "..tostring(ename).." not found for event: "..tostring(sString))
		end
	else
		tFuncs:GetCtrl("ID_tc_debug"):AppendText("*** wxID_GUI_UPDATE No proper ename and/or msg were sent to the GUI: "..tostring(sString))
	end
	collectgarbage("collect")
end)
]]--

tCtrls["app"]:Connect(tCtrls["wxID_EVENT_SHOWDIALOG"],wx.wxEVT_NULL,
function(event)
	local msg = event:DynamicCast("wxCommandEvent"):GetString()
	Notify(msg,"INFO",wx.wxICON_INFORMATION) -- skipped changing to NotifyGUI
	print(msg)
end)

tCtrls["app"]:Connect(tCtrls["wxID_EVENT_PRINT"],wx.wxEVT_NULL,
function(event)
	local msg = event:DynamicCast("wxCommandEvent"):GetString()
	print(msg)
end)

if wx.__WXMSW__ and not _NOGUI then -- On Unix it hides the program...
	tCtrls["app"]:Connect(wx.wxEVT_ICONIZE,
	function(event)
		tCtrls["app"]:Iconize(true)
		tCtrls["app"]:Show(false)
	end)
	tCtrls["taskbar"]:Connect(wx.wxID_ANY,wx.wxEVT_TASKBAR_LEFT_DOWN,
	function(event)
		if tCtrls["app"]:IsIconized() then
			tCtrls["app"]:Show(true)
			tCtrls["app"]:Iconize(false)
		else
			tCtrls["app"]:Iconize(true)
			tCtrls["app"]:Show(false)
		end
	end)
end

tCtrls["app"]:Connect(wx.wxEVT_CLOSE_WINDOW,
function(event)
	if _GOINGDOWN then
		event:Skip()
		return
	else
		_GOINGDOWN = true
	end

	local sure = true
	if tHubSettings.Console.WarnOnClose and not _NOGUI then
		local res = wx.wxMessageBox("This will shutdown the hub, are you sure?", "Shutdown hub", wx.wxYES_NO)
		if res == wx.wxNO then
			sure = false
		end
	end
	if sure then
		OnStop()
	end
end)

--// FlipFlop --//-- Daywalker
