--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tImport = {

["Ptokax"] = function(self, filename)
	local importlog = {["Ok"] = {}, ["Failed"] = {}}
	local f,err = io.open(filename, "r")
	if not f then return false, "Error: Couldn't open file '"..filename.."'. Error: "..(err or "Unknown error").."." end

	local xml = f:read("*a")
	f:close()

	local profname = filename:gsub("RegisteredUsers.xml", "Profiles.xml")

	local p,err = io.open(profname, "r")
	if not p then return false, "Error: Couldn't open file '"..profname.."'. Profiles.xml must be in the same folder as RegisteredUsers.xml" end

	local pxml = p:read("*a")
	p:close()

	--// Get profilenames
	local profnames = {}
	if string.find(pxml, "<Profiles>") then
		for w in string.gmatch(pxml, "<Name>(.-)</Name>") do
			table.insert(profnames, w)
		end
	else
		return false, "Error: File '"..profname.."' doesn't contain any profiles."
	end

	if string.find(xml, "<RegisteredUsers>") then
		for w in string.gmatch(xml, "<RegisteredUser>(.-)</RegisteredUser>") do
			local _,_,nick = w:find("<Nick>(%S+)</Nick>")
			local _,_,pass = w:find("<Password>(%S+)</Password>")
			local _,_,profile = w:find("<Profile>(%S+)</Profile>")
			if nick and pass and profile then
				nick = tFuncs:Escape_XML(nick)
				pass = tFuncs:Escape_XML(pass)
				profile = tFuncs:Escape_XML(profile)

--~ 				--// Check if profilename exists
--~ 				if not tonumber(profile) or not profnames[tonumber(profile)+1] then
--~ 					table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."'. No profilename found for profile: "..profile..".")
--~ 				end

				prof = profnames[tonumber(profile)+1]
				local prof = ProfileManager:GetName("AAAA", prof) or "Reg"
				if prof == "Bot" or prof == "User" then prof = "Reg" end
				local data = ProfileManager:GetData("AAAA", prof)
				local level = data.iLevel or 0

				local res,_,_,err = tCheck:Nick(nil, nil, nick, level)
				if not res then 
					table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : "..tostring(err))
				else
					if tNick:GetReg(nick) then
						table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : Account already exists.")
					else
						local res, err = tCheck:PasswordChars(pass)
						if not res then
							table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : "..err)
						else
							tAccounts[nick:lower()] = {
									["sNick"] = nick,
									["sProfile"] = prof,
									["sBindIP"] = "",
									["sPassword"] = pass,
									["iLastLogin"] = 0,              --// store os.time() here, unformatted for easy sorting
									["iLastSpoke"] = 0,             --// store os.time() here, unformatted for easy sorting
									["iTotalOnlineTime"] = 0,    -- total onlinetime in seconds
									["sReggedBy"] = "Ptokax converter",
									["iRegDate"] = os.time(),              --// store os.time() here, unformatted for easy sorting
									["sComment"] = "",
									["sGreeting"] = "",
							}
							table.insert(importlog.Ok, "Succesfully imported account with nickname: "..nick.."/ Pass: "..pass.." / Original profile: "..profile.." / New profile: "..prof)
						end
					end
				end
			else
				table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : Invalid details.")
			end
		end
--~ 			tFuncs:GetCtrl("ID_bu_viewlog", tCtrls["import"]):Enable()
	else
		return false, "Error: File doesn't contain any accounts."
	end

	local results = "\n"..string.rep("*",120).."\nPtokax account import results: "..filename.."\n\nDate: "..ToDate(os.time()).."\n\n\t"
	results = results..table.concat(importlog.Ok, "\n\t").."\n\n\t"
	results = results..table.concat(importlog.Failed, "\n\t").."\n\n"
	results = results.. "Imported Ptokax accounts: "..#importlog.Ok.." / Skipped: "..#importlog.Failed.."\n"..string.rep("*",120).."\n\n"

	tFuncs:WriteFile("Import", results)

	if #importlog.Ok > 0 then
		tFuncs:Save_File("./settings/HubAccounts.ini",tAccounts,"tAccounts")
		return true, "Imported Ptokax accounts: "..#importlog.Ok.." / Skipped: "..#importlog.Failed
	else
		return false, "Importing failed, skipped: "..#importlog.Failed.." accounts."
	end

end,

["Ynhub"] = function(self, filename)
	local importlog = {["Ok"] = {}, ["Failed"] = {}}
	local f,err = io.open(filename, "r")
	if f then
		local xml = f:read("*a")
		f:close()

		local accounts = {}
		if string.find(xml, "<Protecteds version") then
			for w in string.gmatch(xml, "<Protected>(.-)</Protected>") do
				local _,_,nick = w:find("<Name>(%S+)</Name>")
				local _,_,pass = w:find("<Pass>(%S+)</Pass>")
				local _,_,regdate = w:find("<Created>(.-)</Created>")
				local _,_,onlinedate = w:find("<LastUsed>(.-)</LastUsed>")
				local regdate = tFuncs:DateToEpoch(regdate) or os.time()
				local onlinedate = tFuncs:DateToEpoch(onlinedate) or 0

				if nick and pass then
					nick = tFuncs:Escape_XML(nick)
					pass = tFuncs:Escape_XML(pass)

					local res,_,_,err = tCheck:Nick(nil, nil, nick, 0)
					if not res then 
						table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : "..tostring(err))
					else
						if tNick:GetReg(nick) then
							table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : Account already exists.")
						else
							local res, err = tCheck:PasswordChars(pass)
							if not res then
								table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : "..err)
							else
								tAccounts[nick:lower()] = {
										["sNick"] = nick,
										["sProfile"] = "Reg",
										["sBindIP"] = "",
										["sPassword"] = pass,
										["iLastLogin"] = onlinedate,              --// store os.time() here, unformatted for easy sorting
										["iLastSpoke"] = 0,             --// store os.time() here, unformatted for easy sorting
										["iTotalOnlineTime"] = 0,    -- total onlinetime in seconds
										["sReggedBy"] = "Ynhub converter",
										["iRegDate"] = regdate,              --// store os.time() here, unformatted for easy sorting
										["sComment"] = "",
										["sGreeting"] = "",
								}
								table.insert(importlog.Ok, "Succesfully imported account with nickname: "..nick.."/ Pass: "..pass.." / Original profile: <self registered> / New profile: Reg")
							end
						end
					end
				else
					table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : Invalid details.")
				end
			end

		elseif string.find(xml, "<Accounts version") then

			for w in string.gmatch(xml, "<Account>(.-)</Account>") do
				local _,_,nick = w:find("<Nick>(%S+)</Nick>")
				local _,_,pass = w:find("<Pass>(%S+)</Pass>")
				local _,_,profile = w:find("<Profile>(.-)</Profile>")
				local _,_,by = w:find("<CreatedBy>(.-)</CreatedBy>")
				local _,_,regdate = w:find("<Created>(.-)</Created>")
				local _,_,onlinedate = w:find("<LastUsed>(.-)</LastUsed>")
				local _,_,speakdate = w:find("<LastSpoke>(.-)</LastSpoke>")
				local regdate = tFuncs:DateToEpoch(regdate) or os.time()
				local onlinedate = tFuncs:DateToEpoch(onlinedate) or 0
				local speakdate = tFuncs:DateToEpoch(speakdate) or 0

				if nick and pass and profile then
					nick = tFuncs:Escape_XML(nick):lower()
					pass = tFuncs:Escape_XML(pass)
					profile = tFuncs:Escape_XML(profile)

					local prof = ProfileManager:GetName("AAAA", profile) or "Reg"
					if prof == "Bot" or prof == "User" then prof = "Reg" end
					local data = ProfileManager:GetData("AAAA", prof)
					local level = data.iLevel or 0

					local res,_,_,err = tCheck:Nick(nil, nil, nick, level)
					if not res then 
						table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : "..tostring(err))
					else
						if tNick:GetReg(nick) then
							table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : Account already exists.")
						else
							local res, err = tCheck:PasswordChars(pass)
							if not res then
								table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : "..err)
							else
								tAccounts[nick:lower()] = {
										["sNick"] = nick,
										["sProfile"] = prof,
										["sBindIP"] = "",
										["sPassword"] = pass,
										["iLastLogin"] = onlinedate,              --// store os.time() here, unformatted for easy sorting
										["iLastSpoke"] = speakdate,             --// store os.time() here, unformatted for easy sorting
										["iTotalOnlineTime"] = 0,    -- total onlinetime in seconds
										["sReggedBy"] = by or "Ynhub converter",
										["iRegDate"] = regdate,              --// store os.time() here, unformatted for easy sorting
										["sComment"] = "",
										["sGreeting"] = "",
								}
								table.insert(importlog.Ok, "Succesfully imported account with nickname: "..nick.."/ Pass: "..pass.." / Original profile: "..profile.." / New profile: "..prof)
							end
						end
					end
				else
					table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : Invalid details.")
				end
			end
--~ 			tFuncs:GetCtrl("ID_bu_viewlog", tCtrls["import"]):Enable()
		else
			return false, "Error: File doesn't contain any accounts."
		end

		local results = "\n"..string.rep("*",120).."\nYnhub account import results: "..filename.."\n\nDate: "..ToDate(os.time()).."\n\n\t"
		results = results..table.concat(importlog.Ok, "\n\t").."\n\n\t"
		results = results..table.concat(importlog.Failed, "\n\t").."\n\n"
		results = results.. "Imported Ynhub accounts: "..#importlog.Ok.." / Skipped: "..#importlog.Failed.."\n"..string.rep("*",120).."\n\n"

		tFuncs:WriteFile("Import", results)

		if #importlog.Ok > 0 then
			tFuncs:Save_File("./settings/HubAccounts.ini",tAccounts,"tAccounts")
			return true, "Imported Ynhub accounts: "..#importlog.Ok.." / Skipped: "..#importlog.Failed
		else
			return false, "Importing failed, skipped: "..#importlog.Failed.." accounts."
		end

	else
		return false, "Error: Couldn't open file '"..filename.."'. Error: "..(err or "Unknown error").."."
	end
end,

["ADCH"] = function(self, filename)
	local ctime = os.time()
	local importlog = {["Ok"] = {}, ["Failed"] = {}}

	--// Get profilenames, uses ImportProfiles
	local profname = filename:gsub("users.txt", "adch-profiles.txt")
	local p,err = io.open(profname, "r")
	if not p then return false, "Error: Couldn't open file '"..profname.."'. profiles.txt must be in the same folder as users.txt" end
	p:close()

	local res, err = pcall(dofile, profname)
	if not res then
		tFuncs:Debug("File: '"..profname.."' is invalid. Error: "..tostring(err))
		return false, "Error reading file: "..profname..", error: "..tostring(err)
	else
		local count = 0
		for num, value in pairs(ImportProfiles) do
			if not ProfileManager:GetName("AAAA", value) then return false, "File: "..profname.." contains an unknown profilename: "..value..", use !showprofiles to show valid names." end
			count = count+1
			tFuncs:Debug("Detected level "..num.." with profile: "..value)
		end
		if count < 1 then return false, "File: "..profname.." doesn't contain any valid profiles." end
	end

	local f,err = io.open(filename, "r")
	if f then
		local xml = f:read("*a")
		f:close()

		local accounts = {}
		if string.find(xml, "[{.+}]") then
			for w in string.gmatch(xml, "{\".-\"}[,%]]") do
				local _,_,nick = w:find("\"nick\":\"(%S-)\"")
				local _,_,pass = w:find("\"password\":\"(%S-)\"")
				local _,_,onlinedate = w:find("\"lastofftime\":(%d+)")
				local _,_,by = w:find("\"regby\":\"(%S-)\"")
				local _,_,regdate = w:find("\"regtime\":(%d+)")
				local _,_,proflevel = w:find("\"level\":(%d+)")

				if nick and pass and proflevel then
					nick = tFuncs:Escape_XML(nick):lower()
					pass = tFuncs:Escape_XML(pass)
					profile = ImportProfiles[tonumber(proflevel)]

					local prof = ProfileManager:GetName("AAAA", profile) or "Reg"
					if prof == "Bot" or prof == "User" then prof = "Reg" end
					local data = ProfileManager:GetData("AAAA", prof)
					local level = data.iLevel or 0

					local res,_,_,err = tCheck:Nick(nil, nil, nick, level)
					if not res then 
						table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : "..tostring(err))
					else
						if tNick:GetReg(nick) then
							table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : Account already exists.")
						else
							local res, err = tCheck:PasswordChars(pass)
							if not res then
								table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : "..err)
							else
								tAccounts[nick:lower()] = {
										["sNick"] = nick,
										["sProfile"] = prof,
										["sBindIP"] = "",
										["sPassword"] = pass,
										["iLastLogin"] = tonumber(onlinedate) or 0,              --// store os.time() here, unformatted for easy sorting
										["iTotalOnlineTime"] = 0,    -- total onlinetime in seconds
										["sReggedBy"] = by or "ADCH converter",
										["iRegDate"] = tonumber(regdate) or ctime,              --// store os.time() here, unformatted for easy sorting
										["sComment"] = "",
										["sGreeting"] = "",
								}
								table.insert(importlog.Ok, "Succesfully imported account with nickname: "..nick.."/ Pass: "..pass.." / Original profile: "..proflevel.." / New profile: "..prof)
							end
						end
					end
				else
					table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick)..", pass: "..tostring(pass)..", profile: "..tostring(prof))
				end
			end
		else
			return false, "Error: File doesn't contain any accounts."
		end

		local results = "\n"..string.rep("*",120).."\nADCH account import results: "..filename.."\n\nDate: "..ToDate(os.time()).."\n\n\t"
		results = results..table.concat(importlog.Ok, "\n\t").."\n\n\t"
		results = results..table.concat(importlog.Failed, "\n\t").."\n\n"
		results = results.. "Imported ADCH accounts: "..#importlog.Ok.." / Skipped: "..#importlog.Failed.."\n"..string.rep("*",120).."\n\n"

		tFuncs:WriteFile("Import", results)

		if #importlog.Ok > 0 then
			tFuncs:Save_File("./settings/HubAccounts.ini",tAccounts,"tAccounts")
			return true, "Imported ADCH accounts: "..#importlog.Ok.." / Skipped: "..#importlog.Failed
		else
			return false, "Importing failed, skipped: "..#importlog.Failed.." accounts."
		end

	else
		return false, "Error: Couldn't open file '"..filename.."'. Error: "..(err or "Unknown error").."."
	end
end,

["Aquila"] = function(self, filename)
	if not tSecurity.General.AllowBlankPass then
		return false, "Because the real passwords can not be retreived from Aquila you must allow the use of blank passwords by setting the option !set pass_allowblank true before you can import accounts."
	end
	local ctime = os.time()
	--[[ // aquila accountformat :
	<Account>
		<Name>Nick</Name>
		<Passwd>VSuNaomIcm.yU</Passwd>
		<Rights>None</Rights>
		<Group>friends</Group>
		<Creator>1d.bot</Creator>
		<RegDate>1203717080</RegDate>
		<LastLogin>1212458446</LastLogin>
		<LastIP>nnn.nnn.nnn.nnn</LastIP>
	</Account>
	--]]

	--// Get profilenames, uses ImportProfiles
	local profname = "aquila-profiles.txt"
	local p, err = io.open(_APPPATH.."import/"..profname, "r")
	if not p then return false, "Error: Couldn't open the file '"..profname.."'. It must be in the import folder for the import to succeed" end
	p:close()

	local res, err = pcall(dofile, _APPPATH.."import/"..profname)
	if not res then
		tFuncs:Debug("File: '"..profname.."' is invalid. Error: "..tostring(err))
		return false, "Error reading file: "..profname..", error: "..tostring(err)
	else
		local count = 0
		for num, value in pairs(ImportProfiles) do
			if not ProfileManager:GetName("AAAA", value) then
				tFuncs:Debug("Detected group "..num.." with profile: "..value.." that does NOT exist in the hub.")
				--return false, "File: "..profname.." contains an unknown profilename: "..value..", use !showprofiles to show valid names." 
			else
				count = count+1
				tFuncs:Debug("Detected group "..num.." with matching profile: "..value)
			end
		end
		if count < 1 then return false, "File: "..profname.." doesn't contain any valid profiles." end
	end

	local importlog = {["Ok"] = {}, ["Failed"] = {}}
	local f, err = io.open(filename, "r")
	if f then
		local xml = f:read("*a")
		f:close()

		local accounts = {}
		if string.find(xml, "<Aquila>") then
			for w in string.gmatch(xml, "<Account>(.-)</Account>") do
				local nick = w:match("<Name>(%S+)</Name>")
				local regdate = w:match("<RegDate>(.-)</RegDate>") or ctime
				local onlinedate = w:match("<LastLogin>(.-)</LastLogin>") or 0

				local lastip = w:match("<LastIP>(.-)</LastIP>")
				local profile = w:match("<Group>(.-)</Group>") or "Reg"
				local reggedby = w:match("<Creator>(%S+)</Creator>") or ""

				if nick and lastip then
					nick = tFuncs:Escape_XML(nick)
					local lnick = nick:lower()
					profile = tFuncs:Escape_XML(profile)
					reggedby = tFuncs:Escape_XML(reggedby)

					local trprofile = ImportProfiles[profile] or profile
					local prof = ProfileManager:GetName("AAAA", trprofile) or "Reg"

					local res,_,_,err = tCheck:Nick(nil, nil, nick, 0)
					if not res then 
						table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : "..tostring(err))
					else
						if tNick:GetReg(nick) then
							table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : Account already exists.")
						else
							tAccounts[lnick] = {
								["sNick"] = nick,
								["sProfile"] = prof,
								["sBindIP"] = lastip,
								["sPassword"] = "",
								["iLastLogin"] = tonumber(onlinedate) or 0,
								["iLastSpoke"] = 0,
								["iTotalOnlineTime"] = 0,
								["sReggedBy"] = reggedby or "Aquila importer",
								["iRegDate"] = tonumber(regdate) or ctime,
								["sComment"] = "",
								["sGreeting"] = "",
							}
							table.insert(importlog.Ok, "Succesfully imported account with nickname: "..nick.."/ BindIP: "..lastip.." / Original profile: "..profile.." / New profile: "..prof)
						end
					end
				else
					table.insert(importlog.Failed, "Failed to import account with nickname '"..tostring(nick).."' : Invalid details.")
				end
			end
		else
			return false, "Error: File is not an Aquila setting xml, or doesn't contain any accounts."
		end

		local results = "\n"..string.rep("*",120).."\nAquila account import results: "..filename.."\n\nDate: "..ToDate(ctime).."\n\n\t"
		..table.concat(importlog.Ok, "\n\t").."\n\n\t"
		..table.concat(importlog.Failed, "\n\t").."\n\n"
		.."Imported Aquila accounts: "..#importlog.Ok.." / Skipped: "..#importlog.Failed.."\n"..string.rep("*",120).."\n\n"

		tFuncs:WriteFile("Import", results)

		if #importlog.Ok > 0 then
		--	tFuncs:Save_File("./settings/HubAccounts.ini",tAccounts,"tAccounts")
			return true, "Imported Aquila accounts: "..#importlog.Ok.." / Skipped: "..#importlog.Failed
		else
			return false, "Importing failed, skipped: "..#importlog.Failed.." accounts."
		end

	else
		return false, "Error: Couldn't open file '"..filename.."'. Error: "..(err or "Unknown error").."."
	end
end,

}