--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tInfo = {

["ToClient"] = {
	--// Both ADC and NMDC
	["DE"] = "sDescription", ["SS"] = "iShareSize", ["EM"] = "sEmail", ["SL"] = "iSlots", 
	["PD"] = "sPID", ["ID"] = "sCID", ["SF"] = "iSharedFiles", ["HN"] = "iNormalHubs",  ["HR"] = "iRegHubs" , ["HO"] = "iOpHubs",
	["US"] = "iLlimit",
	
	--// Special NMDC cases
	["NAP"] = "sMode", ["NCO"] = "sConnection", ["NVE"] = "sClientVersion", ["NCL"] = "sClient",  ["NHU"] = "iHubs", ["NTA"] = "sTag",
},

["Required"] = {
	["PD"] = true, ["ID"] = true, ["NI"] = true, ["VE"] = true, ["HN"] = true, ["HR"] = true, ["HO"] = true, ["SS"] = true, ["SL"] = true,
},

["Traffic"] = {
	["DE"] = "DE",
	["VE"] = "TA", ["HN"] = "TA", ["HR"] = "TA", ["HO"] = "TA", ["U4"] = "TA", ["SL"] = "TA",
	["US"] = "CO", ["DS"] = "CO", ["AS"] = "CO", ["AW"] = "CO", 
	["SS"] = "SS", ["SF"] = "SS",
	["EM"] = "EM",
},

["Allowed"] = {
	["LoginOnlyFiltered"] = { ["PD"] = true, ["RF"] = true,},
	["LoginOnly"] = { ["ID"] = true, ["NI"] = true, ["VE"] = true, ["AP"] = true,},
	["Illegal"] = { ["CT"] = true, ["I6"] = true, ["U6"] = true, ["HU"] = true, ["HI"]= true},
	["Ignored"] = { ["TA"] = true, ["I4"] = true},
	["Always"] = { 
			["U4"] = true, ["SS"] = true, ["SF"] = true, ["US"] = true, ["DS"] = true, ["SL"] = true, ["AS"] = true, ["SF"] = true, ["FS"] = true, ["KP"] = true,
			["AM"] = true, ["EM"] = true, ["DE"] = true, ["HN"] = true, ["HR"] = true, ["HO"] = true, ["TO"] = true, ["AW"] = true, ["SU"] = true, ["RF"] = true, ["LC"] = true,
	},
	["Support"] = { ["ADC0"] = true, ["TCP4"] = true, ["UDP4"] = true, ["NAT0"] = true,  ["NATT"] = true, ["SEGA"] = true,},
},

["ParseNMDC"] = function(self, client, sid, msg)
	local o = client.tInfo.tOrig
	local f = client.tInfo.tFiltered
	local n = {}

--~	tFuncs:Debug("NMDC Parsing started on: "..msg, "FI" , -1)

	if not client.tInfo.sOrig or client.tInfo.sOrig ~= msg then
		client.tInfo.tNMDC.sOrig = msg

		--// Nick, Fulldescription, Connection, Email and Sharesize
		local tempdesc = false
		n.NI,tempdesc,n.NCO,n.EM,n.SS = msg:match("%$MyINFO %$ALL (%S+) (.-)%$ %$(.-)%$(.-)%$(%d*)%$")
		if not tempdesc then return false, client, "Missing or malformed data in $MyINFO: "..msg end
		if n.NI ~= client.sNick then return false, client , "Nick spoofing detected: "..msg end

		--// Check for control chars
		if not tCheck:ControlChars(client, sid, n.NI, nil, true) then return false, client end
		if tempdesc then _,tempdesc = tCheck:ControlChars(client, sid, tempdesc) end
		if n.EM and not tCheck:ControlChars(client, sid, n.EM, nil, true) then return false, client end
		if n.SS and not tCheck:ControlChars(client, sid, n.SS, nil, true) then return false, client end

		--// Tag and description
		local temptag
		n.DE,n.NTA = tempdesc:match("(.*)(<.*>)$")
		if not n.NTA then 
			if not tUser:CanOverride(sid, "checkmyinfo") then
				tUser:Disconnect(sid, wrong,  "Invalid tag in MyINFO: "..msg) 
				return false, client
			else 
				n.DE = n.DE or "" 
				n.NTA = ""
			end
		end

		--// Disect the tag
		n.VE,n.NAP,n.NHU,n.NTL = n.NTA:match("<(.*),M:(%S),H:(.-),(.*)>")
		if not n.NHU then
			if not tUser:CanOverride(sid, "checkmyinfo") then
				tUser:Disconnect(sid, wrong,  "Invalid info in tag in MyINFO: "..msg) 
				return false, client
			else 
				n.VE = n.VE or ""
				n.NAP = n.NAP or "P"
				n.NHU = n.NHU or "0"
				n.NTL = n.NTL or ""
			end
		end
		n.NCL = n.VE:match("(%S+)") or ""
		n.NVE = n.VE:match("%S+ (.*)") or ""

		--// Hubcount
		n.HN, n.HR, n.HO = n.NHU:match("(%d+)/(%d+)/(%d+)")
		n.HN = n.HN and tonumber(n.HN) or tonumber(n.NHU) or 0
		n.HR = n.HR and tonumber(n.HR) or 0
		n.HO = n.HO and tonumber(n.HO) or 0
		n.NHU = n.HN + n.HR + n.HO

		--// Sharesize
		if not tonumber(n.SS) then
			if not tUser:CanOverride(sid, "checkmyinfo") then
				tUser:Disconnect(sid, wrong,  "Invalid sharesize in MyINFO: "..msg) 
				return false, client
			else
				n.SS = 0
			end
		end
		n.SS = tonumber(n.SS)

		--// Slots and limits
		n.SL = n.NTL:match("S:(%d+)") or "0"
		n.NLI = n.NTL:match("L:(%d+)")
		if n.NLI then 
			n.US = tonumber(n.NLI)*1024 
		else 
			o.US = ""
			f.US = ""
		end

		local checkpatterns = {["DE"] = "DE", ["EM"] = "EM", ["VE"] = "VE", ["SS"] = "SS" }
		local changed = false

		--// Process changes
		for name, val in pairs(n) do
			if val and o[name] and o[name] == val then
				n[name] = nil		--// Remove existing same values from n
			elseif val then
				--// Store in orig
				o[name] = val
				changed = true

				--// Value changed: perform patterncheck
				local res = true
				if checkpatterns[name] then
					res, val = tCheck:Pattern(sid, name, val)
					if not res then changed = false end
					if tBuffer[sid].bDisconnect then break end
				end

				if res then
					--//  Store in tFiltered
					if name == "SS" then
						if client.bShareHidden or (tHubSettings.tRules.bChatOnly and tHubSettings.tRules.bChatOnlyHideShare) then
							val = 0
						end
						f[name] = val
					elseif name == "NAP" then
						client.sMode = val
						if val == "A" then 
							client.bActive = true 
							client.bProxy = false 
						elseif val == "5" then
							client.bActive = false
							client.bProxy = true
						else
							client.bActive = false
							client.bProxy = false 
						end
						f[name] = val
					else
						f[name] = val
					end

					--// Store changed values in client
					local varname = tInfo.ToClient[name]
					if varname then
						local stype = varname:sub(1,1)
						if stype == "i" then
							client[varname] = tonumber(f[name] or 0) or 0
						elseif stype == "s" then
							client[varname] = tostring(f[name] or "")
						end
					end

				else
					--// Pattern was blocked
				end
			end
		end

		--// Update info strings
		if changed then
			--// Store changes (incrementally, tNew can already contain changes since last MyInfoPool send)
			client.tInfo.tNew = n

			--// Create NMDC $MyINFO strings
			tInfo:CreateNMDC(client, sid)

			--// Create ADC INF strings
			tInfo:CreateADC(client, sid)

			client.tInfo.tNew = {}

--~ 			tFuncs:Debug("NMDC Parsed.", "FI" , -1)
		else
			--// Clear changes
			client.tInfo.tNew = {}
		end
	else
		--// MyINFO didn't change, do nothing
	end

	if not tBuffer[sid].bDisconnect then
		return true, client, n
	else 
		return false, client
	end
end,

["CreateNMDC"] = function(self, client, sid)
	--// Create NMDC strings	--// Generate short NMDC MyINFO
	local f = client.tInfo.tFiltered
	local t = tHubSettings.tTraffic.tShow

	--// Show sharesize 0 to users if protocol isn't NMDC
	local add = ((tHubSettings.tRules.bAllowADC and tHubSettings.tRules.bAllowNMDC) or tHubSettings.tTraffic.tShow.Protocol) and (client.bBot and "" or "["..client.sProtocol..(client.bSecure and "S" or "").."] ") or ""

	client.tInfo.tNMDC.sOperator = "$MyINFO $ALL "..NMDC_E(client.sNick).." "..add..NMDC_E(f.DE or "")..NMDC_E(f.NTA or "").."$ $"..NMDC_E(f.NCO or "\1").."$"..NMDC_E(f.EM or "").."$"..(f.SS or "0").."$"

	local de = NMDC_E((t.DE and f.DE) or (client.bOperator and t.FullOp and f.DE) or (client.bBot and t.FullBot and f.DE) or "")
	local ta = NMDC_E((t.TA and f.NTA) or (client.bOperator and t.FullOp and f.NTA) or (client.bBot and t.FullBot and f.NTA) or "")
	local co = NMDC_E((t.CO and f.NCO)or (client.bOperator and t.FullOp and f.NCO) or (client.bBot and t.FullBot and f.NCO) or "\1")
	local em = NMDC_E((t.EM and f.EM) or (client.bOperator and t.FullOp and f.NTA) or (client.bBot and t.FullBot and f.NTA) or "")

	local sharesize = "0"
	if client.sProtocol == "NMDC" then sharesize = f.SS or "0" end
	local ss = NMDC_E((t.SS and sharesize) or (client.bOperator and t.FullOp and sharesize) or (client.bBot and t.FullBot and sharesize) or "0")

	client.tInfo.tNMDC.sUser = "$MyINFO $ALL "..NMDC_E(client.sNick).." "..add..de..ta.."$ $"..co.."$"..em.."$"..ss.."$"

--~ 	tFuncs:Debug("NMDC Strings created: \nUser: "..tostring(client.tInfo.tNMDC.sUser).."\nOperator: "..tostring(client.tInfo.tNMDC.sOperator), "FI", -1)
end,

["CreateADC"] = function(self, client, sid)
	--// Create ADC strings
	local n = client.tInfo.tNew
	local f = client.tInfo.tFiltered
	local o = client.tInfo.tOrig
	local t = tHubSettings.tTraffic.tShow

	--// Create full INF
 	local cid = client.sCID

	--// Key
	local ct = ""
	if f.CT and f.CT > 0 then ct = " CT"..f.CT end
	if f.CT and f.CT > 3 then ct = ct.." OP1" end

	local add = ((tHubSettings.tRules.bAllowADC and tHubSettings.tRules.bAllowNMDC) or tHubSettings.tTraffic.tShow.Protocol) and (client.bBot and "" or "["..client.sProtocol..(client.bSecure and "S" or "").."] ") or ""
	local base = "BINF "..sid.." ID"..cid..ct.." NI"..ADC_E(client.sNick).." DE"..ADC_E(add..(f.DE or ""))
	local sUser, sOperator, sGui = base, base, base

	local filter = {["ID"] = true, ["CT"] = true, ["OP"] = true, ["DE"] = true, ["NI"] = true}
	for name, value in pairs(o) do
		if #name == 2 and not filter[name] and value ~= "" then		--// Exclude special NMDC flags with 3 chars
			sGui = sGui.." "..name..ADC_E(o[name])
			local fvalue = f[name]
			if fvalue then
				sOperator = sOperator.." "..name..ADC_E(fvalue)
				local tname = tInfo.Traffic[name]	--// Should value be available to users (traffic-reduction rules)
				if not tname or t[tname] or type(t[tname]) ~= "boolean" or (client.bOperator and t.FullOp) or (client.bBot and t.FullBot) then
					if name == "SS" and client.sProtocol ~= "ADC" then fvalue = 0 end
					sUser = sUser.." "..name..ADC_E(fvalue)
				end
			end
		end
	end

	client.tInfo.tADC.sUser = sUser
	client.tInfo.tADC.sOperator = sOperator
	client.tInfo.tADC.sGui = sGui

--~ 	tFuncs:Debug("ADC Strings created: \nUser: "..tostring(client.tInfo.tADC.sUser).."\nOperator: "..tostring(client.tInfo.tADC.sOperator).."\nGui: "..tostring(client.tInfo.tADC.sGui), "FI", -1)

	--// Create update INF:
	if client.iState == 7 then
		local ou = client.tInfo.tADC.tOperatorUpdate
		local uu = client.tInfo.tADC.tUserUpdate
		local gu = client.tInfo.tADC.tGuiUpdate

		for name, val in pairs(n) do
			if #name == 2 and name ~= "ID" then		--// Exclude special NMDC flags with 3 chars
				gu[name] = o[name]	--// Add data to tGuiUpdate
				if f[name] then
					local value = f[name]

					--// Add protocol to description
					if name == "DE" and add ~= "" then value = add.." "..value end
--~ 					if name== "DE" then value = (client.bBot and "" or "["..client.sProtocol..(client.bSecure and "S" or "").."] ")..value end

					ou[name] = value	--// Add data to tOperatorUpdate
					local tname = tInfo.Traffic[name]
					if not tname or t[tname] or type(t[tname]) ~= "boolean" or (client.bOperator and t.FullOp) or (client.bBot and t.FullBot) then
						if name == "SS" and client.sProtocol == "NMDC" then	
							--// Skip sharesize changes of NMDC users to broadcast to ADC users
						else
							uu[name] = value	--// Add data to tUserUpdate
						end
					end
				end
			end
		end
	end
end,

["ParseADC"] = function(self, client, sid, msg, login)
	if login then client.tInfo.tADC.sOrig = msg end
	--// BINF ABCD ....

--~ 	tFuncs:Debug("ADC Parsing started on: "..msg, "FI" , -1)

	local o = client.tInfo.tOrig
	local f = client.tInfo.tFiltered
	local n = {}

	--// Store changes in n
	local retmsg = msg:sub(10)
	for field in retmsg:gmatch(" (%S+)") do
		local name = field:sub(1,2)
		local value = field:sub(3)
		n[name] = ADC_E(value, 1)			--// Store values in n
	end

	--debugging
--~ 	if login then
--~ 		client.tInfo.n = n
--~ 	end
--~ 	if not client.tInfo.Debugprocessed then client.tInfo.Debugprocessed = {} end
--~ 	table.insert(client.tInfo.Debugprocessed, "\n*** Running PARSEADC: ")

	local checkpatterns = {["DE"] = "DE", ["EM"] = "EM", ["VE"] = "VE", ["SS"] = "SS" }
	local changed = false
	local delete = {}

	--// Process changes
	for name, val in pairs(n) do

		--// debugging:
--~ 		table.insert(client.tInfo.Debugprocessed, name.." / "..tostring(val))
--~ 		client.tInfo.lastprocessed = "Last processed: "..name.." / "..tostring(val)

		if o[name] and o[name] == val then
			--// Mark for deletion
			table.insert(delete, name)
		elseif #name == 2 then		--// Skip NMDC special names
			--// Store in orig
			o[name] = val
			changed = true

			--// Value changed: perform patterncheck
			local res = true
			if checkpatterns[name] then
				res, val = tCheck:Pattern(sid, checkpatterns[name], val)
				if not res then changed = false end
				if tBuffer[sid].bDisconnect then break end
			end

			if res then
				--// Store changed values in filtered
				if tInfo.Allowed.Always[name] or (tInfo.Allowed.LoginOnly[name] and login) then
					--// Filter support
					if name == "SU" then
--~ 						local tsup = {}
						for w in string.gmatch(val, "%w+") do
							client.tSupportsInf[w] = true
--~ 							if tInfo.Allowed.Support[w] then	-- Don't block SU fields in INF
--~ 								table.insert(tsup, w)
--~ 							end
						end
						local tsup = {}
						for name,enabled in pairs(client.tSupportsInf) do
							if enabled then table.insert(tsup, name) end
						end
						val = table.concat(tsup, ",")
						f[name] = val
--~ 					elseif name == ["I4"] then				--// External IP is now determined in User:Joined()
--~ 						if val == "0.0.0.0" then val = client.sIP end
--~ 						if tIp:IsLocal(client.sIP) then		--// Replace local IP with hub IP, use hub IP if client sends local IP, otherwise use value sent by client itself.
--~ 							val = tStats.HubIp or val or client.sIP
--~ 						else
--~ 							val = client.sIP
--~ 						end
--~ 						client.sIPEx = val
--~ 						f[name] = val
					elseif name == "SF" then
						if client.bShareHidden or (tHubSettings.tRules.bChatOnly and tHubSettings.tRules.bChatOnlyHideShare) then
							val = 0
						end
						f[name] = val
					elseif name == "SS" then
						if client.bShareHidden or (tHubSettings.tRules.bChatOnly and tHubSettings.tRules.bChatOnlyHideShare) then
							val = 0
						end
						f[name] = val
					elseif name == "U4" then
						if val ~= "" then
							client.bActive = true
							client.sMode = "A"
						else
							client.bActive = false
							client.sMode = "P"
						end
						f[name] = val
					elseif name == "LC" then
						client.sLocale = retmsg:match(" LC(%S+)")
					else
						f[name] = val
					end
					--// If tag changed:
					if tInfo.Traffic[name] and tInfo.Traffic[name] == "TA" then
						local ntag = "<"..(f.VE or "")..",M:"..(client.sMode or "P")..",H:"..(f.HN or 0).."/"..(f.HR or 0).."/"..(f.HO or 0)..",S:"..(f.SL or 0)..">"
						f.NTA = ntag
						client.sTag = ntag
						f.NHU = (f.HN or 0) + (f.HR or 0) + (f.HO or 0)
					end
				elseif tInfo.Allowed.LoginOnlyFiltered[name] then
					if login then
						f[name] = nil
					else
						tUser:Disconnect(sid, wrong, "Illegal field in BINF after login: "..name..val.." / Full BINF: "..msg) 
						return false, client
					end
				elseif tInfo.Allowed.LoginOnly[name] and not login then
					if client.tInfo.tOrig[name] and client.tInfo.tOrig[name] == val then
						tFuncs:SafeDebug(tUser:GetDataLine(sid).." sends "..name..val.." after login, it matches current "..name..", ignoring.", "X")
						return true, client
					else
						tUser:Disconnect(sid, wrong, "Illegal field in BINF after login: "..name..val.." / Full BINF: "..msg) 
						return false, client
					end
				elseif tInfo.Allowed.Illegal[name] then
					tUser:Disconnect(sid, wrong, "Illegal field in BINF: "..name..val.." / Full BINF: "..msg) 
					return false, client
				elseif tInfo.Allowed.Ignored[name] then
					--// Do nothing
				else
					--// Unknown field, ignore
					tFuncs:SafeDebug(tUser:GetDataLine(sid).."Unknown field in BINF, ignored: "..name..val, "FI" , 1)
				end

				--// Store changed values in client
				local varname = tInfo.ToClient[name]
				if varname then
					local stype = varname:sub(1,1)
					if stype == "i" then
						client[varname] = tonumber(val or 0) or 0
					elseif stype == "s" then
						client[varname] = tostring(val)
					end
				end
			else
				--// Value was blocked by patterncheck
			end
		end
	end

	--// Delete same values 
	for _,name in ipairs(delete) do
		n[name] = nil		--// Remove existing same values from n
	end

	--// Update info strings
	if changed then
		client.tInfo.tNew = n

		--// Create NMDC $MyINFO strings
		tInfo:CreateNMDC(client, sid)

		--// Create ADC INF strings
		tInfo:CreateADC(client, sid)

		client.tInfo.tNew = {}

--~ 		tFuncs:Debug("ADC Parsed.", "FI" , -1)

	else
		client.tInfo.tNew = {}
	end

	--// debugging:
--~ 	client.tInfo.n2 = n

	if not tBuffer[sid].bDisconnect then
		return true, client, n
	else 
		return false, client
	end
end,

["SetCT"] = function(self, client, sid)
	--// Calculate CT
	local ct = 0

	--// Is it a bot?
	if client.bBot then ct = ct+1 end

	--// Is it a registered user?
	if (client.iProfile or -1) > -1 then ct = ct + 2 end
	--// Does the user have a key?
	if client.bHasKey then 
		if (client.iProfile or -1) == 10 then 
			ct = ct + 16 
		else
			ct = ct + 4
		end
	end

	--// Finalize
	client.tInfo.tOrig.CT = ct
	client.tInfo.tFiltered.CT = ct

	return ct
end,

}
