--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

ProfileManager = {
	["Protected"] = { ["User"] = 1, ["Reg"] = 1, ["Gui"] = 2, ["Bot"] = 2},	--// 1 = Can't be deleted, 2 = Can't be modified (has all rights and overrides)

	["GetByLevel"] = function(self, ssid, level)
		local profs = {}
		if level then
			for profname, data in pairs(tProfiles) do
				if data.iLevel == level and profname ~= "Gui" and profname ~= "HubBot" then
					table.insert(profs, profname)
				end
			end
			--// Sort for display in Edit Master Accounts:
			table.sort(profs)
			if tonumber(level) == 10 then
				table.insert(profs, "Gui")
			end
			
			if #profs>0 then
				return profs
			else
				return false, "No profiles found for level "..level
			end
		else
			return false, "[ERROR] [ProfileManager] [GetByLevel] No profile level specified."
		end
	end,
	["GetName"] = function(self, ssid, name)
		if name then
			for profname in pairs(tProfiles) do
				if profname:lower() == name:lower() then
					return profname
				end
			end
			for profname in pairs(ProfileManager.Protected) do
				if profname:lower() == name:lower() then
					return profname
				end
			end
			return false, "Profile '"..name.."' not found."
		else
			return false, "[ERROR] [ProfileManager] [GetName] No profilename specified."
		end
	end,
	["GetData"] = function(self, ssid, name)
		local profname, err = ProfileManager:GetName(ssid, name)
		if not profname then
			return false, err
		else
			return tProfiles[profname], profname
		end
	end,
	["Add"] = function(self, ssid, newname, proflevel, source)
		--// Validation
		local sourcename = ProfileManager:GetName(ssid, (source or "Reg"))
		local proflevel = tonumber(proflevel or 0) or 0
		if proflevel < -1 or proflevel > 10 then
			return false, "Invalid profilelevel, use a value between -1 and 10."
		end

		local existingname = ProfileManager:GetName(ssid, newname)
		if not existingname and sourcename then
			tProfiles[newname] = tFuncs:TableCopy(tProfiles[sourcename])
			tProfiles[newname].iLevel = proflevel
			tProfiles[newname].sDescription = ""
			tFuncs:SaveSettings("profiles")
			return true, "Profile '"..newname.."' with level "..tProfiles[newname].iLevel.." created with template settings from profile '"..sourcename.."'."
		elseif existingname then 
			if existingname == "Bot" then
				return false, "Profile '"..existingname.."' is a protected internal profile, you can't use this name."
			else
				return false, "Profile '"..existingname.."' already exists."
			end
		elseif not sourcename then 
			return false, "Profile '"..source.."', to use as template for the new profile, doesn't exist. Choose an existing profile as template" 
		end
	end,
	["Delete"] = function(self, ssid, profilename, replacename)
		local replacename = replacename or "Reg"
		replacename = ProfileManager:GetName(ssid, replacename)
		if not replacename then return false, "Profile '"..replacename.."' to use for accounts registered with profile '"..profilename.."' doesn't exist." end

		local profname = ProfileManager:GetName(ssid, profilename) or false
		local output = ""
		
		if profname and replacename and profname ~= replacename then
			if ProfileManager.Protected[profname] and ProfileManager.Protected[profname] > 0 then
				return false, "This profile is protected and can't be deleted."
			end
		
			for vnick, data in pairs(tAccounts) do
				if data.sProfile == profname then
					local res, line = AccountManager:SetProfile(ssid, vnick, replacename)
					output = "\n\t"..output..line
				end
			end
			tFuncs:SaveSettings("accounts")
			tProfiles[profname] = nil
			tFuncs:SaveSettings("profiles")
			return true, "Profile '"..profname.."' deleted, users with that profile have been moved to profile '"..replacename.."'."..output
		elseif not profname then
			return false, "Profile '"..profilename.."' does not exist."
		elseif profname == replacename then
			return false, "Can't use the same profile to replace the current profile."
		else
			return false, "Unknown error occured, profile not deleted."
		end
	end,
	["Set"] = function(self, ssid, profilename, name, value)
		local name = name:lower()
		local profname, err = ProfileManager:GetName(ssid, profilename)
		local prof, err = ProfileManager:GetData(ssid, profilename)

		if err and not prof then
			return false, Lang(ssid, 0, "You can't change this profile.")
		elseif not prof then
			return false, err	-- huh, how does this work, there are two err's above? fredo
		end

		if not ProfileManager.Settings[name] then	-- added check in modprofiles with proper return message. not sure if its called from somewhere else so doing check here too :P
			return false, "No valid settingname specified."
		else
			local test, val = ProfileManager.Settings[name](ssid, profname, prof, name, value)
			if not test then
				return false, val
			else
				tFuncs:SaveSettings("profiles")
				return true, "Profile '"..profname.."' setting changed: "..name.." = "..tostring(val)
			end
		end
	end,

	["Settings"] = {
		["operator"] = function(ssid, profname, prof, name, value)
			local value = tFuncs:TrueOrFalse(value)
			prof.bOperator = value

			--// Update current users with this profile
			for sid, client in pairs(tClient) do
				if client.sProfile == profname and client.iState == 7 then
					tSend:UpdateRooms(sid)
				end
			end
			return true, value
		end,
		["level"] = function(ssid, profname, prof, name, value)
			if ProfileManager.Protected[profname] then
				return false, "This profile has a fixed level: "..prof.iLevel ..", it can not be changed"
			end
			local value = tonumber(value)
			if value and value >=-1 and value <= 10 then
				prof.iLevel = value

				--// Update current users with this profile
				for sid, data in pairs(tClient) do
					if data.sProfile == profname and data.iState == 7 then
						data.iLevel = value
						tSend:UpdateRooms(sid)
					end
				end

				--// rebuild UserCommand rightclicks
				tUserCmds:Build()
				return true, value
			else
				return false, "Invalid profile level specified, use a number between -1 and 10"
			end
		end,
		["key"] = function(ssid, profname, prof, name, value)
			local value = tFuncs:TrueOrFalse(value)
			prof.bHasKey = value

			--// Update current users with this profile
			for sid, client in pairs(tClient) do
				if client.sProfile == profname and client.iState == 7 then
					client.bHasKey = value
					tInfo:SetCT(client, sid)
					tInfo:CreateADC(client, sid)
					tInfo:CreateNMDC(client, sid)
					tUser:Quit(sid)
					tUser:Announce(sid)
					tSend:OwnKey(sid, true)
				end
			end
			return true, value
		end,
		["fakekey"] = function(ssid, profname, prof, name, value)
			local value = tFuncs:TrueOrFalse(value)
			prof.bHasFakeKey = value

			--// Update current users with this profile
			for sid, client in pairs(tClient) do
				if client.sProfile == profname and client.iState == 7 then
					client.bHasFakeKey = value
					tUser:Quit(sid)
					tUser:Announce(sid)
					tSend:OwnKey(sid, true)
				end
			end
			return true, value
		end,
		["desc"] = function(ssid, profname, prof, name, value)
			local value = value and tostring(value) or ""
			prof.sDescription = value
			return true, value
		end,
		["message"] = function(ssid, profname, prof, name, value)
			local value = value and tostring(value) or ""
			prof.sMessage = value
			return true, value
		end,
		["hideshare"] = function(ssid, profname, prof, name, value)
			local value = tFuncs:TrueOrFalse(value)
			prof.bShareHiddenForced = value

			--// Update current users with this profile
			for sid, client in pairs(tClient) do
				if client.sProfile == profname and client.iState == 7 then
					client.bShareHidden = value
					if value then
						client.tInfo.tNew = { ["SS"] = 0, ["FS"] = 0,}
						client.tInfo.tFiltered.SS = 0
						client.tInfo.tFiltered.FS = 0
					else
						local ss = client.tInfo.tOrig.SS or 0
						client.tInfo.tNew = { ["SS"] = ss}
						client.tInfo.tFiltered.SS = ss
					end

					tInfo:CreateADC(client, sid)
					tInfo:CreateNMDC(client, sid)
					client.tInfo.tNew = {}
					tSend:MyInfUpdate(sid)
				end
			end
			return true, value
		end,
		["minshare"] = function(ssid, profname, prof, name, value)
			local value = value or false
			if not value or value == "false" then
				value = false
			else
				value = tFuncs:ConvertToBytes(value)
				if not value then
					return false, "Invalid value for '"..name.."'"
				end
			end
			if not prof.tRules then prof.tRules = {} end
			prof.tRules.iMinShare = value
			return true, value
		end,
	--	["patternoverride"] = function(ssid, profname, prof, name, value)
	--		local value = tFuncs:TrueOrFalse(value)
	--		prof.bPatternOverride = value
	--		return true, value
	--	end,
		--[""] = function(ssid, profname, prof, name, value)
		--	return
		--end,
	},

	["Allow"] = function(self, ssid, profilename, cmdname, value)
		local cmdname = cmdname:lower()
		local profname, err = ProfileManager:GetName(ssid, profilename)
		if not profname then return false, err end

		if tProfiles[profname].iLevel == 10 then
			return false, "Profiles with level 10 always have access to all hubcommands."
		end

		if not value and type(value) == "Boolean" then
			value = nil --// nil instead of false will remove the entry from ini-file, but will act the same
		elseif not value and type(value) ~= "Boolean" then
			value = true
		elseif value and (value == "false" or value == "0" or value == 0) then
			value = nil	--// nil instead of false will remove the entry from ini-file, but will act the same
		elseif value and (value == "true" or value == "1" or value == 1) then
			value = true
		else
			return false,  "Invalid value, use !allow <profilename> <cmdname> <value>\t\tAccepted values: 1 or 0, true or false, giving no value will default to allow"
		end

		if ProfileManager.Protected[profname] and ProfileManager.Protected[profname] > 1 then
			return false, "This profile is protected and can't be modified."
		end

		if not tProfiles[profname].tCmds then tProfiles[profname].tCmds = {} end

		local UpdateRooms = function(cmdname, profname)
			if cmdname == "opchat" or cmdname == "feed" or cmdname == "all" or cmdname == "none" then
				for sid, data in pairs(tClient) do
					if data.sProfile == profname then
						tSend:UpdateRooms(sid)
					end
				end
			end
		end

		local AllowProfile = function(cmdname, value) 
			if HubCmd[cmdname] or SubSet.ExtraAllows[cmdname] or cmdname == "all" or cmdname == "none" then
				if tProfiles[profname].tCmds.all and cmdname ~= "all" and not value then
					--// Build commandlist with all allowed
					local cmdlist = {}
					for _, name in ipairs(SubSet.ExtraAllows) do
						table.insert(cmdlist, name)
					end
					for name, func in pairs(HubCmd) do
						table.insert(cmdlist, name)
					end		
					table.sort(cmdlist)
					tProfiles[profname].tCmds = {}
					for _, name in ipairs(cmdlist) do
						tProfiles[profname].tCmds[name] = true
					end
					tProfiles[profname].tCmds[cmdname] = nil
					UpdateRooms(cmdname, profname)
					return true, "Rights changed: profile '"..profname.."' is now not allowed to use: "..cmdname
				elseif cmdname == "all" and value then
					if tProfiles[profname].tCmds.all then
						return true, "Profile '"..profname.."' is already allowed to use all commands."
					else
						tProfiles[profname].tCmds = {}
						tProfiles[profname].tCmds.all = true
						UpdateRooms(cmdname, profname)
						return true, "Profile '"..profname.."' is now allowed to use all hubcommands, including setcommands!"
					end
				elseif (cmdname == "all" and not value) or cmdname == "none" then
					if tProfiles[profname].tCmds ~= {} then
						tProfiles[profname].tCmds = {}
						UpdateRooms(cmdname, profname)
						return true, "Profile '"..profname.."' now can't use any hubcommands."
					else
						return true, "Profile '"..profname.."' wasn't allowed to use any commands."
					end
				elseif value then
					if tProfiles[profname].tCmds[cmdname] then
						return true, "Profile '"..profname.."' is already allowed to use: "..cmdname
					else
						tProfiles[profname].tCmds[cmdname] = true
						UpdateRooms(cmdname, profname)
						return true, "Profile '"..profname.."' is now allowed to use: "..cmdname
					end
				elseif not value  then
					if not tProfiles[profname].tCmds[cmdname] then
						return true, "Profile '"..profname.."' wasn't allowed to use: "..cmdname
					else
						tProfiles[profname].tCmds[cmdname] = nil
						UpdateRooms(cmdname, profname)
						return true, "Profile '"..profname.."' is now not allowed to use: "..cmdname
					end
				else
					return false, "Unknown error."
				end
			else
				return false, "Command '"..cmdname.."' doesn't exist."
			end
		end

		local cmdnames = tFuncs:Split(cmdname, ",")
		local result, ok = "", true
		for _,cmdname in ipairs(cmdnames) do
			local res, err = AllowProfile(cmdname, value)
			result = result..err.."\n"
			if not res then ok = false end
		end
		result = result:gsub("\n$", "")	--// Remove the last newline

		tFuncs:SaveSettings("profiles")

		--// rebuild UserCommand rightclicks
		tUserCmds:Build()

		return ok, result
	end,
	["Override"] = function(self, ssid, profilename, cmdname, value)
		local cmdname = cmdname:lower()
		local profname, err = ProfileManager:GetName(ssid, profilename)
		if not profname then return false, err end

		if not value and type(value) == "Boolean" then
			value = nil
		elseif not value and type(value) ~= "Boolean" then
			value = true
		elseif value and (value == "false" or value == "0" or value == 0) then
			value = nil
		elseif value and (value == "true" or value == "1" or value == 1) then
			value = true
		else
			return false,  "Invalid value, use !override <profilename> <cmdname> <value>\t\tAccepted values: 1 or 0, true or false, giving no value will default to allow"
		end

		if ProfileManager.Protected[profname] and ProfileManager.Protected[profname] > 1 then
			return false, "This profile is protected and can't be modified."
		end

		if not tProfiles[profname].tOverrides then tProfiles[profname].tOverrides = {} end

		local OverrideProfile = function(cmdname, value)
			local cmdfound = false
			for _,name in pairs(SubSet.Overrides) do
				if cmdname == name then cmdfound = true break end
			end
			if not cmdfound and cmdname ~= "all" and cmdname ~= "none" then 
				return false, "Override '"..cmdname.."' doesn't exist."
			end

			if tProfiles[profname].tOverrides.all and cmdname ~= "all" and not value then
				--// Build commandlist with all allowed
				local overridelist = {}
				for _, name in ipairs(SubSet.Overrides) do
					table.insert(overridelist, name)
				end
				table.sort(overridelist)
				tProfiles[profname].tOverrides = {}
				for _, name in ipairs(overridelist) do
					tProfiles[profname].tOverrides[name] = true
				end
				tProfiles[profname].tOverrides[cmdname] = nil
				return true, "Rights changed: profile '"..profname.."' is now not allowed to override: "..cmdname	
			elseif cmdname == "all" and value then
				if tProfiles[profname].tOverrides.all then
					return true, "Profile '"..profname.."' already has all overrides."
				else
					tProfiles[profname].tOverrides = {}
					tProfiles[profname].tOverrides.all = true
					tFuncs:SaveSettings("profiles")
					return true, "Profile '"..profname.."' is now allowed to override all checks!"
				end
			elseif (cmdname == "all" and not value) or cmdname == "none" then
				if tProfiles[profname].tOverrides ~= {} then
					tProfiles[profname].tOverrides = {}
					tFuncs:SaveSettings("profiles")
					return true, "Profile '"..profname.."' is now not allowed to override any checks."
				else
					return true, "Profile '"..profname.."' didn't have any overrides"
				end
			elseif value then
				if tProfiles[profname].tOverrides[cmdname] then
					return true, "Profile '"..profname.."' already has override on: "..cmdname
				else
					tProfiles[profname].tOverrides[cmdname] = true
					tFuncs:SaveSettings("profiles")
					return true, "Profile '"..profname.."' is now allowed to override: "..cmdname
				end
			elseif not value then
				if not tProfiles[profname].tOverrides[cmdname] then
					return true, "Profile '"..profname.."' didn't have override on: "..cmdname
				else
					tProfiles[profname].tOverrides[cmdname] = nil
					tFuncs:SaveSettings("profiles")
					return true, "Profile '"..profname.."' is now not allowed to override: "..cmdname
				end
			else
				return false, "Unknown error."
			end
		end

		local cmdnames = tFuncs:Split(cmdname, ",")
		local result, ok = "", true
		for _,cmdname in ipairs(cmdnames) do
			local res, err = OverrideProfile(cmdname, value)
			result = result..err.."\n"
			if not res then ok = false end
		end
		result = result:gsub("\n$", "")	--// Remove the last newline

		tFuncs:SaveSettings("profiles")
		return ok, result
	end,
}
