--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tRedirect = {

["MakeList"] = function(self)
	local names = {["rd_failover"] = "sFailOver", ["rd_adc_disabled"] = "sNoADC", ["rd_nmdc_disabled"] = "sNoNMDC", ["rd_passive"] = "sPassive", ["rd_maxhubs"] = "sMaxHubs", ["rd_minslots"] = "sMinSlots", ["rd_maxslots"] = "sMaxSlots", ["rd_minshare"] = "sMinShare", ["rd_maxshare"] = "sMaxShare", ["rd_flood"] = "sFlood", ["rd_main"] = "sMain", ["rd_regonly"] = "sRegOnly", ["rd_full"] = "sFull", ["rd_deadend"] = "sDeadEnd", ["rd_deadendfail"] = "sDeadEndFail"}
	local list = {}
	for name, varname in pairs(names) do
		local addtype = name:sub(4)
		local value = tHubSettings.tRedirects[varname]
--~ 		tFuncs:Debug("Creating list for addtype: "..addtype.." / name: "..name.." / value: "..tostring(value))
		if value and value ~= "" then 
			list[addtype] = tFuncs:Split(value, ",")
			for num,address in ipairs(list[addtype]) do
				tStats.Redirects.Addresses[address] = tStats.Redirects.Addresses[address] or { ["Offline"] = false, ["Counter"] = 0, ["LastCheck"] = false, ["LastOnline"] = false}
			end
		else
			list[addtype] = {}
		end
	end
	tRedirects.tList = list
end,

["IsOnline"] = function(self, address)
	if tStats.Redirects.Addresses[address] and tStats.Redirects.Addresses[address].Offline then 
		return false
	else 
		return true
	end
end,

["GetStatus"] = function(self, address)
	local stats = tStats.Redirects.Addresses[address]
	if stats and not stats.Offline and stats.LastCheck then
		local status = "ONLINE (since: "..(ToDate(stats.OnlineSince) or "<no data>").." / last check: "..(ToDate(stats.LastCheck) or "<no data>")..")"
		return true, status
	elseif stats and stats.Offline and stats.LastCheck then
		local status = "OFFLINE (since: "..(ToDate(stats.OfflineSince) or "<no data>").." / last check: "..(ToDate(stats.LastCheck) or "<no data>")..")"
		return true, status
	else
		return false, "Not verified"
	end
end,

["GetAddressType"] = function(self, addtype, skip)
	local address = false
	local addies = tRedirects.tList[addtype]
	local stats = tStats.Redirects.List[addtype]
	if not addies or #addies < 1 then return false, false end
	local loops = #addies
	for a=1, loops do
		stats.Number = ((stats.Number+1) <= #addies) and stats.Number+1 or 1
		address = addies[stats.Number]
--~ 		tFuncs:Debug("[Redirect] Getting address for "..addtype.." / Number: "..stats.Number.." / Address: "..address, "X")
		if skip or tHubSettings.tRedirects.sFallback == "none" or tRedirect:IsOnline(address) then
			break
		end
	end
	
	return address, true
end,

["GetAnyAddress"] = function(self)
	local names = {"main", "regonly", "minshare", "maxshare", "maxhubs", "minslots", "maxslots", "full", "passive", "flood"}

	for name in pairs(names) do
		local address = tRedirect:GetAddressType(name)
		if tHubSettings.tRedirects.sFallback == "none" or tRedirect:IsOnline(address) then 
			return address 
		end
	end
	return false
end,

["GetAddress"] = function(self, addtype, nostats)
	local request, path, address, used = addtype, {}, false, false

	if not addtype:find("%.") then
		local skip = false
		if addtype:find("deadend") then skip = true end
		address, used = tRedirect:GetAddressType(addtype, skip)
		table.insert(path, "USED:"..tostring(used).." LIST:"..tostring(address))
		if not skip and not address and used and not tRedirect:IsOnline(address) and tHubSettings.tRedirects.sFallback ~= "none" then
			if tHubSettings.tRedirects.sFallback == "main" then
				address = tRedirect:GetAddressType("main") or tRedirect:GetAddressType("main", true)
				table.insert(path, "OFFLINE FB=MAIN:"..address)
			elseif tHubSettings.tRedirects.sFallback == "any" then
				address = tRedirect:GetAnyAddress() or tRedirect:GetAddressType("main", true)
				table.insert(path, "OFFLINE FB=ANY:"..address)
			end
		end
	else
		address = addtype
		addtype = false
	end

	tFuncs:Debug("[Redirect] [GetAddress] Request: "..tostring(request).." / Path: "..table.concat(path, " / ").." / Answer: "..tostring(address), "X")

	--// Update counters
	if not nostats then
		if addtype then
			tStats.Redirects.List[addtype].Counter = tStats.Redirects.List[addtype].Counter+1
		end
		if address then
			tStats.Redirects.Addresses[address] = tStats.Redirects.Addresses[address] or { ["Offline"] = false, ["Counter"] = 0, ["LastCheck"] = false, ["LastOnline"] = false}
			tStats.Redirects.Addresses[address].Counter = tStats.Redirects.Addresses[address].Counter + 1 --// Address stats
		end
	end

	return address, addtype or ""
end,

["History"] = function(self, sid, client, address, msg)
	local ctime = socket.gettime()
	local name = client.sNick or client.sIP
	local newmsg = false
	local level = (tRedirects.History[name] and tRedirects.History[name].Silent and 4) or 2

	if not name or name == "" then
		tFuncs:Debug(tUser:GetDataLine(sid).."Redirect failed for SID:"..tostring(sid).." Name: "..tostring(name).." Msg: "..tostring(msg))
		return address, msg
	end
	if tRedirects.History[name] then
		--// This is a reoccuring redirect
		if ctime <= (tRedirects.History[name].Time + tHubSettings.tRedirects.iBreakLoopTimeout) then
			tRedirects.History[name].Times = tRedirects.History[name].Times + 1

			--// Maximum allowed reconnects with redirect has occured
			if tRedirects.History[name].Times > tHubSettings.tRedirects.iBreakLoopTimes then
				address = tHubSettings.tRedirects.sDeadEnd

				if tRedirects.History[name].SendDead then 
					--// Was already sent to deadend, sending to deadendfail address now
					if not tRedirects.History[name].Silent then
						address = tHubSettings.tRedirects.sDeadEndFail
						tFuncs:Debug(tUser:GetDataLine(sid).."Redirect 'dead-end address': "..address.." fails, choose an address that doesn't (indirectly) redirects back to this hub. Redirecting user to "..address.." now.", nil, level)
						newmsg = "You have entered a redirect-loop to this hub, you're redirected to "..address.." now."
						tRedirects.History[name].Silent = true
					end
				else
					--// Reoccurring reconnect, but hasn't been sent to dead-end address yet
					tFuncs:Debug(tUser:GetDataLine(sid).."entered a redirect-loop from last redirect: "..tRedirects.History[name].Address.." / last reason: "..tRedirects.History[name].Reason.." / Sending to dead-end address: "..address, nil, level)
					newmsg = "You have entered a redirect-loop to this hub, you're redirected to "..address.." now."
				end
				tRedirects.History[name].Address = address
				tRedirects.History[name].SendDead = true
			end
		else
			tRedirects.History[name].Times = 1
		end
		tRedirects.History[name].Address = address
		tRedirects.History[name].Reason = msg
		tRedirects.History[name].Time = ctime
	else
		--// This is a first redirect
		tRedirects.History[name] = { ["Address"] = address, ["Times"] = 1, ["Time"] = ctime, ["Reason"] = msg }

		--// Reset reconnect timeout once
		ThreadLockMutex(mutexiplog)
		tLogs.IP[tClient[sid].sIP]["$"].CurrentLogin = tLogs.IP[tClient[sid].sIP]["$"].CurrentLogin - (tSecurity.General.ReconnectTimeout)
		ThreadUnlockMutex(mutexiplog)
	end

	return address, newmsg
end,

["Check"] = function(self, ssid, address)	-- thread start, runs from timer
	local result, err = pcall(tRedirect.CheckLoop, nil, ssid, address)
	if not result then 
		tFuncs:SafeDebug("[THREAD] [CheckLoop] Error: "..err)
		tFuncs:UnlockMutexes()
	end
	tThreads.CheckRedirects = nil
	ThreadDestroy()
end,

["CheckLoop"] = function(self, ssid, address)
	local toutput = {}
	local list = {}
	local counter = 0
	if not address then
		tLogs.Redirects.LastRun = os.time()
		local names = {"main", "failover", "regonly", "full", "maxhubs", "minslots", "maxslots", "minshare", "maxshare", "passive", "flood", "adc_disabled", "nmdc_disabled", "deadend", "deadendfail"}
		for num, name in ipairs(names) do
			if #tRedirects.tList[names[num]] > 0 then
				for num, address in ipairs(tRedirects.tList[name]) do
					tStats.Redirects.Addresses[address] = tStats.Redirects.Addresses[address] or { ["Offline"] = false, ["Counter"] = 0, ["LastCheck"] = false, ["LastOnline"] = false}
					local stats = tStats.Redirects.Addresses[address]
					local protocol = address:find("adc.-://") and "ADC" or "NMDC"

					if not stats.LastCheck or stats.LastCheck+60 < socket.gettime() then

						local ctime = socket.gettime()
						tFuncs:SafeDebug("[THREAD] [CheckLoop] [Connect] Name: "..name.." / Checking: "..address.." / PROTOCOL: "..protocol, "X", 0)
						local online, ok, retmsg = tRedirect:Connect(ssid, address, protocol)
						tFuncs:SafeDebug("[THREAD] [CheckLoop] [Connect] Online: "..tostring(online).." / OK: "..tostring(ok).." / MSG: "..retmsg, "X", 0)

						if online then
							if stats.LastCheck and stats.Offline and tHubSettings.tRedirects.bAlertOffline then
								tFuncs:SafeDebug("[Redirect] [Check] Hub is online again: "..address..(stats.OfflineSince and " / Has been offline since "..ToDate(stats.OfflineSince) or ""))
							end
							stats.LastCheck = ctime
							stats.Offline = false
							stats.OnlineSince = stats.OnlineSince or ctime
							stats.OfflineSince = false
							stats.LastResult = retmsg
						else
							if stats.LastCheck and not stats.Offline and tHubSettings.tRedirects.bAlertOffline then
								tFuncs:SafeDebug("[Redirect] [Check] Hub went offline: "..address..(stats.OnlineSince and " / Has been online since "..ToDate(stats.OnlineSince) or ""))
							end
							stats.LastCheck = ctime
							stats.Offline = true
							stats.OfflineSince = stats.OfflineSince or ctime
							stats.OnlineSince = false
							stats.LastResult = retmsg
						end

						if online then online = "ONLINE" else online = "offline" end
						table.insert(toutput, "\trd_"..Padding(name, 30).." \t"..online.."\t\t"..Padding(address, 40).." \t"..retmsg)
					else
						table.insert(toutput, "\trd_"..Padding(name, 30).." \t"..(not stats.Offline and "ONLINE" or "offline").."\t\t"..Padding(address, 40).." \t"..(stats.LastResult or ""))
						tFuncs:SafeDebug("[THREAD] [CheckLoop] Skipping address: "..address.." / Last check was less than 60 seconds ago", "X")
					end
				end
			else
				table.insert(toutput, "\trd_"..Padding(name, 30).." \t\t\tNo address set")
			end
		end
	end

	local output = "\n\n\tRedirect check results:\n\t"..string.rep("=",80).."\n"
	for num,line in ipairs(toutput) do
		output = output..line.."\n"
	end
	tLogs.Redirects.LastResult = output

	if ssid then
		tFuncs:SafeMessage("M", ssid, "AAAA", output)
	end
end,

["Connect"] = function(self, ssid, address, protocol)
	local ctime = os.time()
--	local port = string.gsub(address, "%S+:", "") or 411
	local port = string.match(address, ".-:(%d+)") or 411
	port = tonumber(port)

	local addy = string.gsub(address, ":%d+", "")
	addy = string.gsub(addy, "^.+://", "")
	local online, ok, output = false, false, ""

--~ 	local hubip, err = socket.dns.toip(addy)
--~ 	if not hubip then return false, false, err end

	local sock, err = socket.tcp()
	if not sock then 
--~ 		tFuncs:SafeDebug("[THREAD] [ConnectHub] Socket can't be created: "..err)
		return false, false, "Error: Socket can't be created: "..err 
	else
--~ 		tFuncs:SafeDebug("[THREAD] [ConnectHub] Connecting to "..addy..":"..port)
		sock:settimeout(tHubSettings.tRedirects.iCheckTimeout or 3)
		online, err = sock:connect(addy, port)
	
		if not online then
--~ 			tFuncs:SafeDebug("[THREAD] [ConnectHub] "..addy..":"..port.." is offline")
			output = "Error: Can't connect: "..err
		else
--~ 			tFuncs:SafeDebug("[THREAD] [ConnectHub] "..addy..":"..port.." is online")
			ok = true
			output = "Connection OK"
		end
		
		sock:close()
		sock = nil

		return online, ok, output
	end
end,

}