--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

--// Settings Manager

--// Work in progress

--// Example SetMan:Get("hub_name")	if var can't be found it will return the default value, second argument will be helptext stating possible values
--// Example SetMan:Set("hub_name", "FlexHub Test")	--it will check the validity of the value, if ok returns true, if not, returns false with an errormessage stating possible values

--// SetMan:Get("hub_name")

--~	LocalSettings = {			--// This can be used by Initialize to fill with all settings below, for a fast lookup. Having _G[items[1]] in memory isn't needed then, that will only be used to serialize, save/load then.
--~
--~ }

SetManager = {
["Send"] = function(self)
	local reply = ""

	--// Sort settings:
	local sortlist = {}
	for name, data in pairs(Set.Settings) do
		table.insert(sortlist, name)
	end
	table.sort(sortlist)

	for num,name in ipairs(sortlist) do
		local data = Set.Settings[name]
		if data[2] and data[3] ~= "t" then		--// Skip tables for now
			local items = tFuncs:Split(data[2], "%.")
			if #items < 2 then 
				tFuncs:Debug("[SetManager] [ERROR] Invalid settingspath for '"..name.."' : "..data[2])
				return false, false, "Invalid settingspath: "..data[2] 
			end

			if not _G[items[1]] then _G[items[1]] = {} end

			--// Get current value
			local value
			if #items == 2 then
				value = _G[items[1]][items[2]]
			elseif #items == 3 then
				value = _G[items[1]][items[2]][items[3]]
			elseif #items == 4 then
				value = _G[items[1]][items[2]][items[3]][items[4]]
			elseif #items == 5 then
				value = _G[items[1]][items[2]][items[3]][items[4]][items[5]]
			elseif #items == 6 then
				value = _G[items[1]][items[2]][items[3]][items[4]][items[5]][items[6]]
			elseif #items == 7 then
				value = _G[items[1]][items[2]][items[3]][items[4]][items[5]][items[6]][items[7]]
			end

			reply = reply.."IGUI SV"..ADC_E(name).." TY"..data[3].." VA"..ADC_E(value).."\n"
		end
	end
	if #reply > 0 then return reply else return false end
end,

["Initialize"] = function(self)
	for num, data in ipairs(Set.Files) do
		local res, err = SetManager:CheckSettingFile(data[2], data[3] or {}, data[4])
		if not res then
			tFuncs:SafeDebug(err, "FI", -1) 
			return false, err
		end
	end

	local fresh = false
	if not tHubSettings or not tHubSettings.Main or not tHubSettings.Main.sHubID then
		fresh = true
		tFuncs:SafeDebug("[SetManager:Initialize] No settings avaliable, setting default values for all settings.", "FI", 0)
	end

	--// Check accounts
	local res, err = AccountManager:Check()
	if not res then tFuncs:SafeDebug("[SetManager:Initialize] "..err , "FI", 0) end

	--// Check if all settings exist, if not, they will be created and assigned their default value	- TODO: add typechecking to see if current values still match validation, if not, output to debug and set default value
	for name, data in pairs(Set.Settings) do
		if data[2] then
			local value

			local Invalid = function()
				if not fresh and data[1] ~= "" then	--// Hide debug for values without a category name
					tFuncs:SafeDebug("[SetManager:Initialize] Invalid value for "..name..": "..tostring((value or ""))..", setting to default: "..tostring(data[6]))
				end
			end

			local items = tFuncs:Split(data[2], "%.")
			if #items < 2 then 
				tFuncs:SafeDebug("[SetManager:Initialize] Invalid settingspath for '"..name.."' : "..data[2])
				return false, false, "Invalid settingspath: "..data[2] 
			end

			if #items > 1 and not _G[items[1]] then _G[items[1]] = {} end
			if #items > 2 and not _G[items[1]][items[2]] then _G[items[1]][items[2]] = {} end
			if #items > 3 and not _G[items[1]][items[2]][items[3]] then _G[items[1]][items[2]][items[3]] = {} end
			if #items > 4 and not _G[items[1]][items[2]][items[3]][items[4]] then _G[items[1]][items[2]][items[3]][items[4]] = {} end
			if #items > 5 and not _G[items[1]][items[2]][items[3]][items[4]][items[5]] then _G[items[1]][items[2]][items[3]][items[4]][items[5]] = {} end
			if #items > 6 and not _G[items[1]][items[2]][items[3]][items[4]][items[5]][items[6]] then _G[items[1]][items[2]][items[3]][items[4]][items[5]][items[6]] = {} end

			--// Get current value
			if #items == 2 then
				value = _G[items[1]][items[2]]
			elseif #items == 3 then
				value = _G[items[1]][items[2]][items[3]]
			elseif #items == 4 then
				value = _G[items[1]][items[2]][items[3]][items[4]]
			elseif #items == 5 then
				value = _G[items[1]][items[2]][items[3]][items[4]][items[5]]
			elseif #items == 6 then
				value = _G[items[1]][items[2]][items[3]][items[4]][items[5]][items[6]]
			elseif #items == 7 then
				value = _G[items[1]][items[2]][items[3]][items[4]][items[5]][items[6]][items[7]]
			end

			if (data[3] ~= "b" and not value) or (data[3] == "b" and type(value) ~= "boolean") then 
				Invalid()
				value = data[6]		--// Value doesn't exist yet
				if data[3] == "f" then value = tFuncs:ConvertToBytes(value) end
				local ok, err = SetManager:Set(name, value)
				if not ok then tFuncs:SafeDebug("Value '"..tostring(value).."' not allowed. Error: "..tostring(err)) end
			else	--// A value is set: check if it's ok:
				--// Validator check
				if data[3] == "s" then
					if type(data[4]) == "table" then
						local found = false
						for num, a in ipairs(data[4]) do
							if string.find(value, a) then found = true end
						end
						if not found then Invalid() SetManager:Set(name, nil, true) end
					elseif type(data[4]) == "string" and not string.find(value, data[4]) then
						Invalid() SetManager:Set(name, nil, true)
					end
					if type(data[5]) == "table" then
						for num, a in ipairs(data[5]) do
							if string.find(value, a) then Invalid() SetManager:Set(name, nil, true) end
						end
					elseif type(data[5]) == "string" and string.find(value, data[5]) then
						Invalid() SetManager:Set(name, nil, true)
					end
				elseif data[3] == "i" then
					if data[4] and value < data[4] then Invalid() SetManager:Set(name, nil, true) end
					if data[5] and value > data[5] then
						Invalid()
						SetManager:Set(name, nil, true)
					end
				elseif data[3] == "n" then
					local res,err = tCheck:NickChars(value)
					if not res then Invalid() SetManager:Set(name, nil, true) end
				elseif data[3] == "f" then
					local vmin = tFuncs:ConvertToBytes(data[4])
					local vmax = tFuncs:ConvertToBytes(data[5])
					if vmin and value < vmin then Invalid() SetManager:Set(name, nil, true) end
					if vmax and value > vmax then Invalid() SetManager:Set(name, nil, true) end
				end
			end
		end
	end

	--// Make changes from Set.UpdateSettings:
	local updateoutput, highest = "", tHubSettings.Main.iUpdateSettings
	for name, data in pairs(Set.UpdateSettings) do
		if data.svn > tHubSettings.Main.iUpdateSettings then
			SetManager:Set(name, data.value)
			updateoutput = updateoutput.."\nChanged value '!set "..name.."' to: "..tostring(data.value).." because: "..data.reason
			highest = (data.svn > highest and data.svn) or highest
		end
	end

	-- // automatic renaming of textcommands if same name exists as hub or set command 
	local rename = {}
	local hit = false
	for name in pairs(tCommands) do
		if HubCmd[name] or SetCmd[name] or tCmdAlias[name]then
			rename[name] = true
			hit = true
		end
	end

	if hit then
		for name in pairs(rename) do
			local newname = "x_nameconflict_"..name
			while true do
				if HubCmd[newname] or SetCmd[newname] or tCmdAlias[newname] then
					newname = "x"..newname
				else
					break
				end
			end

			local res, reply = SetCmd["renamecmd"]["cmd"]("AAAA", "!renamecmd "..name.." "..newname, "!renamecmd")

			if res then
				tFuncs:SafeDebug("[SetManager:Initialize] Invalid name for TextCommand '"..name.."', It matches an existing hub command. It has been renamed to ".. newname)
			else
				tFuncs:SafeDebug("[SetManager:Initialize] Invalid name for TextCommand '"..name.."', It matches an existing hub command. Error when renaming : "..reply)
			end
		end
	end

	if updateoutput ~= "" then
		tHubSettings.Main.iUpdateSettings = highest
		if not fresh then
			tFuncs:SafeDebug("An update required some changes to default values: \n"..updateoutput.."\n", "FI", 0)
		end
		tFuncs:SaveSettings("settings")
	end

	return true
end,

["Get"] = function(self, name)
	local name = name or ""
	local lname = name:lower()

	local data = Set.Settings[lname]
	if not data then
		tFuncs:Debug("[SetManager] [ERROR] A setting with name '"..name.."' doesn't exist.")
		return false, false, "A setting with name '"..name.."' doesn't exist." 
	end

	local value = false
	local txt = data[8] or false

	--// Contains a main category, not a variable, return helptext and data
	if not data[2] then return false, false, data[8], data	end

	local items = tFuncs:Split(data[2], "%.")
	if #items < 2 then 
		tFuncs:Debug("[SetManager] [ERROR] Invalid settingspath for '"..name.."' : "..data[2])
		return false, false, "Invalid settingspath: "..data[2] 
	end

		if #items == 2 then
			value = _G[items[1]][items[2]]
		elseif #items == 3 then
			value = _G[items[1]][items[2]][items[3]]
		elseif #items == 4 then
			value = _G[items[1]][items[2]][items[3]][items[4]]
		elseif #items == 5 then
			value = _G[items[1]][items[2]][items[3]][items[4]][items[5]]
		elseif #items == 6 then
			value = _G[items[1]][items[2]][items[3]][items[4]][items[5]][items[6]]
		elseif #items == 7 then
			value = _G[items[1]][items[2]][items[3]][items[4]][items[5]][items[6]][items[7]]
		end

	return value, true, txt, data
end,

["Set"] = function(self, name, value, default, fromhubcmd)
	local msg = false
	local res = true
	local name = name or ""
	local lname = name:lower()

	local data = Set.Settings[lname]
	if not data then 
		tFuncs:Debug("[SetManager:Set] A setting with name '"..name.."' doesn't exist.")
		return false, "A setting with name '"..name.."' doesn't exist." 
	end

	--// If no value was given and default = true, set default value (can be used to reset a certain setting to default)
	if (value and value == "-default") or (not value and default) then
		value = data[6]
	end

	--// handle "" as empty string (tFuncs:Split changes empty string into a string with double quotes, this sets it back)
	if value =="\"\"" or value == "''" then value = "" end

	--// Start validation
	local list = {["s"] = "string", ["b"] = "boolean", ["t"] = "table", ["d"] = "number", ["f"] = "number", ["n"] = "string", ["i"] = "number"}
	local stype = list[data[3]]

	--// Type adjustments
	if data[3] == "f" then
		value = tFuncs:ConvertToBytes(value)
		if not value then
			return false, "Invalid value for '"..name.."'"
		end
	elseif data[3] == "d" then
		value = tFuncs:ConvertToSeconds(value)
		if not value then
			return false, "Invalid value for '"..name.."'"
		end
	end

	if stype == "number" then value = tonumber(value)
	elseif stype == "boolean" and type(value) ~= "boolean" then
		if string.sub(value, 1, 1) == "f" or string.sub(value, 1, 1) == "n" or (tonumber(value) and tonumber(value) < 1) then 
			value = false
		elseif string.sub(value, 1, 1) == "t" or string.sub(value, 1, 1) == "y" or (tonumber(value) and tonumber(value) >= 1) then
			value = true
		else
			return false, "Invalid value for '"..name.."', requires boolean( true/false )"
		end
	elseif type(value) ~= stype then 	--// Type mismatch
		return false, "Invalid value for '"..name.."', requires "..stype..", received: "..type(value)
	end

	--// Validator check
	if data[3] == "s" then
		if type(data[4]) == "table" then
			local found = false
			for num, a in ipairs(data[4]) do
				if string.find(value, a) then found = true end
			end
			if not found then return false, "Value '"..value.."' doesn't match the following patterns: "..table.concat(data[4], " ") end 
		elseif type(data[4]) == "string" and not string.find(value, data[4]) then
			return false, "Incorrect value for '"..name.."', value has to match pattern '"..data[4].."." 
		end
		if type(data[5]) == "table" then
			for num, a in ipairs(data[5]) do
				if string.find(value, a) then return false, "Value '"..value.."' isn't allowed to match the following patterns: "..table.concat(data[5], " ") end
			end
		elseif type(data[5]) == "string" and string.find(value, data[5]) then
			return false, "Incorrect value for '"..name.."', value can't match pattern '"..data[5].."'." 
		end
	elseif data[3] == "i" or data[3] == "d" then
		if data[4] and value < data[4] then return false, "Invalid value, min: "..(data[4] or "-").." / max: "..(data[5] or "-") end
		if data[5] and value > data[5] then return false, "Invalid value, min: "..(data[4] or "-").." / max: "..(data[5] or "-") end
	elseif data[3] == "n" then
		local res, err = tCheck:NickChars(value)
		if not res then return false, err end
	elseif data[3] == "f" then
		local vmin = tFuncs:ConvertToBytes(data[4])
		local vmax = tFuncs:ConvertToBytes(data[5])
		if vmin and value < vmin then return false, "Invalid value, min: "..(tFuncs:ConvertShare(tFuncs:ConvertToBytes(data[4]) or 0)).." / max: "..(tFuncs:ConvertShare(tFuncs:ConvertToBytes(data[5]) or 0)) end
		if vmax and value > vmax then return false, "Invalid value, min: "..(tFuncs:ConvertShare(tFuncs:ConvertToBytes(data[4]) or 0)).." / max: "..(tFuncs:ConvertShare(tFuncs:ConvertToBytes(data[5]) or 0)) end
	end

	local items = tFuncs:Split(data[2], "%.")
	if #items < 2 then 
		tFuncs:Debug("[SetManager:Set] Invalid settingspath for '"..name.."' : "..data[2])
		return false, "Invalid settingspath: "..data[2] 
	end

	--// If it needs another command
	if not fromhubcmd and data[7] and data[7] ~= "GUI" and tStats.Console.Status == "Running" then
		if HubCmd[data[7]] then		--// If a hubcommand exists, the hubcommand has to take care of actually setting the variable and saving settings
			local ok, res = HubCmd[data[7]].cmd("AAAA", data[7].." "..value, data[7])
		--	return ok, false, res
			return ok, res
		elseif SetCmd[data[7]] then		--// If a setcommand exists, the hubcommand has to take care of actually setting the variable and saving settings
			local ok, res = SetCmd[data[7]].cmd("AAAA", data[7].." "..value, data[7])
			return ok, false, res
		elseif Set.Function[data[7]] then	--// After this function the value will be set using the code below, as opposed to how HubCmd and SetCmd work
			res, value, msg = Set.Function[data[7]](name, value)
			if not res then
				return false, value, (msg and "Setting failed: "..msg or "")
			end
		else
			return false, "Use !"..data[7].." to change this value."
		end
	end

	if #items > 2 and not _G[items[1]][items[2]] then _G[items[1]][items[2]] = {} end
	if #items > 3 and not _G[items[1]][items[2]][items[3]] then _G[items[1]][items[2]][items[3]] = {} end
	if #items > 4 and not _G[items[1]][items[2]][items[3]][items[4]] then _G[items[1]][items[2]][items[3]][items[4]] = {} end
	if #items > 5 and not _G[items[1]][items[2]][items[3]][items[4]][items[5]] then _G[items[1]][items[2]][items[3]][items[4]][items[5]] = {} end
	if #items > 6 and not _G[items[1]][items[2]][items[3]][items[4]][items[5]][items[6]] then _G[items[1]][items[2]][items[3]][items[4]][items[5]][items[6]] = {} end

	if #items == 2 then
		_G[items[1]][items[2]] = value
	elseif #items == 3 then
		_G[items[1]][items[2]][items[3]] = value
	elseif #items == 4 then
		_G[items[1]][items[2]][items[3]][items[4]] = value
	elseif #items == 5 then
		_G[items[1]][items[2]][items[3]][items[4]][items[5]] = value
	elseif #items == 6 then
		_G[items[1]][items[2]][items[3]][items[4]][items[5]][items[6]] = value
	elseif #items == 7 then
		_G[items[1]][items[2]][items[3]][items[4]][items[5]][items[6]][items[7]] = value
	end

	tFuncs:SendRawToGui("IGUI SV"..ADC_E(name).." TY"..data[3].." VA"..ADC_E(value))
	tFuncs:SaveSettings(items[1])
	return true, value, msg
end,

["CheckSettingFile"] = function(self, name, value, valuename) -- // only used from SetManager:Initialize, could just as well be in here instead of tFuncs
	name = _APPPATH..name
	local setfile,fileerr = io.open(name, "r") 
	if setfile then
		setfile:close()
		local res, err = pcall(dofile, name)
		if not res then
			tFuncs:Debug("File: '"..name.."' is damaged. Error: "..tostring(err))
			--// Try writing a new one
			setfile, fileerr = io.open(name, "w")
			if not setfile then
				return false, "[Fatal Error] File: '"..name.."' is damaged. Error: "..tostring(err).."\nCan't write to file: "..tostring(fileerr)
			else
				setfile:close()
				tFuncs:Save_File(name,value,valuename)
				tFuncs:Debug("[CheckSettingFile] File: '"..name.."' was damaged. Error: "..tostring(err)..". Settings have been reset.")
				-- because the second return value will never be used when first value is true
				return true, "File: '"..name.."' was damaged. Error: "..tostring(err)..". Settings have been reset." 
			end
		else
			--// Try appending data
			local setfile,fileerr = io.open(name, "a")
			if not setfile then
				return false, "[Fatal Error] Can't write to file: '"..name.."'. Error: "..tostring(fileerr)
			else
				setfile:close()
				return true
			end
		end
	else
		--// File doesn't exist or can't be read, try creating a new file
		tFuncs:Debug("File: '"..name.."' doesn't exist. Using default settings.", "X")
		setfile, fileerr = io.open(name, "w")
		if not setfile then
			return false, "[Fatal Error] Can't write to file: '"..name.."'. Error: "..tostring(fileerr)
		else
			setfile:close()
			tFuncs:Save_File(name,value,valuename)
			--tFuncs:Debug("File: '"..name.."' didn't exist, new file created.")
			return true, "File: '"..name.."' wasn't found, new file created."
		end
	end
end,

}


