--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

 --[[
  RULES :
	* timers that needs own thread _must_ have the same name in tThreads and tTimers!
	* the thread function itself must take care to do ThreadDestroy() and tThreads[name] = nil
 
left to do:
	* rewrite init to make it possible to reinitialize individual timers
	* send ctime as argument to the funcs so each doesnt have to pull it
	* test the actual function on add 
	* finish function check on add and edit

done
	* mutex mod timer settings
  --]]
tTimerFuncs = {

-- should a timer that fails here be disabled, or removed even?
["VerifyTimer"] = function(self, tbl)	-- looks OK for now, only simple checks added
	if type(tbl) == "table" then
		local msg, check = "", true
		local template = {
			["iInterval"] =	{1,"number"},	-- positive number
			["bEnabled"] =	{1, "boolean"},	-- boolean
		--	["bRunning"] =	{2, "boolean"},	-- boolean
			["iLastRun"] =	{2, "number"},	-- positive time
			["iCount"] =	{2, "number"},	-- optional, positive number
			["iStartDelay"] = {2, "number"},-- optional, positive time
			["fFunc"] =		{1, "function"},-- function
		}

		for str, data in pairs(template) do
			if data[1] == 1 and not tbl[str] and not tbl[str] == false then
				check = false
				msg = msg .. str.." is missing from table, "
			elseif tbl[str] and type(tbl[str]) ~= data[2] then
				check = false
				msg = msg .. "the type of "..str.." does not match "..data[2]..", "
			elseif tbl[str] and data[2] == "number" and tbl[str] < 0 then
				check = false
				msg = msg .. str.." must be a positive number, "
			end
		end
		return check, msg
	else
		return false, "not a table"
	end
end,

["Initialize"] = function(self, tbl)	-- run once from Hub-Start to set all table values, will move it to !rs once its made 'secure' :P

	tTimers = {
		["HubIp"] = {			-- done, threaded, no settings
			["bEnabled"] = true,
			["iStartDelay"] = 120,	-- 2 min
			["iInterval"] = 7200,	-- every 2 h
			["fFunc"] = 
				function()
					if not tThreads.HubIp then
						tThreads.HubIp = ThreadCreate(tFuncs.GetHubIP,{1, tHubSettings.Main.sAddress})
						tFuncs:Debug("[FlexTimer] Running thread: GetHubIP with interval "..tFuncs:Duration(tTimers.HubIp.iInterval), "FI", 2)
					end
				end,
		},
		["HubListReg"] = {		-- done, threaded, needs rework
			["iInterval"] = (tHubSettings.tHubLists.AutoInterval or 7200),
			["bEnabled"] = (tHubSettings.tHubLists.AutoInterval > 0 and true or false),
			["iStartDelay"] = 600,	-- 10 mins
			["fFunc"] =
				function()
					local ctime = socket.gettime()
					if not tThreads.HubListReg and tHubSettings.tHubLists.AutoInterval > 0 and tLogs.Hublists.Registration.LastRun + tHubSettings.tHubLists.AutoInterval < ctime and tLogs.Hublists.Registration.LastRun + tHubSettings.tHubLists.MinInterval < ctime then
						tFuncs:Debug("[FlexTimer] Running thread: HubListReg with interval "..tFuncs:Duration(tTimers.HubListReg.iInterval), "FI", 2)
						tThreads.HubListReg = ThreadCreate(tFuncs.RegHub, {1, "AAAA"})
					end
				end,
		},
		["CleanAutoban"] = {	-- done, mutex, no settings
			["bEnabled"] = true,
			["iStartDelay"] = 300,	-- 5 mins
			["iInterval"] = 3600,	-- 1 h
			["fFunc"] =
				function()
					ThreadLockMutex(mutexbanlist)
					local ctime = socket.gettime()
					local counter = 0
					for ip, expires in pairs(tBans.autoban) do
						if ctime > expires then
							tBans.autoban[ip] = nil
							counter = counter + 1
						end
					end
					ThreadUnlockMutex(mutexbanlist)
					if counter > 0 then
						tFuncs:Debug("[FlexTimer] Running: CleanAutoban with interval "..tFuncs:Duration(tTimers.CleanAutoban.iInterval)..", cleared "..counter.." expired autobans.", "FI", 1)
					else
						tFuncs:Debug("[FlexTimer] Running: CleanAutoban with interval "..tFuncs:Duration(tTimers.CleanAutoban.iInterval), "FI", 2)
					end
				end,
		},
		["CleanBans"] = {		-- done, uses mutex, no settings
			["bEnabled"] = true,
			["iStartDelay"] = 310,	-- 5 mins+
			["iInterval"] = 86400,	-- 24 hours
			["fFunc"] =
				function()
					tFuncs:Debug("[FlexTimer] Running: CleanBans with interval "..tFuncs:Duration(tTimers.CleanBans.iInterval), "FI", 2)
					ThreadLockMutex(mutexbanlist)
					local res, reply, level = HubCmd["clearexpiringbans"]["cmd"]("AAAA", "!clearexpiringbans 0 "..tHubSettings.Main.bConfirmCode, "!clearexpiringbans")
					ThreadUnlockMutex(mutexbanlist)
					if res then 
						tFuncs:Debug("[FlexTimer] Automatic banlist cleanup: "..reply, "FI", (level or 0))
					else
						tFuncs:Debug("Error during automatic cleanup of expired bans: "..reply, nil, -1 )
					end
				end,
		},
		["AccountCheck"] = {	-- done
			["bEnabled"] = true,
			["iStartDelay"] = 320,	-- 10 mins+
			["iInterval"] = 86400,	-- 24 hours
			["fFunc"] =
				function()
					if tSecurity.General.AllowBlankPass then
						tFuncs:Debug("[FlexTimer] Running: AccountCheck with interval "..tFuncs:Duration(tTimers.AccountCheck.iInterval), "FI", 2)
						tTimerFuncs:AccountCheck()
					end
				end,
		},
		["PasswordReminder"] = {-- done
			["bEnabled"] = true,
			["iStartDelay"] = 330,	-- 10 mins+
			["iInterval"] = 3600,	-- 1h
			["fFunc"] =
				function()
					if tSecurity.General.AllowBlankPass then
						tFuncs:Debug("[FlexTimer] Running: PasswordReminder with interval "..tFuncs:Duration(tTimers.PasswordReminder.iInterval), "FI", 2)
						tTimerFuncs:PasswordReminder()
					end
				end,
		},
		["Update"] = {			-- done, threaded
			["bEnabled"] = true,
			["iStartDelay"] = 340,	-- 10 minutes+
			["iInterval"] = 172800,	-- 48 hours
			["fFunc"] =
				function()
					if not _NOUPDATE and tHubSettings.Main.bUpdateCheck then
						tFuncs:Debug("[FlexTimer] Running thread: Update with interval "..tFuncs:Duration(tTimers.Update.iInterval), "FI", 2)
						tThreads.Update = ThreadCreate(tUpdate.Thread, {1, "BBBB", "!update timer"})
					end
				end,
		},
		["Commands"] = {		-- done
			["cmds"] = {}, -- where the stuff for each cmd is stored (make cleaning up for this possible too??)
			["bEnabled"] = true,
			["iStartDelay"] = 350,	-- 10 minutes+
			["iInterval"] = 20,		-- dont have to run it every freakin second
			["fFunc"] =
				function()
					--tFuncs:Debug("[FlexTimer] Running: tCommands", "FI", 2)
					local ctime = socket.gettime()
					for name, data in pairs(tCommands) do
						if data.iTimer and data.iTimer > 0 then
							local output = tFuncs:CheckFlag((tCommands[name].iOptions or 0), 1) and "PM" or "mainchat"
							tTimers.Commands.cmds[name] = tTimers.Commands.cmds[name] or { ["iLastRun"] = 0 }
							if tTimers.Commands.cmds[name].iLastRun + data.iTimer <= ctime then
								tTimers.Commands.cmds[name].iLastRun = ctime
								local res, reply = tUser:DoTrigger("AAAA", name)
								if res then
									tFuncs:Debug("[FlexTimer] Executed textcommand '"..name.."' and sent output in "..output.." to all users with profilelevels "..(tCommands[name].iProfile or -1).." to "..(tCommands[name].iMaxProfile or 10).." with interval of "..data.iTimer.." seconds.", "FI", 2)
								else
									tFuncs:Debug("[FlexTimer] Error when executing textcommand: "..name.." with interval of "..data.iTimer.." seconds: "..reply, "FI", -1)
								end
							end
						end
					end
					return true
				end,
		},
		["SaveStats"] = {		-- done
			["cmds"] = {}, -- where the stuff for each cmd is stored (make cleaning up for this possible too??)
			["bEnabled"] = true,
			["iStartDelay"] = 120,	-- 2 mins
			["iInterval"] = 30,		--30 secs
			["fFunc"] = 
				function()
					local ctime = socket.gettime()
					for name, data in pairs(tHubSettings.SaveStats) do
						if data.iTimer and data.iTimer > 0 then
							tTimers.SaveStats.cmds[name] = tTimers.SaveStats.cmds[name] or { ["iLastRun"] = 0 }
							if tTimers.SaveStats.cmds[name].iLastRun + data.iTimer <= ctime then
								tTimers.SaveStats.cmds[name].iLastRun = ctime
								local res = tFuncs:WriteFile(name, tFuncs:FillVars(data.sContent, "AAAA") .."\n") 
								if res then
									tFuncs:Debug("[FlexTimer] Saved statistics to file: "..name..".log with interval of "..data.iTimer.." seconds.", "FI", 2)
								else
									tFuncs:Debug("[FlexTimer] Error when saving statistics to file: "..name.." with interval of "..data.iTimer.." seconds", "FI", -1)
								end
							end
						end
					end
				end,
		},
		["CheckRedirects"] = {	-- threaded, unfinished but working, the if-statement needs a rewrite to utilize timerfuncs
			["bEnabled"] = (tHubSettings.tRedirects.iCheckInterval > 0 and true or false),
			["iStartDelay"] = 400,
			["iInterval"] = tHubSettings.tRedirects.iCheckInterval,
			["fFunc"] =
				function()
					local ctime = socket.gettime()
					if tHubSettings.tRedirects.iCheckInterval > 0 and ((tLogs.Redirects.LastRun or 0) + tHubSettings.tRedirects.iCheckInterval) < ctime then
						tFuncs:Debug("[FlexTimer] Running thread: CheckRedirects with interval "..tFuncs:Duration(tTimers.CheckRedirects.iInterval), "FI", 2)
						tThreads.CheckRedirects = ThreadCreate(tRedirect.Check, {1, ssid})
					end
				end,
		},
		["CleanRedirectHistory"] = {	-- done
			["bEnabled"] = true,
			["iStartDelay"] = 410,
			["iInterval"] = 3600,
			["fFunc"] =
				function()
					tFuncs:Debug("[FlexTimer] Running CleanRedirectHistory with interval "..tFuncs:Duration(tTimers.CleanRedirectHistory.iInterval), "FI", 2)
					local ctime = socket.gettime()
					for name, data in pairs(tRedirects.History) do
						if data.Time + 3600 <= ctime then
							tRedirects.History[name] = nil
						end
					end
				end,
		}
	}

--	["template"] = {	-- lists the fields a timer must have, anything else is optional and not used by DoTimer itself, but can be used to store things between sessions
--		["iInterval"] = 7200,	-- must have
--		["bEnabled"] = true,	-- must have
--		["bRunning"] = false,	-- must have for functions that runs in own thread // skipping this as it makes no sence have it double
								-- instead make sure tTimer function and tThreads function have same name, then check that instead
--		["iLastRun"] = 0,		-- doesnt have to exists, timer will set it after first run
--		["iCount"] = false,		-- optional, if set to a positive number then the function will run that many times, then it will auto stop
--		["iStartDelay"] = 0,	-- optional, if set to a positive number then thats how long it will delay the first run
--		["fFunc"] = {}			-- the function to run

	for name, data in pairs(tTimers) do
		local res, msg = tTimerFuncs:VerifyTimer(data)
		if not res then
			tFuncs:Debug("[FlexTimer] test failed on : "..name.." with message :"..msg.." It has been disabled.")
			if type(data) == "table" then
				data.bEnabled = false
			else
				data = {["bEnabled"] = false,}
			end
	--	else
	--		tFuncs:Debug("[FlexTimer] test OK on : "..name)
		end
	end
end,

["FlexTimer"] = function(self)
	local ctime = socket.gettime()
	local uptime = ctime - tStats.Hub.Started

	for name, data in pairs(tTimers) do
		if data.bEnabled and not tThreads[name] then
			if (not data.iStartDelay or data.iStartDelay < uptime) and (not data.iLastRun or (data.iLastRun + data.iInterval) <= ctime) then 
				tStats.Thread.Listen.Status = "FlexTimer: "..name
				data.iLastRun = ctime

				res, msg = pcall(data.fFunc)
				if not res then
					tFuncs:Debug("[FlexTimer] Error: "..msg, nil, -1)
					tFuncs:UnlockMutexes()
					data.bEnabled = false
					tFuncs:Debug("[FlexTimer] Disabling "..name.." to avoid furter errors", "FI", -1)
				end

				if data.iStartDelay then data.iStartDelay = nil end
				if data.iCount and data.iCount > 1 then
					data.iCount = data.iCount - 1
				elseif data.iCount and data.iCount <= 1 then
					data.bEnabled = false
					tFuncs:Debug("[FlexTimer] Turning off "..name..", its work is finished", "FI", 2)
				end

				-- // 
				if data.iLoopCount then
					data.iLoopCount = nil
				end

			end
		elseif tThreads[name] then	-- temp testing for threads running too long
			if not data.iLastRun or (data.iLastRun + data.iInterval) <= ctime then
			-- ok, if we're here it means thread is still running, cant do much about it except report it
			-- can also mean the command (like hublist regging) is run manually

				if not data.TimeOut or ctime - (data.TimeOut + 120) < 0 then
					data.TimeOut = ctime
					tFuncs:Debug("[FlexTimer] setting TimeOut")
				
					if not data.iLoopCount then
						tFuncs:Debug("[FlexTimer] Possible error on thread "..name..", its runtime is longer than the interval set ("..data.iInterval .." seconds). This is either a bug, 	or a configuration error. ", "FI", 1)
						data.iLoopCount = 1
						data.iLastRunOriginal = data.iLastRun
					else
						data.iLoopCount = data.iLoopCount + 1
						tFuncs:Debug("[FlexTimer] Error : Thread "..name.." is still running; started at "..ToDate(data.iLastRunOriginal) ..",  ", "FI", 1)
					end
					data.iLastRun = ctime
				--else
				--	tFuncs:Debug("[FlexTimer] TimeOut in place")
				end
			end
	--	else
	--		tFuncs:Debug("[FlexTimer] Skipping: "..name..", it is "..(tThreads[name] and "already running." or not data.bEnabled and "disabled."), "FI", 2)
		end
	end
	local endtime = socket.gettime()
	tStats.Thread.Listen.Status = "FlexTimer: Timer Loop Finished in "..(endtime - ctime)
end,

["TStart"] = function(self, name)
	if tTimers[name] then
		ThreadLockMutex(mutextimers)
		tTimers[name].bEnabled = true
		ThreadUnlockMutex(mutextimers)
		return true
	else
		return false
	end
end,

["TStop"] = function(self, name)
	if tTimers[name] then
		ThreadLockMutex(mutextimers)
		tTimers[name].bEnabled = false
		ThreadUnlockMutex(mutextimers)
		return true
	else
		return false
	end
end,

["TGetName"] = function(self, name)	-- // returns true, real name, table or false
	if tTimers[name] then
		return true, name, tTimers[name]
	else
		local llname = name:lower()
		for label, data in pairs (tTimers) do
			if llname == label:lower() then
				return true, label, tTimers[name]
			end
		end
		return false
	end
end,

["TAddTimer"] = function(self, name, tbl)
	local test = tTimerFuncs:TGetName(name)
	if test then
		return false, "A timer with that name already exists. Use another name to create a new timer."
	else
		if tbl.fFunc and type(tbl.fFunc) == "string" then
			local test, err = loadstring(tbl.fFunc)
			if not test then
				return false, "The timer function failed the loadstring check. "..(err or "")
			else
				tbl.fFunc = test
			end
		end

		local test, msg = tTimerFuncs:VerifyTimer(tbl)
		if not test then
			return false, msg
		else

			-- should be safe to add this one
			ThreadLockMutex(mutextimers)
			tTimers[name] = tbl
			ThreadUnlockMutex(mutextimers)
			return true, name
		end
	end
end,

["TDelTimer"] = function(self, name, tbl)
	local test, real = tTimerFuncs:TGetName(name)
	if not test then
		return false, "A timer with that name does not exist."
	elseif tThreads[real] then
		return false, "Timer "..real.." is currently running, it can't be deleted."
	else
		-- should be safe to remove this one now
		ThreadLockMutex(mutextimers)
		tTimers[real] = nil
		ThreadUnlockMutex(mutextimers)
		return true, real
	end
end,

["TEditTimer"] = function(self, name, set, val)
	local test, real, data = tTimerFuncs:TGetName(name)
	tFuncs:Debug("[TEditTimer] Got "..name.."(real : "..real..") and setting : "..set.." to : "..tostring(val), "X")
	if not test then
		tFuncs:Debug("[TEditTimer] returns false, A timer with that name does not exist.", "X")
		return false, "A timer with that name does not exist."
	elseif tThreads[real] then	-- this isnt really neccesary - a running timer doesn't (normally) edit the contents of tTimer anyway
		tFuncs:Debug("[TEditTimer] returns false, Timer "..real.." is currently running, it can't be edited.", "X")
		return false, "Timer "..real.." is currently running, it can't be edited."
	else
		-- // make a test validation
		data[set] = val
		local test, msg = tTimerFuncs:VerifyTimer(data)
		if test then -- we're ok
			tFuncs:Debug("[TEditTimer] test OK", "X")
			-- should be safe to edit this one now
			ThreadLockMutex(mutextimers)
			tTimers[real][set] = val
			ThreadUnlockMutex(mutextimers)
			tFuncs:Debug("[TEditTimer] returns true, "..real, "X")
			return true, real
		else
			tFuncs:Debug("[TEditTimer] returns false, "..msg, "X")
			return false, msg
		end
	end
end,



-- this isnt used any more
["DoTimers"] = function(self)	--// Is called every second from the Listen thread (not accurately every second, but there will be at least one full second between calls)
	local ctime = socket.gettime()

	--// Hub IP
	tStats.Thread.Listen.Status = "DoTimers: Hub IP"
	if (tTimers.HubIp.LastRun + tTimers.HubIp.iInterval) <= ctime then
		tFuncs:Debug("[Timer] Running thread: GetHubIP", "FI", 2)
		tTimers.HubIp.LastRun = ctime
		ThreadCreate(tFuncs.GetHubIP,{1, tHubSettings.Main.sAddress})
	end

	--// Cleanup of expired autobans  
	tStats.Thread.Listen.Status = "DoTimers: Cleanup autobans"
	if (tTimers.CleanAutoban.LastRun + tTimers.CleanAutoban.iInterval) <= ctime then
		ThreadLockMutex(mutexbanlist)
		tTimers.CleanAutoban.LastRun = ctime
		local counter = 0
		for ip,expires in pairs(tBans.autoban) do
			if ctime > expires then
				tBans.autoban[ip] = nil
				counter = counter + 1
			end
		end
		ThreadUnlockMutex(mutexbanlist)
		if counter > 0 then
			tFuncs:Debug("[Timer] Automatic autobanlist cleanup, cleared "..counter.." expired autobans.", "FI", 1)
		end
	end

	--// Cleanup of expired bans  
	tStats.Thread.Listen.Status = "DoTimers: Cleanup bans"
	if (tTimers.CleanBans.LastRun + tTimers.CleanBans.iInterval) <= ctime then
		ThreadLockMutex(mutexbanlist)
		tTimers.CleanBans.LastRun = ctime
		local res, reply, level = HubCmd["clearexpiringbans"]["cmd"]("AAAA", "!clearexpiringbans 0 "..tHubSettings.Main.bConfirmCode, "!clearexpiringbans")
		ThreadUnlockMutex(mutexbanlist)
		if res then 
			tFuncs:Debug("[Timer] Automatic banlist cleanup: "..reply, "FI", (level or 0))
		else
			tFuncs:Debug("Error during automatic cleanup of expired bans: "..reply, nil, -1 )
		end
	end

	--// Commands
	tStats.Thread.Listen.Status = "DoTimers: Timed textcommands"
	for name, data in pairs(tCommands) do
		if data.iTimer and data.iTimer > 0 then
			local output = tFuncs:CheckFlag((tCommands[name].iOptions or 0), 1) and "PM" or "mainchat"
			tTimers.Commands[name] = tTimers.Commands[name] or { ["LastRun"] = 0 }
			if tTimers.Commands[name].LastRun+data.iTimer <= ctime then
				tTimers.Commands[name].LastRun = ctime
				local res, reply = tUser:DoTrigger("AAAA", name)
				if res then
					tFuncs:Debug("[Timer] Executed textcommand '"..name.."' and sent output in "..output.." to all users with profilelevels "..(tCommands[name].iProfile or -1).." to "..(tCommands[name].iMaxProfile or 10).." with interval of "..data.iTimer.." seconds.", "FI", 2)
				else
					tFuncs:Debug("[Timer] Error when executing textcommand: "..name.." with interval of "..data.iTimer.." seconds: "..reply, "FI", -1)
				end
			end
		end
	end

	--// Save stats
	tStats.Thread.Listen.Status = "DoTimers: Save statistics files"
	for name, data in pairs(tHubSettings.SaveStats) do
		if data.iTimer and data.iTimer > 0 then
			tTimers.SaveStats[name] = tTimers.SaveStats[name] or { ["LastRun"] = 0 }
			if tTimers.SaveStats[name].LastRun+data.iTimer <= ctime then
				tTimers.SaveStats[name].LastRun = ctime
				local msg = tFuncs:FillVars(data.sContent, "AAAA")
				local res = tFuncs:WriteFile(name, msg.."\n") 
				if res then
					tFuncs:Debug("[Timer] Saved statistics to file: "..name..".log with interval of "..data.iTimer.." seconds.", "FI", 2)
				else
					tFuncs:Debug("[Timer] Error when saving statistics to file: "..name.." with interval of "..data.iTimer.." seconds", "FI", -1)
				end
			end
		end
	end

	--// Redirect Checker
	tStats.Thread.Listen.Status = "DoTimers: Redirect checker"
	if tHubSettings.tRedirects.iCheckInterval > 0 and ((tLogs.Redirects.LastRun or 0) + tHubSettings.tRedirects.iCheckInterval) < ctime and not tThreads.CheckRedirects then
		tFuncs:Debug("[Timer] Running thread: CheckRedirects", "FI", 2)
		tThreads.CheckRedirects = ThreadCreate(tRedirect.Check, {1, ssid})
	end

	--// Register hub to hublist regservers
	tStats.Thread.Listen.Status = "DoTimers: Reghub"
	if not tThreads.HubListReg and tHubSettings.tHubLists.AutoInterval > 0 and tLogs.Hublists.Registration.LastRun + tHubSettings.tHubLists.AutoInterval < ctime and tLogs.Hublists.Registration.LastRun + tHubSettings.tHubLists.MinInterval < ctime then
		tFuncs:Debug("[Timer] Running thread: HubListReg", "FI", 2)
		tThreads.HubListReg = ThreadCreate(tFuncs.RegHub, {1, "AAAA"})
	end

	--// Update Checker
	tStats.Thread.Listen.Status = "DoTimers: Update check"
	if not _NOUPDATE and tHubSettings.Main.bUpdateCheck and (tTimers.UpdateCheck.LastRun + tTimers.UpdateCheck.iInterval) <= ctime and not tThreads.Update then
		tFuncs:Debug("[Timer] Running thread: Update", "FI", 2)
		tThreads.Update = ThreadCreate(tUpdate.Thread, {1, "BBBB", "!update timer"})
		tTimers.UpdateCheck.LastRun = ctime
	end

	--// Check for passwords without password set and report to all logged in with level 10
	tStats.Thread.Listen.Status = "DoTimers: AccountCheck check"
	if tSecurity.General.AllowBlankPass and (tTimers.AccountCheck.LastRun + tTimers.AccountCheck.iInterval) <= ctime then
		tFuncs:Debug("[Timer] Running: AccountCheck", "FI", 2)
		tTimerFuncs:AccountCheck()
		tTimers.AccountCheck.LastRun = ctime
	end

	--// if we have users with blank passwords we need to 'remind' them to set password
	if tSecurity.General.AllowBlankPass and (tTimers.PasswordReminder.LastRun + tTimers.PasswordReminder.iInterval) <= ctime then
		tStats.Thread.Listen.Status = "DoTimers: PasswordReminder"
		tFuncs:Debug("[Timer] Running: PasswordReminder", "FI", 2)
		tTimerFuncs:PasswordReminder()
		tTimers.PasswordReminder.LastRun = ctime
	end

	tStats.Thread.Listen.Status = "DoTimers: Finished"
end,

["GetTimers"] = function(self)	-- // Not in use
	return tTimer
end,

["AddTimer"] = function(self, name, func, interval)	-- // Not in use
	tTimer[name] = { ["Func"] = loadstring(func), ["Interval"] = tonumber(interval) }
end,

["DelTimer"] = function(self, name)	-- // Not in use
	tTimer[name] = nil
end,

["StartTimer"] = function(self, tID)	-- // Not in use
	tThreads.Timer = ThreadCreate(function()
		tTimer.index = 1
		_STATUS = "Online"
		while _STATUS == "Online" do
			for name,data in pairs(tTimer) do
				if tTimer.index / data.Interval == math.floor(tTimer.index / data.Interval) then
					data.Func()
				end
			end
			ThreadWait(1000)
		end
	end, {})
end,




-- // functions that run on timers

["AccountCheck"] = function(self)
--// for now, the only check that is made here is if there are accounts that has empty passwords but anything could be added, like pruning of old accounts etc
	local res, msg, count = AccountManager:FindNoPassAccounts(ssid)
	if res and count then
		tSend:PmToLevel("This is a result of a check that runs every "..os.date("%H:%M:%S", tTimers.AccountCheck.iInterval) .." listing all regs in the hub that has no password set.\n\n"..msg, ssid, 10)
		tFuncs:SafeDebug("[AccountCheck] There are "..count.." regs without a password set", nil, 0)
	elseif res and not count then
		tFuncs:SafeDebug("[AccountCheck] "..msg, nil, 2)
	end
--	return (res or false), (msg or "an unkown error occured")
	return
end,

["PasswordReminder"] = function(self)	-- function to annoy online users that doesnt have their password set yet
	local client = tClient
	local count = 0
	for sid, data in pairs(client) do
		if data.bNotifySetOwnPass then
			local reg = tNick:GetReg(data.sNick)
			if reg and reg.sPassword == "" then -- we have a hit
				count = count + 1
				tFuncs:SafeMessage("P", sid, "AAAA", "You are logged in with an account without a password set, before you continue you must set your password with !setmypass <your password> . If you need help, please ask an operator in the hub." )
				tFuncs:SafeDebug("[PasswordReminder] Sent reminder to set password to "..data.sNick, "X")
			elseif reg and reg.sPassword ~= "" then -- we made bug and forgot to clear flag
				tFuncs:SafeDebug("[PasswordReminder] [Error] "..data.sNick.." has set password but 'bNotifySetOwnPass' is still true" , "X")
				data.bNotifySetOwnPass = nil
			end
		end
	end
	if count > 0 then
		tFuncs:SafeDebug("[PasswordReminder] Sent reminder to set password to "..count.." user(s).", nil, 1)
	end
	return true
end,

["ChristmasBonus"] = function(self)
	-- first step, find out if its christmas or not
	-- 1324724400 == 2011-12-24 00.00.00
	-- 1324897200 == 2011-12-26 00.00.00

	local now = os.time()
	if now > 1324724400 and now < 1324897200 then
		--second step - make random (timer runs once every 15 mins but we dont want it to spam that often, it should feel like a surprise)
		tTimers.Christmas.iInterval = math.random (1, 16) * 900	-- set next run to anywhere between 15 minutes to 4 hours (in 15 min splits)
		--tTimers.Christmas.iInterval = math.random(1, 16) * 60	-- set next run to anywhere between 15 minutes to 4 hours (in 15 min splits)

		--third step - select what to spam
		local spam = {
		[1] = "A lovely thing about Christmas is that it's compulsory, like a thunderstorm, and we all go through it together. Garrison Keillor",
		[2] = "Anyone who believes that men are the equal of women has never seen a man trying to wrap a Christmas present.",
		[3] = "Christmas is a time when you get homesick - even when you're home.",
		[4] = "Christmas isn't a season. It's a feeling. Edna Ferber",
		[5] = "Christmas makes me happy no matter what time of year it comes around. Bryan White",
		[6] = "Dear Santa,\nI'm writing to let you know that I've been naughty...\nand it was worth it.\nYou fat, judgmental bastard.",
		[7] = "Handmade presents are scary because they reveal that you have too much free time. Doug Coupland",
		[8] = "I love giving gifts and I love receiving them. I really like giving little kids extravagant gifts. You see their little faces light up and they get excited. If it's a really good gift, I love receiving it, like jewels, small islands. Gina Gershon",
		[9] = "I once bought my kids a set of batteries for Christmas with a note on it saying, toys not included. Bernard Manning",
		[10] = "I once wanted to become an atheist, but I gave up - they have no holidays. Henny Youngman",
		[11] = "I stopped believing in Santa Claus when I was six. Mother took me to see him in a department store and he asked for my autograph. Shirley Temple",
		[12] = "If you haven't got any charity in your heart, you have the worst kind of heart trouble. Bob Hope",
		[13] = "If you wake up Christmas day with a funny taste in you mouth just remember Santa only cums once a year",
		[14] = "In the old days, it was not called the Holiday Season\nthe Christians called it 'Christmas' and went to church\nthe Jews called it 'Hanukka' and went to synagogue\nthe atheists went to parties and drank.\n\nPeople passing each other on the street would say 'Merry Christmas!' or 'Happy Hanukka!' or (to the atheists) 'Look out for the wall!'",
		[15] = "Let's be naughty and save Santa the trip. Gary Allan",
		[16] = "Mail your packages early so the post office can lose them in time for Christmas. Johnny Carson",
		[17] = "Never worry about the size of your Christmas tree. In the eyes of children, they are all 30 feet tall.",
		[18] = "Nothing says holidays, like a cheese log. Ellen DeGeneres",
		[19] = "Nothing's as mean as giving a little child something useful for Christmas.",
		[20] = "Once again, we come to the Holiday Season, a deeply religious time that each of us observes, in his own way, by going to the mall of his choice. Dave Barry",
		[21] = "Pets, like their owners, tend to expand a little over the Christmas period. Frances Wright",
		[22] = "Santa Claus has the right idea. Visit people once a year. Victor Borge",
		[23] = "That's the true spirit of Christmas; people being helped by people other than me. Jerry Seinfeld",
		[24] = "The Supreme Court has ruled that they cannot have a nativity scene in Washington, D.C. This wasn't for any religious reasons. They couldn't find three wise men and a virgin. Jay Leno",
		[25] = "The main reason Santa is so jolly is because he knows where all the bad girls live. George Carlin",
		[26] = "The one thing women don't want to find in their stockings on Christmas morning is their husband.",
		[27] = "There are some people who want to throw their arms round you simply because it is Christmas; there are other people who want to strangle you simply because it is Christmas. Robert Lynd",
		[28] = "What I don't like about office Christmas parties is looking for a job the next day. Phyllis Diller",
		[29] = "The FlexHub team wishes everyone a Merry Christmas and a Happy New Year.",
		[30] = "What do you call people who are afraid of Santa Claus? Claustrophobic.",
		[31] = "Did you ever notice that life seems to follow certain patterns? Like I noticed that every year around this time, I hear Christmas music. Tom Sims",
		[32] = "There's nothing sadder in this world than to awake Christmas morning and not be a child. Erma Bombeck",
		[33] = "Let me see if Ive got this Santa business straight. You say he wears a beard, has no discernible source of income and flies to cities all over the world under cover of darkness? You sure this guy isnt laundering illegal drug money? Tom Armstrong",
		[34] = "Q:Why do all the other reindeers have brown noses?\nA:Because they cant stop as quickly as Rudolph!",
		[35] = "Just before Christmas, an honest politician, a generous lawyer and Santa Claus were riding in the elevator of a very posh hotel.\nJust before the doors opened they all noticed a $20 bill lying on the floor.\nWhich one picked it up?\n\nSanta of course, because the other two don't exist!",
		[36] = "Santa runs FlexHub too",
		}

		local num = 0
		while true do
			num = math.random(#spam)
			if num ~= tTimers.Christmas.iLastNum then break end
		end

		tSend:ChatToAll("FlexHub Christmas Bonus:\n\n"..spam[num] .."\n")
		tTimers.Christmas.iLastNum = num
	end
	return
end,

}
