--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

	This file is part of FlexHub.

	FlexHub is free software: you can redistribute it and/or modify
	it under the terms of the GNU Affero General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	FlexHub is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Affero General Public License for more details.

	You should have received a copy of the GNU Affero General Public License
	along with FlexHub. If not, see <http://www.gnu.org/licenses/agpl.txt>.

]]--

tUpdate = {

["Thread"] = function(self, ssid, line)	-- V2 OK
--	local res, result, err = pcall(tUpdate.Init, nil, ssid, line)
--	if not res then 
--		tFuncs:SafeDebug("[THREAD][TEST][Update] Error: "..(err or "no error message"))
--	else
--		tFuncs:SafeDebug("[THREAD][TEST][Update] Msg: "..(err or "no message"))
--	end

	local result, err = pcall(tUpdate.Run, nil, ssid, line)
	if not result then 
		tFuncs:SafeDebug("[THREAD][Update] Error: "..(err or "no error message"))
	end

	if lfs.currentdir().."/" ~= _APPPATH then
		tUpdate:DLog("Error while running update. currentdir is '"..lfs.currentdir() .. "' trying to recover...", nil, -1, true)
		lfs.chdir(_APPPATH)
		tUpdate:DLog("Error while running update. currentdir is '"..lfs.currentdir() .. "'. If this last folder matches the folder where FlexHub.exe is located you should be okay, else its time to restart the hub, seriously...", nil, -1, true)
	end

	tThreads.Update = nil
	ThreadDestroy()
end,

	["Run"] = function(self, ssid, line)
		local line = line:lower() or ""
		local output = ""
		local feedlevel = 1
		local tTextStrings = {
			["NoUpdateNeeded"] = "No update needed, this hub is running the most recent version.",
			["NoListTimer"] = "There is an update avaliable but you must go to the forum to download it. It is not possible to use inline update this time. You will get more information if you run the update command manually",
			["NoList"] = "\n\n\tThere is a new update available for your hub, but inline update is not possible. You need to follow the link\n\tin the changelog message and download a package and manually upgrade your hub software.\n",
			["OnErrorFile"] = "The update file from the server is corrupt or missing, please notify the FlexHub devs about this error. The full information is avaliable in \\logs\\Update.log",
			["OldVersion"] = "There is a new version on the server but it has a lower svn version than you. Please check the FlexHub forum for information if a downgrade is advised.",
			}

		local tUpdateData = {}
		tUpdateData.tUpd = false	-- release or beta table from version.tbl
		tUpdateData.tFiles2Update = {}
		--tFlexVersion = {}			-- is the table inside version.tbl
		--tFlexVersionData = {}		-- is the table inside the list files
		--tFlexVersionlist = {}		-- is the full table inside allversions.tbl

		local bCheckOnly	= true	-- true = only check, false = check and update
		local bAbort		= false	-- true = abort and do nothing
		local bUp			= false	-- true = files need updating
		local bNoList		= false	-- true = there are no files in the version.tbl, user are supposed to download new files from forum
		local bNoUpdate		= false	-- true = no update needed, hub is uptodate
		local bSkip			= false	-- true = there are updates avaliable but we're just reporting it
		local bForce		= false	-- true = user option, means we're forcing an update (overrides tHubSettings.Main.iLastupdate date) , doesnt fully work yet
		local bOldVersion	= false
		local bCompare		= false	-- force a compare between local and remote contents

		tUpdate.Data = {}	-- add stuff here that needs to be avaliable to all the update functions

		local tOut = {}
		if string.find(line, "timer") then
			tUpdate.Data.bTimer = true	-- true = we're running update from a timer, same as bCheckOnly but reporting is done differently
		elseif string.find(line, "list") then	-- this doesnt do anything yet
			table.insert(tOut, "list")
		else
			if string.find(line, "now") then
				table.insert(tOut, "now")
				bCheckOnly = false
			end
			if string.find(line, "compare") then
				table.insert(tOut, "compare")
				bCompare = true
			end
			if string.find(line, "force") then	-- overrides the updatecheck and always does an update even if timestamp matches or if its an older revision
				bForce = true
				table.insert(tOut, "force")
			end
			local sSVN = line:match("^%D-(%d%d%d%d)%D-$")
			if sSVN then
				tUpdate:DLog("found svn tag "..sSVN)
				table.insert(tOut, "svn "..sSVN)	-- this doesnt do anything yet
			end
		end

		tUpdate:DLog(" The update process is started by "..tUser:GetNick(ssid).. ((tOut[2] and " with options "..table.concat(tOut, ", ")) or (tOut[1] and " with option "..tOut[1]) or ""))

		--// Send stats
		local _,_ = pcall(http.request, "http://www.flexhub.org/update/index.php?hub="..(not tHubSettings.tRules.bRegOnly and tHubSettings.Main.sAddress or "private").."&os="..url.escape(_NOGUI and "Linux -nogui" or wx.wxGetOsDescription()).."&svn="..tSettings.General.svnversion.."&upd="..(bCheckOnly and 0 or 1))

		-- 1. Download the version.tbl from server, return values are true on error or false on success
		local test = tUpdate.GetVersionFile()
		if test then
			bAbort = true
			output = output..tTextStrings.OnErrorFile
		else
			tUpdateData.tUpd = tUpdate.UpdateOrNot(tFlexVersion, (bForce or bCompare))
		end

		-- 2. if tUpd is not false the info on the server is newer than iLastupdate date (or we're using force), we continue and get the list stored in tUpd
		if tUpdateData.tUpd == false then	-- we didnt need to update
			output = output.."\t"..tTextStrings.NoUpdateNeeded
			bNoUpdate = true
			bAbort = true
		else	-- we're updating
			local b, b1 = tUpdate.GetFile(tHubSettings.Main.sDownloadFileServerPath ..tUpdateData.tUpd.Source, tUpdateData.tUpd.File)
			if b == true then 
				if string.find(b1, "tFlexVersionData", 1, true) then
					local b2 = pcall(function() loadstring(b1)() end)
					if not b2 then
						output = output..tTextStrings.OnErrorFile
						tUpdate:DLog("[Main] Error: loadstring failed on downloaded list, the file is corrupt.")
						bAbort = true
					elseif tFlexVersionData.sTime == tUpdateData.tUpd.Time then	-- verifying data by comparing the timestamp in the two files
							tUpdate:DLog("[Main] check OK, "..tSettings.Update.sVersionFileName .." time is equal to "..tUpdateData.tUpd.File .." time")
					else
						tUpdate:DLog("[Main] check FAILED, "..tSettings.Update.sVersionFileName .." time is NOT equal to "..tUpdateData.tUpd.File .." time, aborting)")
						output = output..tTextStrings.OnErrorFile
						bAbort = true
					end
				else
					output = output..tTextStrings.OnErrorFile
					tUpdate:DLog("[Main] Error when downloading the update file :"..tUpdateData.tUpd.File)
					bAbort = true
				end
			elseif tonumber(b) and tonumber(b) == 404 then
				tUpdate:DLog("[Main] Error when trying to fetch the update information from server ".. tHubSettings.Main.sDownloadFileServerPath ..tUpdateData.tUpd.Source .." and file name "..tUpdateData.tUpd.File ..": The file is not found.")
				output = output..tTextStrings.OnErrorFile
				bAbort = true
			else
				output = output..tTextStrings.OnErrorFile
				tUpdate:DLog("[Main] Error while GetFile where sVersionFileServerPath is ".. tHubSettings.Main.sDownloadFileServerPath ..tUpdateData.tUpd.Source..", tUpd.File is "..tUpdateData.tUpd.File .." Error is : "..b)
				bAbort = true
			end
		end

		if not bAbort then
			if not tFlexVersionData[1] then	-- we know it doesnt have any numerical indexes.
				bAbort = true
				bNoList = true	-- new switch added, enabling the option to NOT list any files and only link it to the forum for a full download.
			else
				-- test for older version on server than local
				local new = tonumber(tFlexVersionData.sSvnRevision) or 0
				local current = tonumber(tSettings.General.svnversion) or 0
				if current > new and not (bForce or bCompare) then
					bAbort = true
					bOldVersion = true
				end
			end
		end

		if not bAbort then
			-- create the table of files that differ from the server list
			local test, res, tdata = tUpdate:CompareLocalWithList(tFlexVersionData)
			if test then
				output = output..tUpdate.MakeHeader(tFlexVersionData)
				if not tUpdate.Data.bTimer then
					output = output..res
				end
				tUpdateData.tFiles2Update = tdata
				if not bCheckOnly then
					bUp = true					-- set to true if actual file copying is needed
					-- make bUpdateInProgress here
				end
			else
				bNoUpdate = true
				output = output..res
			end
		end

		if bCheckOnly and not bAbort and not bNoUpdate then	-- we don't do any actual updates, only report what we found
			bSkip = true
			output = output.."\n\tNo actual file copying has been done, run '!update now' to download and install updates."
		end

		--3. the actual updating stuff
		if bUp and not bAbort and not bSkip then	-- there are files that need updating
			--a. we start by getting all the files that is needed and verify that the download is ok as we should not do all the other stuff if it isn't
			local g, g1, g2 = tUpdate.UpdateFile(tUpdateData.tFiles2Update, tHubSettings.Main.sDownloadFileServerPath..tUpdateData.tUpd.Source)
			-- g returns true on success and false on failure, g1 is the text to send to output, g2 is the updated table
			if g then	-- all files updated OK
				output = output.."\nDownloading the needed files:\n"..g1
				tUpdateData.tFiles2Update = g2
			else	-- one or more failed and we're aborting
				output = output.."\nDownloading the needed files:\n"..g1.."\n\tOne or more of the downloaded files do not match the TTH in the version information, the update is aborted and no local files are changed.\n\tIf re-running the update command does not resolve this, please notify the FlexHub devs about this error!" 
				bAbort = true
				tUpdate:DLog("[Main] There was an error when downloading the new files, the update process is aborted")
			end

			-- b. as we're still here we continue by creating a temp folder called backups if it doesnt already exist
			if not bAbort then
				tUpdate.ChangeDir(_APPPATH, 200)
				local d, d1 = tFuncs:CheckFolder(tSettings.Update.sLocalPath)
				if d then
					tUpdate:DLog("[Main] Verifying backup folder, "..d1)
				else
					tUpdate:DLog("[Main] There was an error when creating the folder: "..tSettings.Update.sLocalPath .." Message : "..d1)
					bAbort = true
					output = output.."\n\tThere was an error when creating the backups folder, the update is aborted.\n\tError message is :"..d1
				end
			end

			-- c. make a folder inside backups to store the backup copies, tweaked so it always makes a new folder if first folder is not empty
			local sTempfolder = os.date("%Y-%m-%d")
			if not bAbort then
				tUpdate.ChangeDir(tSettings.Update.sLocalPath, 214)	-- now whe're inside backups
				local count = 0
				while true do
					local e = tUpdate.CreateFolder(sTempfolder, true)
					if e and e == 1 then	-- directory is not empty, we need to make another folder
						count = count + 1
						if string.find(sTempfolder, "_", 1, true) then
							sTempfolder = string.sub(sTempfolder, 1, 10).."_"..tostring(count)
						else
							sTempfolder = sTempfolder.."_"..tostring(count)
						end
					elseif e then	-- we failed to create a directory
						tUpdate:DLog("[Main] Error creating folder : "..sTempfolder .." Message : "..e)
						bAbort = true
						output = output.."\n\tThere was an error when creating the backup folder '"..sTempfolder .."', the update is aborted.\n\tThe error message is : "..e
						break
					else
						tUpdate:DLog("[Main] Successfully created sTempfolder : "..sTempfolder)
						break
					end
				end
				tUpdate.ChangeDir(_APPPATH, 235)
			end

			-- d. copy the files that are going to be replaced into the tempfolder above (we're still inside same if-loop so local variable still valid)
			if not bAbort then
				output = output.."\nCreating backup copies of the old files before updating, if something goes wrong with the update the old files are located in " ..tSettings.Update.sLocalPath.."/"..sTempfolder .."\n\n"
				local newpath = ""
				for i, j in ipairs(tUpdateData.tFiles2Update) do
					newpath = tUpdate.FixPath(j.sPath, j.sName)
					--tUpdate:DLog( "(244) newpath = '"..newpath .."' sPath = '"..j.sPath .."' sName = '"..j.sName .."'")
					if tUpdate.CheckFile(newpath) then
						local f, f1 = tUpdate.CopyFile(j.sPath, j.sName, tSettings.Update.sLocalPath.."/"..sTempfolder)
						if f then
							output = output.."\tFile "..newpath.." copied OK\n"
						else
							tUpdate:DLog("[Main] "..newpath.." could not be copied, message = "..f1)
							output = output.."\tThere was an error when trying to copy "..newpath ..", the message is :".. f1
							bAbort = true
							break
						end
					else
						tUpdate:DLog("[Main] newpath is False, the file "..newpath .." does not exist.")
					end
				end
			end

			-- e. the scary last part, we save the new files over whats on disk :D
			if not bAbort then
				tUpdate.ChangeDir(_APPPATH, 263)
				output = output.."\nWriting the new files to the hub folder:\n\n"
				for i, j in ipairs(tUpdateData.tFiles2Update) do
					local h, h1 = tUpdate.SaveFile(j)
					if h then
						output = output.."\tFile "..h1.."\n"
					else
						output = output.."\tFile "..h1.."\n"
						bAbort = true
							-- add logg output here for DLOG and for rollback
						break
					end
				end

				if bAbort then	-- we had an error during save operation, report and tell user what to do to fix it.
					output = output.."\n\nThere were one or more files that failed to be saved over the current files, please try the following steps to fix this error:\n"
					.."\t1) First make sure that you do not have any of the current files open in another application or that they are locked in any other way.\n"
					.."\t2) Make sure that they are not write protected and that the user running the hub has write access to them.\n"
					.."\t3) Move all the files in the '"..tSettings.Update.sLocalPath.."/"..sTempfolder .."' back to their original positions.\n"
					.."\t4) Re-run the update command.\n\n\tIf this does not help please do step 1-3 again before restarting your hub and inform the FlexHub devs about this error."
				end
			end
		end

		-- 4. set timestamp of update file as last updated if we did some updating
		local sRebootMessage = ""
		local sRebootMessageSpam = "\n\t\t(FlexHub is the only hub software that gives you updates without sacrificing your uptime  :-))\n"

		if not bAbort and tUpdateData.tUpd and tFlexVersionData.bRestartHub then
			if bUp then
				sRebootMessage = "\tYou have to restart the hub (use !restart or manually close the hub console and start it again)"
				tSettings.Update.bNeedRestart = true
			else
				sRebootMessage = "\tPlease note that a restart of the hub IS REQUIRED after the update! Use '!update now' to update."
			end
		elseif not bAbort and tUpdateData.tUpd then
			sRebootMessage = "\tA restart of the hub is NOT needed after this update, but remember to run !rs to reload all scripts."
		end

		tUpdate.ChangeDir(_APPPATH, 306)

		if bUp and not bAbort then
			output = output.."\nUpdate finished successfully."
			tUpdate:DLog("[End] The update process ended successfully. The hub software is updated.", nil, 0, true)
			tHubSettings.Main.iLastupdate = tFlexVersionData.sTime
			tFuncs:SaveSettings("settings")
		elseif bNoList and tUpdate.Data.bTimer then
			output = output..tTextStrings.NoListTimer
			tUpdate:DLog("[End] The update process ended successfully. No inline update provided, the update can be found on FlexHub.org.", nil, 0, true)
			-- add pm to opchat here to notify about the update
		elseif bNoList then
			output = output..tTextStrings.NoList..tUpdate.MakeHeader(tFlexVersionData)
			tUpdate:DLog("[End] The update process ended successfully. No inline update provided, the update can be found on FlexHub.org.", nil, 0, true)
		elseif bNoUpdate then
			tUpdate:DLog("[End] The update process ended successfully. No update needed.")
		elseif bCheckOnly and not bAbort then
			tUpdate:DLog("[End] The update process ended successfully. Check only")
		elseif bOldVersion then
			output = output..tTextStrings.OldVersion
			tUpdate:DLog("[End] Local svn is higher than whats on server. If this is not intentional please check FlexHub forum for information.")
		elseif tUpdate.Data.bTimer then
			feedlevel = 0
			-- we want to cut the output short short
			if tFlexVersionData then
				output = output..tUpdate.MakeHeader(tFlexVersionData)
			end
		else
			tUpdate:DLog("[End] The update process was aborted due to errors.", nil, -1, true)
		end

		tFuncs:CloseFile("Update")

		if tUpdate.Data.bTimer then
			tFuncs:SafeDebug("[Update][Timer] "..output..sRebootMessage, nil, feedlevel)
		else
			tFuncs:SafeMessage("M", ssid, "AAAA", "\n\n\tUpdate results:\n"..output.."\n"..sRebootMessage..sRebootMessageSpam)
		end

		-- 5. trying to leave the table cleaned and the dishes done ( might have nicked the silverware when u werent looking tho ;-)
		-- clear bUpdateInProgress here

		tUpdateData = nil
		tFlexVersionData = nil
		tFlexVersion = nil
		tUpdate.Data = nil
	end,

	["CompareLocalWithList"] = function(self, tFlexVersionData)
		--if tFlexVersionData.bRestartHub then
		--	tFuncs:SafeDebug("[THREAD][Update] DEBUG Setting 'tFlexVersionData.bRestartHub' is already TRUE")
		--end
		local res, newpath = "", ""
		local tFiles2Update = {}
		for i, j in ipairs(tFlexVersionData) do
			newpath = tUpdate.FixPath(j.sPath, j.sName)
			local c = tUpdate.GetHash(newpath)
			if not c or c ~= j.sTTH then	-- we need to update
				table.insert(tFiles2Update, j)
				for i, name in ipairs(tSettings.Update.tRestartFiles) do
					if string.lower(name) == string.lower(j.sName) then
		--				tFuncs:SafeDebug("[THREAD][Update] DEBUG Setting 'tFlexVersionData.bRestartHub' TRUE because "..name.." was found in 'tSettings.Update.tRestartFiles'")
						tFlexVersionData.bRestartHub = true
					end
				end
				c = c or "file not found"
				tUpdate:DLog("[TTHCheck] "..newpath.." needs updating, TTH: "..j.sTTH .." does not match local TTH: ".. c)
				res = res.."\t"..newpath.."\n"
			end
		end

		if tFiles2Update[1] then
			tFuncs:OpenLogFile("Update")	-- no need to log unless we are actually going to do something
			return true, "\nThe following "..#tFiles2Update .." of "..#tFlexVersionData .." files need to be updated:\n\n"..res, tFiles2Update
		else	-- all files were ok
			return false, "\tThere are no files that need to be updated."
		end
	end,

	-- changed to use same format as real Debug + one extra parameter to force feed while running on timer
	["DLog"] = function(self, msg, opt, level, force)	-- V2 OK
		tFuncs:WriteFile("Update", "[Update]"..msg .."\n")
		if not tUpdate.Data.bTimer or force then
			tFuncs:SafeDebug("[THREAD][Update] "..msg, (opt or nil), (level or 1))
		end
	end,

-- New function, only deals with getting version.tbl file(s)
	["GetVersionFile"] = function()	--(path is always tHubSettings.Main.sVersionFileServerPath)
		local filename = tSettings.Update.sVersionFileName or "version.tbl"

		-- returns true if successfully got file, else returns error code from web server
		local a, sFileBuffer = tUpdate.GetFile(tHubSettings.Main.sVersionFileServerPath, filename)
		if a == true then
			if string.find(sFileBuffer, "tFlexVersion", 1, plain) then
				tUpdate:DLog("[GetVersionFile] "..filename.." succesfully downloaded from "..tHubSettings.Main.sVersionFileServerPath)
			else
				tUpdate:DLog("[GetVersionFile] "..filename .." succesfully downloaded from "..tHubSettings.Main.sVersionFileServerPath .." but it does NOT look like we expected, aborting")
				return true
			end
		elseif tonumber(a) and tonumber(a) == 404 then
			tUpdate:DLog("[GetVersionFile] ERROR when trying to fetch the update information where server path is ".. tHubSettings.Main.sVersionFileServerPath .." and file name "..filename .." The file is not found.")
			return true
		else
			tUpdate:DLog("[GetVersionFile] ERROR when trying to fetch the update information where sVersionFileServerPath = ".. tHubSettings.Main.sVersionFileServerPath .." sVersionFileName = "..filename)
			return true, "There was an error when trying to fetch the update information from the FlexHub server. The server message is :"..tostring(a).."\n" .. tostring(sFileBuffer)
		end

		local a = pcall(function() loadstring(sFileBuffer)() end)
		if a then
			return false
		else
			tUpdate:DLog("[GetVersionFile] The update file from the server is corrupt, please notify the FlexHub devs about this error.")
			return true
		end
	end,

	["UpdateFile"] = function(tbl, downloadpath)
		local res, sPathName = "\n", ""
		local success, d = true, false

		for i in ipairs (tbl) do
			local sR = ""
		--	tUpdate:DLog( "download (429) : "..i.." : '"..tbl[i].sPath.."' : '"..tbl[i].sName .."'")
			sPathName = tUpdate.FixPath(tbl[i].sPath, tbl[i].sName)
			d, tbl[i].sBuffer = tUpdate.GetFile(downloadpath, sPathName)

			if d and d == true then	-- we're okay
				if ( string.match(tbl[i].sName, ".-lua$") or string.match(tbl[i].sName, ".-txt$") ) and tth.buffer(tbl[i].sBuffer) == tbl[i].sTTH then
					sR = "File "..sPathName .." downloaded OK, TTH confirmed"
					tUpdate:DLog("[UpdateFile] lua file: "..sR)
					success = true
				else 	--// test non lua files differently
					success, sR = tUpdate.SaveTempFile(tbl[i])
					tUpdate:DLog("[UpdateFile] non lua file: "..sR)
					if not success then
						tbl[i].sBuffer = false 
						break
					end
				end
			else
				sR = "There was an error when downloading: "..downloadpath..sPathName..". The message is: "..d
				tUpdate:DLog("[UpdateFile] "..sR)
				success = false
				break
			end
			res = res.."\t"..sR.."\n"
		end
		return success, res, tbl
	end,

	["SaveFile"] = function(tbl)
		local res , sTTH = "", "file not found"
		local sPathName = tUpdate.FixPath(tbl.sPath, tbl.sName)

		if tbl.sPath and tbl.sPath ~= "" then
			local res, mesg = tFuncs:CheckFolder(tbl.sPath)
			if not res then
				tFuncs:SafeDebug("[THREAD] [Update] error from CheckFolder: "..(mesg or "no mesg"))
				return false, mesg 
			end
		end

		local o = io.open(sPathName,"wb")
		if o then
			o:write(tbl.sBuffer)
			o:close()
			sTTH = tUpdate.GetHash(sPathName) or "file not found"
			if sTTH == tbl.sTTH then
				res = sPathName.." saved OK, TTH confirmed"
				tUpdate:DLog("[SaveFile] "..res)
				return true, res
			else
				res = "There was an error when checking '"..sPathName .."'. Local TTH ".. sTTH .." does not match the TTH in the version file : "..tbl.sTTH
			end
		else
			res = "There was an error when trying to write "..sPathName
		end
		tUpdate:DLog("[SaveFile] "..res)
		return false, res
	end,

	["SaveTempFile"] = function(tbl)
		local tempfile = ".tEmpdOwnloadfIle"
		local res , sTTH = "", "file not found"
		local o = io.open(tempfile, "wb")
		if o then
			o:write(tbl.sBuffer)
			o:close()
			sTTH = tUpdate.GetHash(tempfile) or "file not found"
			os.remove(tempfile)

			if sTTH == tbl.sTTH then
				return true, tbl.sName.." downloaded OK, TTH confirmed"
			else
				res = "There was an error when checking '"..tbl.sName .."'. Local TTH ".. sTTH .." does not match the TTH in the version file : "..tbl.sTTH
			end
		else
			res = "There was an error when trying to write the temporary file, TTH could not be verified."
		end
		return false, res
	end,

	["GetFile"] = function(downloadpath, filename)	-- filename contains folder + file name
		if string.sub(downloadpath, -1, -1) ~= "/" then
			downloadpath = downloadpath.."/"
		end
		local x, vReturnCode = false, false
		for retry = 1, 5 do
			x, vReturnCode, _ = http.request(downloadpath..filename)
			if vReturnCode == 200 then	-- returncode 200 means OK
				return true, x
			elseif vReturnCode == 404 then -- its no point continuing
				return tostring(vReturnCode), x
			else
				tUpdate:DLog("[GetFile] Failed to download "..filename.." from server, the error is '"..tostring(vReturnCode) .."'. Waiting to retry ("..retry .."/5).", nil, 2, true)
				ThreadWait(60000) -- 60 secs delay
			end
		end
		return tostring(vReturnCode), x
	end,

	["CopyFile"] = function(path, filename, target)
		local here = lfs.currentdir()
		local mess = ""
		local success = true
		local filewpath = tUpdate.FixPath(path, filename)
		local sTTH = tUpdate.GetHash(filewpath)

		local a = io.open(filewpath, "rb")
		if a then
			local dFile = a:read("*a")
			a:close()

			local c, c1 = tUpdate.ChangeDir(here .."/"..target, 531)
			if not c then
				return false, "[CopyFile] "..c1
			end

			if path and path ~= "" then
				local b, b1 = tFuncs:CheckFolder(path)
				if not b then	-- failure
					mess = "[CopyFile] "..b1
					tUpdate:DLog(mess)
					success = false
					return false, mess
				else
					tUpdate.ChangeDir(path, 544)
				end
			end

			local o = io.open(filename, "wb")
			if o then
				o:write(dFile)
				o:close()
			else
				mess = "[CopyFile] Could not save "..filename.." to "..path
				tUpdate:DLog(mess)
				success = false
				return false, mess
			end

			if sTTH == tUpdate.GetHash(filename) then
				tUpdate:DLog("[CopyFile] "..filewpath.." copied OK.")
			else
				mess = "[CopyFile] Error when verifying "..filewpath ..", the TTH does not match " ..target.."/"..filewpath
				tUpdate:DLog(mess)
				success = false
				return false, mess
			end
		else
			tUpdate:DLog("[CopyFile] "..filewpath .." does not exist, skipping.")
		end
		tUpdate.ChangeDir(here, 570)
		return success, mess
	end,

	-- function to test whether an update is needed based on the settings lastupdate and announcebeta and timestamps in the version file
	["UpdateOrNot"] = function(tVer, bForce)
		local bRelease, bBeta = false, false
		if not tHubSettings.Main.iLastupdate then tHubSettings.Main.iLastupdate = 0 end

		if tHubSettings.Main.bAnnounceBeta and tVer.Beta.Hub.Time ~= 0 and tHubSettings.Main.iLastupdate < tVer.Beta.Hub.Time then
			bBeta = true
		end

		if tVer.Release.Hub.Time ~= 0 and tHubSettings.Main.iLastupdate < tVer.Release.Hub.Time then
			bRelease = true
		end

		if bForce and tVer.Release.Hub.Time > tVer.Beta.Hub.Time then
			tUpdate:DLog("[UpdateOrNot] force or compare are true (release)")
			bRelease = true
		elseif bForce then
			tUpdate:DLog("[UpdateOrNot] force or compare are true (beta)")
			bBeta = true
		end

		if bRelease and bBeta and tVer.Release.Hub.Time > tVer.Beta.Hub.Time then
			tUpdate:DLog("[UpdateOrNot] both are true, release is newer than beta (update is using release)")
			return tVer.Release.Hub
		elseif bRelease and bBeta then
			tUpdate:DLog("[UpdateOrNot] both are true, beta is newer than release (update is using beta)")
			return tVer.Beta.Hub
		elseif bRelease then
			tUpdate:DLog("[UpdateOrNot] release are true (update is using release)")
			return tVer.Release.Hub
		elseif bBeta then
			tUpdate:DLog("[UpdateOrNot] beta are true (update is using beta)")
			return tVer.Beta.Hub
		else 
			tUpdate:DLog("[UpdateOrNot] no update is newer than iLastupdate (update not needed)")
			return false
		end
	end,

	-- get hash, does check file first now, returns TTH or false on failure
	["GetHash"] = function(filewpath)
		if filewpath and tUpdate.CheckFile(filewpath) then
			return tth.file(filewpath)
		else
			return false
		end
	end,

	-- check if file exists, works only with merged paths or from same folder, returns true or false
	["CheckFile"] = function(filewpath)
		local a = io.open(filewpath, "rb")
		if a then
			a:close()
			return true
		else
			return false
		end
	end,

	-- function to take care of merging path and filename correctly
	["FixPath"] = function(path, filename)
		if path and path ~= "" and filename then
			return path.."/"..filename
		elseif filename then
			return filename
		else
			return ""
		end
	end,

	["CreateFolder"] = function(target, empty)
		local here = lfs.currentdir()
		local test = false

		if string.find(target,"/") then	-- there is more than one level in the target path, lets loop it
			tUpdate:DLog("[CreateFolder] target "..target .." contains more than one level.")
			while true do
				local a, _ = string.find(target,"/", 1, true)
				local dir = string.sub(target,1 ,a-1)
				local b = tUpdate.CreateFolder( dir)
				if b then	-- failure
				--	tUpdate:DLog("[CreateFolder] message is "..b)
					tUpdate:DLog("[CreateFolder] error when chdir to "..dir)
					tUpdate.ChangeDir(here, 657)
					return "There was an error creating ".. dir
				else
					tUpdate.ChangeDir(dir, 660)
				--	tUpdate:DLog("[CreateFolder](661): dir is "..dir)
				end
				target = string.sub(target,a +1 , -1)
				--tUpdate:DLog("[CreateFolder] (664): new target is : "..target)
				if not string.find(target,"/") then
				--	tUpdate:DLog("[CreateFolder] (666): bPing is false, all levels done")
					break
				end
			end
		end

		local attr = lfs.attributes(target)
		if attr and attr.mode == "directory" then	-- then the folder exist and we're ready to go
			--tUpdate:DLog("[CreateFolder] attr is true")
			if empty then	-- make sure its empty, if its not return with message
				for file in lfs.dir(target) do
					if file ~= "." and file ~= ".." then
						test = 1
				--		tUpdate:DLog("[CreateFolder] ".. target .. " is not empty")
						break
					end
				end
			end
		else	-- create the folder
		--	tUpdate:DLog( "[CreateFolder] attr is false, making new folder '"..target .."'")
			local a, b = lfs.mkdir(target)
			if a == true then 
			--	tUpdate:DLog("[CreateFolder]a is true, we made new folder")
			else
				tUpdate:DLog("[CreateFolder] error when creating the new folder '"..target.."'")
				test = "Error: It is not possible to create the folder "..target ..". The file system message is : "..b
			end
		end
		tUpdate.ChangeDir(here, 694)
		return test
	end,

	["ChangeDir"] = function(target, line)
		local a, err = lfs.chdir(target)
		if not line then line = "" end
		if not a then
			tUpdate:DLog("[ChangeDir] could not change to "..target..". Message is : "..string.gsub(err, "\n", " "))
--			tUpdate:DLog("[ChangeDir] (703): could not change to "..target..". Message is : "..string.gsub(err, "\n", " ").." Line is "..line)
			return false, err
		else
			return true
		end
	end,

	-- add to output version info from header
	["MakeHeader"] = function(tFlexVersionData)
		local line = string.rep(string.char(175),80)
		local sNewSvn, res, warn = "", "", ".\n"
		if tFlexVersionData.sSvnRevision ~= "" then
			sNewSvn = "svn revision "..tFlexVersionData.sSvnRevision
		end

		local _, _, sNewVersion = string.find(tFlexVersionData.sAppVersion, ".-(%d+.%d+).-")
		sNewVersion = tonumber(sNewVersion) or 0
		local iCurrentSvn = tonumber(tSettings.General.svnversion) or 0
		tFlexVersionData.sSvnRevision = tonumber(tFlexVersionData.sSvnRevision) or 0

	--	tFuncs:SafeDebug("sNewVersion : "..sNewVersion)
	--	tFuncs:SafeDebug("tFlexVersionData.sAppVersion : "..tFlexVersionData.sAppVersion)
	--	tFuncs:SafeDebug("sCurrentVersion : "..sCurrentVersion)
	--	tFuncs:SafeDebug("_APPVER : ".._APPVER)
	--	tFuncs:SafeDebug("iCurrentSvn : "..iCurrentSvn)
	--	tFuncs:SafeDebug("tFlexVersionData.sSvnRevision : "..tFlexVersionData.sSvnRevision)
	--	tFuncs:SafeDebug("tFlexVersionData.sTime : "..tFlexVersionData.sTime)
	--	tFuncs:SafeDebug("tHubSettings.Main.iLastupdate : "..tHubSettings.Main.iLastupdate)

		if tUpdate.Data.bTimer and tFlexVersionData.sSvnRevision < iCurrentSvn then
			res = " There is a new update avaliable, it will update your hub to "..tFlexVersionData.sAppName.." "..tFlexVersionData.sAppVersion.." "..sNewSvn.." (you have ".._APPVER.." svn revision "..iCurrentSvn..") "
		else
			if tFlexVersionData.sSvnRevision < iCurrentSvn then
				warn = ", and you should\n\tnote that the version in the update looks like it is an older version than yours\n\t(you have "..tFlexVersionData.sAppName.." ".._APPVER .." svn revision "..iCurrentSvn ..").\n\n"
			end

			res = "\n\tThis release will update your hub to "..tFlexVersionData.sAppName.." "..tFlexVersionData.sAppVersion .." "..sNewSvn ..warn
			if tFlexVersionData.sChangelog ~= "" then
				res = res.."\nThe changelog :\n"..line.."\n".. tFlexVersionData.sChangelog .."\n\n"
			end
			if tFlexVersionData.sForumLink ~= "" then
				res = res.."\tTalk about the release here : ".. tFlexVersionData.sForumLink .."\n"
			end
			if tFlexVersionData.sBugReport ~= "" then
				res = res.."\tReport bugs here : ".. tFlexVersionData.sBugReport .."\n"
			end
			res = res..line.."\n"
		end

		return res
	end,
}
--tFuncs:CheckFolder(foldername)