--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tUserCmds = {

["Replace"] = function(self, cmdstring, cmdname, nmdc)
	local replace = {}
--~ 	tFuncs:Debug("[tUserCmds:Replace] "..tostring(cmdstring).." / "..tostring(cmdname).." / "..tostring(nmdc))

	if nmdc then						--// Backwards compatibility, replace %[userNI] style vars with %[nick] style ones
		cmdstring = "$<%[mynick]> "..cmdstring

		replace = {
			{"%%s", (cmdname or "")},
			{"%%%[userNI%]", "%%[nick]"},
			{"%%%[userI4%]", "%%[ip]"},
			{"%%%[myNI%]", "%%[mynick]"},
		}
	else
		cmdstring = ADC_E(ADC_E(cmdstring))
		replace = {
			{"%%s", (cmdname or "")},
		}
	end

	for i, v in ipairs(replace) do
		cmdstring = cmdstring:gsub(v[1],v[2])
	end

	return cmdstring
end,

["Build"] = function(self)
	tUserCmds.NMDC = {}
	tUserCmds.ADC = {}
	tUserCmds.allow = {}

	local cmdprefix = string.sub((tHubSettings.Main.sCmdPrefixes or "!"), 1, 1)

	--// build total commandlist
	local totallist = {}  
	for name, func in pairs(HubCmd) do
		totallist[name] = true
	end
	for name, func in pairs(SetCmd) do
		totallist[name] = true
	end

	--// Add categorized commands
	for _, cat in ipairs(SubSet.HubCommands) do
		for _, name in ipairs(cat.cmds) do
			totallist[name] = nil
			if HubCmd[name] and HubCmd[name]["ucmd"] then
				for num, data in ipairs(HubCmd[name]["ucmd"]) do
					table.insert(tUserCmds.NMDC, "$UserCommand "..data[1].." "..data[2].." "..NMDC_E(tHubSettings.Main.sMenuname or "FlexHub Commands").."\\"..NMDC_E(cat.text).."\\"..NMDC_E(data[3])..tUserCmds:Replace(data[4], name, true).."&#124;|")
					table.insert(tUserCmds.ADC, "ICMD "..ADC_E((tHubSettings.Main.sMenuname or "FlexHub Commands")).."/"..ADC_E(cat.text).."/"..ADC_E(data[3]).." TTHMSG\\s"..tUserCmds:Replace(data[4], name) .."\\n CO1 CT"..data[2].."\n")
					table.insert(tUserCmds.allow, {["cmd"] = "HubCmd", ["name"] = name})
				end
			end
		end
	end
	for _, cat in ipairs(SubSet.SetCommands) do
		for _, name in ipairs(cat.cmds) do
			totallist[name] = nil
			if SetCmd[name] and SetCmd[name]["ucmd"] then
				for num, data in ipairs(SetCmd[name]["ucmd"]) do
					table.insert(tUserCmds.NMDC, "$UserCommand "..data[1].." "..data[2].." "..NMDC_E(tHubSettings.Main.sMenuname or "FlexHub Commands").."\\Settings\\"..NMDC_E(cat.text).."\\"..NMDC_E(data[3])..tUserCmds:Replace(data[4], name, true).."&#124;|")
					table.insert(tUserCmds.ADC, "ICMD "..ADC_E((tHubSettings.Main.sMenuname or "FlexHub Commands")).."/Settings/"..ADC_E(cat.text).."/"..ADC_E(data[3]).." TTHMSG\\s"..tUserCmds:Replace(data[4], name) .."\\n CO1 CT"..data[2].."\n")
					table.insert(tUserCmds.allow, {["cmd"] = "SetCmd", ["name"] = name})
				end
			end
		end
	end

	--// Add miscellaneous commands
	local restlist = {}
	for name, res in pairs(totallist) do table.insert(restlist, name) end --// build list of commands that weren't categorized
	table.sort(restlist)

	for _, name in ipairs(restlist) do
		if HubCmd[name] and HubCmd[name]["ucmd"] then
			for num, data in ipairs(HubCmd[name]["ucmd"]) do
				table.insert(tUserCmds.NMDC, "$UserCommand "..data[1].." "..data[2].." "..NMDC_E(tHubSettings.Main.sMenuname or "FlexHub Commands").."\\Miscellaneous\\"..NMDC_E(data[3])..tUserCmds:Replace(data[4], name, true).."&#124;|")
				table.insert(tUserCmds.ADC, "ICMD "..ADC_E((tHubSettings.Main.sMenuname or "FlexHub Commands")).."/Miscellaneous/"..ADC_E(data[3]).." TTHMSG\\s"..tUserCmds:Replace(data[4], name).."\\n CO1 CT"..data[2].."..\n")
				table.insert(tUserCmds.allow, {["cmd"] = "HubCmd", ["name"] = name})
			end
		elseif SetCmd[name] and SetCmd[name]["ucmd"] then
			for num, data in ipairs(SetCmd[name]["ucmd"]) do
				table.insert(tUserCmds.NMDC, "$UserCommand "..data[1].." "..data[2].." "..NMDC_E(tHubSettings.Main.sMenuname or "FlexHub Commands").."\\Settings\\Miscellaneous\\"..NMDC_E(data[3])..tUserCmds:Replace(data[4], name, true).."&#124;|")
				table.insert(tUserCmds.ADC, "ICMD "..ADC_E((tHubSettings.Main.sMenuname or "FlexHub Commands")).."/Settings/Miscellaneous/"..ADC_E(data[3]).." TTHMSG\\s"..tUserCmds:Replace(data[4], name).."\\n CO1 CT"..data[2].."..\n")
				table.insert(tUserCmds.allow, {["cmd"] = "SetCmd", ["name"] = name})
			end
		end
	end

	--// Add trigger rightclicks
	local jointriggers = {}
	for name, data in pairs(tCommands) do
		if not tFuncs:CheckFlag((data.iOptions or 0), 32) and not tFuncs:CheckFlag((data.iOptions or 0), 8) then
			table.insert(jointriggers, name)
		end
	end

	table.sort(jointriggers, 	--// Sort alfabetically
		function(a,b)
			local ah = nil
			local bh = nil
			if tCommands[a].sHelpText and tCommands[a].sHelpText ~= "" then ah = string.lower(tCommands[a].sHelpText) end
			if tCommands[b].sHelpText and tCommands[b].sHelpText ~= "" then bh = string.lower(tCommands[b].sHelpText) end

			if (ah or a) < (bh or b) then return true
			else return false
			end
		end
	)

	for num, name in ipairs(jointriggers) do
		local cmdname = (tCommands[name].sHelpText or name)
		if cmdname == "" then cmdname = name end
		cmdname = string.gsub(cmdname, "[\t\n]", "")
		table.insert(tUserCmds.NMDC, "$UserCommand 1 3 "..NMDC_E(tHubSettings.Main.sMenuname or "FlexHub Commands").."\\"..NMDC_E("Textcommands\\"..cmdname)..tUserCmds:Replace(cmdprefix..NMDC_E(name), nil, true).."&#124;|")
		table.insert(tUserCmds.ADC, "ICMD "..ADC_E((tHubSettings.Main.sMenuname or "FlexHub Commands")).."/"..ADC_E("Textcommands/")..ADC_E(cmdname).." ".."TTHMSG\\s"..cmdprefix..name.."\\n CT3\n")
		table.insert(tUserCmds.allow, {["cmd"] = "jointriggers", ["name"] = name})
	end
end,

["CreateUC"] = function(self, ssid, protocol)
	if not tUserCmds.allow then tFuncs:Debug("[ERROR] [tUserCmds] [CreateUC] tUserCmds.allow doesn't exist.") return {}, "" end

	local tReply = {}
	for indx, data in ipairs(tUserCmds.allow) do
		if data.cmd == "HubCmd" and HubCmd[data.name]["allowed"](ssid, "", data.name) then
			table.insert(tReply, tUserCmds[protocol][indx])
		elseif data.cmd == "SetCmd" and SetCmd[data.name]["allowed"](ssid, "", data.name) then
			table.insert(tReply, tUserCmds[protocol][indx])
		elseif data.cmd == "jointriggers" and (tClient[ssid].iProfile >= (tCommands[data.name].iUseProfile or tCommands[data.name].iProfile or -1)) then
			table.insert(tReply, tUserCmds[protocol][indx])
		end
	end
	return table.concat(tReply, "")
end,

}
