--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tSend = {

["Debug"] = function(self, msg, debugtype, debuglevel)
	tFuncs:Debug(msg, debugtype, debuglevel)
end,

["UpdateRooms"] = function(self, ssid)	--// To be used when a user changes profile, or a profilelevel changes
	local client = tClient[ssid]
	for vsid, data in pairs(tClient) do
		if data.bBot and vsid ~= "AAAA" then
			--// Quit the room
			if client.sProtocol == "ADC" then
				tSend:Raw(ssid, "IQUI "..vsid.."\n")
			elseif client.sProtocol == "NMDC" then
				tSend:Raw(ssid, "$Quit "..data.sNick.."|")
			end

			--// Send the allowed room
			if tUser:IsVisibleTo(vsid, ssid) then
				local inf = tUser:GetMyInfo(vsid, ssid)
				if inf then
					tSend:Raw(ssid, inf)
				end
			end
		end
	end
end,

["ADCUserIps"] = function(self, sid, clear)
	if tClient[sid] and tClient[sid].sProtocol == "ADC" then
		if tUser:IsOp(sid) then
			for vsid, data in pairs(tClient) do
				if data.iState == 7 and not data.bBot then
					tSend:Raw(sid, "BINF "..vsid.." I4"..(data.sIPEx or data.sIP).."\n", nil, nil, "Delay")
				end
			end
			return true, "User is operator, IP's sent."
		elseif not clear and not tHubSettings.tRules.bChatOnly then
			for vsid, data in pairs(tClient) do
				if data.iState == 7 and not data.bBot and data.sProtocol == "ADC" then
					tSend:Raw(sid, "BINF "..vsid.." I4"..(data.sIPEx or data.sIP).."\n", nil, nil, "Delay")
				end
			end
			return true, "User will receive all ADC IP's"
--~ 		elseif tUser:CanDownload(sid) then
--~ 			for vsid, data in pairs(tClient) do
--~ 				if data.iState == 7 and not data.bBot and not data.bShareHidden and not data.bHidden and data.sProtocol == "ADC" then
--~ 					tSend:Raw(sid, "BINF "..vsid.." I4"..data.sIP.."\n", nil, nil, "Delay")
--~ 				end
--~ 			end
--~ 			return true, "User is allowed to download, IP's sent."
		elseif clear then
			for vsid, data in pairs(tClient) do
				if data.iState == 7 and not data.bBot then
					tSend:Raw(sid, "BINF "..vsid.." I4\n", nil, nil, "Delay")
				end
			end
			return false, "User can't download, IP's cleared."
		end
		return false, "User can't download, no IP's sent."
	else
		return false, "NMDC user."
	end
end,

["OwnKey"] = function(self, sid, clear)
	local client = tClient[sid]
	local ct = client.tInfo.tFiltered.CT or 0
	
	if client then
		--// Send key
		if client.sProtocol == "ADC" then
			if ct > 3 then
				tSend:Raw(sid, "BINF "..sid.." CT"..ct.." OP1\n")
			elseif ct > 0 then
				tSend:Raw(sid, "BINF "..sid.." CT"..ct..(clear and " OP" or "").."\n")
			elseif clear then
				tSend:Raw(sid, "BINF "..sid.." CT OP\n")
			end
		elseif client.sProtocol == "NMDC" then
			if ct > 3 then
				tSend:Raw(sid, "$OpList "..NMDC_E(client.sNick).."|")
			end
		end
		
		--// Send fake key
		if client.bHasFakeKey then
			for vsid, data in pairs(tClient) do
				if client.sIP == data.sIP and data.iState == 7 then
					if tClient[vsid].sProtocol == "ADC" then
						tSend:Raw(vsid, "BINF "..sid.." CT16 OP1\n")
					elseif tClient[vsid].sProtocol == "NMDC" then
						tSend:Raw(vsid, "$OpList "..NMDC_E(client.sNick).."|")
					end
				end
			end
		end
	end
end,

["BroadcastKey"] = function(self, sid, clear)	--// Not finished yet
	local client = tClient[sid]
	if client then
		--// Send key
		if client.sProtocol == "ADC" then
			local ct = client.tInfo.tFiltered.CT
			if ct > 3 then
				tSend:Raw(sid, "BINF "..sid.." CT"..ct.." OP1\n")
			elseif ct > 0 then
				tSend:Raw(sid, "BINF "..sid.." CT"..ct.."\n")
			elseif clear then
				tSend:Raw(sid, "BINF "..sid.." CT\n")
			end
		elseif client.sProtocol == "NMDC" then
			if ct > 3 then
				tSend:Raw(sid, "$OpList "..NMDC_E(client.sNick).."|")
			end
		end
		
		--// Send fake key
		if client.bHasFakeKey then
			for vsid, data in pairs(tClient) do
				if client.sIP == data.sIP then
					if tClient[vsid].sProtocol == "ADC" then
						tSend:Raw(vsid, "BINF "..sid.." CT16 OP1\n")
					elseif tClient[vsid].sProtocol == "NMDC" then
						tSend:Raw(vsid, "$OpList "..NMDC_E(client.sNick).."|")
					end
				end
			end
		end
	end
end,

["IntervalUserMessage"] = function(self, msg, stime, stype, fromsid, tosid, roomsid, ...)		--// ... = fromsid, tosid, roomsid
end,

["IntervalMessage"] = function(self, msg, stime, target, fromsid, tosid, ...)		--// ... = fromsid, tosid, roomsid
	--// This handler will display same messages only if it hasn't been displayed in the past 'stime' seconds
	--// Make sure this handler isn't called from any of the functions mentioned in the targettypes, this will create a loop and stack overflow

	--// Format: tSend:IntervalMessage( msg, stime, stype, fromsid, tosid or roomsid )

	--// stype: MU / MA / MO / PU / PA / CR / FR / DE
	
	--// The message can contain the following variables:
	--// %s - The added amount of arguments after the tosid, will be replaced in the message for the %s occurances. This way same messages with different variables can be filtered
	--// %c - Amount of messages received since last display
	--// %t - Total amount of these messages received during hub uptime
	--// %a - Average message per minute since last display
	
	--// Use stime = 0 for displaying immediately
	--// Use stime = -1 for resetting the current counter and displaying immediately
	--// Use stime = < -1 to remove all messages older than stime seconds
	--// Use stime = nil to remove all messages
	
	--// Declaration
	local targettypes = {
		["MU"] = {"ChatToUser", 1, 4, 5}, ["MA"] = {"ChatToAll", 1, 4}, ["MO"] = {"ChatToOps", 1, 4}, 
		["PU"] = {"PmToUser", 1, 4, 5},	["PA"] = {"PmToAll", 1, 4}, 
		["CR"] = {"ChatRoom", 1, 4, 5}, ["FR"] = {"Feedroom", 1, 4, 5},
		["DE"] = {"Debug", 1, 4, 5},
	}

--~ 	if not stime then
--~ 		tInterval.Messages = {} 
--~ 		tFuncs:Debug("[IntervalMessage] [Reset] All message history removed.") 
--~ 		return true, "All message history removed."
--~ 	end
--~ 	
	--// Validation
	local stime = tonumber(stime) or 60
	if not msg or msg == "" then return false, "No message found" end
	if not targettypes[target] then return false, "Messagetype "..tostring(target).." not found for message: "..msg end

--~ 	if stime < -1 then 
--~ 		for name, data in pairs(tInterval.Messages) do
--~ 			if os.time + stime > tInterval.Messages[msg].iTime then
--~ 				tInterval.Messages[name] = nil
--~ 			end
--~ 		end
--~ 		tFuncs:Debug("[IntervalMessage] [Clean] All message older than "..string.sub(tostring(stime), 2).." removed.") 
--~ 	end

	--// Execution
	if not tInterval.Messages[msg] or (tInterval.Messages[msg].iTime + stime) <= os.time() then
		local dcount = 1
		local average = 0

		--// Make local
		local msgtable = { ["iTime"] = os.time(), ["iCount"] = 1, ["iCountDisplay"] = 1, ["tType"] = targettypes[target]}

		--// Update counters
		if tInterval.Messages[msg] then
			msgtable = tInterval.Messages[msg]
			msgtable.iCount = msgtable.iCount+1
			dcount = msgtable.iCount - msgtable.iCountDisplay
			local dtime = os.time() - msgtable.iTime
			if dtime > 0 then average = dcount/(msgtable.iTime/60) end
		end
		msgtable.iCountDisplay = msgtable.iCount
		
		--// Format endstring
		if msg:find("%%s") and #arg<1 then
			msg = msg:gsub("(%%s)", "")
		else
			for a=1, #arg do
				msg = msg:gsub("(%%s)", tFuncs:EscapePercent(arg[a]), 1)
			end
		end
		msg = msg:gsub("(%%a)", average, 1)
		msg = msg:gsub("(%%c)", dcount, 1)
		msg = msg:gsub("(%%t)", msgtable.iCount, 1)

		--// Put back to global
		tInterval.Messages[msg] = msgtable

		--// Output endstring
		local result, ok = pcall(tSend[targettypes[target][1]], nil, msg, fromsid, tosid)
		if not result then tFuncs:Debug("[IntervalMessage] [ERROR] "..ok) end
	else 
		tInterval.Messages[msg].iCount = tInterval.Messages[msg].iCount + 1
		return false
	end
end,

["Pm"] = function(self, client, sid, tosid, retmsg, me)
--~ 	tFuncs:Debug("PM received")
	if not tClient[sid] or not tClient[tosid] then return false, client end

	--// Bot PM
	if tosid == "AAAA" then 
		--// Hubcommands
		if string.find(retmsg, "^["..tHubSettings.Main.sCmdPrefixes.."].+$") then    -- hubcommand found
			local res, msg = tFuncs:HubCommand(sid, retmsg, tosid) --// use the unchanged line in case Deflood changed it, filtermsg = sLine

			--// Add hubcommand to cmdhistory
			table.insert(tLogs.HubCommands.History, {socket.gettime(), client.sNick, client.sIP, res, retmsg})
			
			if msg and msg~= "" then
				tSend:PmToUser(msg, tosid, sid) 
			end
			client = tClient[sid]	--// WORKAROUND ALERT!!! HubCommands don't pass the client, they make changes directly to tClient, don't return client, but return tClient[sid]
			tLogs.Counters[client.sProtocol].HubCommands = tLogs.Counters[client.sProtocol].HubCommands + 1
			return res, client

		--// Chat to bot
		else
			tSend:PmToUser("This is a bot, your message is ignored.", "AAAA", sid) 
			if not tHubSettings.Main.bBotTalk then -- defaults to false and message will be sent
				tFuncs:Debug(tUser:GetDataLine(sid).."talks to bot: "..retmsg, nil, 1)
			end
			tLogs.Counters[client.sProtocol].PrivateMessage = tLogs.Counters[client.sProtocol].PrivateMessages + 1
			return false, client
		end	

	--// Dont talk in FeedRoom
	elseif tosid == "CCCC" and not tHubSettings.Main.bFeedTalk then 
		tSend:PmToUser("You can't talk here, this is the FeedRoom.", tosid, sid)

	--// ChatRoom PM
	elseif tosid and tUser:IsChatRoom(tosid) then
		local allow = (tosid=="BBBB" and "opchat") or (tosid=="CCCC" and "feed") or false
		
		if (not allow and tUser:IsEqual(sid, tosid)) or (allow and tUser:IsAllowed(sid, allow)) or (tClient[tosid].bPrivate and tUser:IsVisibleTo(tosid, sid)) then
			local pmcommands = { ["history"] = true, ["list"] = true, ["invite"] = true, ["uninvite"] = true, ["leave"] = true, ["join"] = true, }
			local _,_,cmd = string.find(retmsg,  "^["..tHubSettings.Main.sCmdPrefixes.."](%S+)")
			--// OpChat or FeedRoom  Commands
			if (tosid == "BBBB" or tosid == "CCCC") and string.find(retmsg, "^["..tHubSettings.Main.sCmdPrefixes.."].+$") then    -- hubcommand found	
				local res, msg = tFuncs:HubCommand(sid, retmsg, tosid) --// use the unchanged line in case Deflood changed it, filtermsg = sLine

				--// Add hubcommand to cmdhistory
				table.insert(tLogs.HubCommands.History, {socket.gettime(), client.sNick, client.sIP, res, retmsg})
				
				if msg and msg ~= "" then tSend:ChatRoom(msg, "AAAA", tosid)  end
				client = tClient[sid]	--// WORKAROUND ALERT!!! HubCommands don't pass the client, they make changes directly to tClient, don't return client, but return tClient[sid]
				return res, client
			--// General chatRoom command
			elseif cmd and pmcommands[cmd] then	--// exception for pm command
				local res, msg = tFuncs:HubCommand(sid, retmsg, tosid) --// use the unchanged line in case Deflood changed it, filtermsg = sLine

				--// Add hubcommand to cmdhistory
				table.insert(tLogs.HubCommands.History, {socket.gettime(), client.sNick, client.sIP, res, retmsg})

				if msg and msg ~= "" then tSend:PmToUser(msg, tosid, sid) end
				client = tClient[sid]	--// WORKAROUND ALERT!!! HubCommands don't pass the client, they make changes directly to tClient, don't return client, but return tClient[sid]
				return res, client
			--// Normal chatroom PM
			else
				tSend:ChatRoom(retmsg, sid, tosid, me)
			end
		else
			tSend:PmToUser("You are not allowed to use this chatroom.", tosid, sid)
		end

	--// Normal PM
	elseif tosid then
		tSend:PmToUser(retmsg, sid, tosid, nil, me)
		tLogs.Counters[client.sProtocol].PrivateMessage = tLogs.Counters[client.sProtocol].PrivateMessages + 1
	--// Receiver offline
	else
		return false, client
	end
	return true, client
end,

["Mainchat"] = function(self, client, sid, retmsg, me)
	--// Format: tSend:IntervalMessage( msg, stime, target, fromsid, tosid or roomsid )
	if not retmsg then return client end
	local protocol = (client.sProtocol ~= "Bot" and client.sProtocol) or "ADC"

	local res, output = true, ""
	--// HubCommand
	if string.find(retmsg, "^["..tHubSettings.Main.sCmdPrefixes.."].+$") then    -- hubcommand found

		tSend:ChatToUser("*** "..retmsg, "AAAA", sid)
--~ 		tSend:Raw(sid, "*** "..retmsg)     -- send typed line back to the one who typed it in mainchat  won't work for ADC

		--// workaround for REport pattern checking
		if string.find(retmsg:lower(), "^.report%s+") then 
			res, retmsg = tCheck:Pattern(sid, "RE", retmsg)
			if not res then 
				tSend:ChatToUser("Report denied.", "AAAA", sid, me)
				return client 
			end
		end

		local origmsg = retmsg
		res, retmsg, output = tFuncs:HubCommand(sid, retmsg)

		--// Add hubcommand to cmdhistory
		table.insert(tLogs.HubCommands.History, {socket.gettime(), client.sNick, client.sIP, res, origmsg})

		if retmsg and retmsg ~= "" then
			if output == "PM"  then
				tSend:ChatToUser("The result has been sent in PM from <"..tUser:GetNick("AAAA")..">.", "AAAA", sid)
				tSend:PmToUser(retmsg, "AAAA", sid)
			else
				tSend:ChatToUser(retmsg, "AAAA", sid)
			end
		end

		client = tClient[sid]	--// WORKAROUND ALERT!!! HubCommands don't pass the client, they make changes directly to tClient, don't return client, but return tClient[sid]
		tLogs.Counters[protocol].HubCommands = tLogs.Counters[protocol].HubCommands + 1
	--// Normal mainchat
	else
		tSend:ChatToAll(retmsg, sid, me)
		if sid ~= "AAAA" then
			tLogs.Counters[protocol].MainchatMessages = tLogs.Counters[protocol].MainchatMessages + 1
		end
	end
	return client
end,

["Inform"] = function(self, msg, fromsid, tosid, handlemsg)     --// This is used for sending a mainchat line to one user
	handlemsg = handlemsg or msg
	--// Repeattimer
	local trepeat = tHubSettings.Main.iInformTime or 3600

	--// If msg was handled in last 'trepeat' secs:
	if tClient[tosid].tHandledMsg[handlemsg] then
		if trepeat == -1 or (tClient[tosid].tHandledMsg[handlemsg] + trepeat) >= os.time() then
			return
		end
	end
	tClient[tosid].tHandledMsg[handlemsg] = os.time()
	tSend:ChatToUser(msg, fromsid, tosid)
end,

["ChatToUser"] = function(self, msg, fromsid, tosid)     --// This is used for sending a mainchat line to one user
	if msg then
		local fromsid = (fromsid or "AAAA")
		local fromnick = "<"..(tClient[fromsid].sNick or "unknown").."> "

		msg = string.gsub(msg, "\n%[", "\n- [")     --// add a - in front of newlines beginning with a bracket to make it clear in client that it's a continued message instead of a new one

		if tClient[tosid].sProtocol == "NMDC" then
			local smsg = fromnick..NMDC_E(msg).."|"
			tSend:Raw(tosid, smsg, fromsid, 1)
		elseif tClient[tosid].sProtocol == "ADC" then
			local smsg = "BMSG "..fromsid.." "..ADC_E(msg)
			tSend:Raw(tosid, smsg, fromsid, 1)
		end
	end

	--// Send to GUI
--    tFuncs:SendToGui("ID_tc_monitor_mainchat", Clock().."*** "..tonick.."is being informed: "..msg.."\n")
end,

["ChatToAll"] = function(self, msg, fromsid, me)	-- takes msg as protocol independent, unformatted, plain text
	fromsid = fromsid or "AAAA"
	msg = string.gsub(msg, "\n%[", "\n- [")	--// add a - in front of newlines beginning with a bracket to make it clear in client that it's a continued message instead of a new one

	local fromnick =  (tClient[fromsid].sNick or "unknown")
	local fakenick = tHubSettings.Main.sHiddenUser or "Hidden User"

	local reg, vnick, lnick = tUser:GetReg(fromsid)
	if tAccounts[lnick] then 
		tAccounts[lnick].iLastSpoke = os.time() 
		tAccounts[lnick].sLastSpoke = msg
	end

	tFuncs:SendToConsoleChat("["..ToDate().."] <"..fromnick.."> "..msg)

	if tUser:IsVisible(fromsid) then
		--// Send to all users
		if me then 
			tSend:Raw("P_NMDC", "** "..fromnick.." "..NMDC_E(msg), fromsid, 1)
			tSend:Raw("P_ADC", "BMSG "..fromsid.." "..ADC_E(msg).." ME1\n", fromsid, 1)
		else
			tSend:Raw("P_NMDC", "<"..fromnick.."> "..NMDC_E(msg), fromsid, 1)
			tSend:Raw("P_ADC", "BMSG "..fromsid.." "..ADC_E(msg).."\n", fromsid, 1)
		end
	else
		local own = "[Hidden:"..((tClient[fromsid].iProfile or -1) - 1).."] "
		local nmsg = "<"..fromnick.."> "..NMDC_E(msg)
		local ownnmsg = "<"..fromnick.."> "..NMDC_E(own..msg)
		local hnmsg = "<"..fakenick.."> "..NMDC_E(msg)
		local amsg = "BMSG "..fromsid.." "..ADC_E(msg)
		local ownamsg = "BMSG "..fromsid.." "..ADC_E(own..msg)
		local fakesid = tNick:GetSid(fakenick) or tFuncs:CreateSid()
		local fakebinf = "BINF "..fakesid.." ID"..dc.hashpid(fakesid).." NI"..ADC_E(fakenick)
		local hamsg = "BMSG "..fakesid.." "..ADC_E(msg)

		if me then 
			own = "[Hidden:"..((tClient[fromsid].iProfile or -1) - 1).."] "
			nmsg = "** "..fromnick.." "..NMDC_E(msg)
			ownnmsg = "** "..fromnick.." "..NMDC_E(own..msg)
			hnmsg = "** "..fakenick.." "..NMDC_E(msg)
			amsg = "BMSG "..fromsid.." "..ADC_E(msg).." ME1"
			ownamsg = "BMSG "..fromsid.." "..ADC_E(own..msg).." ME1"
			fakesid = tNick:GetSid(fakenick) or tFuncs:CreateSid()
			fakebinf = "BINF "..fakesid.." ID"..dc.hashpid(fakesid).." NI"..ADC_E(fakenick)
			hamsg = "BMSG "..fakesid.." "..ADC_E(msg).." ME1"
		end

		for vsid, data in pairs(tClient) do
			if data.iState == 7 and not data.bBot and tUser:IsVisibleTo(fromsid, vsid) then
				if data.sProtocol == "NMDC" then
					if vsid == fromsid then 
						tSend:Raw(vsid, ownnmsg, fromsid)
					else
						tSend:Raw(vsid, nmsg, fromsid)
					end
				elseif data.sProtocol == "ADC" then
					if vsid == fromsid then 
						tSend:Raw(vsid, ownamsg, fromsid)
					else
						tSend:Raw(vsid, amsg, fromsid)
					end
				end
			elseif data.iState == 7 and not data.bBot then
				if data.sProtocol == "NMDC" then
					tSend:Raw(vsid, hnmsg, fromsid)
				elseif data.sProtocol == "ADC" then
					if tClient[fakesid] then
						tSend:Raw(vsid, hamsg, fakesid)
					else
						tSend:Raw(vsid, fakebinf, fakesid)
						tSend:Raw(vsid, hamsg, fakesid)
						tSend:Raw(vsid, "IQUI "..fakesid, fakesid)
					end
				end
			end
		end
	end
	if not tUser:IsVisible(fromsid) then fromnick = fakenick end

	tFuncs:AddLog("Mainchat", nil, msg, fromsid, fromnick)

	--// Send to GUI
--    tFuncs:SendToGui("ID_tc_monitor_mainchat", Clock()..fromnick..msg.."\n")
end,

["ChatToOps"] = function(self, msg, fromsid)
	for tosid,data in pairs(tClient) do
		if data.iState == 7 and data.bOperator then
			tSend:ChatToUser(msg, fromsid, tosid)
		end
	end
end,

["ChatToProf"] = function(self, msg, fromsid, profile)
	for tosid, data in pairs(tClient) do
		if data.iState == 7 and data.sProfile == profile then
			tSend:ChatToUser(msg, fromsid, tosid)
		end
	end
end,

["ChatToProfLevel"] = function(self, msg, fromsid, level)
	for tosid, data in pairs(tClient) do
		if data.iState == 7 and data.iProfile >= level then
			tSend:ChatToUser(msg, fromsid, tosid)
		end
	end
end,

["ChatFromNick"] = function(self, msg, fromsid, fakenick)     -- takes msg as protocol independent, unformatted, plain text
	local fakesid = tNick:GetSid(fakenick) or tFuncs:CreateSid() 

	msg = string.gsub(msg, "\n%[", "\n- [")     --// add a - in front of newlines beginning with a bracket to make it clear in client that it's a continued message instead of a new one

	--// Send to NMDC Users
	local smsg = "<"..fakenick.."> "..NMDC_E(msg)     -- add function, trying tFuncs:Escape for now
	tSend:Raw("P_NMDC", smsg, fromsid, 1)

	--// Send to ADC Users
	local smsg = "BMSG "..fakesid.." "..ADC_E(msg)     -- currenty prefix can only be added AFTER the nick, unless the BINF contains the prefix
	if not tClient[fakesid] then
		tSend:Raw("P_ADC", "BINF "..fakesid.." ID"..dc.hashpid(fakesid).." NI"..ADC_E(fakenick), fakesid)
		tSend:Raw("P_ADC", smsg, fakesid)
		tSend:Raw("P_ADC", "IQUI "..fakesid, fakesid)
	else
		tSend:Raw("P_ADC", smsg, fakesid, 1)
	end

	tFuncs:AddLog("Mainchat", nil, msg, fromsid, fakenick)

	tFuncs:Debug("User: "..tClient[fromsid].sNick.." spoke as <"..fakenick.."> saying: "..msg.."\n")
end,

["PmToUser"] = function(self, msg, fromsid, tosid, roomsid, me)     -- takes msg as protocol independent, unformatted, plain text
	local fromsid = fromsid or "AAAA"
	local tonick = tUser:GetNick(tosid)
	local fromnick = tUser:GetNick(fromsid)
	local roomsid = roomsid or fromsid
	local roomnick = tUser:GetNick(roomsid)
	local debugtype

	if tUser:GetProtocol(tosid) and tUser:GetProtocol(tosid) == "NMDC" then
		if not me then 
			tSend:Raw(tosid, "$To: "..tonick.." From: "..roomnick.." $<"..fromnick.."> "..NMDC_E(msg))
		else
			tSend:Raw(tosid, "$To: "..tonick.." From: "..roomnick.." $** "..fromnick.." "..NMDC_E(msg))
		end

		if tUser:GetProtocol(fromsid) == "NMDC" then debugtype = "PNN" else debugtype = "PNA" end		--// Remove before beta
--~ 		tFuncs:Debug(tUser:GetDataLine(fromsid).. "$To: "..tonick.." From: "..roomnick.." $<"..fromnick.."> "..NMDC_E(msg), debugtype, 1)	--// Remove before beta
	elseif tUser:GetProtocol(tosid) and tUser:GetProtocol(tosid) == "ADC" then
		if not me then 
			tSend:Raw(tosid, "EMSG "..fromsid.." "..tosid.." "..ADC_E(msg).." PM"..roomsid)
		else
			tSend:Raw(tosid, "EMSG "..fromsid.." "..tosid.." "..ADC_E(msg).." PM"..roomsid.." ME1")
		end

		if tUser:GetProtocol(fromsid) == "ADC" then debugtype = "PAA" else debugtype = "PAN" end	--// Remove before beta
--~ 		tFuncs:Debug(tUser:GetDataLine(fromsid).."EMSG "..fromsid.." "..tosid.." "..ADC_E(msg).." PM"..roomsid, debugtype, 1)	--// Remove before beta
	else
		-- user not available, could be a try to send it to a bot, messages can't currently be send from the hub to a hubbot
	end
end,

["PmToAll"] = function(self, msg, fromsid)
	for tosid,data in pairs(tClient) do
		if data.iState == 7 and not data.bBot then
			tSend:PmToUser(msg, fromsid, tosid)
		end
	end
end,

["PmToOps"] = function(self, msg, fromsid)
	for tosid,data in pairs(tClient) do
		if data.iState == 7 and tUser:IsOp(tosid) then
			tSend:PmToUser(msg, fromsid, tosid)
		end
	end
end,

["PmToProf"] = function(self, msg, fromsid, profile)
	for tosid,data in pairs(tClient) do
		if data.iState == 7 and not data.bBot and data.sProfile == profile then
			tSend:PmToUser(msg, fromsid, tosid)
		end
	end
end,

["PmToLevel"] = function(self, msg, fromsid, level)
	for tosid,data in pairs(tClient) do
		if data.iState == 7 and data.iProfile >= level then
			tSend:PmToUser(msg, fromsid, tosid)
		end
	end
end,

["ChatRoom"] =  function(self, msg, fromsid, roomsid, me)     -- takes msg as protocol independent, unformatted, plain text
	local fromnick = tClient[fromsid].sNick
	local allow = (roomsid=="BBBB" and "opchat") or (roomsid=="CCCC" and "feed") or false
	
	if tClient[roomsid].bPrivate then
		for nick, value in pairs(tClient[roomsid].tMembers) do
			local tosid = tNick:GetSid(nick)
			if tosid and tosid ~= fromsid then
				tSend:PmToUser(msg, fromsid, tosid, roomsid, me)
			end
		end
		for nick, value in pairs(tClient[roomsid].tMembersTemp) do
			local tosid = tNick:GetSid(nick)
			if tosid and tosid ~= fromsid then
				tSend:PmToUser(msg, fromsid, tosid, roomsid, me)
			end
		end
	else
		for tosid,data in pairs(tClient) do
			if data.iState >= 7 and tosid ~= fromsid and not data.bBot then
				if (not allow and tUser:IsEqual(tosid, roomsid)) or (allow and tUser:IsAllowed(tosid, allow)) then
					tSend:PmToUser(msg, fromsid, tosid, roomsid, me)
				end
			end
		end
		tFuncs:AddLog("ChatRooms", roomsid, msg, fromsid, fromnick)
	end
end,

["Feedroom"] =  function(self, msg, fromsid)     -- takes msg as protocol independent, unformatted, plain text
	local fromsid = fromsid or "AAAA"
	local roomsid = "CCCC"
	local fromnick = tUser:GetNick(fromsid)
	for tosid,data in pairs(tClient) do
		if not tClient[tosid].iState then
			Notify("[Feedroom] [ERROR] No iState for SID: "..tosid.." / Nick: "..(tClient[tosid].sNick or "<none>"))
		end
		if not data.bBot and data.iState >= 7 and tosid ~= fromsid and tUser:IsAllowed(tosid, "feed") then
			 tSend:PmToUser(msg, fromsid, tosid, roomsid)
		end
	end

	tFuncs:AddLog("ChatRooms", "CCCC", msg, fromsid, fromnick)
end,

["MyInfPool"] = function(self, sid)
	if tClient[sid] then
		tBuffer[sid].MyInfo.SendAll = true
		return true
	else
		return false
	end
end,

["MyInfUpdate"] = function(self, fromsid)
	local client = tClient[fromsid]

	--// Localize frequently used functions 
--~ 	local IsVisibleTo = tUser.IsVisibleTo
--~ 	local GetMyInfo = tUser.GetMyInfo
--~ 	local Raw = tSend.Raw
--~ 	local ou = client.tInfo.tADC.tOperatorUpdate
--~ 	local gu = client.tInfo.tADC.tGuiUpdate
--~ 	local uu = client.tInfo.tADC.tUserUpdate
	
	--// Create ADC update strings
	local uOperator = "BINF "..fromsid
	for name, value in pairs(client.tInfo.tADC.tOperatorUpdate) do
		uOperator = uOperator.." "..name..ADC_E(value)
	end
	client.tInfo.tADC.sOperatorUpdate = (#uOperator>9 and uOperator or false)

	local uUser = "BINF "..fromsid
	for name, value in pairs(client.tInfo.tADC.tUserUpdate) do
		uUser = uUser.." "..name..ADC_E(value)
	end
	client.tInfo.tADC.sUserUpdate = (#uUser>9 and uUser or false)
	
	local uGui = "BINF "..fromsid
	for name, value in pairs(client.tInfo.tADC.tGuiUpdate) do
		uGui = uGui.." "..name..ADC_E(value)
	end
	client.tInfo.tADC.sGuiUpdate = (#uGui>9 and uGui or false)
	
	--// Send changes
	for vsid,data in pairs(tClient) do
		if data.iState == 7 and not data.bBot and tUser:IsVisibleTo(fromsid, vsid) then        --// only send MyINFO to users with iState = 7
			local inf,err = tUser:GetMyInfo(fromsid, vsid, true)
			if inf then
				if data.sProtocol ~= "NMDC" or data.tSupports.NoHello or data.tSupports.NoGetINFO then
					tSend:Raw(vsid, inf)
				end
			else
--~ 				tFuncs:Debug("[tSend:MyInf] [ERROR] inf from SID: "..fromsid.." for SID: "..vsid.." not found. "..err)
			end
		end
	end
	
	tClient[fromsid].tInfo.tADC.sGuiUpdate = false
	tClient[fromsid].tInfo.tADC.sUserUpdate = false
	tClient[fromsid].tInfo.tADC.sOperatorUpdate = false
end,

["MyInf"] = function(self, fromsid, tosid, toself, stype)			--// MyInf must check, if fromsid has iState < 7 then tBuffer[sid].Login and LoginString are the target
	if not toself then toself = 0 end
	local fromclient = tClient[fromsid] or false
	local toclient = tClient[tosid] or false

	--// Send from ALL to tosid		--// Includes sending own info to tosid
	if fromsid == "ALL" and toclient then

		local inf = "sUser"
		if toclient.sProfile == "Gui" then
			inf = "sGui"
		elseif toclient.bOperator or tUser:CanOverride(tosid, "hideinfo") then
			inf = "sOperator"
		end
		
		--// Localize frequently used functions
--~ 		local IsVisibleTo = tUser.IsVisibleTo
--~ 		local bShowFullOp = tHubSettings.tTraffic.bShowFullOp
--~ 		local GetMyInfo = tUser.GetMyInfo		
--~ 		local Raw = tSend.Raw

		for vsid,data in pairs(tClient) do
			if data.iState == 7 and vsid ~= tosid and tUser:IsVisibleTo(vsid, tosid) and (not tBuffer[vsid] or not tBuffer[vsid].bDisconnect) then        --// only send MyINFO from users with iState = 7, and skip self
				local inf = tUser:GetMyInfo(vsid, tosid)
				if inf then 
					tSend:Raw(tosid, inf)
				end
--~ 			
--~ 				if toclient.sProfile == "Gui" then
--~ 					tSend:ExtraDataToGui(vsid, data)
--~ 				end
			end
		end
		--// Send own info last
		local inf = tUser:GetMyInfo(tosid, tosid)
		tSend:Raw(tosid, inf)
		
	--// Send from fromsid to ALL
	elseif fromclient and (not tBuffer[fromsid] or not tBuffer[fromsid].bDisconnect) and tosid == "ALL" then

		--// Localize frequently used functions 
--~ 		local IsVisibleTo = tUser.IsVisibleTo
--~ 		local GetMyInfo = tUser.GetMyInfo
--~ 		local Raw = tSend.Raw

		for vsid,data in pairs(tClient) do
			if data.iState == 7 and not data.bBot and tUser:IsVisibleTo(fromsid, vsid) and (not tBuffer[vsid] or not tBuffer[vsid].bDisconnect) then        --// only send MyINFO to users with iState = 7

				local inf, err = tUser:GetMyInfo(fromsid, vsid)
				if inf then 
					tSend:Raw(vsid, inf)
				else
					--// There's no myinfo (update) for this user
--~ 					tFuncs:Debug("[tSend:MyInf] [ERROR] inf from SID: "..fromsid.." for SID: "..vsid.." not found. "..err)
				end
--~ 				if data.sProfile == "Gui" then
--~ 					tSend:ExtraDataToGui(fromsid, fromclient)
--~ 				end				
			end
		end

	--// Send from fromsid to tosid -- will even send 'hidden' myinfo's?
	elseif fromclient and toclient then
		if fromclient.iState == 7 and not toclient.bBot and tUser:IsVisibleTo(fromsid, tosid) then
			local inf = tUser:GetMyInfo(fromsid, tosid)
			if inf then
				tSend:Raw(tosid, inf)
			end
--~ 			if toclient.sProfile == "Gui" then
--~ 				tSend:ExtraDataToGui(fromsid, fromclient)
--~ 			end
		end
	else
		--// tClient[fromsid] or tClient[tosid] are non-existent, ignore
	end
end,

["ExtraDataToGui"] = function(self, sid, data)
	if data and tSettings.Gui.Enable and tServer.GuiSID then
		if tClient[tServer.GuiSID] and tClient[tServer.GuiSID].sProfile == "Gui" then
			tSend:Raw(tServer.GuiSID, "IGUI XI "..sid.." SU"..data.sSupports.." PT"..data.sProtocol.." PF"..data.sProfile.." SS"..tostring(data.iShareSize or 0).." MO"..(data.sMode or "P").."\n")
		end
	end
end,

["SearchResult"] = function(self, msg, fromsid, tosid)
	--~ tSend:SearchResult(msg, fsid, "AAAA")

end,

["Raw"] = function(self, target, msg, senderid, toself, stype)
	--// target: ALL / P_<protocol> / <sid>    for example: "P_NMDC" or "P_ADC"
	--// senderid:  if a senderid is given by default the message won't get send to self, 'if not senderid then' message will be send to self
	--// toself: when this is given and senderid is given, the message will be send to self
	local senderid = senderid or "AAAA"
	local msg = msg or "tSend:Raw was used without a msg"
	local toself = toself or 0

	local msgn = string.gsub(msg, "|$", "").."|"    -- add endpipe  --// NMDC_E is done separately for chat and pm
	local msga = string.gsub(msg, "\n$", "").."\n"    -- add newline

	if not target then 
		target = "CCCC"	--// redirect to feed
		Notify("Error when sending RAW, no target. Msg: "..msg.." / Sender SID: "..(senderid or ""),"ERROR",wx.wxICON_ERROR) 
	end

	local _,_,protocol = target:find("^P_(.*)")
	if target and protocol then
		if protocol == "NMDC" then msg = msgn elseif protocol == "ADC" then _, msg = tFuncs:ToUTF8(msga) end
		for tsid,data in pairs(tClient) do
			if data.iState >= 7 then	-- only send broadcasts to users that are logged in properly
				if data.sProtocol == protocol then
					if senderid and toself == 0 and tsid == senderid then
						-- skip sending to self
					else
						if stype and stype == "Delay" then	--// This is data that may be sent delayed, like Searchresults, new MyINFO's and BINF's, this allows for the data-buffer to collect several and compress them
							tBuffer[tsid].Send.Data[#tBuffer[tsid].Send.Data+1] = msg
						else
							tBuffer[tsid].Send.Chat[#tBuffer[tsid].Send.Chat+1] = msg
						end
					end
				end
			end
		end
	elseif target and tClient[target] then
		if tClient[target].sProtocol == "NMDC" then
			msg = msgn
		elseif tClient[target].sProtocol == "ADC" then
			local a = ""
			a, msg = tFuncs:ToUTF8(msga)
		end
		if tClient[target].xHandler == "Login" then  --and tClient[target].iState < 7
			tBuffer[target].Send.Login[#tBuffer[target].Send.Login+1] = msg
		elseif stype and stype == "Delay" then
			tBuffer[target].Send.Data[#tBuffer[target].Send.Data+1] = msg
		elseif tClient[target].xHandler == "Main" then
			tBuffer[target].Send.Chat[#tBuffer[target].Send.Chat+1] = msg
		else
			Notify("Error when sending RAW, target handler: "..tostring(tClient[target].xHandler),"ERROR",wx.wxICON_ERROR)
--~ 			tFuncs:Debug("Error when sending RAW, target handler: "..tostring(tClient[target].xHandler)) --// Fatal mistake, debug uses Raw indirectly, can create loop
		end
	else
		if tBuffer[target] then
			tBuffer[target].Send.Login[#tBuffer[target].Send.Login+1] = msg
		else
			Notify(tUser:GetDataLine(sid).."Failed sending data to user without tBuffer: "..msg,"ERROR",wx.wxICON_ERROR)
--~ 			tFuncs:Debug(tUser:GetDataLine(sid).."Failed sending data to user without tBuffer: "..msg)	--// Fatal mistake, debug uses Raw indirectly, can create loop
		end
	end
end,
}


tHub = {
["GetBotName"] = function(self)
	return tHubSettings.Main.sBotName
end,
}

tUser = {

--// These two functions should replace DisconnectSelf, DisconnectUser and Redirectuser

["Delete"] = function(self, sid)	--// Used when hub stops, to prevent Quits being sent to users
	local client = tClient[sid]
	if client then 
		tClient[sid].bDisconnectSilent = true
		tBuffer[sid].bDisconnect = true              -- don't use client.active = 0 because then the last messages to this user won't be send...
		tBuffer[sid].bDisconnectTime = os.time() + (tHubSettings.tTcp.DisconnectTimeout or 5)
		return true
	else
		return false, "User with SID: "..tostring(sid).." not found."
	end
end,

["Disconnect"] = function(self, sid, reason, rreason, debugtype, debuglevel)
	local debugtype = debugtype or "FI"
	if rreason then debuglevel = debuglevel or 0 end
	local debuglevel = debuglevel or 2
	local reason = reason or ""
	local client = tClient[sid]
	if client then 
		client.bDisconnect = true
		if tBuffer[sid].bDisconnect then
--~ 			tBuffer[sid].bDisconnectTime = os.time() + (tHubSettings.tTcp.DisconnectTimeout or 5)
			return false, "User is already being disconnected"
		end

		tUser:Quit(sid)

		local _,_,errorcode,msg = string.find(reason, "^(%d%d%d) (.*)")
		if errorcode and msg then reason = msg end

		if reason ~= "" then
			if client.sProtocol == "ADC" then
				if client.iState < 2 then
					if tHubSettings.tRules.bRegOnly then
						tSend:Raw(sid,"IINF NI\\s CT32 HU1 HI1\n")   -- Empty hubname and topic
					else
						tSend:Raw(sid,"IINF NI"..ADC_E(tHubSettings.Main.sName).." DE"..ADC_E(tHubSettings.Main.sTopic).." VE"..ADC_E(_APPNAME).."\\s"..ADC_E(_APPVER).." CT32 HU1 HI1\n")   -- Hubname and topic
					end
				end
				if errorcode then
					tSend:Raw(sid, "ISTA "..errorcode.." "..ADC_E(reason))
				else
					tSend:Raw(sid, "ISTA 200 "..ADC_E(reason))
				end
			elseif client.sProtocol == "NMDC" then
				tSend:ChatToUser(reason, "AAAA", sid)
			else
				tSend:Raw(sid, reason)
			end
		end
		tBuffer[sid].bDisconnect = true		-- don't use client.active = 0 because then the last messages to this user won't be send...
		tBuffer[sid].bDisconnectTime = os.time() + (tHubSettings.tTcp.DisconnectTimeout or 3)

		if rreason then tFuncs:Debug(tUser:GetDataLine(sid).."was disconnected because: "..tostring(rreason), debugtype, debuglevel) end
		tFuncs:AddDisconnect(sid, reason, rreason)

		return (client.sNick or "<nick unknown>").." is disconnected because: "..reason
	else
		return false, "User with SID: "..tostring(sid).." not found."
	end
end,

["Autoban"] = function(self, sid)
	local client = tClient[sid]
	if client and not tUser:IsProtected(sid) then
		tBans.autoban[client.sIP] = os.time() + 7200	--// Add a timeout for this IP of 2 hours
	end
end,

["Redirect"] = function(self, sid, msg, address, rreason, debugtype, debuglevel, hide)
	local client = tClient[sid]
	if not client then	return false, "User with SID: "..sid.." not found." end

	local ctime = os.time()
	local debugtype = debugtype or "FI"
	local debuglevel = debuglevel or 2
	local newmsg = false
	local reason = msg
	local addtype = ""
	local msg = msg or ""
	local addmsg = ""
	if msg ~= "" then addmsg = " because: " end

	--// Get online redirect address
	address, addtype = tRedirect:GetAddress(address)
	if not address or address == "" then
		tBuffer[sid].bDisconnect = true		-- don't use client.active = 0 because then the last messages to this user won't be send
		tBuffer[sid].bDisconnectTime = os.time() + (tHubSettings.tTcp.DisconnectTimeout or 5)
		tFuncs:AddDisconnect(sid, reason, rreason, "No redirect address specified")

		tFuncs:Debug("[Redirect] "..addtype..tUser:GetDataLine(sid).."redirect failed, no address specified", "X")
		return false, tUser:GetShortData(sid).." disconnected, no redirect address found", false
	end

	--// Check if in history
	if tHubSettings.tRedirects.bBreakLoop then
		address, newmsg = tRedirect:History(sid, client, address, msg)
		msg = newmsg or msg
	end

	if client.sProtocol == "ADC" then
		if hide then msg = "" end
		if msg ~= "" then msg = ADC_E("MS"..msg).." " end
		if tHubSettings.tRules.bRegOnly then
			if client.iState < 2 then
				tSend:Raw(sid,"IINF NI\\s CT32 HU1 HI1\n")   -- Empty hubname and topic
			end
			tSend:Raw(sid, "IQUI "..sid.." "..msg.."TL-1 RD"..ADC_E(address).." DI\n")
		else
			if client.iState < 2 then
				tSend:Raw(sid, "ISID "..sid.."\n".."IINF NI"..ADC_E(tHubSettings.Main.sName).." CT32\n")   -- Hubname and topic
				tSend:MyInf("AAAA", sid)    -- Hubbot
			end
			tSend:Raw(sid, "IQUI "..sid.." IDAAAA "..msg.."TL-1 RD"..ADC_E(address).." DI\n")
		end
	else
		if not hide and msg ~= "" then tSend:ChatToUser(msg, "AAAA", sid) end
		tSend:Raw(sid, "$ForceMove "..address.."|")
	end

	tBuffer[sid].bDisconnect = true			-- don't use client.active = 0 because then the last messages to this user won't be send
	tBuffer[sid].bDisconnectTime = os.time() + (tHubSettings.tTcp.DisconnectTimeout or 5)
	tFuncs:AddDisconnect(sid, reason, rreason, address)

--~ 		if rreason then tFuncs:Debug(tUser:GetDataLine(sid).."was redirected to: "..address.." because: "..rreason, debugtype, debuglevel) end
	local reasontxt = rreason or reason
	reasontxt = (reasontxt and reasontxt ~= "" and " because: "..reasontxt) or ""
	tFuncs:Debug("[Redirect] "..addtype..tUser:GetDataLine(sid).."was redirected to: "..address..reasontxt, debugtype, debuglevel)

	return true, tUser:GetShortData(sid).." redirected to "..address, address
end,

["Spy"] = function(self, client, sid, msg)
	if client.sMonitor then
		local tosid = tNick:GetSid(client.sMonitor) or "CCCC"
		if tosid then
			nick = tUser:GetNick(sid) or tUser:GetIp(sid)
			if tClient[tosid].sProtocol == "ChatRoom" then
				tSend:ChatRoom("[Monitoring "..nick.." ] "..msg, "AAAA", tosid)
			else
				tSend:PmToUser("[Monitoring "..nick.." ] "..msg, "AAAA", tosid)
			end
		else
			tFuncs:Debug("[Monitor] Error: No sid for targetnick: <"..client.sMonitor.."> found.")
		end
	end
end,

["DoTrigger"] = function(self, sid, name, line)
	if tCommands[name] and tClient[sid].iProfile >= (tCommands[name].iUseProfile or tCommands[name].iProfile or -1) then
		local cmd = tCommands[name]
		local msg = false

		-- this is how i want it, but this breaks current settings so will leave that out for now
		-- since a fix needs to be added somewhere to re-process all textcmds first
	--[[
		if cmd.bHasParm then -- %[parm] exists and FillVars must be run with "single" parameter
			local tbl = tFuncs:Split(line, " ")
			table.remove(tbl, 1)
			msg = tFuncs:FillVars(cmd.sContent, sid, "single", tbl)
		elseif cmd.bHasFillVars then -- other FillVars exists
			msg = tFuncs:FillVars(cmd.sContent, sid)
		else	-- no need to run FillVars if no variables exists
			msg = cmd.sContent
		end
	--]]
		if cmd.bHasParm then -- %[parm] exists and FillVars must be run with "single" parameter
			local tbl = tFuncs:Split(line, " ")
			table.remove(tbl, 1)
			if tbl[cmd.bHasParm] then -- the commandline has at least as many words as the parm-count in the text command
				msg = tFuncs:FillVars(cmd.sContent, sid, "single", tbl)
			else
				return false, "Text command '"..name.."' failed, it needs "..cmd.bHasParm.." parameters to run. Please try again"
			end
		else
			msg = tFuncs:FillVars(cmd.sContent, sid)
		end

	--	local reply = ""

--		local options = {
--			"Displays text in PM instead mainchat", 
--			"Shows text to all",
--			"Shows text when user logged in",
--			"Doesn't allow the command to be activated in mainchat",
--			"Dooesn't create an item in !help for this command",
--			"Doesn't create a rightclick item for this command",
--			"Text will be send in raw format",
--			"Will only be sent to ADC user(s)",
--			"Will only be sent to NMDC user(s)",
--		}

		if msg and msg ~= "" then
			--// Send to all
			local output = "mainchat"
			if tFuncs:CheckFlag((cmd.iOptions or 0), 2) then
				for vsid, vdata in pairs(tClient) do
					if vdata.iState == 7 and vdata.sProfile and tProfiles[vdata.sProfile] and tProfiles[vdata.sProfile].iLevel >= (cmd.iProfile or -1) and tProfiles[vdata.sProfile].iLevel <= (cmd.iMaxProfile or 10) and not tBuffer[vsid].bDisconnect then

						 -- only send login messages for hidden users to same or larger levels
						if not tClient[sid].bHidden or (tClient[sid].bHidden and tClient[sid].iProfile <= vdata.iProfile ) then
							--// PM
							if tFuncs:CheckFlag((cmd.iOptions or 0), 1) then
								output = "PM"
								if (not tFuncs:CheckFlag((cmd.iOptions or 0), 128) and vdata.sProtocol == "NMDC") or (not tFuncs:CheckFlag((cmd.iOptions or 0), 256) and vdata.sProtocol == "ADC") then  --// Don't send only to ADC
									tSend:PmToUser(msg, "AAAA", vsid)
								end
							--// Mainchat
							else
								if (not tFuncs:CheckFlag((cmd.iOptions or 0), 128) and vdata.sProtocol == "NMDC") or (not tFuncs:CheckFlag((cmd.iOptions or 0), 256) and vdata.sProtocol == "ADC") then  --// Don't send only to ADC
									if tFuncs:CheckFlag((cmd.iOptions or 0), 64) then
										tSend:Raw(vsid, msg)
									else
										tSend:ChatToUser(msg, "AAAA", vsid)
									end
								end
							end
					--	else
					--		tFuncs:Debug("[DoTriggers] message "..name.." blocked on "..vdata.sNick)
						end
					end
				end
--~ 				return true, "Textcommand '"..name.."' sent in "..output.." to all users with profilelevels "..(cmd.iProfile or -1).." to "..(cmd.iMaxProfile or 10).."."
				return true --, "Textcommand '"..name.."' sent in "..output.." to all users with profilelevels "..(cmd.iProfile or -1).." to "..(cmd.iMaxProfile or 10).."."
			--// Reply to the sender only
			else
				--// PM
				if tFuncs:CheckFlag((cmd.iOptions or 0), 1) then
					if (not tFuncs:CheckFlag((cmd.iOptions or 0), 128) and tClient[sid].sProtocol == "NMDC") or (not tFuncs:CheckFlag((cmd.iOptions or 0), 256) and tClient[sid].sProtocol == "ADC") then  --// Don't send only to ADC
						tSend:PmToUser(msg, "AAAA", sid)
						return true, "You received a PM from <"..tClient["AAAA"].sNick..">."
					end
				--// Mainchat
				else
					if (not tFuncs:CheckFlag((cmd.iOptions or 0), 128) and tClient[sid].sProtocol == "NMDC") or (not tFuncs:CheckFlag((cmd.iOptions or 0), 256) and tClient[sid].sProtocol == "ADC") then  --// Don't send only to ADC
						if tFuncs:CheckFlag((cmd.iOptions or 0), 64) then
							tSend:Raw(sid, msg)
							return true --, reply
						else
							tSend:ChatToUser(msg, "AAAA", sid)
							return true
						end
					end
				end
			end
		elseif not msg then
			return false, "[DoTrigger] [ERROR] Trigger didn't result in valid message."
		else
			return true	--// fail silently, message was empty
		end
	else
		return false, "You are not allowed to use this command."
	end
	return false, "Something went wrong"
end,

["Joined"] = function(self, client, sid)
	--// Updating state to fully logged in
	client.iState = 7
	tClient[sid].iState = 7

	--// Determine external IP
	client.sIPEx = (not tIp:IsLocal(client.sIP) and client.sIP) or tStats.HubIp or client.sIP

	tFuncs:UpdateUserCount(1, client.sProtocol)

	client.iLoginDuration = os.time() - client.iLoginTime
	local nick = client.sNick

	--// Send protocol specific information
	if client.sProtocol == "NMDC" then
		tSend:Raw(sid, "$Debug Start of NMDC Join function")
		if client.bSecure then tSend:ChatToUser("You are now connected through TLS, traffic to and from the hub is encrypted.", "AAAA", sid) end

		local nicklist, oplist, botlist = NMDC_CreateNickList(sid)

		--// Send Join stuff to user
		tSend:Raw(sid, "$HubName "..NMDC_E(tHubSettings.Main.sName).." - "..NMDC_E(tHubSettings.Main.sTopic).."|")

		if client.tSupports.NoHello or client.tSupports.NoGetINFO then
			--// Send all MyINFO's to user
			tSend:MyInf("ALL", sid)
		else
			tSend:Raw(sid, nicklist.."|")
		end

		--// Create and send NickList and OpList
		tSend:Raw(sid, oplist.."|"..((client.tSupports.BotList and botlist.."|") or ""))

		--// If user is allowed to use $UserIP send all UserIP's
		if tUser:IsOp(sid) then
			local useriplist = NMDC_CreateUserIPList(sid)
			if useriplist then 
				tSend:Raw(sid, useriplist)
			end
		end

		if tUser:Supports(sid,"UserIP2") then			--// Send users own IP to be used by client
			tSend:Raw(sid, "$UserIP "..nick.." "..client.sIPEx.."|")
		end

		--// Send UserCommands (rightclick menu)
		if not tHubSettings.Main.bSuppressUsercommands and tUser:Supports(sid, "UserCommand") then
			tSend:Raw(sid, tUserCmds:CreateUC(sid, client.sProtocol))
		end

		tSend:Raw(sid, "$Debug End of NMDC Join function")

	elseif client.sProtocol == "ADC" then
		if tUser:Supports(sid, "ADFGUI") then
			tSend:Raw(sid, "ISTA 000 "..ADC_E("Start BINF").."\n")
		end

		--// Send all MyINFO's to user
		tSend:MyInf("ALL", sid)	

		--// Send user IP's
		tSend:ADCUserIps(sid)

		--// Send own key
		tSend:OwnKey(sid)

		if client.bSecure then tSend:ChatToUser("You are now connected through TLS, traffic to and from the hub is encrypted.", "AAAA", sid) end

		--~--// Send UserCommands (rightclick menu)
		if not tHubSettings.Main.bSuppressUsercommands and (tUser:Supports(sid, "ADUCMD") or tUser:Supports(sid, "ADUCM0")) then
			tSend:Raw(sid, tUserCmds:CreateUC(sid, client.sProtocol).."\n")
		end

		--// Send end of login to GUI
		if tUser:Supports(sid, "ADFGUI") then
			tSend:Raw(sid, "ISTA 000 "..ADC_E("End BINF").."\n")
		end
	else

	end
	tSend:ChatToUser(_APPNAME.." ".._APPVER.." svn: "..(tSettings.General.svnversion or "Not available")..(not tSettings.General.svnpublic and " (private)" or "").." - Made by Daywalker and FlipFlop.", "AAAA", sid)

	tLogs.Counters[client.sProtocol].Logins = tLogs.Counters[client.sProtocol].Logins + 1
	--// Check if there are triggers to execute on join: flag contains 4
	local jointriggers = {}
	for name, data in pairs(tCommands) do
		if tFuncs:CheckFlag((data.iOptions or 0), 4) and (client.iProfile >= (data.iProfile or -1) and client.iProfile <= (data.iMaxProfile or 10)) then
			jointriggers[#jointriggers+1] = name
		end
	end
	table.sort(jointriggers)
	--// Execute the join triggers
	for num, name in ipairs(jointriggers) do
		local res, retmsg = tUser:DoTrigger(sid, name)
	end

	--// Announce hidden
	if client.bHidden then
		tSend:ChatToUser("Warning: You are hidden to everyone below level: "..client.iProfile.." ("..client.sProfile..")", "AAAA", sid)
	end

	--// Announce locked main
	if tUser:IsOp(sid) and tSecurity.General.MainLocked then
		tSend:ChatToUser("Note: Mainchat is locked, only operators are allowed to speak.", "AAAA", sid)
	end

	local dtype
	if tUser:GetProtocol(sid) == "NMDC" then dtype = "LNJ" else dtype = "LAJ" end
	tFuncs:Debug(tUser:GetDataLine(sid).."Logged in succesfully.", dtype, 2)

	if not client.bBot and client.iPort and tServer.Port[client.iPort] then
		tServer.Port[client.iPort].loginsok = tServer.Port[client.iPort].loginsok+1
	end

	--// added check for logged in user without pass
	if client.bNotifySetOwnPass then
		tSend:PmToUser("Note: You have logged in with an account without a password set, before you continue you must set your password with !setmypass <your password>\nIf you need help, please ask an operator in the hub.", "AAAA", sid)
	end

--~ 	TestApi()
	return client
end,

["AnnouncePinger"] = function(self, sid)
	if not tHubSettings.tHubLists.sPingerMsg or not tHubSettings.tHubLists.sPingerOutput then return false, "No message or output for pinger set" end
	local report = tFuncs:FillVars(tHubSettings.tHubLists.sPingerMsg, sid)
	if tHubSettings.tHubLists.sPingerOutput:lower() == "mainchat" then
		tSend:ChatToProfLevel(report, "AAAA", tHubSettings.tHubLists.iPingerLevel or 8)
	elseif tHubSettings.tHubLists.sPingerOutput:lower() == "feed" then
		tSend:ChatRoom(report, "AAAA", "CCCC")
	elseif tHubSettings.tHubLists.sPingerOutput:lower() == "opchat" then
		tSend:ChatRoom(report, "AAAA", "BBBB")
	end
end,

["Announce"] = function(self, sid)
	--// Send OpList / UserIP / MyINFO to all allowed users   NOT TO SELF
	if tClient[sid] then
		local vclient = tClient[sid]
		local nick = vclient.sNick
		local ip = vclient.sIPEx or vclient.sIP

--~ 		if vclient.sProtocol == "NMDC" and tUser:HasFakeKey(sid) then		--// Moved to Joined
--~ 			tSend:Raw(sid, "$OpList "..nick.."|")
--~ 		end

		--// Announce pinger
		if (vclient.tSupports["BotINFO"] or vclient.tSupports["ADPING"]) and tHubSettings.tHubLists.bPingerAnnounce then
			tUser:AnnouncePinger(sid)
		end

		for vsid, data in pairs(tClient) do
			if data.sProtocol == "NMDC" and data.iState == 7 and not data.bBot and tUser:IsVisibleTo(sid, vsid) and sid ~= vsid and not tBuffer[vsid].bDisconnect then
--~ 			if data.sProtocol == "NMDC" and data.iState == 7 and not data.bBot and sid ~= vsid and tUser:IsVisibleTo(sid, vsid) then

				--// Send $Hello first
				if (not data.tSupports.NoHello) or (vclient.bBot and not data.tSupports.BotList) then
					tSend:Raw(vsid, "$Hello "..nick.."|", nil, nil, "Delay")
				end

				--// Send $MyINFO
				if data.tSupports.NoHello or data.tSupports.NoGetINFO then
					tSend:MyInf(sid, vsid, nil, "Delay")
				end

				--// Send $Botlist
				if vclient.bBot and data.tSupports.BotList then
					tSend:Raw(vsid, "$BotList "..nick.."|", nil, nil, "Delay")
				end
				
				--// Send $OpList
				if tUser:HasKey(sid) then
					tSend:Raw(vsid, "$OpList "..nick.."|", nil, nil, "Delay")
				--// Send fake key to other users with same IP and self
				elseif tUser:HasFakeKey(sid) and data.sIP == vclient.sIP then
					tSend:Raw(vsid, "$OpList "..nick.."|", nil, nil, "Delay")
				end

				--// Send UserIP to all allowed
				if tUser:IsOp(vsid) and not vclient.bBot then		--// Send UserIP to all
					tSend:Raw(vsid, "$UserIP "..nick.." "..ip.."|", nil, nil, "Delay")
				end

			elseif data.sProtocol == "ADC" and data.iState == 7 and not data.bBot and tUser:IsVisibleTo(sid, vsid) and sid ~= vsid and not tBuffer[vsid].bDisconnect then
--~ 			elseif data.sProtocol == "ADC" and data.iState == 7 and not data.bBot and sid ~= vsid and tUser:IsVisibleTo(sid, vsid) then
				--// Send MyINFO , NoHello and NoGetINFO are assumed		--// This one needs to be send before Oplist BotList and UserIP
				tSend:MyInf(sid, vsid, nil, "Delay") 

				--// Send fake key to other users with same IP and self
				if tUser:HasFakeKey(sid) and data.sIP == vclient.sIP then
					tSend:Raw(vsid, "BINF "..sid.." CT16 OP1\n") 
				end

				--// Send UserIP to all allowed
				if (tUser:IsOp(vsid) or tUser:CanDownload(vsid)) and not vclient.bBot then		--// Send UserIP to all
					tSend:Raw(vsid, "BINF "..sid.." I4"..ip.."\n", nil, nil, "Delay")
				end
			end
		end
	else
		tFuncs:Debug("[ERROR] [tUser:Announce] SID: "..sid.." doesn't exist.", "FI")
	end
end,

["Quit"] = function(self, sid, noquit)
	--// Send Quit to all
	local client = tClient[sid]
	if client and client.iState == 7 and not client.bQuit then        --// only send Quit if the disconnecting user has reached proper login state
		tSend:Raw("P_NMDC", "$Quit "..client.sNick.."|", sid)
		tSend:Raw("P_ADC", "IQUI "..sid.."\n", sid)
		if not client.bBot and not noquit then 
			client.bQuit = true
--~ 			client.iState = 0
			tFuncs:UpdateUserCount(-1, client.sProtocol)
		end
	end
end,

["Hide"] = function(self, sid)
	--// Send Quit to all
	local nick = tClient[sid].sNick
	local hidelevel = tClient[sid].iProfile
	for vsid, data in pairs(tClient) do
		if data.iState == 7 and data.iProfile and data.iProfile < hidelevel then
			if data.sProtocol == "NMDC" then
				tSend:Raw(vsid, "$Quit "..nick.."|", sid)
			elseif data.sProtocol == "ADC" then
				tSend:Raw(vsid, "IQUI "..sid.."\n", sid)
			end
		end
	end
end,

["Supports"] = function(self, sid, msg)
	local data = tClient[sid]
	if data and data.tSupports and data.tSupports[msg] then
		return true
	else
		return false
	end
end,

["HasKey"] = function(self, sid)
	local data = tClient[sid]
	if data and (data.bHasKey or tProfiles[data.sProfile]["bHasKey"]) then
		return true
	else
		return false
	end
end,

["HasFakeKey"] = function(self, sid)
	local data = tClient[sid]
	if data and not data.bBot and (data.bHasFakeKey or tProfiles[data.sProfile].bHasFakeKey) then
		return true
	else
		return false
	end
end,

["CanDownload"] = function(self, sid, notify)
	local result, msg = false, ""
	if tHubSettings.tRules.bChatOnly then		--//and not tUser:CanOverride(sid, "blockdownload") 
		msg = "Searching and downloading is disabled in this hub."
	elseif not tHubSettings.tRules.bUnreggedDownload and tClient[sid].iProfile < 0 then
		if tUser:IsAllowed(sid, "regme") then
			msg = "You need to be registered to be able to download and search."
		else
			msg = "Searching and downloading is disabled."
		end
	elseif tBlocks.tConnects[tClient[sid].sIP] then
		msg = "You're not allowed to search and download."
	elseif not tClient[sid].bActive and not tHubSettings.tRules.bPassiveDownload and not tUser:CanOverride(sid, "blockpassive") then
		msg = "You're connected in passive mode, you're not allowed to search and download. Get active!"
	elseif not tUser:CanOverride(sid, "blockdownload") and (tClient[sid].iShareSize or 0) < tHubSettings.tRules.iMinShareDownload then
		msg = "You need to share at least "..tFuncs:ConvertShare(tHubSettings.tRules.iMinShareDownload or 0).." to be able to search and download."
	else
		result = true
	end

	if not result and notify then tSend:Inform(msg, "AAAA", sid, "Searching and downloading is blocked.") end

	return result, msg
end,

["CanUpload"] = function(self, sid, notify)
	local result, msg = false, ""
	if tHubSettings.tRules.bChatOnly and not tUser:CanOverride(sid, "blockupload") then
		return false, "The hub is set to chat-only, upload not allowed."
	elseif tBlocks.tConnects[tClient[sid].sIP] then
		msg = "You're not allowed to upload."
	else
		result = true
	end

	if not result and notify then tSend:Inform(msg, "AAAA", sid, "Searching and downloading is blocked.") end

	return result, msg
end,

["IsBot"] = function(self, sid)
	local data = tClient[sid]
	if data and data.bBot then
		return true
	else
		return false
	end
end,

["IsVisible"] = function(self, sid)
	local data = tClient[sid]
	if data and not data.bHidden then
		return true
	else
		return false
	end
end,

["IsProtected"] = function(self, sid, nick, ip)
	local user = tClient[sid]
	if sid and user then
		local onick = user.sNick or ""
		local nick = onick:lower()
		local ip = user.sIP
		if user.bBot then 
			return true, "The nickname <"..onick.."> is protected because it's in use by a bot"
		elseif tProtects.Ip[ip] or (tStats.HubIp and tStats.HubIp == ip) then
			return true, "The IP: "..ip.." of <"..onick.."> is protected."
		elseif tProtects.Nick[nick] then
			return true, "The nickname <"..onick.."> is protected."
		elseif tIp:IsLocal(ip) then
			return true, "The IP: "..ip.." of <"..onick.."> is protected because it matches a local address."
		elseif ip ~= "" then
			--// Is there a protected nickname on this IP?
			local sids = tIp:GetSids(ip)
			for _,vsid in ipairs(sids) do
				if tClient[vsid].bBot then
					return true, "User <"..onick.."> with IP "..ip.." is indirectly protected because <"..tClient[vsid].sNick.."> on this IP is a bot."
				elseif tClient[vsid].sNick and tProtects.Nick[tClient[vsid].sNick:lower()] then
					return true, "User <"..onick.."> with IP: "..ip.." is indirectly protected because nickname <"..tClient[vsid].sNick.."> on this IP is protected."
				end
			end
		end
	end
	if nick then
		local onick = nick
		nick = nick:lower()
		--// Nick protected
		if tProtects.Nick[nick] then
			return true, "Nickname <"..onick.."> is protected."
		else
			--// Is nick online? Then get IP to see if it's protected
			local vsid = tNick:GetSid(nick)
			if vsid then
				local ip = tClient[vsid].sIP
				--// Nick protected because it's a bot
				if tClient[vsid].bBot then 
					return true, "Nickname <"..onick.."> is protected because it's in use by a bot."
				--// Nick protected because it's IP is protected
				elseif tProtects.Ip[ip] then
					return true, "IP: "..ip.." of <"..onick.."> is protected."
				elseif tIp:IsLocal(ip) then
					return true, "IP: "..ip.." of <"..onick.."> is protected because it matches a local address."
				end
			end
		end
	end
	if ip then
		--// IP protected
		if tProtects.Ip[ip] or (tStats.HubIp and tStats.HubIp == ip) then
			return true, "IP: "..ip.." is protected."
		elseif tIp:IsLocal(ip) then
			return true, "IP: "..ip.." is protected because it matches a local address."
		elseif ip ~= "" then
			--// Is there a protected nickname on this IP?
			local sids = tIp:GetSids(ip)
			if sids then
				for _,vsid in ipairs(sids) do
					tFuncs:Debug("[IsProtected:IP] vsid"..vsid, "X")
					local data = tClient[vsid]
					if data and data.bBot then
						return true, "IP "..ip.." is indirectly protected because <"..data.sNick.."> on this IP is a bot."
					elseif data and data.sNick and tProtects.Nick[data.sNick:lower()] then
						return true, "IP: "..ip.." is indirectly protected because nickname <"..data.sNick.."> on this IP is protected."
					end
				end
			end
		end
	end
	--tFuncs:Debug("[IsProtected] return false - not protected", "X")
	return false
end,

["IsVisibleTo"] = function(self, sid, vsid)
	local data = tClient[sid]
	local vdata = tClient[vsid]

	if (data and not data.bHidden) or sid == vsid then
		return true
	elseif data and data.bHidden and not data.bPrivate and data.iState == 7 and data.iProfile and vdata.iProfile then
		if sid == "BBBB" then 
			return tUser:IsAllowed(vsid, "opchat")
		elseif sid == "CCCC" then
			return tUser:IsAllowed(vsid, "feed")
		elseif data.iProfile <= vdata.iProfile then
			return true
		end
	elseif data and data.sProtocol == "ChatRoom" and data.bPrivate then
		if data.tMembers and data.tMembers[string.lower(vdata.sNick or "")] then
			return true
		elseif data.tMembersTemp and data.tMembersTemp[string.lower(vdata.sNick or "")] then
			return true
		end
	end

	return false
end,

["IsAllowed"] = function(self, sid, cmd, informuser)
	if not tClient[sid] then return false end
	local prof = tClient[sid].sProfile or "User"
	if tClient[sid].iProfile == 10 or (tProfiles[prof].tCmds and (tProfiles[prof].tCmds[cmd] or tProfiles[prof].tCmds.all)) then
		return true
	else
		if informuser then tSend:ChatToUser(prof.."s are not allowed to use this command.", "AAAA", sid) end
		return false
	end
end,

["CanOverride"] = function(self, sid, cmd, informuser)
	if sid and tClient[sid] then
		if tClient[sid].bBot then return true end
		local prof = tClient[sid].sProfile or "User"

		if not tProfiles[prof] then
			tFuncs:Debug("[tUser:CanOverride] [ERROR] Can't find profile "..prof.. " to determine override on SID: "..sid.." / Nick: "..tUser:GetNick(sid).."/ cmd: "..cmd.." / Override failed.")
			return false
		end
		if tClient[sid].iProfile == 10 or (tProfiles[prof].tOverrides and (tProfiles[prof].tOverrides[cmd] or tProfiles[prof].tOverrides.all)) then
			return true
		else
			if informuser then	tSend:ChatToUser("You are not allowed to override '"..cmd.."'.", "AAAA", sid) end
		end
	end
	return false
end,

["IsOp"] = function(self, sid)
	if tClient[sid] and tClient[sid].iState == 7 and tClient[sid].bOperator then
		return true
	else
		return false
	end
end,

["IsReg"] = function(self, sid)
	if tClient[sid] and tClient[sid].iProfile and tClient[sid].iProfile > -1 then
		return true
	else
		return false
	end
end,

["IsChatRoom"] = function(self, sid)
	if tClient[sid] and tClient[sid].sProtocol == "ChatRoom" then
		return true
	else
		return false
	end
end,

["IsEqual"] = function(self, sid, targetsid)
	if tClient[targetsid] and (tUser:GetProfileLevel(sid) or -2) >= (tUser:GetProfileLevel(targetsid) or -2) then
		return true
	else
		return false
	end
end,

["IsHigher"] = function(self, sid, tsid, informuser)
	if tClient[sid] and tClient[tsid] then
		if tClient[sid].iProfile and (tClient[tsid].iProfile and tClient[sid].iProfile > tClient[tsid].iProfile) or tClient[sid].iProfile == 10 then
			return true
		elseif not tClient[tsid].iProfile then
			return true
		end
	end
	if informuser then
		tSend:ChatToUser("This user is too powerful, better rethink your actions.", "AAAA", sid)
	end
	return false
end,

["IsHigherUser"] = function(self, sid, tsid, informuser)
	if tClient[sid] and tClient[tsid] and not tClient[tsid].bBot and sid ~= tsid then
		if (tClient[sid].iProfile or -2) > (tClient[tsid].iProfile or -2) or tClient[sid].iProfile == 10 then
			return true
		end
	end
	if informuser then
		tSend:ChatToUser("This user is too powerful, better rethink your actions.", "AAAA", sid)
	end
	return false
end,

["GetMyInfo"] = function(self, fromsid, tosid, update)
	if tClient[fromsid] and tClient[tosid] then
		if tClient[tosid].sProtocol == "NMDC" then
			if tClient[tosid].bOperator or tUser:CanOverride(tosid, "hideinfo") or tClient[fromsid].bBot or (tClient[fromsid].bOperator and tHubSettings.tTraffic.tShow.FullOp) then 
				return tClient[fromsid].tInfo.tNMDC.sOperator
			else
				return tClient[fromsid].tInfo.tNMDC.sUser
			end
		elseif tClient[tosid].sProtocol == "ADC" then 
			if tClient[tosid].sProfile == "Gui" then
				if update then
					return tClient[fromsid].tInfo.tADC.sGuiUpdate
				else
					return tClient[fromsid].tInfo.tADC.sGui
				end
			elseif tClient[tosid].bOperator or tUser:CanOverride(tosid, "hideinfo") or (tClient[fromsid].bBot and tHubSettings.tTraffic.tShow.FullBot) or (tClient[fromsid].bOperator and tHubSettings.tTraffic.tShow.FullOp) then 
				if update then
					return tClient[fromsid].tInfo.tADC.sOperatorUpdate
				else
					return tClient[fromsid].tInfo.tADC.sOperator
				end
			elseif update then
				return tClient[fromsid].tInfo.tADC.sUserUpdate
			else
				return tClient[fromsid].tInfo.tADC.sUser
			end
		else
			return false, "No such protocol: "..tClient[tosid].sProtocol
		end
	else
		if not tClient[fromsid] then
			return false, "[GetMyInfo] [ERROR] There's no user with fromsid: "..fromsid
		else
			return false, "[GetMyInfo] [ERROR] There's no user with tosid: "..tosid
		end
	end
end,

["GetProfileName"] = function(self, profilename)
	if profilename then
		for name,data in pairs(tProfiles) do
			if name:lower() == profilename:lower() then
				return name
			end
		end
	end
	return false
end,

["GetProfileData"] = function(self, profilename)
	local result = false
	for name,data in pairs(tProfiles) do
		if name:lower() == profilename:lower() then
			result = tProfiles[name]
		end
	end
	if not result then tFuncs:Debug("[ERROR] Profile '"..profilename.." doesn't exist (anymore).") end
	return result
end,

["GetProfile"] = function(self, sid)
	if tClient[sid] and tClient[sid].sProfile then
		if tProfiles[tClient[sid].sProfile] then
			return tClient[sid].sProfile
		elseif tClient[sid] then
			tFuncs:Debug("[ERROR] Profile '"..tClient[sid].sProfile.." for user "..tClient[sid].sNick.." doesn't exist (anymore), using 'User' profile.")
			tClient[sid].sProfile = "User"
			tClient[sid].iProfile = -1
			return "User"
		else
			return false
		end
	end
end,

["GetHighProfiles"] = function(self, sidlevel)		--// Includes own profilelevel
	local level = tClient[sidlevel] and tClient[sidlevel].iProfile or (tonumber(sidlevel) or -1)
	local result = {}
	for name, data in pairs(tProfiles) do
		if data.iLevel >= level then
			table.insert(result, name)
		end
	end
	return result
end,

["GetProfileLevel"] = function(self, sid)
	if tClient[sid] then
		return tClient[sid].iProfile
	end
	return false
end,

["GetInfo"] = function(self,sid)
	local u = tClient[sid]
	if u then
		local user = {
			["sNick"] = u.sNick or "<unknown>",
			["sProtocol"] = u.sProtocol or "<unknown>",
			["sProfile"] = u.sProfile or "<unknown>",
			["sTag"] = u.sTag or "<unknown>",
			["sDescription"] = u.sDescription or "",
			["sEmail"] = u.sEmail or "",
			["iShareSize"] = u.iShareSize or 0,
			["sIP"] = u.sIP or "<unknown>",
		}
		return user
	else
		return false
	end
end,

["GetNick"] = function(self, sid)
	if tClient[sid] then
		return tClient[sid].sNick
	else
		return false
	end
end,

["GetIp"] = function(self, sid)
	if tClient[sid] then
		return tClient[sid].sIP
	end
	return false
end,

["GetReg"] = function(self, sid)
	local nick = tClient[sid] and tClient[sid].sNick or false
	if nick then
		local reg, rnick, lnick = tNick:GetReg(nick)
		if reg and rnick then
			reg.iLevel = tProfiles[reg.sProfile].iLevel
			return reg, rnick, lnick
		end
	end
	return false, false, false
end,

["GetOp"] = function(self, sid)
	if tUser:IsAllowed(sid, "bIsOP") then
		return true
	else
		return false
	end
end,

["IsOnline"] = function(self, sid)
	if tClient[sid] and tClient[sid].iState == 7 then
		return true
	else
		return false
	end
end,

["GetPort"] = function(self, sid)
	if tClient[sid] and tClient[sid].iPort then
		return tClient[sid].iPort
	elseif tClient[sid] and tClient[sid].iPort then
		return tClient[sid].iPort
	else
		return false
	end
end,

["GetProtocol"] = function(self, sid)
	if tClient[sid] and tClient[sid].sProtocol then
		return tClient[sid].sProtocol
--~ 	elseif tClientLogin[sid] and tClientLogin[sid].sProtocol then
--~ 		return tClientLogin[sid].sProtocol
	else
		return false
	end
end,

["GetDataLine"] = function(self, sid)
	local ip, nick
	if sid and tClient[sid] then
		ip = tClient[sid].sIP or "IP unknown"
		nick = tClient[sid].sNick or tClient[sid].sNickTemp or "nick unknown"
	else
		sid, ip, nick = "AAAA", "Hub", ""
	end
--~ 	local msg = "SID: "..sid.." IP: "..ip.." <"..nick.."> "
	local msg = "IP: "..ip.." <"..nick.."> "
	return msg
end,

["GetShortData"] = function(self, sid)
	if sid and tClient[sid] then
		return "<"..(tClient[sid].sNick or tClient[sid].sNickTemp or "nick unknown")..">".." IP: "..(tClient[sid].sIP or "IP unknown").." "
	else
		return "<no userdata>"
	end
end,

}

tNick = {

["GetProfile"] = function(self, nick)
	local reg, nick = tNick:GetReg(nick)
	if reg then
		return tProfiles[reg.sProfile], nick
	else
		return false
	end
end,

["SetReg"] = function(self, nick, var, value)
	local result = false
	local reg = tAccounts[nick:lower()]
	if reg then 
		reg[var] = value
		tFuncs:SaveSettings("accounts")
		return reg
	end
	local err = "[NICK:SETREG] [ERROR] User "..nick.." isn't registered."
	tFuncs:Debug(err)
	return false, err
end,

["GetReg"] = function(self, nick)
	--// Must return reg, nick (unlowercased)
	if nick then 
		local lnick = nick:lower()
		local reg = tAccounts[lnick]
		if not reg then return false, "User "..nick.." isn't registered.", lnick end

		local prof, err = ProfileManager:GetData("AAAA", reg.sProfile)
		if prof then
			reg.iLevel = prof.iLevel
			return reg, reg.sNick, lnick
		else
			local err = "[tNick:GetReg] [ERROR] "..err..", using profile 'Reg'."
			reg.sProfile = "Reg"
			reg.iLevel = tProfiles[reg.sProfile].iLevel
			return reg, reg.sNick, lnick
		end
	end
	return false, "No nick specified", ""
end,

["GetSidFromLNick"] = function (self, nick)
	if nick and tLNICKs[nick] then
		return tLNICKs[nick].sid, tLNICKs[nick].sNick
	else
		return false
	end
end,

["GetSid"] = function (self, nick)
	if nick then
		local lnick = nick:lower()
		for vsid, data in pairs(tClient) do
			if data.sNick and string.lower(data.sNick) == lnick then
				return vsid, data.sNick
			end
		end
	end
	return false
end,

}

tIp = {

["Autoban"] = function(self, ip, btype, msg, clearlog)
	--// omit btype for a silent autoban (ddos connections for example), if btype is given, autoban will show in feed
	if not tUser:IsProtected(nil, nil, ip) then
		ThreadLockMutex(mutexbanlist)
		if not tBans.autoban[ip] and btype then
			tSend:IntervalMessage("["..btype.."] IP: "..ip.." is autobanned for "..tFuncs:Duration(tSecurity.General.DefaultAutoban)..". Message: %s", 300, "FR", "AAAA", nil, msg)
		end
		tBans.autoban[ip] = os.time() + tSecurity.General.DefaultAutoban	--// Add a timeout for this IP of 2 hours
		ThreadUnlockMutex(mutexbanlist)
		if clearlog then
			ThreadLockMutex(mutexiplog)
			tLogs.IP[ip] = nil	--// Remove this IP from log to prevent table overflow with a lot of IP's connecting, NOTE: this might give errors if IP is in hub with normal status
			ThreadUnlockMutex(mutexiplog)
		end
		return true
	else
		return false, "IP is protected."
	end
end,

["IsLocal"] = function(self, ip)
	--// NOTE: If protect_local (tSecurity.General.ProtectLocal) = false, addresses 192.168.*.* won't return true for this function, will assume a LAN with uprotected IP's from users in that range.

	if not ip then return false end
	if ip:find("^127%.0%.0%.1$") or (tSecurity.General.ProtectLocal and ip:find("^10%.%d+%.%d+%.%d+$")) or (tSecurity.General.ProtectLocal and ip:find("^192%.168%.%d+%.%d+$")) then
		return true
	end
end,

["GetSids"] = function (self, ip)
	local sids = {}
	for sid, client in pairs(tClient) do
		if client and client.sIP == ip then
			sids[#sids+1] = sid
		end
	end
	if #sids > 0 then
		return sids
	else
		return false
	end
end,

["GetNick"] = function(self, ip)
	--// Gets the nick with highest profile, returns a table with nicks as second argument, returns a string with nicks as third argument
	local nick, tnicks, snicks, level = false, {}, "", -2

	local sids = tIp:GetSids(ip)
	if sids then
		for num, vsid in ipairs(sids) do
			local vnick = tUser:GetNick(sid)
			local data = tClient[vsid]
			if data then 
				if (data.iProfile or -1) > level then
					level = data.iProfile or -1
					nick = data.sNick or "nick unknown"
				end
				table.insert(tnicks, data.sNick or "nick unknown")
			end
		end
		snicks = table.concat(tnicks, ", ")
	end

	return nick, tnicks, snicks
end,

["ResetTimeout"] = function(self, ip)
	ThreadLockMutex(mutexiplog)
	if tLogs.IP[ip] then
		local ctime = socket.gettime()
--~ 		if tLogs.IP[ip]["$"].CurrentLogin + tSecurity.General.ReconnectTimeout >= ctime or then
			tLogs.IP[ip]["$"].DoubleLogins = 0
			tLogs.IP[ip]["$"].CurrentLogin = ctime - tSecurity.General.ReconnectTimeout
			return true, "Reconnection timeout for IP: "..ip.." has been reset."
--~ 		else
--~ 			return true, "Reconnection timeout didn't need resetting, last login for this IP was: "..ToDate(tLogs.IP[ip]["$"].CurrentLogin)
--~ 		end
	else
		return false, "IP hasn't been logged, no timeout set."
	end
	ThreadUnlockMutex(mutexiplog)
end,

["GetHighestReg"] = function(self, ip)
	local vnick, vdata, vlevel = false, false, -1
	if ip then 
		for nick, data in pairs(tAccounts) do
			if data.sLastIP == ip then
				if tProfiles[data.sProfile] and tProfiles[data.sProfile].iLevel > vlevel then
					vnick = data.sNick
					vdata = data
					vdata.iLevel = tProfiles[data.sProfile].iLevel
					vlevel = tProfiles[data.sProfile].iLevel
					result = true
				end
			end
		end
	end
	if result then return vnick, vdata else return false end
end,

["GetRegs"] = function(self, ip)
	local regs, result, nregs = {}, false, {}
	if ip then 
		for vnick, data in pairs(tAccounts) do
			if data.sLastIP and data.sLastIP == ip then
				table.insert(nregs, data.sNick)
				regs[vnick] = data
				result = true
			end
		end
	end
	if result then return regs, nregs else return false, false end
end,

["HighestProfile"] = function(self, ip)
	local level = -2
	local vnick, vprof = false, false
	
	if ip then 
		local sids = tIp:GetSids(ip)
		if not sids then return false end
		for _,sid in ipairs(sids) do
			if (tClient[sid].iProfile or -2) > level then		
				level = tClient[sid].iProfile
				vnick = tClient[sid].sNick
				vprof = tClient[sid].sProfile
			end
		end
	end
	return level, vnick, vprof
end,

}