--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

--[[ states

state 1 = user connected, hub sends $Lock
state 2 = after client sends valid $Supports
state 3 = after client sends valid $Key
state 4 = after client sends valid $ValidateNick and is regged, needs a pass, $GetPass sent
state 5 = after client sends valid $ValidateNick and wasn't regged, or after valid $MyPass was sent
state 6 = after proper $Version
state 7 = after proper $MyINFO

]]--

function NMDC_DataArrival(client,sid,msg)
--	local wrong = "Malformed NMDC message received."
	client.sStatus = "[RECV] "..msg

	--// function execution
	local _,_,cmd = msg:find("^(%S*)")
	if cmd and cmd:sub(1,1) == "<" then cmd = "Mainchat" end

	--// Add stats to user and hub
	tFuncs:AddStats(sid, "NMDC", cmd, 1, msg)

	--// Check for control chars
	if cmd ~= "$MyINFO" and cmd ~= "$Key" and cmd ~= "$SR" and cmd ~= "$OpForceMove" and cmd ~= "$Search" then
		local res
		res, msg = tCheck:ControlChars(client, sid, msg, "[\n\t\r\14]")
	end

	if NMDC_Commands[cmd] then
		return NMDC_Commands[cmd](client, sid, msg)
	elseif msg:sub(1,1) == "$" then
		return NMDC_Commands["Unknown"](client, sid, msg)
	else
		return NMDC_Commands["Malformed"](client, sid, msg)
	end
	return client
end

function NMDC_E(msg, invert)
	if invert and msg then
		msg = string.gsub( msg, "&#36;", "$")
		msg = string.gsub( msg, "&#124;", "|")
--~ 		msg = string.gsub( msg, "&#38;", "&")
		msg = string.gsub( msg, "&amp;", "&")
	elseif msg then
		msg = string.gsub( msg, "&", "&amp;")
		msg = string.gsub( msg, "%$", "&#36;")
		msg = string.gsub( msg, "|", "&#124;")
	end
	return msg
end

function NMDC_CreateNickList(sid)
	local ip = tUser:GetIp(sid)
	local oplist = {"$OpList "}
	local botlist = {"$BotList "}
	local nicklist = {"$NickList "}  --// Nicklist not supported, not used anymore
	for vsid,data in pairs(tClient) do
		if data.iState >= 7 and tUser:IsVisibleTo(vsid, sid) then
			local nick = data.sNick.."$$"
			table.insert(nicklist, nick)
			if data.bBot then
				table.insert(botlist, nick)
			end
			if tUser:HasKey(vsid) or (tUser:HasFakeKey(vsid) and ip == data.sIP) then
				table.insert(oplist, nick)
			end
		end
	end
	return table.concat(nicklist), table.concat(oplist), table.concat(botlist)
end

function NMDC_CreateUserIPList(sid)
	local reply = false
	local iplist = {"$UserIP "}

	for vsid, data in pairs(tClient) do
		if not data.bBot and data.iState == 7 and tUser:IsVisibleTo(vsid, sid) then
			table.insert(iplist, data.sNick.." "..(data.sIPEx or "0.0.0.0"))
		end
	end

	if #iplist > 1 then
		return table.concat(iplist, "$$").."|"
	else 
		return false
	end
end

--~ http://www.dchublist.com/hub/search?m=adv&keyword=&SearchHubForm[adc]=1&yt0=Search+hub+%C2%BB
NMDC_Commands = {
	["$Supports"] = function(client, sid, msg)
		local res, retmsg = false, msg

		--// Basic validation
		if client.iState == 1 then
			res = true
		else
			tFuncs:AddStats(sid, "NMDC", "$Supports", -1, msg)
			tUser:Disconnect(sid, wrong, "Wrong login sequence, received $Supports in state "..client.iState..": "..msg)
			return client
		end

		--// Deflood not available - $Supports is only allowed in state 1 and automatically moved to state 2 after receival

		--// Script call
--~ 		if res then res, retmsg = tApi:Main(sid, "SUP", msg) end

		--// Execution
		client.iState = 2	-- update state
		local _,_,supstr = msg:find("$Supports (.+)")
		supstr = supstr or ""

		--// Possible $Supports: UserCommand, NoGetINFO, NoHello, UserIP2, TTHSearch, ZPipe0, ZPipe, BotList, sid, SecuredExecutor, GetZBlock, ChatOnly, QuickList, TTHHAsh, NickChange, IN
		local tSup = {
			-- // supported
			["Cid"] = 1, ["SaltPass"] = 1, ["BotINFO"] = 1, ["BotList"] = 1, ["FailOver"] = 1, ["UserCommand"] = 1, ["NickChange"] = 0, ["TTHSearch"] = 1, ["GetZBlock"] = 1,
			["ChatOnly"] = 1, ["NoHello"] = 1, ["NoGetINFO"] = 1, ["ZPipe0"] = 1, ["ZPipe"] = 1, ["UserIP2"] = 1,  ["DHT0"] = 1, ["TLS"] = 1, ["MCTo"] = 1,
			-- // not supported

			--["SecuredExecutor"] = 2, ["ClientID"] = 2, 	-- Zion
			--[""] = 2, [""] = 2, [""] = 2,
			}

		client.sSupports = ""
		local userSendSupp = "$Supports "
		for w in string.gmatch(supstr, "%S+") do
			client.sSupports = client.sSupports..w..","
			if tSup[w] and tSup[w] == 1 then
				userSendSupp = userSendSupp..w.." " 
				client.tSupports[w] = true
			else
				tFuncs:SafeDebug("[XXX] unknown $Supports received from "..(client.sNick or client.sNickTemp or "unknown nick") .." : "..w, "X")
			end
		end
		client.sSupports = client.sSupports:sub(1,-2)
		userSendSupp = string.gsub(userSendSupp, " $", "") -- remove the last space

		if client.tSupports["BotINFO"] and tHubSettings.tRules.bRegOnly then
			tUser:Disconnect(sid, "This hub is for registered users only, pingers not allowed.")
			return client
		elseif  client.tSupports["BotINFO"] and not tHubSettings.tRules.bAllowPinger then
			tUser:Disconnect(sid, "Pingers are not allowed.")
			return client
		end

		-- store the support string in tClient
--~ 			client.sSupports = userSendSupp

		-- send the $Supports
		tSend:Raw(sid, userSendSupp)

		--// If FailOver is supported, send the FailOver command
		if client.tSupports.FailOver then
			if tHubSettings.tRedirects.sFailOver == "" then
				tSend:Raw(sid, "$FailOver|")
			else
				tSend:Raw(sid, "$FailOver "..tHubSettings.tRedirects.sFailOver.."|")
			end
		end
		
		tFuncs:Debug(tUser:GetDataLine(sid).."$Supports OK, Hub sends: "..userSendSupp, "LNJ", 2)

		return client
	end,
	["$BotINFO"] = function(client, sid, msg)
		if client.iState >= 2 then
			if client.tSupports["BotINFO"] then
				tSend:Raw(sid, "$HubINFO "..tHubSettings.Main.sName.."$"..tHubSettings.Main.sAddress.."$"..tHubSettings.Main.sDescription.."$"..tHubSettings.tRules.iMaxUsers.."$"..tHubSettings.tRules.iMinShare.."$"..tHubSettings.tRules.iMinSlots.."$"..tHubSettings.tRules.iMaxHubs.."$".._APPNAME.._APPVER.."$"..tHubSettings.Main.sEmail.."|")
			elseif tUser:CanOverride(sid, "checkmyinfo") then
				tFuncs:AddStats(sid, "NMDC", "$BotINFO", -1, msg)
				return client
			else
				tFuncs:AddStats(sid, "NMDC", "$BotINFO", -1, msg)
				tUser:Disconnect(sid, wrong, "BotINFO received without BotINFO in $Supports, user is not a pinger.")
				return client
			end
		else
			tFuncs:AddStats(sid, "NMDC", "$BotINFO", -1, msg)
			tUser:Disconnect(sid, wrong,"Wrong login sequence, received $BotINFO in state: "..client.iState)
			return client
		end
		return client
	end,
	["$Kick"] = function(client, sid, msg)
		local res, retmsg, output = false, msg, ""
		local _,_,vnick = string.find(msg, "^$Kick (%S+)$")

		--// Basic validation
		if client.iState ~= 7 then
			tFuncs:AddStats(sid, "NMDC", "$Kick", -1, msg)
			tUser:Disconnect(sid, wrong, "HACK WARNING! Not logged in user tries to kick: "..msg)
			return client
		elseif not tUser:IsAllowed(sid, "kick") then
			tFuncs:AddStats(sid, "NMDC", "$Kick", -1, msg)
			tUser:Disconnect(sid, wrong, "Unauthorized user tried to kick: "..msg)
			return client
		elseif not vnick then 
			tFuncs:AddStats(sid, "NMDC", "$Kick", -1, msg)
			tFuncs:Debug(tUser:GetDataLine(sid).."$Kick doesn't contain nick, message ignored: "..msg)
			return client
		else 
			res = true
		end

		--// Execution
		if res and retmsg then 
			--// HubCommand
			tSend:Raw(sid, "*** "..retmsg)     -- send typed line back to the one who typed it in mainchat

			local action = "!kick "..vnick

			res, retmsg, output = tFuncs:HubCommand(sid, action)

			--// Add hubcommand to cmdhistory
			table.insert(tLogs.HubCommands.History, {socket.gettime(), client.sNick, client.sIP, res, msg})

			if retmsg then
				if output == "PM" then
					tSend:ChatToUser("The result has been sent in PM from <"..tUser:GetNick("AAAA")..">.", "AAAA", sid)
					tSend:PmToUser(retmsg, "AAAA", sid)
				else
					tSend:ChatToUser(retmsg, "AAAA", sid)
				end
			end
		end
		return client
	end,
	["$Key"] = function(client, sid, msg) -- no need to support clients that have no support so key after support is ok^^
		local res, retmsg = true, msg

		--// Basic validation
		if client.iState ~= 2 then
			tFuncs:AddStats(sid, "NMDC", "$Key", -1, msg)
			tUser:Disconnect(sid, wrong, "Wrong login sequence, received $Key in state: "..client.iState)	
			return client
		end

		--// Deflood not available - $Key is only allowed in state 2 and automatically moved to state 3 after receival

		--// Script call
--~ 		if res then res, retmsg = tApi:Main(sid, "KEY", msg) end

		--// Execution
		if res then		--// TODO: check key here
			local key = dc.computeAccessKey("EXTENDEDPROTOCOL_FLEXHUB_MULTIPROTOCOL_"..sid) 
			if key == msg:match("^$Key (.*)$") or not tSecurity.General.bCheckKey or tUser:IsProtected(sid) then
				client.iState = 3
				tFuncs:Debug(tUser:GetDataLine(sid).."$Key OK.", "LNJ", 4)
			else
				tFuncs:AddStats(sid, "NMDC", "$Key", -1, msg)
--~ 				tFuncs:Debug(tUser:GetDataLine(sid).."Wrong $Key, user disconnected: "..msg.." Expected: "..key, "LN", 0)
				tUser:Disconnect(sid, wrong, "Wrong key received after lock: "..msg)
				
				tFuncs:Debug("Lock was: EXTENDEDPROTOCOL_FLEXHUB_MULTIPROTOCOL_"..sid, "FI", 1)
				tFuncs:Debug("Proper key: "..key, "FI", 1)
			end
		end

		return client
	end,
	["$ValidateNick"] = function(client, sid, msg)
		local _,_,nick = string.find(msg, "%$ValidateNick (%S+)$")
		local res, reg, err, inuse = false,false,false,false

		--// Check if nick is on Monitor
		if tMonitor[nick] then
			client.sMonitor = tMonitor[nick]
			tFuncs:Debug("Monitored nick is connecting "..tUser:GetShortData(sid)..", all traffic outputs to: "..tMonitor[nick])
		end

		--// Validation
		if client.iState == 3 then
			if nick and nick ~= "" then
				--// Charset test:
				local ok, convnick, errmsg = tFuncs:ToUTF8(nick)
				local sok, orgnick, errmsg = tFuncs:ToLocale(convnick)

				if not ok or not sok or nick ~= orgnick then
					tFuncs:AddStats(sid, "NMDC", "$ValidateNick", -1, msg)
					tUser:Disconnect(sid, "Your nickname contains characters that can't be converted to the hub's local charset, please use alphanumeric characters.", "[ValidateNick] Nick conversion error: "..(errmsg or ""), "LI",2) 
					return client
				end

				res, client, reg, err, inuse = tCheck:Nick(client, sid, nick)		--// Ban / pattern and IP-binding check

				--// Registered user
				if res and reg then
					if client.tSupports.SaltPass then
						client.sPassSalt = tFuncs:MakeSalt() --// Store random for comparison on pass-arrival

						if tSecurity.General.AllowBlankPass and reg.sPassword == "" then	--// CHECK
							tFuncs:Debug("[$ValidateNick] check 1 blankpass", "X")	-- leave these in until its really confirmed it works on nmdc
							client.bNotifySetOwnPass = true
						else
							tFuncs:Debug("[$ValidateNick] check 1 normal", "X")
							tSend:Raw(sid, "$GetPass "..client.sPassSalt.."|")
						end
					else
						tSend:Raw(sid, "$GetPass|")
					end

					-- // new stuff for blankpass, since a pass request wont be sent, we jump directly to $GetPass from here
					if client.bNotifySetOwnPass then
						client = NMDC_Commands["$MyPass"](client, sid, msg)
					end

					-- add protect for level 10 users, temporary measure
					if reg.iLevel == 10 then
						local save = false
						local lnick = string.lower(client.sNick or client.sNickTemp)
						local reason = "Level 10 user automaticly protected by hub"

						if not tProtects["Nick"][lnick] then
							tProtects["Nick"][lnick] = { ["By"] = tUser:GetNick("AAAA"),  ["Date"] = os.time(), ["Reason"] = reason }
							save = true
							tFuncs:Debug("[Login] added to tProtects: "..lnick, "X")
						end

						if not tProtects["Ip"][client.sIP] then
							tProtects["Ip"][client.sIP] = { ["By"] = tUser:GetNick("AAAA"),  ["Date"] = os.time(), ["Reason"] = reason }
							save = true
							tFuncs:Debug("[Login] added to tProtects: "..client.sIP, "X")
						end

						if save then
							tFuncs:SaveSettings("protects")
						end
					end


				--// Unregistered user
				elseif res then
					--// Create and check CID
					local cid, pid = tFuncs:CalcCid(client.sNick, client.sIP)
					local res, client = tCheck:Cid(client, sid, pid, cid)		--// Ban / pattern and IP-binding check
					if not res then tFuncs:AddStats(sid, "NMDC", "$ValidateNick", -1, msg) return client end

					tSend:Raw(sid, "$Hello "..nick.."|")
				elseif not res and inuse then
					tFuncs:AddStats(sid, "NMDC", "$ValidateNick", -1, msg)
					tSend:Raw(sid, "$ValidateDenide|")
					tUser:Disconnect(sid, "Nickname <"..nick.."> is in use, change nickname before reconnecting")
					return client
				else
					tFuncs:AddStats(sid, "NMDC", "$ValidateNick", -1, msg)
					if not tBuffer[sid].bDisconnect then
						tUser:Disconnect(sid, err)
					end
					return client
				end
			else
				tFuncs:AddStats(sid, "NMDC", "$ValidateNick", -1, msg)
				tUser:Disconnect(sid, wrong, "Malformed $ValidateNick received in state: "..client.iState..": "..msg)
				tIp:Autoban(client.sIP, "Invalid $ValidateNick", msg)
				return client
			end
		else
			tFuncs:AddStats(sid, "NMDC", "$ValidateNick", -1, msg)
			tUser:Disconnect(sid, wrong, "Wrong login sequence, received $ValidateNick in state: "..client.iState..": "..msg)
			return client
		end

--~ 		tFuncs:Debug("CID generated for NMDC client: "..client.sCID, "FI", -1)
		return client
	end,
	["$MyPass"] = function(client, sid, msg)
		--// Basic validation
		if client.iState == 4 then
			if not client.bNotifySetOwnPass then	-- normal
				local retmsg = msg
				local _,_,pass = retmsg:find("$MyPass (%S+)")     --// get the password
				local pass = pass or ""
				
--~ 				if not pass then
--~ 					tFuncs:AddStats(sid, "NMDC", "$MyPass", -1, msg)
--~ 					tUser:Disconnect(sid, wrong, "Malformed $MyPass received: "..msg)
--~ 					return client
--~ 				else
					--// Check password
					local res, client, err = tCheck:Pass(client, sid, pass)
					if not res then
						tFuncs:AddStats(sid, "NMDC", "$MyPass", -1, msg)
						tSend:Raw(sid, "$BadPass|")   --// send $Badpass..
						if tSecurity.General.WrongPassMsg and tSecurity.General.WrongPassMsg ~= "" then
							tUser:Disconnect(sid, tFuncs:FillVars(tSecurity.General.WrongPassMsg), err)
						else
							tUser:Disconnect(sid, "You provided a wrong password.", err)
						end
						return client
					end
--~ 				end
			else
				--tFuncs:Debug("[$MyPass] blankpass", "X")
				local res, client, err = tCheck:Pass(client, sid, "")
				if not res then 
					tFuncs:AddStats(sid, "NMDC", "$MyPass", -1, msg)
					if tSecurity.General.WrongPassMsg and tSecurity.General.WrongPassMsg ~= "" then
						tUser:Disconnect(sid, tFuncs:FillVars(tSecurity.General.WrongPassMsg), err)
					else
						tUser:Disconnect(sid, "You provided a wrong password.", err)
					end
					return client
				end
			end

		else	-- if not iState 4
			tFuncs:AddStats(sid, "NMDC", "$MyPass", -1, msg)
			tUser:Disconnect(sid, wrong, "Wrong login sequence, received $MyPass in state: "..client.iState)
			return client
		end

		--// Create and check CID
		local cid, pid = tFuncs:CalcCid(client.sNick, client.sIP)
		local res, client = tCheck:Cid(client, sid, pid, cid)		--// Ban / pattern and IP-binding check
		if not res then tFuncs:AddStats(sid, "NMDC", "$ValidateNick", -1, msg) return client end

		--// Send LogedIn or Hello
		client.iState = 5

		if tProfiles[client.sProfile]["bHasKey"] or tProfiles[client.sProfile]["bHasFakeKey"]then
			tSend:Raw(sid, "$LogedIn|$Hello "..client.sNick.."|")
		else
			tSend:Raw(sid, "$Hello "..client.sNick.."|")
		end
		
		tFuncs:Debug(tUser:GetDataLine(sid).."$MyPass OK", "LNJ", 3)

		return client
	end,
	["$Version"] = function(client, sid, msg)
		local DoVersionArrival = function(sid,msg) return true end      --// TODO: replace with actual function
		if client.iState == 5 then
			if DoVersionArrival(sid, msg) then
				local _,_,version = string.find(msg, "%$Version (%S+)")
				if version then
					client.sVersion = version
					client.iState = 6 -- not much to do here just upgrade iState
					tFuncs:Debug(tUser:GetDataLine(sid).."$Version OK.", "LNJ", 3)
				else
					tFuncs:AddStats(sid, "NMDC", "$Version", -1, msg)
					tUser:Disconnect(sid, wrong, "Malformed $Version received: "..msg)
					return client
				end
			else
				tFuncs:AddStats(sid, "NMDC", "$Version", -1, msg)
				tUser:Disconnect(sid, wrong, "Wrong login sequence, received $Version in state: "..client.iState..": "..msg)
				return client
			end
		end
		return client
	end,
	["$MyINFO"] = function(client, sid, msg)		-- $MyINFO $ALL Daywalker  <RSX++ V:1.00,M:A,H:3/3/16,S:3,L:19>$ $0.005$$75142726994$|
		local res, retmsg, err = false, msg, false

		--// Deflood check
		res, retmsg = tDeflood:Main(sid, "MN", retmsg)

		--// Basic validation
		if (client.iState < 6 and not tUser:IsProtected(sid)) then
			tFuncs:AddStats(sid, "NMDC", "$MyINFO", -1, msg)
			tUser:Disconnect(sid, wrong, "Wrong login sequence, received $MyINFO in state: "..client.iState..": "..msg)			
			return client
		end

		--// Execution
		if res and client.iState == 7 then	--// this user is already logged in, just update it's myinfo
			if retmsg ~= client.tInfo.tNMDC.sOrig then
				res, client, err = tInfo:ParseNMDC(client, sid, retmsg)
				if not res then tFuncs:AddStats(sid, "NMDC", "$MyINFO", -1, msg) return client end

				--// Rulecheck
				res, retmsg = tCheck:Rules(sid)
				if not res then tFuncs:AddStats(sid, "NMDC", "$MyINFO", -1, msg) return client end

				--// Script execution
--~				res, retmsg = tApi:Main(sid, "MY", retmsg) end
--~				if not res then return client end

				--// Send MyINFO
				tSend:MyInfPool(sid)
				tFuncs:Debug(tUser:GetDataLine(sid).."$MyINFO update OK, resent to all users if myinfointerval has passed.", "TN", 8)
			end
		elseif res and (client.iState == 6 or (client.iState == 5 and tUser:IsProtected(sid)))then        --// this user is in the final stage of logging in
			res, client, err = tInfo:ParseNMDC(client, sid, msg)
			if not res then tFuncs:AddStats(sid, "NMDC", "$MyINFO", -1, msg) tUser:Disconnect(sid, wrong, err) return client end

			--// Set client settings
			local reg, rnick = tNick:GetReg(nick)
			if reg and reg.bHidden then
				client.bHidden = true
			end
			
			if (reg and reg.bShareHidden) or tProfiles[client.sProfile].bShareHiddenForced then
				client.bShareHidden = true
			end

			--// Rulecheck
			res, err = tCheck:Rules(sid)
			if not res then tFuncs:AddStats(sid, "NMDC", "$MyINFO", -1, msg) return client end

			-- // wanky bot fix
			if tUser:CanOverride(sid, "checkmyinfo") then
				if client.sSupports and client.sSupports == "" or not client.sSupports then
					tFuncs:Debug("[$MyINFO] a connecting client override myinfo check and has no supports, faking it", "X")
					client.tSupports = {
						["UserIP2"] = true,
						["NoHello"] = true,
						["NoGetINFO"] = true,
						}
					client.sSupports = "NoGetINFO,NoHello,UserIP2"
				end
			end

			--// Script execution
			--~if res then res, client, retmsg = tApi:Main(sid, "MY", retmsg) end

			--// TO USER
			tUser:Joined(client, sid)        --// send Welcome, MOTD, etc

			--// TO OTHERS
			tUser:Announce(sid)		--// Sends MyINFO / UserIP / Oplist to all users
		
			if client.tSupports["BotINFO"] then
--~ 					tUser:Disconnect(sid)
			end

			--// Send Debug to GUI
			tFuncs:Debug(tUser:GetDataLine(sid).."$MyINFO OK.", "LNJ", 3)
		end
		return client
	end,
	["$GetNickList"] = function(client, sid, msg)	-- Do we need to support this???? Seems like a command to obfuscate client- or hubbugs by just resending entire userlist
		local res, msg = tDeflood:Main(sid, "NL", msg)
		if not res then tFuncs:AddStats(sid, "NMDC", "$GetNickList", -1, msg) return client end

		if client.iState == 7 or (tUser:IsProtected(sid) and not tUser:Supports(sid, "NoGetINFO")) then    -- only sends the nick and oplist if nick is allowed in (unregged or with proper pass), can respond before MyINFO is received and checked...
--~ 			--// Create NickList and OpList
			local nicklist, oplist, botlist = NMDC_CreateNickList(sid)

			--~ 			--// Send the BotList and OpList

			if tUser:Supports(sid, "NoGetINFO") then
				tSend:MyInf("ALL", sid)
			else
				tSend:Raw(sid, nicklist.."|")
--~ 				-- do nothing, let user send GetINFO to get the MyINFO's he wants
			end

			tSend:Raw(sid, botlist.."|"..oplist.."|")

			--// Debug to GUI
--~ 			tFuncs:Debug(tUser:GetDataLine(sid).."Hub sends NickList and OpList: "..nicklist.."|"..oplist, "LNJ", 9)
--~ 			tFuncs:Debug(tUser:GetDataLine(sid).."$GetNickList OK. NickList and OpList sent", "LNJ",3)
		elseif client.iState == 6 then
			-- skip $GetNickList during login, NickList and all MyINFO's will be send when iState 7 is reached
		else
			tUser:Disconnect(sid, wrong, "Wrong login sequence, received $GetNickList in state: "..client.iState..": "..msg)
			return client
		end

		return client
   end,
	["$GetINFO"] = function(client, sid, msg)		-- Do we need to support this???? Hubs sends all needed info-changes
		local res, msg = tDeflood:Main(sid, "XX", msg)
		if not res then tFuncs:AddStats(sid, "NMDC", "$GetINFO", -1, msg) return client end

		local _,_,swho,sfrom = string.find(msg, "$GetINFO (%S+) (%S+)")

		if not tUser:Supports(sid, "NoGetINFO") then
			if sfrom and sfrom == client.sNick then
				if client.iState == 7 then
--~ 					tUser:Disconnect(sid, "Your client is too old, update to a recent client.", "Old client: sends $GetINFO, doesn't support NoGetINFO")
				
--~ 					usid = tNick:GetSid(swho)
--~ 					if usid then
--~ 						tSend:MyInf(usid, sid)
--~ 					end
--~ 					tFuncs:Debug(tUser:GetDataLine(sid).."$GetINFO: "..msg, "LNJ", 3) -- later on
				elseif client.iState == 6 then
					-- skip during login
				else
					tFuncs:AddStats(sid, "NMDC", "$GetINFO", -1, msg)
					tUser:Disconnect(sid, wrong, "Wrong login sequence, received $GetINFO in state: "..client.iState..": "..msg)
					return client
				end
			else
				--// Nick spoofed
				tFuncs:AddStats(sid, "NMDC", "$GetINFO", -1, msg)
				tUser:Disconnect(sid, wrong, "Nick spoofed in $GetINFO: "..msg)
				return client					
			end
		else
			--// User supports 'NoGetINFO' but sends NoGetINFO anyway
			tFuncs:AddStats(sid, "NMDC", "$GetINFO", -1, msg)
			tUser:Disconnect(sid, wrong, "Client supports NoGetINFO but sends $GetINFO: "..msg)
			return client
		end

		return client
   end,
	["$To:"] = function(client, sid, msg)			-- $To: <othernick> From: <nick> $<<nick>> <message>|
		local res = false
		local _,_,tonick,fromnick,chatnick,retmsg = msg:find("$To: (%S+) From: (%S+) $<(%S+)> (.+)")
		local tosid = tNick:GetSid(tonick)
		local hcmd = false
		local me = false

		--// Basic validation
		if client.iState == 7 then
			if tonick and retmsg then
				if fromnick and fromnick == client.sNick then
					if fromnick == chatnick then
						if retmsg:find("^[/%+][Mm][Ee] ") then 
							me = true
							retmsg = retmsg:gsub("^[/%+][Mm][Ee] ", "")
						end

						if tosid then
							hcmd = string.find(retmsg, "^["..tHubSettings.Main.sCmdPrefixes.."].+$")
							res = true
						else
							tDeflood:Main(sid, "PM", retmsg, tosid)
							tFuncs:AddStats(sid, "NMDC", "$To:", -1, msg)
							return client
							--// User offline, only count for deflood
						end
					else
						tFuncs:AddStats(sid, "NMDC", "$To:", -1, msg)
						tUser:Disconnect(sid, wrong, "Nick spoofed: "..msg)
						return client
					end
				else
					tFuncs:AddStats(sid, "NMDC", "$To:", -1, msg)
					tUser:Disconnect(sid, wrong, "Nick spoofed: "..msg)
					return client
				end
			else
				tFuncs:AddStats(sid, "NMDC", "$To:", -1, msg)
				tUser:Disconnect(sid, wrong, "Malformed PM: "..msg)
				return client
			end
		else
			tFuncs:AddStats(sid, "NMDC", "$To:", -1, msg)
			tUser:Disconnect(sid, wrong, "Wrong login sequence, received $To: in state: "..client.iState..": "..msg)
			return client
		end

		retmsg = NMDC_E(retmsg, true)
		local original = retmsg

		--// Deflood check
		if res and hcmd and (tosid == "AAAA" or tosid == "BBBB" or tosid == "CCCC") then
			tFuncs:AddStats(sid, "NMDC", "HC", 1, msg)
			res, retmsg = tDeflood:Main(sid, "HC", retmsg)
			if not res then tFuncs:AddStats(sid, "NMDC", "HC", -1, msg) tFuncs:AddStats(sid, "NMDC", "$To:", -1, msg) return client end
		else
			tFuncs:AddStats(sid, "NMDC", "PM", 1, msg)
			res, retmsg = tDeflood:Main(sid, "PM", retmsg, tosid)
			if not res then tFuncs:AddStats(sid, "NMDC", "PM", -1, msg) tFuncs:AddStats(sid, "NMDC", "$To:", -1, msg) return client end
			if res then res, retmsg = tCheck:Pattern(sid, "PM", retmsg) end
			if not res then tFuncs:AddStats(sid, "NMDC", "PM", -1, msg) tFuncs:AddStats(sid, "NMDC", "$To:", -1, msg) return client end
			if res then res, retmsg = tCheck:Filter(sid, "PM", retmsg) end
			if not res then tFuncs:AddStats(sid, "NMDC", "PM", -1, msg) tFuncs:AddStats(sid, "NMDC", "$To:", -1, msg) return client end
		end

		--// Script call
--~ 		if res then res, sLine = tApi:Main(sid, "PM", sLine, tosid)

		--//Execution
		if res and retmsg and retmsg ~= original then
			tSend:PmToUser("[Message filtered] <"..client.sNick.."> "..retmsg, "AAAA", sid, tosid)
		end

		if res and retmsg then
			res, client = tSend:Pm(client, sid, tosid, retmsg)
		end

		return client
	end,
	["$MCTo:"] = function(client, sid, msg)			-- $MCTo: <target_nick> $<sender_nick]> <message>| http://neisep.com/dc/index.php?title=$MCTo
	-- [2012/02/14 00:11:25] <[-Atlantis-]> [Monitoring <TestBOT>] [RCV] $MCTo: [OP]endator $<TestBOT> Number of unread releases 5

		local res = false
		local _,_,tonick, nick, retmsg = msg:find("$MCTo: (%S+) $<(%S+)> (.+)")
		local tosid = tNick:GetSid(tonick)

		--// Basic validation
		if not client.tSupports["MCTo"] then
			tFuncs:AddStats(sid, "NMDC", "$MCTo", -1, msg)
			tFuncs:Debug(tUser:GetDataLine(sid).."Trying to use unsupported command, ignored: "..msg, nil, 3)
			return client
		elseif client.iState == 7 then
			if nick and nick == tClient[sid].sNick then	-- Checks if <nick> in msg matches usernick to prevent spoofing
				if retmsg then
					res = true
				else
					tDeflood:Main(sid, "MC", retmsg)	--// Empty message will be ignored but will be checked for flooding
					tFuncs:AddStats(sid, "NMDC", "$MCTo", -1, msg)
					return client
				end
			elseif nick then
				tFuncs:AddStats(sid, "NMDC", "$MCTo", -1, msg)
				tBan:KickUser(sid, "Nick spoofing is not allowed.")
				return client
			elseif not tosid then
				tFuncs:AddStats(sid, "NMDC", "$MCTo", -1, msg)
				tFuncs:Debug(tUser:GetShortData(sid).."Sent: "..retmsg.." to unknown user: "..tonick, nil , 3)
				return client
			else
				if string.find(msg, "^<> ") then
					tFuncs:AddStats(sid, "NMDC", "$MCTo", -1, msg)
					tUser:Disconnect(sid, "You're using a client too old to work with these rightclicks, update your client to make them work.", "Client too old to support current rightclicks, disconnected: "..msg)
				else
					tFuncs:AddStats(sid, "NMDC", "$MCTo", -1, msg)
					tUser:Disconnect(sid, wrong, "Malformed $MCTo message: "..msg)
				end
				return client
			end
		else
			tFuncs:AddStats(sid, "NMDC", "$MCTo", -1, msg)
			tUser:Disconnect(sid, wrong, "Wrong login sequence, received $MCTo state "..client.iState..": "..msg)
			return client
		end

		retmsg = NMDC_E(retmsg, true)

		--// Deflood check, using MC rules as message is sent to MainChat
		if res and retmsg then
			res, retmsg = tDeflood:Main(sid, "MC", retmsg)

			if not res then tFuncs:AddStats(sid, "NMDC", "$MCTo", -1, msg) return client end
			if res then res, retmsg = tCheck:Pattern(sid, "MC", retmsg) end
			if not res then tFuncs:AddStats(sid, "NMDC", "$MCTo", -1, msg) return client end
			if res then res, retmsg = tCheck:Filter(sid, "MC", retmsg) end
			if not res then tFuncs:AddStats(sid, "NMDC", "$MCTo", -1, msg) return client end
		end

		--// Execution
		tSend:ChatToUser(retmsg, sid, tosid)
		return client
	end,
	["$ConnectToMe"] = function(client, sid, msg)
		-- $ConnectToMe <SenderNick> <RemoteNick> <SenderIp>:<SenderPort>
		local _,_,tonick,fromip,fromport = string.find(msg, "$ConnectToMe (%S+) (%d+%.%d+%.%d+%.%d+):(%d+)")
		local _,_,nat = string.find(msg, "$ConnectToMe %S+ %d+%.%d+%.%d+%.%d+:%d+([NR])$")
		local _,_,natnick = string.find(msg, "$ConnectToMe %S+ %d+%.%d+%.%d+%.%d+:%d+.- (%S+)$")
--~ 		local xports = { ["80"] = true, ["21"] = true, ["25"] = true, ["110"] = true, ["411"] = true }

		--// Basic validation
		if client.iState == 7 then
			if tonick and fromip and fromport then
				client.sActivePort = fromport
				local realip = client.sIP
				
				if not tClient[sid].bActive and not tUser:IsOp(sid) and not nat and not tUser:CanOverride(sid, "modecheck") then
					local continue,_,actionok,actionresult = tDeflood.Action[tSecurity.General.sWrongModeAction](sid, "Sending active connection command while stating passive mode in tag (fake tag)",nil, true)
					local debuglevel = SubSet.ActionDebug[tSecurity.General.sWrongModeAction] or 0
					if actionok and actionresult then tFuncs:Debug("[Security] [wrongmode_action] "..tUser:GetShortData(sid).."Sending active connection command while stating passive mode in tag (fake tag) / "..actionresult, "FI", debuglevel) end
					if not continue then
						tFuncs:AddStats(sid, "NMDC", "$ConnectToMe", -1, msg)
						return client
					end
				end

				if nat and natnick and natnick ~= client.sNick then
					tUser:Disconnect(sid, wrong, "Nick spoofing detected: "..msg)
				end

				--// Replace local IP for connecting hubowners
				if tIp:IsLocal(realip) then
					msg = "$ConnectToMe "..tonick.." "..(client.sIPEx or client.sIP)..":"..fromport
				elseif fromip ~= client.sIP then
--~ 					if not xports[fromport] then
						--// Only allowed to specify another IP in CTM if connecting locally
						if not tIp:IsLocal(realip) then
							msg = "$ConnectToMe "..tonick.." "..realip..":"..fromport
							tSend:ChatToUser("Your client sends a wrong IP, make sure you have set your IP correctly in your client's settings: "..realip, "AAAA", sid)
							tFuncs:Debug(tUser:GetDataLine(sid).."Wrong IP in $ConnectToMe ("..fromip.."), replaced with real IP: "..msg, "CN", 3)
						end
--~ 					else
--~ 						tFuncs:Debug(tUser:GetDataLine(sid).."Maliciously spoofed $ConnectToMe received, user disconnected: "..msg)
--~ 						return tUser:Disconnect(client, sid, "Malformed message received, get a proper client.")
--~ 					end
				end
			else
				tFuncs:AddStats(sid, "NMDC", "$ConnectToMe", -1, msg)
				tUser:Disconnect(sid, wrong, "Malformed $ConnectToMe received: "..msg)
				return client
			end
		else
			tFuncs:AddStats(sid, "NMDC", "$ConnectToMe", -1, msg)
			tUser:Disconnect(sid, wrong, "Wrong login sequence, received $ConnectToMe in state "..client.iState..": "..msg)
			return client
		end

		--// Deflood
		local res, retmsg = tDeflood:Main(sid, "CT", msg)
		if not res then tFuncs:AddStats(sid, "NMDC", "$ConnectToMe", -1, msg) return client end

		--// Does remote user exist?
		local tosid = tNick:GetSid(tonick)
		if not tosid or tClient[tosid].iState < 7 then
			tFuncs:AddStats(sid, "NMDC", "$ConnectToMe", -1, msg)
			tFuncs:Debug(tUser:GetDataLine(sid).."Unknown/offline nick in $ConnectToMe: "..retmsg, "CN", 4)
			return client
		end

		--// Share hidden
		if tClient[tosid].bShareHidden then
			tFuncs:AddStats(sid, "NMDC", "$ConnectToMe", -1, msg)
			tFuncs:Debug(tUser:GetDataLine(sid).."Trying to download a hidden share, $ConnectToMe dropped: "..retmsg, "CN", 3)
			return client 
		end

		--// Execution
		if tClient[tosid].sProtocol == "NMDC" then
			tSend:Raw(tosid, retmsg)
		else
			tFuncs:AddStats(sid, "NMDC", "$ConnectToMe", -1, msg)
			tFuncs:Debug(tUser:GetDataLine(sid).."$ConnectToMe to ADC user attempted, message dropped: "..msg, "CN", 5)
			return client
		end

		return client
	end,
	["$RevConnectToMe"] = function(client, sid, msg)
		-- $RevConnectToMe <ClientNick> <TargetNick>|
		local _,_,fromnick,tonick = msg:find("$RevConnectToMe (%S+) (%S+)")
		local res = false

		--// Basic validation
		if client.iState == 7 then
			if fromnick and tonick then
				if fromnick == client.sNick then
					if tClient[sid].bActive and not tUser:IsOp(sid) and not tUser:CanOverride(sid, "modecheck") then
						local continue,_,actionok, actionresult = tDeflood.Action[tSecurity.General.sWrongModeAction](sid, "Sending passive connection command while stating active mode in tag (fake tag)", nil, true)
						local debuglevel = SubSet.ActionDebug[tSecurity.General.sWrongModeAction] or 0
						if actionok and actionresult then tFuncs:Debug("[Security] [wrongmode_action] "..tUser:GetShortData(sid).."Sending passive connection command while stating active mode in tag (fake tag) / "..actionresult, "FI", debuglevel) end
						if not continue then
							tFuncs:AddStats(sid, "NMDC", "$RevConnectToMe", -1, msg) 
							return client
						end
					end
					res = true
				else	
					tFuncs:AddStats(sid, "NMDC", "$RevConnectToMe", -1, msg)
					tUser:Disconnect(sid, wrong, "Nick spoofing detected: "..msg)
					return client
				end
			else
				tFuncs:AddStats(sid, "NMDC", "$RevConnectToMe", -1, msg)
				tUser:Disconnect(sid, wrong, "Malformed  $RevConnectToMe received: "..msg)
				return client
			end
		else
			tFuncs:AddStats(sid, "NMDC", "$RevConnectToMe", -1, msg)
			tUser:Disconnect(sid, wrong, "Wrong login sequence, received $RevConnectToMe in state "..client.iState..": "..msg)
			return client
		end

		--// Deflood
		local res, retmsg = tDeflood:Main(sid, "RC", msg)
		if not res then tFuncs:AddStats(sid, "NMDC", "$RevConnectToMe", -1, msg) return client end

		--// Does remote user exist?
		local tosid = tNick:GetSid(tonick)
		if not tosid or tClient[tosid].iState < 7 then
			tFuncs:AddStats(sid, "NMDC", "$RevConnectToMe", -1, msg)
			tFuncs:Debug(tUser:GetDataLine(sid).."Unknown/offline nick in $ConnectToMe: "..retmsg, "CN", 4)
			return client
		end

		--// Share hidden
		if tClient[tosid].bShareHidden then
			tFuncs:AddStats(sid, "NMDC", "$RevConnectToMe", -1, msg)		
			tFuncs:Debug(tUser:GetDataLine(sid).."Trying to download a hidden share, $ConnectToMe dropped: "..retmsg, "CN", 3)
			return client 
		end

		--// Execution
		if res then
			if tClient[tosid].sProtocol == "NMDC" and tClient[tosid].bActive then
				tSend:Raw(tosid, retmsg.."|")
			elseif tClient[tosid].sProtocol == "ADC" then
				tFuncs:AddStats(sid, "NMDC", "$RevConnectToMe", -1, msg)		
				tFuncs:Debug(tUser:GetDataLine(sid).."$RevConnectToMe to ADC user attempted, message dropped: "..msg, "CN", 5)
				return client
			else
				tFuncs:AddStats(sid, "NMDC", "$RevConnectToMe", -1, msg)		
				tFuncs:Debug(tUser:GetDataLine(sid).."$RevConnectToMe to passive user attempted, message dropped: "..msg, "CN", 4)
				return client
			end
		end

		return client
	end,
	["$Search"] = function(client, sid, msg)
		--// $Search 84.240.39.180:54291 F?T?0?9?TTH:I2MGEYIJJHHZNNWXO5JJ3IOOBGZFMKMZP2Q53QQ
		--// $Search 84.160.56.137:1026 F?T?0?7?some$movie
		local searchtype = "SP"
		local active,ip,port,tth, search,changed, nick = false, false, false, false, false, false, false
		local _,_,source,middle, search = msg:find("^%S+ (%S+)%s+(.%?.%?.*%?.*%?)(.*)$")

		--// Basic validation
		if source and search then
			_,_,ip,port = source:find("^(%d+%.%d+%.%d+%.%d+):(%--%d+)$")
			_,_,nick = source:find("^Hub:(%S+)$")
			_,_,tth = string.find(search, "^TTH:(.*)$")
			if ip and port then active = true end

			local debuglevel = 2
			if tth then debuglevel = 3 end

			if client.iState == 7 then
				if not active and nick then
					--// Check proper mode
					if tClient[sid].bActive and not tUser:IsOp(sid) and not tUser:CanOverride(sid, "modecheck") then
						local continue, _, actionok, actionresult = tDeflood.Action[tSecurity.General.sWrongModeAction](sid, "Sending passive search while stating active mode in tag (fake tag)", nil, true)
						local debuglevel = SubSet.ActionDebug[tSecurity.General.sWrongModeAction] or 0
						if actionok and actionresult then tFuncs:Debug("[Security] [wrongmode_action] "..tUser:GetShortData(sid).."Sending passive search while stating active mode in tag (fake tag) / "..actionresult, "FI", debuglevel) end
						if not continue then
							tFuncs:AddStats(sid, "NMDC", "$Search", -1, msg)
							return client
						end
					end

					if tHubSettings.Feed.bShowSearches then 
						tFuncs:Debug(tUser:GetDataLine(sid)..msg, "SNP", debuglevel)
					end

					if nick ~= client.sNick then
						tFuncs:AddStats(sid, "NMDC", "$Search", -1, msg)
						tUser:Disconnect(sid, wrong, "Nick spoofing detected: "..msg)
						return client
					end
				elseif active and ip and port then
					searchtype = "SA"
					client.sSearchPort = port

					--// Check proper mode
					if not tClient[sid].bActive and not tUser:IsOp(sid) and not tUser:CanOverride(sid, "modecheck") then
						local continue, _, actionok, actionresult = tDeflood.Action[tSecurity.General.sWrongModeAction](sid, "Sending active search while stating passive mode in tag (fake tag)", nil, true)
						local debuglevel = SubSet.ActionDebug[tSecurity.General.sWrongModeAction] or 0
						if actionok and actionresult then tFuncs:Debug("[Security] [wrongmode_action] "..tUser:GetShortData(sid).."Sending active search while stating passive mode in tag (fake tag) / "..actionresult, "FI", debuglevel) end
						if not continue then
							tFuncs:AddStats(sid, "NMDC", "$Search", -1, msg)
							return client
						end
					end

					if tHubSettings.Feed.bShowSearches then 
						tFuncs:Debug(tUser:GetDataLine(sid)..msg, "SNA", debuglevel)
					end

					--// Use proper IP
					if ip ~= client.sIPEx and not tIp:IsLocal(client.sIP) then tFuncs:Debug(tUser:GetDataLine(sid).."IP spoofing detected, replaced with actual IP: "..msg, "FI", 3) end
					source = client.sIPEx..":"..port
					
				else
					tFuncs:AddStats(sid, "NMDC", "$Search", -1, msg)
--~ 					tFuncs:Debug(tUser:GetDataLine(sid).."[Malformed search] "..msg, "SI", 2)
					tUser:Disconnect(sid, wrong, "Malformed search detected: "..msg)
					return client
				end
			else
				tFuncs:AddStats(sid, "NMDC", "$Search", -1, msg)
				tUser:Disconnect(sid, wrong, "Wrong login sequence, received $Search in state "..client.iState..": "..msg)
				return client
			end
		else
			tFuncs:AddStats(sid, "NMDC", "$Search", -1, msg)
			tFuncs:Debug(tUser:GetDataLine(sid).."[Malformed search] "..msg, "SI", 0)
			tUser:Disconnect(sid, wrong, "Malformed search detected: "..msg)
			return client
		end

		tFuncs:AddStats(sid, "NMDC", searchtype, 1, msg)

		--/ Deflood
		search = search:gsub("%$", " ")
		local res, search = tDeflood:Main(sid, searchtype, search)
		if not res then tFuncs:AddStats(sid, "NMDC", "$Search", -1, msg) tFuncs:AddStats(sid, "NMDC", searchtype, -1, msg) return client end

		--// Pattern check
		if not tth then 
			res, search = tCheck:Pattern(sid, "SE", search)
			if not res then tFuncs:AddStats(sid, "NMDC", "$Search", -1, msg) tFuncs:AddStats(sid, "NMDC", searchtype, -1, msg) return client end
		end
		
		--// Script call
--~ 		if res then res, search = tApi:Main(sid, "SE", search) end

		--/ Execution
		if res then 
			search = search:gsub("%s+", "$")
			local send = "$Search "..source.." "..middle..search.."|"
			if not active then 	--// Passive search, only send to active users
				for vsid, vdata in pairs(tClient) do
					if vdata.iState == 7 and vdata.sProtocol == "NMDC" and vdata.iShareSize and vdata.iShareSize > 0 and vdata.bActive and (not vdata.bShareHidden) then
						tSend:Raw(vsid, send)
					end
				end
			else
				for vsid, vdata in pairs(tClient) do
					if vdata.iState == 7 and vdata.sProtocol == "NMDC" and vdata.iShareSize and vdata.iShareSize > 0 and (not vdata.bShareHidden) then
						tSend:Raw(vsid, send)
					end
				end
--~ 				tSend:Raw("P_NMDC", send, sid)
			end
		end
		return client
	end,
	["$SR"] = function(client, sid, msg)			-- Example: $SR plop documentaries_7\9-11 - The Greatest Lie Ever Sold by Anthony Hilder 0/3FlexHub Test - FlexHub News Flash: i can see all bugs in buglist now without scrolling!!! (85.223.76.125:8000)nmdctester
		local res, retmsg = false, msg
		if not tCheck:ControlChars(client, sid, msg, "\05") then return client end
		local fromnick, tonick, tosid = false, false, false
		local _,_, fromnick, result, fslots, tslots, hub, tonick = msg:find("^%$SR (%S+) (.*) (%d+)/(%d+)\05(.*)\05(%S+)$")

		--// Basic validation
		if fromnick and tonick then
 			if fromnick == client.sNick then
				if client.iState == 7 then
					tosid = tNick:GetSid(tonick)
					local ok, convnick = tFuncs:ToLocale(tonick)
					if not tosid then tosid = tNick:GetSid(convnick) end
					if tosid and tClient[tosid].iState == 7 then		--// this is the only possible allowance of the command to proceed to deflood / script etc
						res = true
					else
						tFuncs:Debug("[$SR] "..tUser:GetDataLine(sid).."User send an $SR for non-existing nick <"..tonick.."> / <"..convnick.."> : "..msg, "SNR", 9)	--// this should somehow still go through deflood... before discarding
						res, retmsg = tDeflood:Main(sid, "SR", retmsg)
						tFuncs:AddStats(sid, "NMDC", "$SR", -1, msg)
						return client
					end
				else
					tFuncs:AddStats(sid, "NMDC", "$SR", -1, msg)
					tUser:Disconnect(sid, wrong, "Wrong login sequence, received $SR in state "..client.iState..": "..msg)	
					return client
				end
			else
				tFuncs:AddStats(sid, "NMDC", "$SR", -1, msg)
				tUser:Disconnect(sid, wrong, "Nick spoofing detected: "..msg)
				return client
			end
		elseif tUser:IsOp(sid) then
			tFuncs:AddStats(sid, "NMDC", "$SR", -1, msg)
			tFuncs:Debug(tUser:GetDataLine(sid).."Couldn't parse $SR: "..msg.."\n"..tostring(fromnick).." / "..tostring(result).." / "..tostring(fslots).." / "..tostring(tslots).." / "..tostring(hubname).." / "..tostring(hubaddress).." / "..tostring(tonick))
			return client
		else
			tFuncs:AddStats(sid, "NMDC", "$SR", -1, msg)
			tFuncs:Debug(tUser:GetDataLine(sid).."Couldn't parse $SR: "..msg.."\n"..tostring(fromnick).." / "..tostring(result).." / "..tostring(fslots).." / "..tostring(tslots).." / "..tostring(hubname).." / "..tostring(hubaddress).." / "..tostring(tonick))
			tUser:Disconnect(sid, wrong, "Couldn't parse $SR: "..msg)
			return client
		end

		--// Deflood check
		local res, retmsg = tDeflood:Main(sid, "SR", retmsg)
		if not res then tFuncs:AddStats(sid, "NMDC", "$SR", -1, msg) return client end

		--// Pattern check
		local res, retmsg = tCheck:Pattern(sid, "SR", retmsg)
		if not res then tFuncs:AddStats(sid, "NMDC", "$SR", -1, msg) return client end

		--// Check if target user is allowed to receive these results
		local res, retmsg = tDeflood:Main(tosid, "PR", retmsg)
		if not res then tFuncs:AddStats(sid, "NMDC", "$SR", -1, msg) return client end

		--// Script call
--~ 		res, retmsg = tAPi:Main(sid, "SR", msg) 

		--// Execution
		local retmsg = retmsg:gsub("\05%S-$", "|")
		tSend:Raw(tosid, retmsg)

		return client
	end,
	["$Quit"] = function(client, sid, msg)
--~ 		tSend:ChatToAll(tUser:GetNick(sid).." left the hub.", "AAAA")
		tUser:Disconnect(sid, "You've send a quit-command.", "User sent $Quit command")
		return client
	end,
	["$UserIP"] = function(client, sid, msg)
		if client.iState == 7 then
			if tUser:IsAllowed(sid, "userip") then		--// $UserIP is ignored for operators, if allowed 'userip' then userip's will be sent automatically on login
--~ 				local parse = string.gsub(msg, ".UserIP ", "$$").."$$"
--~ 				local reply = "$UserIP "
--~ 				local match = false
--~ 				
--~ 				for nick in string.gmatch(parse, "%$%$(.*)%$%$") do
--~ 					local vsid = tNick:GetSid(nick)
--~ 					if vsid and tUser:IsHigher(sid, vsid) then
--~ 						match = true
--~ 						reply = reply..nick.." "..tUser:GetIp(vsid).."$$"
--~ 					else
--~ 						tSend:ChatToUser("*** Can't retrieve IP for nick: "..nick, "AAAA", sid)
--~ 					end
--~ 				end

--~ 				if match then
--~ 					reply = string.sub(reply, 1, -3).."|"
--~ 					tSend:Raw(sid, reply)				
--~ 				else
--~ 					tSend:ChatToUser("*** Can't retrieve IP's for $UserIP request: "..msg, "AAAA", sid)
--~ 				end
			else
				if not tUser:IsOp(sid) then	--// If non-operator uses $UserIP then...
--~ 					client = tUser:Disconnect(client, sid, "You are not allowed to use $UserIP.")
--~ 					tFuncs:Debug(tUser:GetDataLine(sid).."Received $UserIP from user, user dropped.")
				end
			end
		else
			tUser:Disconnect(sid, wrong, "Wrong login sequence, received $UserIP in state "..client.iState..": "..msg)
			return client
		end
		return client
	end,
	["$OpForceMove"] = function(client, sid, msg)
		if client.iState == 7 then
			if tUser:IsAllowed(sid, "redirect") then
				local s,e,nick,where,data = msg:find("%$Who:([^$]+)%$Where:([^$]+)%$Msg:(.*)$")
				local me = tUser:GetNick(sid)
				local olu = tNick:GetSid(nick)
				local usr = tUser:IsOnline(olu)
				if nick then
					if usr then
						local send = "$To: "..nick.." From: "..me.." $<"..me.."> You are being redirected to "..where.." because: "..data.."|$ForceMove "..where.."|"
						tFuncs:Debug("[OPFORCEMOVE] <"..tostring(nick).."> is redirected by <"..tostring(me).."> to: "..tostring(where).." because: "..tostring(data))
						tSend:Raw(olu, send, sid)
					end
				end
			elseif tUser:IsOp(sid) then
				tFuncs:AddStats(sid, "NMDC", "$OpForceMove", -1, msg)
				tFuncs:Debug("[WARNING] Operator: "..tUser:GetDataLine(sid).."tried to redirect: "..msg)
				tSend:ChatToUser("You are not allowed to redirect.", "AAAA", sid)
			else
				tFuncs:AddStats(sid, "NMDC", "$OpForceMove", -1, msg)
				tFuncs:Debug("[WARNING] "..tUser:GetDataLine(sid).."tried to redirect: "..msg)
				tBan:KickUser(sid, "You're not allowed to redirect.")
				return client
			end
		else
			tFuncs:AddStats(sid, "NMDC", "$OpForceMove", -1, msg)
			tFuncs:Debug("[WARNING] "..tUser:GetDataLine(sid).."tried to redirect: "..msg)
			tBan:KickUser(sid, "You're not allowed to redirect.")
			return client
		end
		return client
	end,
	------------------------------- ################## --------------------------------
	["Unknown"] = function(client, sid, msg)
		tFuncs:AddStats(sid, "NMDC", "Unknown", -1, msg)
		if tUser:IsOp(sid) then
			tFuncs:Debug(tUser:GetDataLine(sid).."Unknown command received, ignored: "..msg)
		elseif msg:find("^%$MultiSearch ") or msg:find("^%$MultiConnectToMe ") then
			tFuncs:Debug(tUser:GetDataLine(sid).."Unsupported command received, ignored: "..msg)
		else
			tUser:Disconnect(sid, wrong, "Unknown command received: "..msg)
			return client
		end
		return client
	end,
	["Mainchat"] = function(client,sid, msg)        -- mainchat function
		local _,_,nick,retmsg = string.find(msg, "^<(%S+)> (.+)$")
		local res = false

		--// Basic validation
		if client.iState == 7 then
			if nick and nick == tClient[sid].sNick then		-- Checks if <nick> in msg matches usernick to prevent spoofing
				if retmsg then
					res = true
				else
					tDeflood:Main(sid, "MC", retmsg) 		--// Empty message will be ignored but will be checked for flooding
					tFuncs:AddStats(sid, "NMDC", "Mainchat", -1, msg)
					return client
				end
			elseif nick then
				tFuncs:AddStats(sid, "NMDC", "Mainchat", -1, msg)
				tBan:KickUser(sid, "Nick spoofing is not allowed.")
				return client
			else
				if string.find(msg, "^<> ") then
					tFuncs:AddStats(sid, "NMDC", "Mainchat", -1, msg)
					tUser:Disconnect(sid, "You're using a client too old to work with these rightclicks, update your client to make them work.", "Client too old to support current rightclicks, disconnected: "..msg)
					return client
				else
					tFuncs:AddStats(sid, "NMDC", "Mainchat", -1, msg)
					tUser:Disconnect(sid, wrong, "Malformed mainchat message: "..msg)
					return client
				end
				return client
			end
		else
			tFuncs:AddStats(sid, "NMDC", "Mainchat", -1, msg)
			tUser:Disconnect(sid, wrong, "Wrong login sequence, received mainchat line in state "..client.iState..": "..msg)
			return client
		end

		local me = false
		if retmsg:find("^[/%+][Mm][Ee] ") then 
			me = true
			retmsg = retmsg:gsub("^[/%+][Mm][Ee] ", "")
		end

		retmsg = NMDC_E(retmsg, true)

		--// Deflood check
		if res and retmsg then
			if string.find(retmsg, "^["..tHubSettings.Main.sCmdPrefixes.."].+$") then    -- hubcommand found
				tFuncs:AddStats(sid, "NMDC", "HC", 1, msg)
				res, retmsg = tDeflood:Main(sid, "HC", retmsg)
				if not res then tFuncs:AddStats(sid, "NMDC", "HC", -1, msg) tFuncs:AddStats(sid, "NMDC", "Mainchat", -1, msg) return client end
			else
				tFuncs:AddStats(sid, "NMDC", "MC", 1, msg)
				res, retmsg = tDeflood:Main(sid, "MC", retmsg)
				if not res then tFuncs:AddStats(sid, "NMDC", "MC", -1, msg) tFuncs:AddStats(sid, "NMDC", "Mainchat", -1, msg) return client end
				if res then res, retmsg = tCheck:Pattern(sid, "MC", retmsg) end
				if not res then tFuncs:AddStats(sid, "NMDC", "MC", -1, msg) tFuncs:AddStats(sid, "NMDC", "Mainchat", -1, msg) return client end
				if res then res, retmsg = tCheck:Filter(sid, "MC", retmsg) end
				if not res then tFuncs:AddStats(sid, "NMDC", "MC", -1, msg) tFuncs:AddStats(sid, "NMDC", "Mainchat", -1, msg) return client end
			end
		end

		--// Script call
--~ 		if res then res, retmsg = tApi:Main(sid, "MC", smsg) end

		--// Execution
		client = tSend:Mainchat(client, sid, retmsg, me)

		return client
	end,
	["Malformed"] = function(client, sid, msg)
		tFuncs:AddStats(sid, "NMDC", "Malformed", -1, msg)
		local nick = client.sNick or "<unknown>"
		tUser:Disconnect(sid, wrong, "Malformed message received in state "..client.iState..": "..msg)
		return client
	end,
}

--// FlipFlop --//-- Daywalker
