--[[

	FlexHub is a server-software for the Direct Connect P2P Network.

	Copyright (C) 2011 Daywalker and FlipFlop, dev@flexhub.org

    This file is part of FlexHub.

    FlexHub is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FlexHub is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with FlexHub.  If not, see <http://www.gnu.org/licenses/agpl.txt>.


]]--

tCheck = {
--~ 	["CheckList"] = { 
--~ 		["MC"] = {"1", "mainchat message"}, ["PM"] = {"2", "private message"}, ["SE"] = {"4", "search"}, ["SR"] = {"8", "searchresult"},
--~ 		["NI"] = {"16", "nickname", "sNick"}, ["DE"] = {"32", "description", "sDescription"}, ["EM"] = {"64", "email", "sEmail"},
--~ 		["TA"] = {"128", "tag", "sTag"}, ["SS"] = {"256", "sharesize", "sShareSize"}, ["RE"] = {"512", "report"},
--~ 	},

	--// The order of this list is linked to the bitwise operation of iActions: 2^(num-1)
--~ 	["ActionList"] = { "Ignore", "Notify", "Warn", "Mute", "Disconnect", "Kick", "Ban", "Redirect", "NotifyOpchat", "LockMain" },
--~ 	["ActionOrder"] = { "NotifyOpchat", "LockMain", "Ban", "Kick", "Redirect", "Disconnect", "Mute", "Warn", "Notify", "Ignore"},

	--// CURRENT SETTING

	--// Check for control characters
	["ControlChars"] = function(self, client, sid, msg, exceptions, ban)
		--// Best pattern to catch control chars other than \n \t \r : "[^\9\10\13\32-\126]"
		for cchar in string.gmatch(msg, "(%c)") do
			if not exceptions or not string.find(cchar, exceptions) then
				msg = string.gsub(msg, cchar, "*")
				if ban then 
					if not tUser:IsProtected(nil, nil, client.sIP) and not tUser:IsOp(sid) then
--~ 						tBans.autoban[client.sIP] = os.time() + tSecurity.General.DefaultAutoban
						tFuncs:Debug("[ControlChars] Sending illegal control character: "..string.byte(cchar).. " / "..tUser:GetDataLine(sid).."/ "..msg)
					end
					tUser:Disconnect(sid, "", "Possible exploit attempt, sending control characters", "FI", 0)
				elseif client and sid then
					if not tUser:IsProtected(nil, nil, client.sIP) and not tUser:IsOp(sid) then 
						tFuncs:Debug(tUser:GetDataLine(sid).."Sending illegal control character, replaced with *: "..string.byte(cchar).. " / "..tUser:GetDataLine(sid).."/ "..msg, "FI", 1)
					end
				end
				return false, msg
			end
		end
		return true, msg
	end,

	["NickChars"] = function(self, nick)
		if #nick < (tSecurity.NI.MinLength or 2) then
			return false, "Nickname is too short, minimum length: "..tSecurity.NI.MinLength
		elseif #nick > (tSecurity.NI.MaxLength or 128) then
			return false, "Nickname is too long, maximum length: "..tSecurity.NI.MaxLength
		end

		local invalidchars = {[""] = "non-breaking-space", ["\""] = "\"", ["%s"] = "space", ["|"] = "endpipe", ["\\"] = "\\", ["&"] = "&", ["<"] = "<", [">"] = ">", ["%$"] = "$", ["\n"] = "newline"}
		for char, chartxt in pairs(invalidchars) do
			if string.find(nick, char) then
				return false, "Invalid character(s) in nickname: "..chartxt
			end
		end
		return true, nick
	end,

	["PasswordChars"] = function(self, pass)
		--// CHECK FOR INVALID CHARS
		local invalidchars = {[""] = "non-breaking-space", ["\""] = "\"", ["%s"] = "space", ["|"] = "|", ["\\"] = "\\", ["%$"] = "$", ["\n"] = "line-break"}
		for char, chartxt in pairs(invalidchars) do
			if string.find(pass, char) then
				return false, "Invalid character(s) in password: "..chartxt
			end
		end

		if #pass < 3 then
			return false, "Password too short, use at least 3 characters." 
		elseif #pass > 64 then 
			return false, "Password too long, use max 64 characters." 
		end

		--// CHECK FOR PROPER CONVERSION BETWEEN UTF AND LOCALE
		local ok, convpass, errmsg = tFuncs:ToUTF8(pass, true)
		local sok, orgpass, errmsg = tFuncs:ToLocale(convpass, true)
		if not ok or not sok or pass ~= orgpass then 
			local charset = SetManager:Get("char_set")
			return false, "Your password can't be properly converted to/from this hub's local charset.\nPlease use characters from "..charset..": http://en.wikipedia.org/wiki/"..charset
		end
		return true
	end,

	["Bind"] = function(self, client, reg, nick)
		--// CHECK FOR IP BINDING
		local bindok = true
		if client and reg and reg.sBindIP and reg.sBindIP ~= "" then
			bindok = false
			--// IP is local
			if client.sIP == reg.sBindIP or client.sIP == "127.0.0.1" or string.find(client.sIP, "^192%.168%.%d+%.%d+") or string.find(client.sIP, "^10%.%d+%.%d+%.%d+$") then	--// BindIP override for local connections
				bindok = true
			--// IP matches pattern
			elseif reg.sBindIP then
				local bindip = reg.sBindIP..";"
				bindip = bindip:gsub(",",";")
				for range in string.gmatch(bindip, "(.-);") do
					range = range:gsub("%.", "%%%.")
					range = "^"..range:gsub("%*", ".*").."$"
					if client.sIP:find(range) then
						bindok = true
					end
				end
			end
		end
		if not bindok then
			tFuncs:Debug("[Check] [Nick] User with IP: "..client.sIP.." tried logging in as <"..nick.."> but IP ("..client.sIP..") doesn't match IP for this account: "..reg.sBindIP, "LI", 0)
--~ 			tUser:Disconnect(sid, "Your IP doesn't match the IP for this account.")
			return false, "IP doesn't match the IP for this account."
		end

		return bindok
	end,

	["State"] = function(self, client, sid, state, msg)
		local msg = msg or ""
		local state = state or 7
		if client.iState < state then
			tUser:Disconnect(sid, wrong, "[LOGIN] Wrong login sequence, message not allowed in state: "..client.iState.." / "..msg)
			return false
		else
			return true
		end
	end,

	["Sid"] = function(self, sid, msg)
		local vsid = msg:match("%S+ (%S+)")

		if vsid and vsid == sid then
			return true
		else
			tFuncs:Debug(tUser:GetDataLine(sid).."Hack attempt, SID spoofed: "..msg, "FI", 0)
			tUser:Disconnect(sid, "Malformed ADC message received.", "Hack attempt: SID spoofed: "..msg)
			return false
		end
	end,

	["Cid"] = function(self, client, sid, pid, cid)
		--// Check for CID (and PID)?
		if pid and cid then
			--// Check PID
			local pidhash = dc.hashpid(pid)
			if pidhash ~= cid then
					tUser:Disconnect(sid, "Public ID doesn't match the client ID, get a proper client.", "Hashed PID ("..pidhash..") doesn't match CID ("..cid.."), user disconnected.", "LI", 0)
					return false, client
			end
			--// Check CID
			for vsid, data in pairs(tClient) do
				if cid == data.sCID and vsid ~= sid then
					if data.sIP == client.sIP then
						tUser:Disconnect(vsid, "Another client connected from your IP with same Client ID (CID), this connection is closed.", "New client connected with same IP and Client ID (CID)", "LI", 2)
					else
						tUser:Disconnect(sid, "Client ID is already in use, use another client or change your client's PID.", "Client ID already in use by: "..(data.sNick or "<nick unknown>")..", user disconnected", "LI", 2)
						return false, client
					end
				end
			end
			client.sCID = cid
			client.sPID = pid
			return true, client
		else
			tUser:Disconnect(sid, "Malformed message received, get a proper client.", "Missing PID or CID during login, user disconnected,", "LI", 1)
			return false, client
		end
	end,

	["Filter"] = function(self, sid, stype, data)
		local client = tClient[sid]
		if not client then return false, "User not found." end
		local ip = client.sIP

		--// Kenny check
		if tBlocks.tKenny[ip] then
			data = tChange:Kenny(sid, data)
		end

		--// Muppet check
		if tBlocks.tMuppet[ip] then
			data = tChange:Muppet(sid,data)
		end

		return true, data
	end,

--		["ActionList"] =	{"Ignore", "Notify", "Warn", "Mute", "Disconnect", "Kick", "Ban", "Redirect", "NotifyOpchat", "LockMain", "Replace", "ReplaceLine"},
--		["ActionOrder"] =	{"NotifyOpchat", "LockMain", "Ban", "Kick", "Redirect", "Disconnect", "Mute", "Warn", "Notify", "Ignore", "Replace", "ReplaceLine"},
--	new	["ActionOrder"] =	{"NotifyOpchat", "LockMain", "Ban", "Kick", "Redirect", "Disconnect", "Mute", "Warn", "Notify", "Ignore", "ReplaceLine", "Replace"},
--~ 	{"64","32","128","16","512","8","4","256","2","1"}

	["PatternPriority"] = function(self, stype)
		--// Prioritize patternlist
		local patorder = {}
		local priority = tPatterns["Priority"] or {"64","32","128","16","512","8","4","256","2","1"}	-- not sure there is a 'tPatterns["Priority"]' anywhwere ?? fredo
		local patlist = {}
		if not tPatterns[stype] then
			return {}
		end

		for pat, _ in pairs(tPatterns[stype]) do
			patlist[pat] = true
		end
		for num, flag in ipairs(priority) do
			for pat, patdata in pairs(tPatterns[stype]) do
				if patlist[pat] and tFuncs:CheckFlag(patdata.iActions, tonumber(flag)) then
					table.insert(patorder, pat)
					patlist[pat] = nil
				end
			end
		end
		for pat, patdata in pairs(tPatterns[stype]) do
			if patdata.iActions == 0 then
				table.insert(patorder, pat)
			end
		end

		if not tPatterns.Sorted then tPatterns.Sorted = {} end
		tPatterns.Sorted[stype] = patorder
		return patorder
	end,

	["Pattern"] = function(self, sid, stype, value, level)
--		tFuncs:Debug("Performing patterncheck, SID: "..(sid or "<no sid>")..", stype: "..stype..", value: "..(value or "<no value>")..", level: "..(level or "<no level>"), "X")

		local checklist = SubSet.CheckList
		local actionlist = SubSet.ActionList
		local client = tClient[sid]
		if client then level = level or client.iProfile or client.iProfileTemp or -1 end
		local level = level or -1

		if sid and not client then
			return false, "User with SID: "..sid.." not found."
		elseif not checklist[stype] then
			return false, "[tCheck:Pattern] [ERROR] No such stype in checklist: "..stype
		end

		local original = false
		if client and checklist[stype][3] and not value then
			original = (tClient[sid][checklist[stype][3]] or "")
		elseif value then
			original = value
		else
			tFuncs:Debug("[tCheck:Pattern] [ERROR] Sid: "..sid.." Type: "..stype.." Value: "..tostring(value))
			return false, "[tCheck:Pattern] [ERROR] No string to match pattern."
		end

		local endresult = true
	--	local candidate = tostring(original)
		local r = {}
		r.patterns = {}
		r.msg = tostring(original)

		--// Patterncheck
		if not tPatterns["Sorted"] then tPatterns["Sorted"] = {} end
		local patorder = tPatterns["Sorted"][stype] or tCheck:PatternPriority(stype)
		local actionorder = SubSet.ActionOrder

		for num, pat in pairs(patorder) do
			local patdata = tPatterns[stype][pat]
			local match = false
			if level <= patdata.iProfile and not patdata.bDisabled and string.find(r.msg, pat) then
			--	tFuncs:Debug(".1 '" .. pat .. "' is a match.", "X")	-- // DEBUG
				match = true
				table.insert(r.patterns, pat)

				--// Perform action(s) - only if client is still available
				if client and match then
					local actions, actionsconcat = {}, ""

					--// Build actionlist
					for num, action in ipairs(actionlist) do
						if patdata.iActions > 0 and tFuncs:CheckFlag(patdata.iActions, (2^(num-1))) then
							actions[action] = true
							--tFuncs:Debug(".2 adding '" .. action .. "' to actions.", "X")	-- // DEBUG
							actionsconcat = actionsconcat.." "..action
						end
					end

					r.outputreason =  "Your "..checklist[stype][2].." matches a forbidden pattern: "..r.msg

					if patdata.iReplace and patdata.iReplace == 1 then
						-- // think this is the correct order to process it in :)
						r.rmsg = string.gsub((r.endreply or r.rmsg or r.msg), pat, (tFuncs:Escape(tFuncs:FillVars(patdata.sReplace, sid)) or ""))
						actions["Replace"] = true
						actionsconcat = actionsconcat.." ".."Replace"
					elseif patdata.iReplace and patdata.iReplace == 2 then
						actions["ReplaceLine"] = true
						actionsconcat = actionsconcat.." ".."ReplaceLine"
						r.rmsg = tFuncs:FillVars(patdata.sReplace, sid)
						r.outputreason = r.rmsg
					end

					for num, actionname in ipairs(actionorder) do
						r.actionok = false
						r.actionresult = "No result."

						if actions[actionname] then
							if actionname == "NotifyOpchat" then
								r.opmsg = "User: <"..(client.sNick or "nick unknown").."> with IP: "..client.sIP.." sent a "..checklist[stype][2].." matching pattern(s): \""..table.concat(r.patterns, " / ").."\" / Message: "..r.msg.." / Actions:"..actionsconcat
							end

							-- // action
							r = tCheck.Actions[actionname](sid, r, pat)

							--tFuncs:Debug("endreply is \t"..(r.endreply or "nil"), "X")
							--tFuncs:Debug("rmsg is \t"..(r.rmsg or "nil"), "X")
							--tFuncs:Debug("outputreason is \t"..(r.outputreason or "nil"), "X")

							--// Output actionresult to feedroom
							tFuncs:Debug("[Pattern] ["..stype.."] "..tUser:GetShortData(sid).." sent a "..checklist[stype][2].." matching pattern: "..pat.." / Message: "..r.msg.." / Output: "..tostring(r.endreply or r.rmsg or r.msg).." / Action: "..actionname..(r.actionok and " OK: " or " FAILED :")..tostring(r.actionresult), "FI", (r.debuglevel or 0))

							--// Exception for defloodactions on NIckname, they need to return false to prevent wrong logins.
							if not r.result and (stype ~= "NI" or (actionname ~= "Warn" and actionname ~= "Mute")) then
								--tFuncs:Debug("we ended up in here, actionname is "..actionname, "X")	-- DEBUG
								endresult = false 
								r.endreply = r.outputreason
							end
						end
					end

					if r.endreply then r.msg = r.endreply end
				end
			end
			if not endresult then
				--tFuncs:Debug(".5 breaking", "X")	-- DEBUG
				break 
			end
		end
		if type(original) == "number" then r.msg = (tonumber(r.msg) or 0) end
		return endresult, r.msg
	end,

	["Actions"] = {
		["Disabled"] = function(sid, r)	-- does nothing
		--	r.debuglevel = 1 
			return r
		end,
		["Ignore"] = function(sid, r)
			r.debuglevel = 2 
			r.result = false
			r.actionok = true
			r.actionresult = "Message ignored"
			return r
		end,
		["Notify"] = function(sid, r)
			r.debuglevel = 1 
			r.result = false
			r.endreply = r.outputreason
			r.actionok = true
			r.actionresult = "User notified."
			tSend:ChatToUser(r.outputreason, "AAAA", sid)
			return r
		end,
		["Warn"] = function(sid, r)
			r.actionok, r.actionresult = tBan:WarnUser(sid, r.outputreason, nil, true)
			r.result = false
			r.endreply = r.outputreason
			return r
		end,
		["Mute"] = function(sid, r)
			--local vnick = (tClient[sid].sNick or "<unknown>")
			--local vip = tClient[sid].sIP
			r.actionok, r.actionresult = tBlock:AddBlock("AAAA", "mute", (tClient[sid].sNick or "<unknown>"), tClient[sid].sIP)
			r.endreply = r.outputreason
			r.result = false
			return r
		end,
		["Disconnect"] = function(sid, r)
			r.actionok, r.actionresult = tBan:DropUser(sid, r.outputreason, nil, true)
			r.result = false
			r.endreply = r.outputreason
			return r
		end,
		["Kick"] = function(sid, r)
			r.actionok, r.actionresult = tBan:KickUser(sid, r.outputreason, nil, true)
			r.result = false
			r.endreply = r.outputreason
			return r
		end,
		["Ban"] = function(sid, r)
			r.actionok, r.actionresult = tBan:BanIp(tClient[sid].sIP, r.outputreason, 0, nil, true)
			r.result = false
			r.endreply = r.outputreason
			return r
		end,
		["Redirect"] = function(sid, r)
			r.actionok, r.actionresult = tBan:RediUser(sid, r.outputreason, nil, "flood", true)
			r.result = false
			r.endreply = r.outputreason
			return r
		end,
		["NotifyOpchat"] = function(sid, r) -- done
			r.result = true
			r.actionok = true
			r.actionresult = "Opchat notified."
			tSend:ChatRoom(r.opmsg, "AAAA", "BBBB")	-- if/when combining opchat and feed add variable here
			return r
		end,
		["LockMain"] = function(sid, r)
			SetManager:Set("lock_main", true)
			tSend:ChatToAll("Mainchat is now locked.")
			tSend:ChatRoom("User caused a mainchat lock, check feed for details", "AAAA", "BBBB")	-- if/when combining opchat and feed add variable here
			r.result = false
			r.actionok = true
			r.endreply = r.outputreason
			r.actionresult = "User caused a mainchat lock."
			return r
		end,
		["Replace"] = function(sid, r, pat)
			r.debuglevel = 1
			r.result = true
			r.actionok = true
			r.actionresult = "Match replaced."
			r.endreply = string.gsub((r.rmsg or r.msg), pat, (r.sReplace and tFuncs:Escape(r.sReplace) or ""))
			return r
		end,
		["ReplaceLine"] = function(sid, r)
			r.debuglevel = 1
			r.result = true
			r.actionok = true
			r.actionresult = "Message replaced."
			r.endreply = (r.rmsg or r.msg or "")
			return  r
		end,
	},

	["Rules"] = function(self, sid)
		local res, checkname, result, resultmsg = tCheck:RulesExecute(sid)
		if res then
			--// Check if redirector is in other mode than "n"ormal
			if tHubSettings.tRedirects.sMode ~= "normal" then
				if not tUser:CanOverride(sid, "redirect") then
					res, retmsg = tUser:Redirect(sid, "", "main", nil, "FI", (tHubSettings.Feed.RuleLevel or 1))
					return false, "[Redirect] "..retmsg
				else
					return true
				end
			else
				return true
			end
		else
			if result then
--~ 				tFuncs:Debug("[Rules] "..checkname..resultmsg, "FI", (tHubSettings.Feed.RuleLevel or 1))
				return false, resultmsg
			else
				tFuncs:Debug("[Rules] "..checkname.." / FAILED: "..resultmsg, "FI", (tHubSettings.Feed.RuleLevel or 1))
				return false, resultmsg	--// Should this return true if the redirect failed for some reason??
			end
		end
	end,

	["RulesExecute"] = function(self, sid)
		local debugtype, debuglevel = "FI", (tHubSettings.Feed.RuleLevel or 0)
		local res, retmsg = false, "Check completed succesfully"
		local client = tClient[sid]
		if not client then return false, "No client to check" end

		--~ 		tUser:Redirect(sid, msg, address, rreason, debugtype, debuglevel, hide)

		--// Proxy - disabled by default, enable override per profile
		if client.bProxy and (not tHubSettings.tRules.bAllowProxy) and (not tUser:CanOverride(sid, "blockproxy")) then
			res, retmsg = tUser:Redirect(sid, "No proxy connection allowed, please connect normally.", "flood", nil, debugtype, debuglevel)
			return false, "[allowproxy:false] ", res, retmsg
		end

		--// Minshare - rule for each profile
		if not tUser:CanOverride(sid, "minshare") then
			local minshare = (tProfiles[client.sProfile].tRules.iMinShare or tHubSettings.tRules.iMinShare or 0)
--~ 			minshare = tFuncs:ConvertToBytes(minshare)
			if (client.iShareSize or 0) < minshare then
				res, retmsg = tUser:Redirect(sid, "You're not sharing enough, you have: "..tFuncs:ConvertShare(client.iShareSize)..", minimum for your profile is: "..tFuncs:ConvertShare(minshare), "minshare", nil, debugtype, debuglevel)
				return false, "[minshare:"..tFuncs:ConvertShare(minshare).." user:"..tFuncs:ConvertShare(client.iShareSize).."] ", res, retmsg
			end
		end

		--// Maxshare - rule for each profile
		if not tUser:CanOverride(sid, "maxshare") then
			local maxshare = (tProfiles[client.sProfile].tRules.iMaxShare or tHubSettings.tRules.iMaxShare or 10e+16)
--~ 			maxshare = tFuncs:ConvertToBytes(maxshare)
			if (client.iShareSize or 0) > maxshare then
				res, retmsg = tUser:Redirect(sid, "You're sharing too much, you have: "..tFuncs:ConvertShare(client.iShareSize)..", maximum for your profile is: "..tFuncs:ConvertShare(maxshare), "maxshare", nil, debugtype, debuglevel)
				return false, "[maxshare:"..tFuncs:ConvertShare(maxshare).." user:"..tFuncs:ConvertShare(client.iShareSize).."] ", res, retmsg
			end
		end

		--// Slots - general hubrule, can only override based on profile
		if (client.iSlots or 0) < tHubSettings.tRules.iMinSlots and not tUser:CanOverride(sid, "minslots") then
			local res, retmsg = tUser:Redirect(sid, "Too few slots open, you have: "..(client.iSlots or "<none>")..", the hub minimum is: "..tHubSettings.tRules.iMinSlots, "minslots", nil, debugtype, debuglevel)
			return false, "[minslots:"..tHubSettings.tRules.iMinSlots.." user: "..(client.iSlots or "0").."] ", res, retmsg
		end
		if (client.iSlots or 0) > tHubSettings.tRules.iMaxSlots and not tUser:CanOverride(sid, "maxslots") then
			res, retmsg = tUser:Redirect(sid, "Too many slots open, you have: "..(client.iSlots or "<none>")..", the hub maximum is: "..tHubSettings.tRules.iMaxSlots, "maxslots", nil, debugtype, debuglevel)
			return false, "[maxslots: "..tHubSettings.tRules.iMaxSlots.." user: "..(client.iSlots).."] ", res, retmsg
		end

		--// Hubs - general hubrules, can only override based on profile
 		if (client.iHubs or 0) > tHubSettings.tRules.iMaxHubs and not tUser:CanOverride(sid, "maxhubs") then
			res, retmsg = tUser:Redirect(sid, "Connected to too many hubs, you have: "..(client.iHubs or "<none>")..", the hub maximum is: "..tHubSettings.tRules.iMaxHubs, "maxhubs", nil, debugtype, debuglevel)
			return false, "[maxhubs:"..tHubSettings.tRules.iMaxHubs.." user:"..(client.iHubs).."] ", res, retmsg
		end

 		if (client.iNormalHubs or 0) > tHubSettings.tRules.iMaxPublicHubs and not tUser:CanOverride(sid, "maxpublichubs") then
			res, retmsg = tUser:Redirect(sid, "Connected to too many public hubs, you have: "..(client.iNormalHubs or "<none>")..", the hub maximum is: "..tHubSettings.tRules.iMaxPublicHubs, "maxhubs", nil, debugtype, debuglevel)
			return false, "[maxpublichubs:"..tHubSettings.tRules.iMaxPublicHubs.." user:"..(client.iNormalHubs).."] ", res, retmsg
		end

 		if (client.iRegHubs or 0) > tHubSettings.tRules.iMaxRegHubs and not tUser:CanOverride(sid, "maxreghubs") then
			res, retmsg = tUser:Redirect(sid, "Connected to too many hubs as registered user, you have: "..(client.iRegHubs or "<none>")..", the hub maximum is: "..tHubSettings.tRules.iMaxRegHubs, "maxhubs", nil, debugtype, debuglevel)
			return false, "[maxreghubs:"..tHubSettings.tRules.iMaxRegHubs.." user:"..(client.iRegHubs).."] ", res, retmsg
		end

 		if (client.iOpHubs or 0) > tHubSettings.tRules.iMaxOpHubs and not tUser:CanOverride(sid, "maxophubs") then
			res, retmsg = tUser:Redirect(sid, "Connected to too many hubs as operator, you have: "..(client.iOpHubs or "<none>")..", the hub maximum is: "..tHubSettings.tRules.iMaxOpHubs, "maxhubs", nil, debugtype, debuglevel)
			return false, "[maxophubs:"..tHubSettings.tRules.iMaxOpHubs.." user:"..(client.iOpHubs).."] ", res, retmsg
		end

		--// ADC connection allowed?
		if client.sProtocol == "ADC" and (not tHubSettings.tRules.bAllowADC) and (not tUser:CanOverride(sid, "blockadc")) then
			res, retmsg = tUser:Redirect(sid, "ADC Protocol is disabled in this hub.", "adc_disabled", nil, debugtype, debuglevel)
			return false, "[allowadc:false] ", res, retmsg
		end

		--// NMDC connection allowed?
		if client.sProtocol == "NMDC" and (not tHubSettings.tRules.bAllowNMDC) and (not tUser:CanOverride(sid, "blocknmdc")) then
			res, retmsg = tUser:Redirect(sid, "NMDC Protocol is disabled in this hub.", "nmdc_disabled", nil, debugtype, debuglevel)
			return false, "[allownmdc:false] ", res, retmsg
		end

		--// Active / Passive - disabled by default, enable override per profile
		if not client.bActive and (not tHubSettings.tRules.bAllowPassive) and (not tUser:CanOverride(sid, "blockpassive")) then
			res, retmsg = tUser:Redirect(sid, "No Passive mode allowed, please change to Active. Take a look @ http://www.portforward.com", "passive", nil, debugtype, debuglevel)
			return false, "[allowpassive:false] ", res, retmsg
		end

		--// DHT - enabled by default, enable override per profile
		if not tHubSettings.tRules.bAllowDHT and (tUser:Supports(sid, "ADDHT0") or tUser:Supports(sid, "DHT0")) and not tUser:CanOverride(sid, "blockdht") then
			res, retmsg = tUser:Redirect(sid, "Please turn off DHT in 'Settings/Advanced/Publish shared files on DHT...' before reconnecting", "flood", nil, debugtype, debuglevel)
			return false, "[allowdht:false] ", res, retmsg
		end

		return true
	end,

	["Hammer"] = function(self, sid, ip)		--// Careful, this function is called from the Wait-thread, don't use Debug from here.
		local ctime = socket.gettime()

		--// Check if reconnecting too fast
		if not tUser:IsProtected(nil, nil, ip) then
			--// Check 
			ThreadLockMutex(mutexiplog)
			local iplog = tLogs.IP[ip]["$"]
			local timeout = (iplog.DoubleLogins >= tSecurity.General.DoubleLoginMax and tSecurity.General.DoubleLoginTimeout) or (tSecurity.General.ReconnectTimeout or 20)
			local waittime = timeout - ( ctime - iplog.LastLogin)
			ThreadUnlockMutex(mutexiplog)
			if waittime > 0 then
				return false, "Reconnecting too fast, you have to wait "..tostring(timeout).." seconds before reconnecting."
			end
		end

		return true
	end,

	["IP"] = function(self, sid, ip)
		local ctime = socket.gettime()
		local client = tClient[sid]

--~ 		--// Check if reconnecting too fast
--~ 		if not tUser:IsProtected(sid) then
--~ 			local timeout = (tSecurity.General.ReconnectTimeout or 20)
--~ 			ThreadLockMutex(mutexiplog)
--~ 			local waittime = timeout - ( ctime - tLogs.IP[ip]["$"].LastLogin)
--~ 			ThreadUnlockMutex(mutexiplog)
--~ 			if waittime > 0 then
--~ 				return false, "Reconnecting too fast, you have to wait "..tostring(timeout).." seconds before reconnecting."
--~ 			end
--~ 		end

		--// Check if banned
		local ban = tBans.ipban[ip]
		if ban then				--// ipban
			if ban.iExpires <= 0 or ban.iExpires > ctime then
				local banuntil = "permanently"
				if ban.iExpires > 0 then
					banuntil = "until: "..ToDate(ban.iExpires)
				end
				if client then
					tFuncs:Debug("[Check] [IP] User with banned IP: "..ip.." was redirected because: "..ban.sReason, "LI", 3)
					tUser:Redirect(sid, "You are kicked/banned from "..tHubSettings.Main.sAddress.." "..banuntil.." because: "..ban.sReason, "main") 
				end
				return false, "You are kicked/banned from "..tHubSettings.Main.sAddress.." "..banuntil.." because: "..ban.sReason
			else
				if ban.iExpires and ban.iExpires > 0 and ctime > ban.iExpires and (ban.sType ~= "kick" or (ban.iExpires+(3600*24*30)) < ctime) then		--// kicks will stay in list for 30 days even if expired, to allow ban after number of kicks
					--// Remove expired ban
					tBan:Unban(ip, "IPban expired.")
				end
			end
		else					--// rangeban
			local ipdec = tFuncs:IP2DEC(ip)
			for name,data in pairs(tBans.rangeban) do
				if  ipdec >= data.sStart and ipdec <= data.sEnd then
					if data.iExpires <= 0 or data.iExpires > ctime then
						local banuntil = "permanently"
						if data.iExpires > 0 then
							banuntil = "until: "..ToDate(data.iExpires)
						end
						if client then
							tFuncs:Debug("[Check] [IP] User with rangebanned IP: "..ip.." was disconnected because: "..data.sReason, "LI", 3)
							tUser:Redirect(sid, "You are banned from "..tHubSettings.Main.sAddress.." "..banuntil.." because: "..data.sReason, "main") 
						end
						return false, "You are kicked/banned from "..tHubSettings.Main.sAddress.." "..banuntil.." because: "..data.sReason
					else
						--// Remove expired ban
						tBan:Unban(name, "Rangeban expired.")
					end
				end
			end
		end

		--// Check if max clones is not exceeded
		if client then
			local maxclones = tSecurity.General.ClonesMax or 2
			local clones = tIp:GetSids(ip)
			if clones and #clones > maxclones then
				local override = false
				for _,vsid in ipairs(clones) do	--// checks if there are users online with this IP who may override maxclones
					if tUser:CanOverride(vsid, "maxclones") then
						tFuncs:Debug("[Check] [IP] Clone ("..#clones..") allowed from IP: "..ip.." based on profile of <"..tUser:GetNick(vsid)..">", "LI", 1)
						override = true
					end
				end
				if not override then
					local nicks = ""
					for _,sid in ipairs(clones) do
						nicks = nicks..(tClient[sid].sNick or "<nick unknown / login>").." "
					end
					local msg = "You've reached the maximum number of connections from this IP."
					if tSecurity.General.ClonesBan then
						local res, err = tBan:BanIp(ip, msg, tSecurity.General.ClonesBanTime)
						if res then
							tFuncs:Debug("[Check] [IP] Maximum clones reached from IP: "..ip.." and nickname(s): "..nicks..". IP banned.", "FI", 0)
						else
							tFuncs:Debug("[Check] [IP] Maximum clones reached from IP: "..ip.." and nickname(s): "..nicks..". IP ban failed: "..err, "FI", 0)
						end
					else
						tFuncs:Debug("[Check] [IP] Maximum clones reached from IP: "..ip.." and nickname(s): "..nicks..". New connection dropped.", "FI", 1)
						tUser:Redirect(sid, msg, "flood")
					end
					return false, msg
				end
			elseif clones then
				tFuncs:Debug("[Check] [IP] Clone ("..#clones..") allowed from IP: "..ip.."", "LI", 3)
			end
		end

		return true
	end,

	["Ghost"] = function(self, client, sid, ip, usid)
--~ 		tBuffer[usid].Send.iFailed = os.time()-tHubSettings.tTcp.SendTimeout+3
		if client.sProtocol == "NMDC" then
			tSend:Raw(sid, "|")
		elseif client.sProtocol == "ADC" then
			tSend:Raw(sid, "\n")
		end
	end,

	["Nick"] = function(self, client, sid, nick, level)
		if client then client.sNickTemp = nick end 	--// To be able to get nick displayed in feedinfo before full login
		local ctime = os.time()
		local result, msg = true, false
		local reason, banuntil, tuntil = false,false, false
		local ip = tUser:GetIp(sid)

		if not nick then return false, "No nickname to check" end

		--// CHECK NICKLENGTH AND CHARS
		local res, err = tCheck:NickChars(nick)
		if not res then
			if client then
				tUser:Disconnect(sid, err)
				return false, client, nil, err
			else
				return false, nil, nil, err
			end
		end

		--// CHECK FOR NICKBAN
		local snick = nick:lower()
		local ban = tBans.nickban[snick]
		if ban then
			if ban.iExpires <= 0 or ban.iExpires > ctime then
				local banuntil = "permanently"
				if ban.iExpires > 0 then
					banuntil = "until: "..ToDate(ban.iExpires)
				end

				if client then
					tFuncs:Debug("[Check] [Nick] User with banned nick: "..nick.." was disconnected because: "..ban.sReason, "FI", 3)
					tUser:Redirect(sid, "You are kicked/banned from "..tHubSettings.Main.sAddress.." "..banuntil.." because: "..ban.sReason, "main")
					return false, client, nil, msg
				else
					return false, nil, nil, "This nickname is banned, unban first"
				end
			else
				--// Remove expired ban
				local result, retmsg = tBan:Unban(nick, "Nickban expired.")
				if not result then tFuncs:Debug("[Check] [Nick] "..retmsg) end
			end
		end

		--// CHECK FOR REG
		local reg = tNick:GetReg(nick)  --// get the RegInfo for this nick
		if client then 
			if not reg and tHubSettings.tRules.bRegOnly then
--~ 				if tHubSettings.tRedirects.sRegOnly ~= "" then
					tFuncs:Debug("[Check] [Nick] "..tUser:GetDataLine(sid).."User isn't registered, redirected to rd_regonly="..tostring(tHubSettings.tRedirects.sRegOnly), "X")
					tUser:Redirect(sid, "This hub is for registered users only", "regonly")
--~ 				else
--~ 					tFuncs:Debug("[Check] [Nick] "..tUser:GetDataLine(sid).."User isn't registered, but redirect addy is empty, disconnecting", "X")
--~ 				end
				return false, client, nil, "This hub is for registered users only."
			end
			if reg then
				--// Check IP binding
				local bindok, reason = tCheck:Bind(client, reg, nick)
				if not bindok then return false, client, nil, reason end
			end
		end

		--// CHECK FOR NICK IN USE
		local usid = tNick:GetSid(nick)

		--// CHECK FOR BOT/ROOM
		if usid and tClient[usid].bBot then return false, client, nil,  "Nickname <"..nick.."> is already in use", true end

		--// CHECK FOR NONREGGED INUSE
		if client and not reg and usid then 
			return false, client, nil,  "Nickname <"..nick.."> is already in use", true
		end

--[[
		if client and usid and ip and tClient[usid].sIP == ip and tHubSettings.tTraffic.bDropExisting then
			if tNick:GetReg(nick) then
				--// Only allow regged to login when a 'ghost' or same nick from same ip is connected to join
--~ 				client.iWaitTime = ctime + 2	--// Add a waittime to allow the other client to gracefully disconnect
--~ 				tClient[sid].iWaitTime = client.iWaitTime
				tClient[usid].sCID = false		--// Remove the other clients CID to enable to enter with 'same CID' before the other client is completely removed from hub
				tUser:Disconnect(usid, "You're connecting with the same nickname <"..nick.."> twice on "..tHubSettings.Main.sAddress..", this connection is closed.")
			else
				tCheck:Ghost(client, sid, ip, usid)
				return false, client, nil,  "Nickname <"..nick.."> is already in use", true
			end
		elseif client and usid then
			tCheck:Ghost(client, sid, ip, usid)
			return false, client, nil,   "Nickname <"..nick.."> is already in use", true
		end
]]--

		--// CHECK FOR FULL HUB
		if client and tStats.Hub.Users.Total >= tHubSettings.tRules.iMaxUsers then	--// Soft limit
			-- borrow setting from presumed profile (even if pass isn't checked yet to allow login to proceed)
			if reg and tProfiles[reg.sProfile] and (tProfiles[reg.sProfile].tOverrides and (tProfiles[reg.sProfile].tOverrides.all or tProfiles[reg.sProfile].tOverrides.maxusers))then
				if tStats.Hub.Users.Total >= (tHubSettings.tRules.iMaxUsersHard or 20000) then --// Hard limit reached, need to drop one unregistered user for the operator to join
					local userdropped = false
					--// Try to drop a user
					for vsid, data in pairs(tClient) do
						if data.sProfile == "User" and data.iState == 7 then
							tUser:Redirect(vsid, "", "full")
							userdropped = vsid
							break
						end
					end
					--// Try to drop a regged user
					if not userdropped then
						for vsid, data in pairs(tClient) do
							if data.sProfile == "Reg" and data.iState == 7 then
								tUser:Redirect(vsid, "", "full")
								userdropped = vsid
								break
							end
						end
					end
					if not userdropped then
						tUser:Redirect(sid, "Hub is currently full, try again later.", "full")
						return false, client, nil, "Hub is currently full, try again later."
					else
						tFuncs:Debug("Hub is currently full (host limit), user <"..tClient[userdropped].sNick.."> was dropped to allow <"..nick.."> to connect. Lower !set maxusers to prevent this.", "FI", 0)
					end
				else
					--// Allow user to continue with login
				end
			else
				tUser:Redirect(sid, "Hub is currently full, try again later.", "full")
				return false, client, nil, "Hub is currently full, try again later."
			end
		end

		--// CHECK FOR INVALID CHARS
		local res, err = tCheck:NickChars(nick)
		if not res then
			if client then
				tUser:Disconnect(sid, err)
				return false, client, nil, err
			else
				return false, nil, nil, err
			end
		end

 		--// CHECK FOR FORBIDDEN NICKPATTERN
		if reg then level = (reg.iLevel or -1) end
		result, msg = tCheck:Pattern(sid, "NI", nick, level)
		if not result then 	
			--// If uses isn't set to be disconnected by the patterncheck: redirect!
			if not tBuffer[sid].bDisconnect then
				local res, err = tUser:Redirect(sid, msg, "main")
			end
			return false, client, nil, "This nickname is not allowed: "..nick 
		end 	--// only return the values, Check:Pattern handles kicks/bans/disconnects itself.

--~ 		--// CHECK FOR FORBIDDEN NICK
--~ 		if tFuncs:IsForbidden(nick, "Nick") then		--//Todo: add tFuncs:IsForbidden 
--~ 			return false, client, nil, "This is a protected nickname."
--~ 		end

--~ 		--// CHECK FOR PROTECTED NICK
--~ 		if tFuncs:IsProtected(nick, "Nick") then
--~ 			return false, client, nil, "This is a protected nickname."
--~ 		end

		--// NICKCHECK COMPLETE - UPDATE CLIENT
		if reg and client then
			client.sNickTemp = nick
			client.iState = 4     --// user was regged, update to state 4 and send $GetPass
			client.iProfileTemp = reg.iLevel or -1
		elseif client then
			client.sNickTemp = nil
			client.sNick = nick
			--tNicks[nick] = sid
			tLNicks[snick] = {["sid"] = sid, ["sNick"] = nick}
			client.iState = 5
			client.sProfile = "User"
			client.iProfile = -1
		end

		return true, client, reg, nick
	end,

	["Pass"] = function(self, client, sid, pass)
		local nick = client.sNick or client.sNickTemp
		local reg, rnick, lnick = tNick:GetReg(nick)  --// get the RegInfo for this nick
		local passmatch = false
		local droppedexisting = false

		local addnewpass = false
		local letuserin = false

		if not reg then
			return false, client, "User not registered."
		elseif reg.sPassword ~= "" then
			passmatch = reg.sPassword
			if client.sProtocol == "ADC" or client.tSupports.SaltPass then
				if client.sPassSalt then 
					passmatch = dc.hashpas(passmatch,client.sPassSalt)
				else
					return false, client, "No password salt was set."
				end
			end
		elseif reg.sPassword == "" and tSecurity.General.AllowBlankPass then
			tFuncs:Debug(tUser:GetDataLine(sid).."connecting, using reg with empty pwd", nil, 1)

			if (client.sProtocol == "ADC" or client.tSupports.SaltPass) then
				if not client.sPassSalt then 
					return false, client, "No password salt was set."
				end
				-- we should let the user in with limited level
				letuserin = true 
			else
				local test, err = tCheck:PasswordChars(pass)
				if not test then
					return false, client, err
				end

				passmatch = pass
				addnewpass = true
			end
		end

		-- Correct password
		if reg and (passmatch == pass or letuserin) then      --// if pass is Ok then state = 5 else send $BadPass| and disconnect user
			--// DISCONNECT EXISTING USER WITH THIS NICK
			local usid = tNick:GetSid(nick)
			if usid then 
				tFuncs:Debug(tUser:GetDataLine(sid).."Double connected, dropping existing user with nick <"..nick.."> and IP: "..tClient[usid].sIP, "X")
				droppedexisting = true
				tClient[usid].sCID = false		--// Remove the other clients CID to enable to enter with 'same CID' before the other client is completely removed from hub
				tUser:Disconnect(usid, "A connection from IP: "..client.sIP.." is made with your nick and password, this connection is closed.")
			end

			client.sNick = nick
			client.sNickTemp = nil
			--tNicks[nick] = sid
			tLNicks[lnick] = {["sid"] = sid, ["sNick"] = nick}

			--// Reset bad password counter
			if reg.tFailed then tAccounts[lnick].tFailed.iTimes = 0 end  

			--// Store details in reg
			tAccounts[lnick].iLastLogin = client.iLoginTime
			tAccounts[lnick].sLastIP = client.sIP

			--// Assign special values to client based on reg and profile
			if reg.bShareHidden then client.bShareHidden = true end
			if reg.bHidden then client.bHidden = true end

			if letuserin then
				client.sProfile = "Reg"	-- // temp hardcoded profile for all that connects without password and cant set new on login
				client.bNotifySetOwnPass = true
			else
				client.sProfile = reg.sProfile
			end

			local prof, err = ProfileManager:GetData(sid, client.sProfile)
			if not prof then tFuncs:Debug("Profile not found", "FI", -2) end
			if not prof then return false, client, err end

			client.iProfile = (letuserin and 1) or prof.iLevel	-- CHECK  
			client.bOperator = prof.bOperator or (prof.iLevel == 10 and true)
			client.bHasKey = prof.bHasKey
			client.bHasFakeKey = prof.bHasFakeKey

			--// Set back last login to allow fast reconnects
			if client.bOperator then
				ThreadLockMutex(mutexiplog)
				local iplog = tLogs.IP[client.sIP]["$"]
				iplog.CurrentLogin = iplog.CurrentLogin - tSecurity.General.ReconnectTimeout
				if droppedexisting then tFuncs:Debug(tUser:GetDataLine(sid).."Double reconnected, but profile is operator, resetting counter for IP: "..client.sIP, "X") end
				iplog.DoubleLogins = 0
				ThreadUnlockMutex(mutexiplog)
			elseif droppedexisting then
				--// Make sure non operators can't keep reconnecting (for example with 2 clients with same favs)
				if not tUser:IsProtected(nil, nil, client.sIP) then
					ThreadLockMutex(mutexiplog)
					local iplog = tLogs.IP[client.sIP]["$"]
					if (iplog.CurrentLogin - iplog.LastLogin) < tSecurity.General.DoubleLoginTime then		--// Two IP's can keep logging in alternating, for one IP the reconnect frequency will then be max 2 times the usual clientwaittime = 2 x 3 minutes
						iplog.DoubleLogins = iplog.DoubleLogins + 1
						tFuncs:Debug(tUser:GetDataLine(sid).."Double reconnected too quick, counter for IP: "..client.sIP.." set to: "..iplog.DoubleLogins, "X")
					else
						iplog.DoubleLogins = 0
						tFuncs:Debug(tUser:GetDataLine(sid).."Double reconnected, but enough time between logins, resetting counter for IP: "..client.sIP, "X")
					end
					if iplog.DoubleLogins == tSecurity.General.DoubleLoginMax then 
						tFuncs:Debug(tUser:GetDataLine(sid).."Double reconnects maximum reached, reconnect timeout set to "..tSecurity.General.DoubleLoginTimeout.." seconds.", "FI", 1)
						tSend:ChatToUser("You have reached the maximum amount of double logins in a short time, you have to wait  "..tSecurity.General.DoubleLoginTimeout.." seconds before connecting again.", "AAAA", sid)
					end

					ThreadUnlockMutex(mutexiplog)
--~ 				tLogs.IP[client.sIP]["$"].LastLogin --// last login time
--~ 				tLogs.IP[client.sIP]["$"].CurrentLogin --// real login time
				else
					tFuncs:Debug(tUser:GetDataLine(sid).."Double reconnected, but IP is protected, not updating counter.", "X")
				end

			end

			tInfo:SetCT(client, sid)

			--// If GUI logged in:
			if client.sProfile == "Gui" then
				tServer.GuiSID = sid
			end

			-- // add for addnewpass here
			if addnewpass then
				tAccounts[lnick].sPassword = pass
				tFuncs:SaveSettings("accounts")
				tFuncs:Debug(tUser:GetDataLine(sid).."Successfully set a new password on login.", nil , 0)

				if tSecurity.General.CleanBindIP and tAccounts[lnick].sBindIP then
					tAccounts[lnick].sBindIP = nil
					tFuncs:Debug(tUser:GetDataLine(sid).."Successfully cleared BindIP on new password.", nil , 0)
				end
			end
			return true, client

		--// Wrong password
		elseif reg and rnick then
			local failed = reg.tFailed or {}
			failed.iDate = os.time()
			failed.sIp = client.sIP
			failed.sReason = "Wrong pass used: "..pass
			failed.iTimes = (failed.iTimes or 0) + 1
			tAccounts[lnick].tFailed = failed

			if client.tSupports.SaltPass then tFuncs:Debug(tUser:GetDataLine(sid).."[Wrong pass] Salt: "..client.sPassSalt.." Salted Pass: "..passmatch.." Client sends: "..pass, "X") end

			if failed.iTimes >= tSecurity.General.WrongPassMax then
				tAccounts[lnick].tFailed.iTimes = 0
				local res, retmsg = tBan:BanIp(client.sIP, "Maximum password attempts reached, you are banned for "..tFuncs:Duration(tSecurity.General.WrongPassBan)..".", tSecurity.General.WrongPassBan)
				if res then
					tFuncs:Debug("Maximum password attempts reached for user <"..nick.."> with IP: "..client.sIP..", IP banned for "..tFuncs:Duration(tSecurity.General.WrongPassBan)..".")
				else
					tFuncs:Debug("Maximum password attempts reached for user <"..nick.."> with IP: "..client.sIP..", "..retmsg)
				end
				return false, client, "Wrong pass entered by "..nick..", IP banned"
			elseif tProfiles[reg.sProfile] and tProfiles[reg.sProfile].bOperator then --// Only output failed operator logins to feed
				tFuncs:Debug("Wrong password ("..failed.iTimes..") entered for nickname <"..nick.."> with IP: "..client.sIP.." and profile: "..reg.sProfile)
			end

			tFuncs:SaveSettings("accounts")
			return false, client, "Wrong pass entered by "..nick..", total attempts: "..failed.iTimes

		--// ERROR
		else
			tFuncs:Debug(tUser:GetDataLine(sid).."[Check:Pass] [ERROR] "..rnick..", assigned profile:User")
			client.sProfile = "User"
			client.iProfile = -1
			return false, client, "Account doesn't exist"
		end
	end,

}

tChange = {

["Kenny"] = function (self, sid, data)
	local kenny = "***uummmffllhn"
	local result = ""
	math.randomseed(os.time())
	math.random(100)
	for character in string.gmatch(data, ".") do
		if character == " " then 
			result = result.." "
		else
			local ran = math.random(1, #kenny)
			result = result..kenny:sub(ran, ran)
		end
	end
	tFuncs:Debug("[Kenny] "..tUser:GetDataLine(sid).."said: "..data.." / Output: "..result)
	return result
end,

["Muppet"] = function (self, sid, data)
	local replaced = data
	local muppets = tFilters.Muppets
	local num = false

	--// Quotes by 3rd person
	num = math.random(1, ((muppets.QuotesChance or 1)*(#muppets.Quotes or 1)))
	if muppets.Quotes[num] then 
		tSend:ChatFromNick(muppets.Quotes[num][2], "AAAA", muppets.Quotes[num][1])
		tFuncs:Debug("[Muppet] "..tUser:GetDataLine(sid).."said: "..data.." / Output by <"..muppets.Quotes[num][1].."> :"..muppets.Quotes[num][2])
		return false
	end

	--// Random line replacement
	num = math.random(1,((muppets.LinesChance or 1)*(#muppets.Lines or 1)))
	local result = muppets.Lines[num]
	if result then
		tFuncs:Debug("[Muppet] "..tUser:GetDataLine(sid).."said: "..data.." / Output: "..result)
		return result
	end

	--// Text Replacements
	for num,value in ipairs(muppets.Replacements) do
		replaced = replaced:gsub(value[1], tFuncs:Escape(value[2]))
	end

	--// Random word replacement
	local result = ""
	for w in string.gmatch(replaced, "%S+") do
		local num = math.random(1,((muppets.WordsChance or 1)*(#muppets.Words or 1)))
		result = result..(muppets.Words[num] or w).." "
	end

	tFuncs:Debug("[Muppet] "..tUser:GetDataLine(sid).."said: "..data.." / Output: "..result)
	return result
end,

}
